/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-async-queue.c - Async message passing between threads.
 *
 * Copyright (C) 2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 *
 * Based on GAsyncQueue from the GLib library.
 */

/*
 * MT safe
 */

#include "config.h"

#include <glib.h>
#include "tnt-async-queue.h"

#define RETURN_QUEUE_MAX_LENGTH 128

static void
hash_action (TntAsyncQueue *queue, gpointer action, GList *elem)
{
  g_hash_table_insert (queue->action_hash, action, elem);
}

static void
unhash_action (TntAsyncQueue *queue, gpointer action)
{
  g_hash_table_remove (queue->action_hash, action);
}

static void
destroy_action (TntAsyncQueue *queue, gpointer action)
{
  unhash_action (queue, action);
  if (queue->action_destroy_func)
    queue->action_destroy_func (action);
}

/**
 * tnt_async_queue_new:
 * 
 * Creates a new asynchronous queue with the initial reference count of 1.
 * 
 * Return value: the new #TntAsyncQueue.
 **/
TntAsyncQueue*
tnt_async_queue_new (TntDispatchFunc *return_dispatch_func,
                     GHashFunc action_hash_func,
                     GEqualFunc action_equal_func,
                     GDestroyNotify action_destroy_func,
                     GDestroyNotify return_destroy_func,
                     gpointer callback_data)
{
  TntAsyncQueue *retval;

  g_return_val_if_fail (action_hash_func != NULL, NULL);
  g_return_val_if_fail (action_equal_func != NULL, NULL);

  retval = g_new (TntAsyncQueue, 1);
  retval->mutex = g_mutex_new ();
  retval->cond = g_cond_new ();
  retval->dispatch_cond = g_cond_new ();
  retval->action_queue = g_queue_new ();
  retval->return_queue = g_queue_new ();
  retval->action_hash = g_hash_table_new (action_hash_func, action_equal_func);
  retval->dispatching_for_action = NULL;
  retval->last_popped_action = NULL;
  retval->cancelling_popped_action = FALSE;
  retval->discard_returns = FALSE;
  retval->waiting_threads = 0;
  retval->ref_count = 1;
  retval->return_dispatch_func = return_dispatch_func;
  retval->action_hash_func = action_hash_func;
  retval->action_equal_func = action_equal_func;
  retval->action_destroy_func = action_destroy_func;
  retval->return_destroy_func = return_destroy_func;
  retval->return_dispatch_id = 0;
  retval->callback_data = callback_data;

  return retval;
}

/**
 * tnt_async_queue_ref:
 * @queue: a #TntAsyncQueue.
 *
 * Increases the reference count of the asynchronous @queue by 1.
 **/
void 
tnt_async_queue_ref (TntAsyncQueue *queue)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  
  g_mutex_lock (queue->mutex);
  queue->ref_count++;
  g_mutex_unlock (queue->mutex);
}

/**
 * tnt_async_queue_ref_unlocked:
 * @queue: a #TntAsyncQueue.
 * 
 * Increases the reference count of the asynchronous @queue by 1. This
 * function must be called while holding the @queue's lock.
 **/
void 
tnt_async_queue_ref_unlocked (TntAsyncQueue *queue)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  
  queue->ref_count++;
}

/**
 * tnt_async_queue_unref_and_unlock:
 * @queue: a #TntAsyncQueue.
 * 
 * Decreases the reference count of the asynchronous @queue by 1 and
 * releases the lock. This function must be called while holding the
 * @queue's lock. If the reference count went to 0, the @queue will be
 * destroyed and the memory allocated will be freed. So you are not
 * allowed to use the @queue afterwards, as it might have disappeared.
 * The obvious asymmetry (it is not named
 * tnt_async_queue_unref_unlocked(<!-- -->)) is because the queue can't be
 * unlocked after unreffing it, as it might already have disappeared.
 **/
void 
tnt_async_queue_unref_and_unlock (TntAsyncQueue *queue)
{
  gboolean stop;

  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);

  queue->ref_count--;
  stop = (queue->ref_count == 0);
  g_mutex_unlock (queue->mutex);
  
  if (stop)
  {
    GList *l;

    g_return_if_fail (queue->waiting_threads == 0);
    g_mutex_free (queue->mutex);
    g_cond_free (queue->cond);
    g_cond_free (queue->dispatch_cond);

    if (queue->return_dispatch_id)
      g_source_remove (queue->return_dispatch_id);

    if (queue->action_destroy_func)
    {
      if (queue->last_popped_action)
        queue->action_destroy_func (queue->last_popped_action);

      for (l = queue->action_queue->head; l; l = g_list_next (l))
        queue->action_destroy_func (l->data);
    }

    if (queue->return_destroy_func)
    {
      for (l = queue->return_queue->head; l; l = g_list_next (l))
        queue->return_destroy_func (l->data);
    }

    g_queue_free (queue->action_queue);
    g_queue_free (queue->return_queue);
    g_hash_table_destroy (queue->action_hash);
    g_free (queue);
  }
}

/**
 * tnt_async_queue_unref:
 * @queue: a #TntAsyncQueue.
 * 
 * Decreases the reference count of the asynchronous @queue by 1. If
 * the reference count went to 0, the @queue will be destroyed and the
 * memory allocated will be freed. So you are not allowed to use the
 * @queue afterwards, as it might have disappeared.
 **/
void 
tnt_async_queue_unref (TntAsyncQueue *queue)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);

  g_mutex_lock (queue->mutex);
  tnt_async_queue_unref_and_unlock (queue);
}

/**
 * tnt_async_queue_lock:
 * @queue: a #TntAsyncQueue.
 * 
 * Acquires the @queue's lock. After that you can only call the
 * <function>tnt_async_queue_*_unlocked()</function> function variants on that
 * @queue. Otherwise it will deadlock.
 **/
void
tnt_async_queue_lock (TntAsyncQueue *queue)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);

  g_mutex_lock (queue->mutex);
}

/**
 * tnt_async_queue_unlock:
 * @queue: a #TntAsyncQueue.
 * 
 * Releases the queue's lock.
 **/
void 
tnt_async_queue_unlock (TntAsyncQueue *queue)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);

  g_mutex_unlock (queue->mutex);
}

static gboolean
dispatch_return (TntAsyncQueue *queue)
{
  gpointer dispatch_data;

  g_mutex_lock (queue->mutex);

  g_assert (queue->return_dispatch_func);

  while ((dispatch_data = g_queue_pop_tail (queue->return_queue)))
  {
    gpointer action;

    action = queue->dispatching_for_action = *((gpointer *) dispatch_data - 1);

    if (!queue->discard_returns)
    {
      g_mutex_unlock (queue->mutex);
      queue->return_dispatch_func (dispatch_data, queue->callback_data);
      g_mutex_lock (queue->mutex);
    }

    /* queue->dispatching_for_action gets set to NULL if the action being dispatched for
     * was cancelled by the dispatchee */
    if (queue->dispatching_for_action)
    {
      /* Unref the action it points to */
      action = *((gpointer *) dispatch_data - 1);
      (*((guint *) action - 1))--;

      if (*((guint *) action - 1) == 0 && action != queue->last_popped_action)
        destroy_action (queue, action);
    }

    if (queue->return_destroy_func)
      queue->return_destroy_func (dispatch_data);
  }

  queue->return_dispatch_id = 0;
  queue->dispatching_for_action = NULL;
  g_cond_signal (queue->dispatch_cond);
  g_mutex_unlock (queue->mutex);
  return FALSE;
}

void
tnt_async_queue_push_return (TntAsyncQueue *queue, gpointer data, gpointer action)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);
  g_return_if_fail (queue->return_dispatch_func != NULL);

  g_mutex_lock (queue->mutex);
  tnt_async_queue_push_return_unlocked (queue, data, action);
  g_mutex_unlock (queue->mutex);
}

void
tnt_async_queue_push_return_unlocked (TntAsyncQueue *queue, gpointer data, gpointer action)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);
  g_return_if_fail (queue->return_dispatch_func != NULL);

  if (!action)
    action = queue->last_popped_action;

  /* If we're cancelling, don't accept return values */
  if (queue->discard_returns ||
      (queue->cancelling_popped_action && action == queue->last_popped_action))
  {
    if (queue->return_destroy_func)
      queue->return_destroy_func (data);
    return;
  }

  /* Give it a pointer to current action, and increment action's ref count */
  *((gpointer *) data - 1) = action;
  (*((guint *) action - 1))++;

  /* Queue it */
  g_queue_push_head (queue->return_queue, data);

  if (!queue->return_dispatch_id)
  {
    queue->return_dispatch_id = g_idle_add_full (G_PRIORITY_LOW,
                                                 (GSourceFunc) dispatch_return,
                                                 queue, NULL);
  }

  /* If master is handling returned data slowly, block here */
  if (queue->return_queue->length >= RETURN_QUEUE_MAX_LENGTH)
    g_cond_wait (queue->dispatch_cond, queue->mutex);
}

/**
 * tnt_async_queue_push:
 * @queue: a #TntAsyncQueue.
 * @data: @data to push into the @queue.
 *
 * Pushes the @data into the @queue. @data must not be %NULL.
 **/
void
tnt_async_queue_push_action_head (TntAsyncQueue* queue, gpointer data)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);

  g_mutex_lock (queue->mutex);
  tnt_async_queue_push_action_head_unlocked (queue, data);
  g_mutex_unlock (queue->mutex);
}

/**
 * tnt_async_queue_push_unlocked:
 * @queue: a #TntAsyncQueue.
 * @data: @data to push into the @queue.
 * 
 * Pushes the @data into the @queue. @data must not be %NULL. This
 * function must be called while holding the @queue's lock.
 **/
void
tnt_async_queue_push_action_head_unlocked (TntAsyncQueue* queue, gpointer data)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);

  g_queue_push_head (queue->action_queue, data);
  hash_action (queue, data, queue->action_queue->head);
  g_cond_signal (queue->cond);
}

/**
 * tnt_async_queue_push:
 * @queue: a #TntAsyncQueue.
 * @data: @data to push into the @queue.
 *
 * Pushes the @data into the @queue. @data must not be %NULL.
 **/
void
tnt_async_queue_push_action_tail (TntAsyncQueue* queue, gpointer data)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);

  g_mutex_lock (queue->mutex);
  tnt_async_queue_push_action_tail_unlocked (queue, data);
  g_mutex_unlock (queue->mutex);
}

/**
 * tnt_async_queue_push_unlocked:
 * @queue: a #TntAsyncQueue.
 * @data: @data to push into the @queue.
 * 
 * Pushes the @data into the @queue. @data must not be %NULL. This
 * function must be called while holding the @queue's lock.
 **/
void
tnt_async_queue_push_action_tail_unlocked (TntAsyncQueue* queue, gpointer data)
{
  g_return_if_fail (queue);
  g_return_if_fail (queue->ref_count > 0);
  g_return_if_fail (data);

  g_queue_push_tail (queue->action_queue, data);
  hash_action (queue, data, queue->action_queue->tail);
  g_cond_signal (queue->cond);
}

static gpointer
tnt_async_queue_pop_action_intern_unlocked (TntAsyncQueue* queue, gboolean try, 
                                            GTimeVal *end_time)
{
  gpointer retval;

  if (queue->last_popped_action && *((guint *) queue->last_popped_action - 1) == 0)
    destroy_action (queue, queue->last_popped_action);

  queue->last_popped_action = NULL;
  queue->cancelling_popped_action = FALSE;

  if (!g_queue_peek_tail (queue->action_queue))
  {
    if (try)
      return NULL;
    if (!end_time)
    {
      queue->waiting_threads++;
      while (!g_queue_peek_tail (queue->action_queue))
        g_cond_wait (queue->cond, queue->mutex);
      queue->waiting_threads--;
    }
    else
    {
      queue->waiting_threads++;
      while (!g_queue_peek_tail (queue->action_queue))
        if (!g_cond_timed_wait (queue->cond, queue->mutex, end_time))
          break;
      queue->waiting_threads--;
      if (!g_queue_peek_tail (queue->action_queue))
        return NULL;
    }
  }

  retval = g_queue_pop_tail (queue->action_queue);
  g_assert (retval);

  hash_action (queue, retval, NULL);

  queue->last_popped_action = retval;

  return retval;
}

/**
 * tnt_async_queue_pop:
 * @queue: a #TntAsyncQueue.
 * 
 * Pops data from the @queue. This function blocks until data become
 * available.
 *
 * Return value: data from the queue.
 **/
gpointer
tnt_async_queue_pop_action (TntAsyncQueue* queue)
{
  gpointer retval;

  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  g_mutex_lock (queue->mutex);
  retval = tnt_async_queue_pop_action_intern_unlocked (queue, FALSE, NULL);
  g_mutex_unlock (queue->mutex);

  return retval;
}

/**
 * tnt_async_queue_pop_unlocked:
 * @queue: a #TntAsyncQueue.
 * 
 * Pops data from the @queue. This function blocks until data become
 * available. This function must be called while holding the @queue's
 * lock.
 *
 * Return value: data from the queue.
 **/
gpointer
tnt_async_queue_pop_action_unlocked (TntAsyncQueue* queue)
{
  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  return tnt_async_queue_pop_action_intern_unlocked (queue, FALSE, NULL);
}

/**
 * tnt_async_queue_try_pop:
 * @queue: a #TntAsyncQueue.
 * 
 * Tries to pop data from the @queue. If no data is available, %NULL is
 * returned.
 *
 * Return value: data from the queue or %NULL, when no data is
 * available immediately.
 **/
gpointer
tnt_async_queue_try_pop_action (TntAsyncQueue* queue)
{
  gpointer retval;

  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  g_mutex_lock (queue->mutex);
  retval = tnt_async_queue_pop_action_intern_unlocked (queue, TRUE, NULL);
  g_mutex_unlock (queue->mutex);

  return retval;
}

/**
 * tnt_async_queue_try_pop_unlocked:
 * @queue: a #TntAsyncQueue.
 * 
 * Tries to pop data from the @queue. If no data is available, %NULL is
 * returned. This function must be called while holding the @queue's
 * lock.
 *
 * Return value: data from the queue or %NULL, when no data is
 * available immediately.
 **/
gpointer
tnt_async_queue_try_pop_action_unlocked (TntAsyncQueue* queue)
{
  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  return tnt_async_queue_pop_action_intern_unlocked (queue, TRUE, NULL);
}

/**
 * tnt_async_queue_timed_pop:
 * @queue: a #TntAsyncQueue.
 * @end_time: a #GTimeVal, determining the final time.
 *
 * Pops data from the @queue. If no data is received before @end_time,
 * %NULL is returned.
 *
 * To easily calculate @end_time a combination of g_get_current_time()
 * and g_time_val_add() can be used.
 *
 * Return value: data from the queue or %NULL, when no data is
 * received before @end_time.
 **/
gpointer
tnt_async_queue_timed_pop_action (TntAsyncQueue* queue, GTimeVal *end_time)
{
  gpointer retval;

  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  g_mutex_lock (queue->mutex);
  retval = tnt_async_queue_pop_action_intern_unlocked (queue, FALSE, end_time);
  g_mutex_unlock (queue->mutex);

  return retval;  
}

/**
 * tnt_async_queue_timed_pop_unlocked:
 * @queue: a #TntAsyncQueue.
 * @end_time: a #GTimeVal, determining the final time.
 *
 * Pops data from the @queue. If no data is received before @end_time,
 * %NULL is returned. This function must be called while holding the
 * @queue's lock.
 *
 * To easily calculate @end_time a combination of g_get_current_time()
 * and g_time_val_add() can be used.
 *
 * Return value: data from the queue or %NULL, when no data is
 * received before @end_time.
 **/
gpointer
tnt_async_queue_timed_pop_action_unlocked (TntAsyncQueue* queue, GTimeVal *end_time)
{
  g_return_val_if_fail (queue, NULL);
  g_return_val_if_fail (queue->ref_count > 0, NULL);

  return tnt_async_queue_pop_action_intern_unlocked (queue, FALSE, end_time);
}

gboolean
tnt_async_queue_cancel_action (TntAsyncQueue *queue, gpointer data)
{
  gboolean result;

  g_return_val_if_fail (queue != NULL, FALSE);
  g_return_val_if_fail (queue->ref_count > 0, FALSE);
  g_return_val_if_fail (data != NULL, FALSE);

  g_mutex_lock (queue->mutex);
  result = tnt_async_queue_cancel_action_unlocked (queue, data);
  g_mutex_unlock (queue->mutex);

  return result;
}

static void
remove_action_results (TntAsyncQueue *queue, gpointer action)
{
  GList *l, *next;

  for (l = queue->return_queue->head; l; l = next)
  {
    gpointer result = l->data;

    next = g_list_next (l);

    if (*((gpointer *) result - 1) == action)
    {
      (*((guint *) action - 1))--;

      if (queue->return_destroy_func)
        queue->return_destroy_func (result);

      if (queue->return_queue->tail == l)
        queue->return_queue->tail = l->prev;

      queue->return_queue->head = g_list_delete_link (queue->return_queue->head, l);
      queue->return_queue->length--;

      if (*((guint *) action - 1) == 0)
        break;
    } 
  }
}

gboolean
tnt_async_queue_cancel_action_unlocked (TntAsyncQueue *queue, gpointer data)
{
  gpointer  action;
  GList    *l;

  g_return_val_if_fail (queue != NULL, FALSE);
  g_return_val_if_fail (queue->ref_count > 0, FALSE);
  g_return_val_if_fail (data != NULL, FALSE);

  if (queue->last_popped_action && queue->action_equal_func (data, queue->last_popped_action))
  {
    remove_action_results (queue, queue->last_popped_action);
    queue->cancelling_popped_action = TRUE;
    return TRUE;
  }

  if (!g_hash_table_lookup_extended (queue->action_hash, data, &action, (gpointer *) &l))
    return FALSE;

  if (l)
  {
    /* Unlink from action queue */

    if (queue->action_queue->tail == l)
      queue->action_queue->tail = l->prev;

    queue->action_queue->head = g_list_delete_link (queue->action_queue->head, l);
    queue->action_queue->length--;
  }

  remove_action_results (queue, action);

  if (*((guint *) action - 1) == 0)
    destroy_action (queue, action);

  if (queue->dispatching_for_action == action)
    queue->dispatching_for_action = NULL;

  return TRUE;
}

gboolean
tnt_async_queue_cancelling_current_action (TntAsyncQueue *queue)
{
  gboolean result;

  g_mutex_lock (queue->mutex);
  result = tnt_async_queue_cancelling_current_action_unlocked (queue);
  g_mutex_unlock (queue->mutex);

  return result;
}

gboolean
tnt_async_queue_cancelling_current_action_unlocked (TntAsyncQueue *queue)
{
  return queue->cancelling_popped_action;
}

/**
 * tnt_async_queue_length:
 * @queue: a #TntAsyncQueue.
 * 
 * Returns the length of the queue, negative values mean waiting
 * threads, positive values mean available entries in the
 * @queue. Actually this function returns the number of data items in
 * the queue minus the number of waiting threads. Thus a return value
 * of 0 could mean 'n' entries in the queue and 'n' thread waiting.
 * That can happen due to locking of the queue or due to
 * scheduling.  
 *
 * Return value: the length of the @queue.
 **/
gint
tnt_async_queue_get_pending_action_length (TntAsyncQueue* queue)
{
  gint retval;

  g_return_val_if_fail (queue, 0);
  g_return_val_if_fail (queue->ref_count > 0, 0);

  g_mutex_lock (queue->mutex);
  retval = queue->action_queue->length - queue->waiting_threads;
  g_mutex_unlock (queue->mutex);

  return retval;
}

/**
 * tnt_async_queue_length_unlocked:
 * @queue: a #TntAsyncQueue.
 * 
 * Returns the length of the queue, negative values mean waiting
 * threads, positive values mean available entries in the
 * @queue. Actually this function returns the number of data items in
 * the queue minus the number of waiting threads. Thus a return value
 * of 0 could mean 'n' entries in the queue and 'n' thread waiting.
 * That can happen due to locking of the queue or due to
 * scheduling. This function must be called while holding the @queue's
 * lock.
 *
 * Return value: the length of the @queue.
 **/
gint
tnt_async_queue_get_pending_action_length_unlocked (TntAsyncQueue* queue)
{
  g_return_val_if_fail (queue, 0);
  g_return_val_if_fail (queue->ref_count > 0, 0);

  return queue->action_queue->length - queue->waiting_threads;
}

void
tnt_async_queue_begin_shutdown (TntAsyncQueue *queue)
{
  g_return_if_fail (queue != NULL);
  g_return_if_fail (queue->ref_count > 0);

  g_mutex_lock (queue->mutex);
  queue->discard_returns = TRUE;

  /* If worker is already blocking, wake it up */
  g_cond_signal (queue->dispatch_cond);
  g_mutex_unlock (queue->mutex);
}

/* Action stores a simple ref count from results, and result stores a
 * pointer to action that generated it */

gpointer
tnt_async_queue_new_action_real (gint size)
{
  gpointer action;

  action = g_malloc0 (size + sizeof (guint));
  return ((guint *) action + 1);
}

void
tnt_async_queue_free_action (gpointer action)
{
  g_free ((guint *) action - 1);
}

gpointer
tnt_async_queue_new_return_real (gint size)
{
  gpointer result;

  result = g_malloc0 (size + sizeof (gpointer));
  return ((gpointer *) result + 1);
}

void
tnt_async_queue_free_return (gpointer result)
{
  g_free ((gpointer *) result - 1);
}

void
tnt_async_queue_ref_action_unlocked (TntAsyncQueue *queue, gpointer action)
{
  (*((guint *) action - 1))++;
}

void
tnt_async_queue_ref_action (TntAsyncQueue *queue, gpointer action)
{
  g_mutex_lock (queue->mutex);
  tnt_async_queue_ref_action_unlocked (queue, action);
  g_mutex_unlock (queue->mutex);
}

void
tnt_async_queue_unref_action_unlocked (TntAsyncQueue *queue, gpointer action)
{
  (*((guint *) action - 1))--;

  if (*((guint *) action - 1) == 0)
  {
    destroy_action (queue, action);
    if (queue->last_popped_action == action)
      queue->last_popped_action = NULL;
  }
}

void
tnt_async_queue_unref_action (TntAsyncQueue *queue, gpointer action)
{
  g_mutex_lock (queue->mutex);
  tnt_async_queue_unref_action_unlocked  (queue, action);
  g_mutex_unlock (queue->mutex);
}
