/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-avi.c - AVI stream metadata extraction.
 * Adapted from avilib, a part of transcode.
 *
 * Copyright (C) 2005 Joakim Ziegler
 * Copyright (C) 2001-2002 Thomas Oestreich
 * Copyright (C) 1999 Rainer Johanni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Joakim Ziegler <joakim@avmaria.com>
 *          Thomas Oestreich
 *          Rainer Johanni <rainer@johanni.de>
 */

#include "config.h"
#include "tnt.h"

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <glib.h>
#include <limits.h>
#include <errno.h>
#include <limits.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# define PACKED_STRUCT_ATTR __attribute__((__packed__))
#else
# include <io.h>
# define PACKED_STRUCT_ATTR
#endif

#define PAD_EVEN(x) ( ((x)+1) & ~1 )

#define AVI_MAX_TRACKS 8
#define AVI_ERR_READ         3
#define AVI_ERR_NO_MEM       8
#define AVI_ERR_NO_AVI       9     /* Not an AVI file */
#define AVI_ERR_NO_HDRL     10     /* AVI file has no has no header list,
                                      corrupted ??? */
#define AVI_ERR_NO_MOVI     11     /* AVI file has no has no MOVI list,
                                      corrupted ??? */

#define AVI_ERR_NO_VIDS     12     /* AVI file contains no video data */

#define AVI_MODE_READ   1

typedef struct
{
  gint64 key;
  gint64 pos;
  gint64 len;
} video_index_entry;

typedef struct
{
  gint64 pos;
  gint64 len;
  gint64 tot;
} audio_index_entry;

typedef struct _avisuperindex_entry {
  guint64 qwOffset;
  guint32 dwSize;
  guint32 dwDuration;
} avisuperindex_entry;

typedef struct _avistdindex_entry {
  guint32 dwOffset;
  guint32 dwSize;
} avistdindex_entry;

typedef struct _avistdindex_chunk {
  gchar   fcc[4];
  guint32 dwSize;
  guint16 wLongsPerEntry;
  guint8  bIndexSubType;
  guint8  bIndexType;
  guint32 nEntriesInUse;
  char    dwChunkId[4];
  guint64 qwBaseOffset;
  guint32 dwReserved3;
  avistdindex_entry *aIndex;
} avistdindex_chunk;

typedef struct _avisuperindex_chunk {
  gchar   fcc[4];
  guint32 dwSize;
  guint16 wLongsPerEntry;
  guint8  bIndexSubType;
  guint8  bIndexType;
  guint32 nEntriesInUse;
  gchar   dwChunkId[4];
  guint32 dwReserved[3];
  avisuperindex_entry *aIndex;
  avistdindex_chunk **stdindex;
} avisuperindex_chunk;

typedef struct track_s
{
  long   a_fmt;             /* Audio format, see #defines below */
  long   a_chans;           /* Audio channels, 0 for no audio */
  long   a_rate;            /* Rate in Hz */
  long   a_bits;            /* bits per audio sample */
  long   mp3rate;           /* mp3 bitrate kbs*/
  long   a_vbr;             /* 0 == no Variable BitRate */
  long   padrate;	          /* byte rate used for zero padding */
  long   audio_strn;        /* Audio stream number */
  gint64 audio_bytes;       /* Total number of bytes of audio data */
  long   audio_chunks;      /* Chunks of audio data in the file */
  char   audio_tag[4];      /* Tag of audio data */
  long   audio_posc;        /* Audio position: chunk */
  long   audio_posb;        /* Audio position: byte within chunk */
  gint64 a_codech_off;       /* absolut offset of audio codec information */ 
  gint64 a_codecf_off;       /* absolut offset of audio codec information */ 
  audio_index_entry *audio_index;
  avisuperindex_chunk *audio_superindex;
} track_t;

typedef struct
{
  guint32  bi_size;
  guint32  bi_width;
  guint32  bi_height;
  guint16  bi_planes;
  guint16  bi_bit_count;
  guint32  bi_compression;
  guint32  bi_size_image;
  guint32  bi_x_pels_per_meter;
  guint32  bi_y_pels_per_meter;
  guint32  bi_clr_used;
  guint32  bi_clr_important;
} alBITMAPINFOHEADER;

#pragma pack (1)

typedef struct PACKED_STRUCT_ATTR
{
  guint16  w_format_tag;
  guint16  n_channels;
  guint32  n_samples_per_sec;
  guint32  n_avg_bytes_per_sec;
  guint16  n_block_align;
  guint16  w_bits_per_sample;
  guint16  cb_size;
} alWAVEFORMATEX;

typedef struct PACKED_STRUCT_ATTR
{
  guint32 fcc_type; 
  guint32 fcc_handler; 
  guint32 dw_flags; 
  guint32 dw_caps; 
  guint16 w_priority;
  guint16 w_language;
  guint32 dw_scale;
  guint32 dw_rate;
  guint32 dw_start;
  guint32 dw_length;
  guint32 dw_initial_frames;
  guint32 dw_suggested_buffer_size;
  guint32 dw_quality;
  guint32 dw_sample_size;
  guint16 dw_left;
  guint16 dw_top;
  guint16 dw_right;
  guint16 dw_bottom;
} alAVISTREAMHEADER;

#pragma pack ()

typedef struct
{
  
  long   fdes;              /* File descriptor of AVI file */
  long   mode;              /* 0 for reading, 1 for writing */
  
  long   width;             /* Width  of a video frame */
  long   height;            /* Height of a video frame */
  double fps;               /* Frames per second */
  gchar  compressor[8];     /* Type of compressor, 4 bytes + padding for 0 byte */
  gchar  compressor2[8];    /* Type of compressor, 4 bytes + padding for 0 byte */
  long   video_strn;        /* Video stream number */
  long   video_frames;      /* Number of video frames */
  char   video_tag[4];      /* Tag of video data */
  long   video_pos;         /* Number of next frame to be read (if index present) */
  alAVISTREAMHEADER video_stream_header;
  guint32 max_len;          /* maximum video chunk present */
  track_t track[AVI_MAX_TRACKS];  /* up to AVI_MAX_TRACKS audio tracks supported */
  gint64  pos;              /* position in file */
  long    n_idx;             /* number of index entries actually filled */
  long    max_idx;           /* number of index entries actually allocated */
  gint64  v_codech_off;     /* absolut offset of video codec (strh) info */ 
  gint64  v_codecf_off;     /* absolut offset of video codec (strf) info */ 
  guint8 (*idx)[16];       /* index entries (AVI idx1 tag) */
  video_index_entry *video_index;
  avisuperindex_chunk *video_superindex;  /* index of indices */
  gint    is_opendml;           /* set to 1 if this is an odml file with multiple index chunks */
  gint64  last_pos;         /* Position of last frame written */
  guint32 last_len;         /* Length of last frame written */
  gint    must_use_index;       /* Flag if frames are duplicated */
  gint64  movi_start;
  gint    total_frames;         /* total number of frames if dmlh is present */
  
  gint    anum;                 /* total number of audio tracks */
  gint    aptr;
  gint    comment_fd;
  gchar  *index_file;
  
  alBITMAPINFOHEADER *bitmap_info_header;
  alWAVEFORMATEX     *wave_format_ex[AVI_MAX_TRACKS];
  alAVISTREAMHEADER   stream_headers[AVI_MAX_TRACKS];
  void*		extradata;
  unsigned long	extradata_size;
} avi_t;

static long AVI_errno = 0;

static guint64
str2ullong (unsigned char *str)
{
   guint64 r = (str[0] | (str[1]<<8) | (str[2]<<16) | (str[3]<<24));
   guint64 s = (str[4] | (str[5]<<8) | (str[6]<<16) | (str[7]<<24));
   return ((s << 32) & G_GINT64_CONSTANT (0xffffffff00000000U)) | (r & G_GINT64_CONSTANT (0xffffffffU));
}

static guint32
str2ulong (unsigned char *str)
{
   return ( str[0] | (str[1]<<8) | (str[2]<<16) | (str[3]<<24) );
}

static guint32
str2ushort (unsigned char *str)
{
   return ( str[0] | (str[1]<<8) );
}

static void
long2str (unsigned char *dst, gint32 n)
{
   dst[0] = (n    )&0xff;
   dst[1] = (n>> 8)&0xff;
   dst[2] = (n>>16)&0xff;
   dst[3] = (n>>24)&0xff;
}

static guint32
str2ulong_len (unsigned char *str)
{
   return str2ulong (str) & 0x7fffffff;
}

static guint32
str2ulong_key (unsigned char *str)
{
  guint32 c = str2ulong (str);
  c &= 0x80000000;
  if (c == 0) return 0x10;
  else return 0;
}

static gint
xio_close (int fd)
{
  return close(fd);
}

static off_t
xio_lseek (int fd, off_t offset, int whence)
{
  return lseek (fd, offset, whence);
}

static gssize
xio_read (int fd, void *buf, gsize count)
{
  return read (fd, buf, count);
}

static gint
avi_sampsize (avi_t *AVI, int j)
{
   int s;
   s = ((AVI->track[j].a_bits+7)/8)*AVI->track[j].a_chans;
   if(s<4) s=4; /* avoid possible zero divisions */ 
   return s;
}

static gint
AVI_close (avi_t *AVI)
{
   int j;

   if (AVI->comment_fd>0)
       xio_close(AVI->comment_fd);
   AVI->comment_fd = -1;
   xio_close (AVI->fdes);
   if (AVI->idx) free(AVI->idx);
   if (AVI->video_index) free(AVI->video_index);
   if (AVI->video_superindex) {
       if(AVI->video_superindex->aIndex) free(AVI->video_superindex->aIndex);
       free(AVI->video_superindex);
   }

   for (j=0; j<AVI->anum; j++) 
   {
       if (AVI->track[j].audio_index) free(AVI->track[j].audio_index);
       if (AVI->track[j].audio_superindex) {
	       if(AVI->track[j].audio_superindex->aIndex) free(AVI->track[j].audio_superindex->aIndex);
         free(AVI->track[j].audio_superindex);
       }
   }

   if (AVI->bitmap_info_header)
     free(AVI->bitmap_info_header);
   for (j = 0; j < AVI->anum; j++)
     if (AVI->wave_format_ex[j])
       free(AVI->wave_format_ex[j]);

   free(AVI);
   AVI=NULL;

   return 0;
}

#define ERR_EXIT(x) \
{ \
   AVI_close(AVI); \
   AVI_errno = x; \
   return 0; \
}

static gssize
avi_read (int fd, char *buf, gsize len)
{
   gssize n = 0;
   gssize r = 0;

   while (r < len) {
      n = xio_read (fd, buf + r, len - r);
      if (n == 0)
	  break;
      if (n < 0) {
	  if (errno == EINTR)
	      continue;
	  else
	      break;
      } 

      r += n;
   }

   return r;
}

static gint
avi_add_index_entry (avi_t *AVI, unsigned char *tag, long flags, unsigned long pos, unsigned long len)
{
   void *ptr;

   if(AVI->n_idx>=AVI->max_idx) {
     ptr = realloc((void *)AVI->idx,(AVI->max_idx+4096)*16);
     
     if(ptr == 0) {
       AVI_errno = AVI_ERR_NO_MEM;
       return -1;
     }
     AVI->max_idx += 4096;
     AVI->idx = (unsigned char((*)[16]) ) ptr;
   }
   
   /* Add index entry */
   memcpy(AVI->idx[AVI->n_idx],tag,4);
   long2str(AVI->idx[AVI->n_idx]+ 4,flags);
   long2str(AVI->idx[AVI->n_idx]+ 8, pos);
   long2str(AVI->idx[AVI->n_idx]+12, len);
   
   /* Update counter */
   AVI->n_idx++;

   if(len>AVI->max_len) AVI->max_len=len;

   return 0;
}

static gint
avi_parse_index_from_file (avi_t *AVI, char *filename)
{
    char data[100];
    FILE *fd = NULL;
    gint64 pos, len, f_pos, tot_chunks[AVI_MAX_TRACKS];
    int key=0, type;
    int vid_chunks=0, aud_chunks[AVI_MAX_TRACKS];
    long line_count=0;
    char *c, d;
    int i,j;

    for (i=0; i<AVI_MAX_TRACKS; i++) aud_chunks[i] = 0;
    i=0;

    /* already have an index -- may be incomplete */
    if (AVI->video_index) {
      free(AVI->video_index);
      AVI->video_index = NULL;
    }

    for(j=0; j<AVI->anum; ++j) {
      if(AVI->track[j].audio_index) {
        free(AVI->track[j].audio_index);
      }
      AVI->track[j].audio_index = NULL;
      AVI->track[j].audio_chunks = 0;
    }   

    if (!(fd = fopen(filename, "r"))) { perror ("avi_parse_index_from_file: fopen"); return -1; }

    /* read header */
    fgets(data, 100, fd);

    if ( g_strncasecmp(data, "AVIIDX1", 7) != 0) {
      return -1;
    }

    /* read comment */
    fgets(data, 100, fd);
    f_pos = ftell(fd);
    while (fgets(data, 100, fd)) {
      d = data[5] - '1';
      if        (d == 0) {
        vid_chunks++;
      } else if (d == 1 || d == 2 || d == 3 || d == 4 ||
                 d == 5 || d == 6 || d == 7 || d == 8  ) {
        aud_chunks[d-1]++;
      } else 
        continue;
      
      line_count++;
    }

    AVI->video_frames = vid_chunks;
    for (j=0; j<AVI->anum; ++j) AVI->track[j].audio_chunks = aud_chunks[j];

    if (AVI->video_frames==0) ERR_EXIT(AVI_ERR_NO_VIDS);
    AVI->video_index = (video_index_entry *) malloc(vid_chunks*sizeof(video_index_entry));
    if (AVI->video_index==0) ERR_EXIT(AVI_ERR_NO_MEM);

    for (j=0; j<AVI->anum; ++j) {
      if(AVI->track[j].audio_chunks) {
        AVI->track[j].audio_index = (audio_index_entry *) malloc(aud_chunks[j]*sizeof(audio_index_entry));
        if(AVI->track[j].audio_index==0) ERR_EXIT(AVI_ERR_NO_MEM);
      }
    }   

    /* reset after header */
    fseek(fd, f_pos, SEEK_SET);

    vid_chunks = 0;
    for(j=0; j<AVI->anum; ++j) aud_chunks[j] = tot_chunks[j] = 0;

    while (fgets(data, 100, fd)) {
      c     = strchr (data, ' ');
      type  = strtol(c+1, &c, 10);
      c = strchr(c+1, ' ');
      c = strchr(c+1, ' ');
      pos   = strtol(c+1, &c, 10);  /* FIXME: Should be strtoll, but that doesn't work on win32 */
      len   = strtol(c+1, &c, 10);
      key   = strtol(c+1, &c, 10);
      i = type-1;
      
      switch (i) {
	    case 0: /* video */
        AVI->video_index[vid_chunks].key = (gint64)(key?0x10:0);
        AVI->video_index[vid_chunks].pos = pos+8;
        AVI->video_index[vid_chunks].len = len;
        vid_chunks++;
        break;
	    case 1: case 2: case 3: case 4:
	    case 5: case 6: case 7: case 8:
        j=i-1;
        AVI->track[j].audio_index[aud_chunks[j]].pos = pos+8;
        AVI->track[j].audio_index[aud_chunks[j]].len = len;
        AVI->track[j].audio_index[aud_chunks[j]].tot = tot_chunks[j];
        tot_chunks[j] += AVI->track[j].audio_index[aud_chunks[j]].len;
        aud_chunks[j]++;
        break;
	    default:
        continue;
      }
    }
    for(j=0; j<AVI->anum; ++j) AVI->track[j].audio_bytes = tot_chunks[j];

    fclose (fd);

    return 0;
}

static gint
avi_parse_input_file (avi_t *AVI, gboolean getIndex)
{
  long i, rate, scale, idx_type;
  gint64 n;
  unsigned char *hdrl_data;
  long header_offset=0, hdrl_len=0;
  long nvi, nai[AVI_MAX_TRACKS], ioff;
  long tot[AVI_MAX_TRACKS];
  int j;
  int lasttag = 0;
  int vids_strh_seen = 0;
  int vids_strf_seen = 0;
  int auds_strh_seen = 0;
  int num_stream = 0;
  char data[256];
  gint64 oldpos=-1, newpos=-1;
  
  /* Read first 12 bytes and check that this is an AVI file */

  if( avi_read(AVI->fdes,data,12) != 12 ) ERR_EXIT(AVI_ERR_READ)
  if( g_strncasecmp(data  ,"RIFF",4) !=0 ||
      g_strncasecmp(data+8,"AVI ",4) !=0 ) ERR_EXIT(AVI_ERR_NO_AVI)

   /* Go through the AVI file and extract the header list,
      the start position of the 'movi' list and an optionally
      present idx1 tag */
  hdrl_data = 0;

  while (1)
  {
    if( avi_read(AVI->fdes,data,8) != 8 ) break; /* We assume it's EOF */
    newpos=xio_lseek(AVI->fdes,0,SEEK_CUR);
    if(oldpos==newpos) {
      /* This is a broken AVI stream... */
      return -1;
    }
    oldpos=newpos;
    
    n = str2ulong((unsigned char *)data+4);
    n = PAD_EVEN(n);
    
    if(g_strncasecmp(data,"LIST",4) == 0)
      {
        if( avi_read(AVI->fdes,data,4) != 4 ) ERR_EXIT(AVI_ERR_READ)
        n -= 4;
        if(g_strncasecmp(data,"hdrl",4) == 0)
        {
            hdrl_len = n;
            hdrl_data = (unsigned char *) malloc(n);
            if(hdrl_data==0) ERR_EXIT(AVI_ERR_NO_MEM);
            
            /* offset of header */
	    
            header_offset = xio_lseek(AVI->fdes,0,SEEK_CUR);
				 
            if( avi_read(AVI->fdes,(char *)hdrl_data,n) != n ) ERR_EXIT(AVI_ERR_READ)
         }
         else if(g_strncasecmp(data,"movi",4) == 0)
         {
            AVI->movi_start = xio_lseek(AVI->fdes,0,SEEK_CUR);
            if (xio_lseek(AVI->fdes,n,SEEK_CUR)==(gint64)-1) break;
         }
         else
            if (xio_lseek(AVI->fdes,n,SEEK_CUR)==(gint64)-1) break;
      }
      else if(g_strncasecmp(data,"idx1",4) == 0)
      {
         /* n must be a multiple of 16, but the reading does not
            break if this is not the case */

         AVI->n_idx = AVI->max_idx = n/16;
         AVI->idx = (unsigned  char((*)[16]) ) malloc(n);
         if(AVI->idx==0) ERR_EXIT(AVI_ERR_NO_MEM)
         if(avi_read(AVI->fdes, (char *) AVI->idx, n) != n ) {
	     free ( AVI->idx); AVI->idx=NULL;
	     AVI->n_idx = AVI->max_idx = 0;
	 }
      }
      else
         xio_lseek(AVI->fdes,n,SEEK_CUR);
   }

   if(!hdrl_data      ) ERR_EXIT(AVI_ERR_NO_HDRL)
   if(!AVI->movi_start) ERR_EXIT(AVI_ERR_NO_MOVI)

   /* Interpret the header list */

   for(i=0;i<hdrl_len;)
   {
      /* List tags are completly ignored */
      if(g_strncasecmp((char *)hdrl_data+i,"LIST",4)==0) { i+= 12; continue; }

      n = str2ulong(hdrl_data+i+4);
      n = PAD_EVEN(n);

      /* Interpret the tag and its args */
      if(g_strncasecmp((char *)hdrl_data+i,"strh",4)==0)
      {
         i += 8;
         if(g_strncasecmp((char *)hdrl_data+i,"vids",4) == 0 && !vids_strh_seen)
         {
            memcpy(AVI->compressor,hdrl_data+i+4,4);
            AVI->compressor[4] = 0;

            AVI->v_codech_off = header_offset + i+4;

            scale = str2ulong(hdrl_data+i+20);
            rate  = str2ulong(hdrl_data+i+24);
            if(scale!=0) AVI->fps = (double)rate/(double)scale;
            AVI->video_frames = str2ulong(hdrl_data+i+32);
            AVI->video_strn = num_stream;
            AVI->max_len = 0;
            vids_strh_seen = 1;
            lasttag = 1; /* vids */
            memcpy(&AVI->video_stream_header, hdrl_data + i,
                   sizeof(alAVISTREAMHEADER));
         }
         else if (g_strncasecmp ((char *)hdrl_data+i,"auds",4) ==0 && ! auds_strh_seen)
         {
           /* inc audio tracks */
           AVI->aptr=AVI->anum;
           ++AVI->anum;
	   
           if(AVI->anum > AVI_MAX_TRACKS) {
             return(-1);
           }
	   
           AVI->track[AVI->aptr].audio_bytes = str2ulong(hdrl_data+i+32)*avi_sampsize(AVI, 0);
           AVI->track[AVI->aptr].audio_strn = num_stream;
           
           AVI->track[AVI->aptr].a_vbr = !str2ulong(hdrl_data+i+44);
           
           AVI->track[AVI->aptr].padrate = str2ulong(hdrl_data+i+24);
           memcpy(&AVI->stream_headers[AVI->aptr], hdrl_data + i,
                  sizeof(alAVISTREAMHEADER));

           lasttag = 2; /* auds */
           AVI->track[AVI->aptr].a_codech_off = header_offset + i;
         }
         else if (g_strncasecmp (hdrl_data+i,"iavs",4) ==0 && ! auds_strh_seen) {
           return (-1);
         }
         else
           lasttag = 0;
         num_stream++;
      }
      else if (g_strncasecmp(hdrl_data+i,"dmlh",4) == 0) {
        AVI->total_frames = str2ulong(hdrl_data+i+8);
        i += 8;
      }
      else if(g_strncasecmp((char *)hdrl_data+i,"strf",4)==0)
      {
         i += 8;
         if(lasttag == 1)
         {
           alBITMAPINFOHEADER bih;
           memcpy(&bih, hdrl_data + i, sizeof(alBITMAPINFOHEADER));
           AVI->bitmap_info_header = (alBITMAPINFOHEADER *)
             malloc(str2ulong((unsigned char *)&bih.bi_size));
           if (AVI->bitmap_info_header != NULL)
             memcpy(AVI->bitmap_info_header, hdrl_data + i,
                    str2ulong((unsigned char *)&bih.bi_size));
           
           AVI->width  = str2ulong(hdrl_data+i+4);
           AVI->height = str2ulong(hdrl_data+i+8);
           vids_strf_seen = 1;
           AVI->v_codecf_off = header_offset + i+16;
           
           memcpy(AVI->compressor2, hdrl_data+i+16, 4);
           AVI->compressor2[4] = 0;
           
         }
         else if(lasttag == 2)
         {
           alWAVEFORMATEX *wfe;
           char *nwfe;
           int wfes;
           
           if ((hdrl_len - i) < sizeof(alWAVEFORMATEX))
             wfes = hdrl_len - i;
           else
             wfes = sizeof(alWAVEFORMATEX);
           wfe = (alWAVEFORMATEX *)malloc(sizeof(alWAVEFORMATEX));
           if (wfe != NULL) {
             memset(wfe, 0, sizeof(alWAVEFORMATEX));
             memcpy(wfe, hdrl_data + i, wfes);
             if (str2ushort((unsigned char *)&wfe->cb_size) != 0) {
               nwfe = (char *)
                 realloc(wfe, sizeof(alWAVEFORMATEX) +
                         str2ushort((unsigned char *)&wfe->cb_size));
               if (nwfe != 0) {
                 gint64 lpos = xio_lseek(AVI->fdes, 0, SEEK_CUR);
                 xio_lseek(AVI->fdes, header_offset + i + sizeof(alWAVEFORMATEX),
                           SEEK_SET);
                 wfe = (alWAVEFORMATEX *)nwfe;
                 nwfe = &nwfe[sizeof(alWAVEFORMATEX)];
                 avi_read(AVI->fdes, nwfe,
                          str2ushort((unsigned char *)&wfe->cb_size));
                 xio_lseek(AVI->fdes, lpos, SEEK_SET);
               }
             }
             AVI->wave_format_ex[AVI->aptr] = wfe;
           }
           
           AVI->track[AVI->aptr].a_fmt   = str2ushort(hdrl_data+i  );
           AVI->track[AVI->aptr].a_codecf_off = header_offset + i;
           AVI->track[AVI->aptr].a_chans = str2ushort(hdrl_data+i+2);
           AVI->track[AVI->aptr].a_rate  = str2ulong (hdrl_data+i+4);
           AVI->track[AVI->aptr].mp3rate = 8*str2ulong(hdrl_data+i+8)/1000;
           AVI->track[AVI->aptr].a_bits  = str2ushort(hdrl_data+i+14);
         }
      }
      else if(g_strncasecmp(hdrl_data+i,"indx",4) == 0) {
        char *a;
        int j;
        
        if(lasttag == 1) /* Video */
          { 
            a = hdrl_data+i;
            
            AVI->video_superindex = (avisuperindex_chunk *) malloc (sizeof (avisuperindex_chunk));
            memcpy (AVI->video_superindex->fcc, a, 4);             a += 4;
            AVI->video_superindex->dwSize = str2ulong(a);          a += 4;
            AVI->video_superindex->wLongsPerEntry = str2ushort(a); a += 2;
            AVI->video_superindex->bIndexSubType = *a;             a += 1;
            AVI->video_superindex->bIndexType = *a;                a += 1;
            AVI->video_superindex->nEntriesInUse = str2ulong(a);   a += 4;
            memcpy (AVI->video_superindex->dwChunkId, a, 4);       a += 4;

            a += 4; a += 4; a += 4;
            
            AVI->video_superindex->aIndex = 
              malloc (AVI->video_superindex->wLongsPerEntry * AVI->video_superindex->nEntriesInUse * sizeof (guint32));
            
            /* position of ix## chunks */
            for (j=0; j<AVI->video_superindex->nEntriesInUse; ++j) {
              AVI->video_superindex->aIndex[j].qwOffset = str2ullong (a);  a += 8;
              AVI->video_superindex->aIndex[j].dwSize = str2ulong (a);     a += 4;
              AVI->video_superindex->aIndex[j].dwDuration = str2ulong (a); a += 4;  
            }
            AVI->is_opendml = 1;            
          }
        else if(lasttag == 2) /* Audio */
          {

            a = hdrl_data+i;
            
            AVI->track[AVI->aptr].audio_superindex = (avisuperindex_chunk *) malloc (sizeof (avisuperindex_chunk));
            memcpy (AVI->track[AVI->aptr].audio_superindex->fcc, a, 4);             a += 4;
            AVI->track[AVI->aptr].audio_superindex->dwSize = str2ulong(a);          a += 4;
            AVI->track[AVI->aptr].audio_superindex->wLongsPerEntry = str2ushort(a); a += 2;
            AVI->track[AVI->aptr].audio_superindex->bIndexSubType = *a;             a += 1;
            AVI->track[AVI->aptr].audio_superindex->bIndexType = *a;                a += 1;
            AVI->track[AVI->aptr].audio_superindex->nEntriesInUse = str2ulong(a);   a += 4;
            memcpy (AVI->track[AVI->aptr].audio_superindex->dwChunkId, a, 4);       a += 4;

            a += 4; a += 4; a += 4;
            
            AVI->track[AVI->aptr].audio_superindex->aIndex = 
              malloc (AVI->track[AVI->aptr].audio_superindex->wLongsPerEntry * 
                      AVI->track[AVI->aptr].audio_superindex->nEntriesInUse * sizeof (guint32));
            
            /* position of ix## chunks */
            for (j=0; j<AVI->track[AVI->aptr].audio_superindex->nEntriesInUse; ++j) {
              AVI->track[AVI->aptr].audio_superindex->aIndex[j].qwOffset = str2ullong (a);  a += 8;
              AVI->track[AVI->aptr].audio_superindex->aIndex[j].dwSize = str2ulong (a);     a += 4;
              AVI->track[AVI->aptr].audio_superindex->aIndex[j].dwDuration = str2ulong (a); a += 4;
            }
          }
        i += 8;
      }
      else if((g_strncasecmp(hdrl_data+i,"JUNK",4) == 0) ||
              (g_strncasecmp(hdrl_data+i,"strn",4) == 0) ||
              (g_strncasecmp(hdrl_data+i,"vprp",4) == 0)){
        i += 8;
        /* do not reset lasttag */
      } else
        {
          i += 8;
          lasttag = 0;
        }  
      i += n;
   }

   free(hdrl_data);

   if (!vids_strh_seen || !vids_strf_seen) ERR_EXIT(AVI_ERR_NO_VIDS)

   AVI->video_tag[0] = AVI->video_strn/10 + '0';
   AVI->video_tag[1] = AVI->video_strn%10 + '0';
   AVI->video_tag[2] = 'd';
   AVI->video_tag[3] = 'b';
   
   /* Audio tag is set to "99wb" if no audio present */
   if(!AVI->track[0].a_chans) AVI->track[0].audio_strn = 99;

   { 
     int i=0;
     for(j=0; j<AVI->anum+1; ++j) {
       if (j == AVI->video_strn) continue;
       AVI->track[i].audio_tag[0] = j/10 + '0';
       AVI->track[i].audio_tag[1] = j%10 + '0';
       AVI->track[i].audio_tag[2] = 'w';
       AVI->track[i].audio_tag[3] = 'b';
       ++i;
     }
   }
   
   xio_lseek(AVI->fdes,AVI->movi_start,SEEK_SET);
   
   /* get index if wanted */

   if(AVI->index_file && !getIndex) {
     int ret;
     
     ret = avi_parse_index_from_file(AVI, AVI->index_file);
     
     /* Reposition the file */
     
     xio_lseek(AVI->fdes,AVI->movi_start,SEEK_SET);
     AVI->video_pos = 0;
     return (ret);
     
   }
   if(!getIndex) return(0);
   
   /* if the file has an idx1, check if this is relative
      to the start of the file or to the start of the movi list */
   
   idx_type = 0;
   
   if(AVI->idx)
     {
       gint64 pos, len;
       
       /* Search the first videoframe in the idx1 and look where
          it is in the file */
       
      for (i=0;i<AVI->n_idx;i++)
         if (g_strncasecmp((char *)AVI->idx[i],(char *)AVI->video_tag,3)==0 ) break;
      if (i>=AVI->n_idx) ERR_EXIT(AVI_ERR_NO_VIDS)
                          
      pos = str2ulong(AVI->idx[i]+ 8);
      len = str2ulong(AVI->idx[i]+12);
      
      xio_lseek(AVI->fdes,pos,SEEK_SET);
      if (avi_read(AVI->fdes,data,8)!=8) ERR_EXIT(AVI_ERR_READ)
      if (g_strncasecmp(data,(char *)AVI->idx[i],4)==0 && str2ulong((unsigned char *)data+4)==len )
      {
        idx_type = 1; /* Index from start of file */
      }
      else
      {
        xio_lseek(AVI->fdes,pos+AVI->movi_start-4,SEEK_SET);
        if (avi_read(AVI->fdes,data,8)!=8) ERR_EXIT(AVI_ERR_READ)
        if (g_strncasecmp(data,(char *)AVI->idx[i],4)==0 && str2ulong((unsigned char *)data+4)==len )
        {
          idx_type = 2; /* Index from start of movi list */
        }
      }
      /* idx_type remains 0 if neither of the two tests above succeeds */
   }

   if(idx_type == 0 && !AVI->is_opendml && !AVI->total_frames)
   {
     /* we must search through the file to get the index */
     xio_lseek(AVI->fdes, AVI->movi_start, SEEK_SET);

     AVI->n_idx = 0;
     
     while(1)
     {
       if (avi_read(AVI->fdes,data,8) != 8 ) break;
       n = str2ulong ((unsigned char *)data+4);
       
       /* The movi list may contain sub-lists, ignore them */
       if(g_strncasecmp(data,"LIST",4)==0)
       {
         xio_lseek(AVI->fdes,4,SEEK_CUR);
         continue;
       }

       /* Check if we got a tag ##db, ##dc or ##wb */
       if( ( (data[2]=='d' || data[2]=='D') &&
             (data[3]=='b' || data[3]=='B' || data[3]=='c' || data[3]=='C') )
           || ( (data[2]=='w' || data[2]=='W') &&
                (data[3]=='b' || data[3]=='B') ) )
       {
         avi_add_index_entry(AVI,(unsigned char *)data,0,xio_lseek(AVI->fdes,0,SEEK_CUR)-8,n);
       }
       xio_lseek(AVI->fdes,PAD_EVEN(n),SEEK_CUR);
     }
     idx_type = 1;
   }

   /* read extended index chunks */
   if (AVI->is_opendml) {
     guint64 offset = 0;
     int hdrl_len = 4+4+2+1+1+4+4+8+4;
     char *en, *chunk_start;
     int k = 0, audtr = 0;
     guint32 nrEntries = 0;
     
     AVI->video_index = NULL;
     
     nvi = 0;
     for(audtr=0; audtr<AVI->anum; ++audtr) nai[audtr] = tot[audtr] = 0; 

     for (j=0; j<AVI->video_superindex->nEntriesInUse; j++) {
       /* read from file */
       chunk_start = en = malloc (AVI->video_superindex->aIndex[j].dwSize+hdrl_len);

       if (xio_lseek(AVI->fdes, AVI->video_superindex->aIndex[j].qwOffset, SEEK_SET) == (gint64)-1) {
         free(chunk_start);
         continue;
       }

       if (avi_read(AVI->fdes, en, AVI->video_superindex->aIndex[j].dwSize+hdrl_len) <= 0) {
         free(chunk_start);
         continue;
       }

       nrEntries = str2ulong(en + 12);
       offset = str2ullong(en + 20);

       /* skip header */
       en += hdrl_len;
       nvi += nrEntries;
       AVI->video_index = (video_index_entry *) realloc (AVI->video_index, nvi * sizeof (video_index_entry));
       if (!AVI->video_index) {
         exit(1);
       }

       while (k < nvi) {
         
         AVI->video_index[k].pos = offset + str2ulong(en); en += 4;
         AVI->video_index[k].len = str2ulong_len(en);
         AVI->video_index[k].key = str2ulong_key(en); en += 4;
         
         /* completely empty chunk */
         if (AVI->video_index[k].pos-offset == 0 && AVI->video_index[k].len == 0) {
           k--;
           nvi--;
         }
         
         k++;
       }

       free(chunk_start);
     }
     
     AVI->video_frames = nvi;
     /* this should deal with broken 'rec ' odml files. */
     if (AVI->video_frames == 0) {
       AVI->is_opendml=0;
       goto multiple_riff;
      }
 
      for(audtr=0; audtr<AVI->anum; ++audtr) {
        k = 0;
        if (!AVI->track[audtr].audio_superindex) {
          continue;
        }
        for (j=0; j<AVI->track[audtr].audio_superindex->nEntriesInUse; j++) {
          
          /* read from file */
          chunk_start = en = malloc (AVI->track[audtr].audio_superindex->aIndex[j].dwSize+hdrl_len);

          if (xio_lseek(AVI->fdes, AVI->track[audtr].audio_superindex->aIndex[j].qwOffset, SEEK_SET) == (gint64)-1) {
            free(chunk_start);
            continue;
          }
          
          if (avi_read(AVI->fdes, en, AVI->track[audtr].audio_superindex->aIndex[j].dwSize+hdrl_len) <= 0) {
            free(chunk_start);
            continue;
          }
          
          nrEntries = str2ulong(en + 12);
          offset = str2ullong(en + 20);
          
          /* skip header */
          en += hdrl_len;
          nai[audtr] += nrEntries;
          AVI->track[audtr].audio_index = (audio_index_entry *) realloc (AVI->track[audtr].audio_index, nai[audtr] * sizeof (audio_index_entry));
          
          while (k < nai[audtr]) {

            AVI->track[audtr].audio_index[k].pos = offset + str2ulong(en); en += 4;
            AVI->track[audtr].audio_index[k].len = str2ulong_len(en); en += 4;
            AVI->track[audtr].audio_index[k].tot = tot[audtr];
            tot[audtr] += AVI->track[audtr].audio_index[k].len;
            
            ++k;
          }
          free(chunk_start);
        }
	 
        AVI->track[audtr].audio_chunks = nai[audtr];
        AVI->track[audtr].audio_bytes = tot[audtr];
      }
   }

   else if (AVI->total_frames && !AVI->is_opendml && idx_type==0) {

     long aud_chunks = 0;
   multiple_riff:
     xio_lseek(AVI->fdes, AVI->movi_start, SEEK_SET);
     
     AVI->n_idx = 0;
     
     /* Number of frames; only one audio track supported */
     nvi = AVI->video_frames = AVI->total_frames;
     nai[0] = AVI->track[0].audio_chunks = AVI->total_frames;
     for(j=1; j<AVI->anum; ++j) AVI->track[j].audio_chunks = 0;
     
     AVI->video_index = (video_index_entry *) malloc(nvi*sizeof(video_index_entry));
     
     if(AVI->video_index==0) ERR_EXIT(AVI_ERR_NO_MEM);
     
     for(j=0; j<AVI->anum; ++j) {
       if(AVI->track[j].audio_chunks) {
         AVI->track[j].audio_index = (audio_index_entry *) malloc((nai[j]+1)*sizeof(audio_index_entry));
         memset(AVI->track[j].audio_index, 0, (nai[j]+1)*(sizeof(audio_index_entry)));
         if(AVI->track[j].audio_index==0) ERR_EXIT(AVI_ERR_NO_MEM);
       }
     }   

     nvi = 0;
     for(j=0; j<AVI->anum; ++j) nai[j] = tot[j] = 0;
     
     aud_chunks = AVI->total_frames;
     
     while(1)
     {
       if (nvi >= AVI->total_frames) break;
       
       if( avi_read(AVI->fdes,data,8) != 8 ) break;
       n = str2ulong((unsigned char *)data+4);
       
       j=0;

       if (aud_chunks - nai[j] -1 <= 0) {
         aud_chunks += AVI->total_frames;
         AVI->track[j].audio_index = (audio_index_entry *) 
           realloc( AVI->track[j].audio_index, (aud_chunks+1)*sizeof(audio_index_entry));
         if (!AVI->track[j].audio_index) {
           AVI_errno = AVI_ERR_NO_MEM;
           return -1;
         }
       }
       
       /* Check if we got a tag ##db, ##dc or ##wb */
	 
       /* VIDEO */
       if( 
          (data[0]=='0' || data[1]=='0') &&
          (data[2]=='d' || data[2]=='D') &&
          (data[3]=='b' || data[3]=='B' || data[3]=='c' || data[3]=='C') ) {
         
         AVI->video_index[nvi].key = 0x0;
         AVI->video_index[nvi].pos = xio_lseek(AVI->fdes,0,SEEK_CUR);
         AVI->video_index[nvi].len = n;
         
         nvi++;
         xio_lseek(AVI->fdes,PAD_EVEN(n),SEEK_CUR);
       } 
       
       /* AUDIO */
       else if( 
               (data[0]=='0' || data[1]=='1') &&
               (data[2]=='w' || data[2]=='W') &&
               (data[3]=='b' || data[3]=='B') ) {
 
         AVI->track[j].audio_index[nai[j]].pos = xio_lseek(AVI->fdes,0,SEEK_CUR);
         AVI->track[j].audio_index[nai[j]].len = n;
         AVI->track[j].audio_index[nai[j]].tot = tot[j];
         tot[j] += AVI->track[j].audio_index[nai[j]].len;
         nai[j]++;

         xio_lseek(AVI->fdes,PAD_EVEN(n),SEEK_CUR);
       }
       else {
         xio_lseek(AVI->fdes,-4,SEEK_CUR);
       }
       
     }
     if (nvi < AVI->total_frames) {
     }
     AVI->video_frames = nvi;
     AVI->track[0].audio_chunks = nai[0];
     
      for(j=0; j<AVI->anum; ++j) AVI->track[j].audio_bytes = tot[j];
      idx_type = 1;
   }
   
   else
   {   
     /* Now generate the video index and audio index arrays */
     nvi = 0;
     for(j=0; j<AVI->anum; ++j) nai[j] = 0;
     
     for(i=0;i<AVI->n_idx;i++) {
       if(g_strncasecmp((char *)AVI->idx[i],AVI->video_tag,3) == 0) nvi++;
       
       for(j=0; j<AVI->anum; ++j) if(g_strncasecmp((char *)AVI->idx[i], AVI->track[j].audio_tag,4) == 0) nai[j]++;
     }
     
     AVI->video_frames = nvi;
     for(j=0; j<AVI->anum; ++j) AVI->track[j].audio_chunks = nai[j];
     
     if(AVI->video_frames==0) ERR_EXIT(AVI_ERR_NO_VIDS);
     AVI->video_index = (video_index_entry *) malloc(nvi*sizeof(video_index_entry));
     if(AVI->video_index==0) ERR_EXIT(AVI_ERR_NO_MEM);
     
     for(j=0; j<AVI->anum; ++j) {
       if(AVI->track[j].audio_chunks) {
         AVI->track[j].audio_index = (audio_index_entry *) malloc((nai[j]+1)*sizeof(audio_index_entry));
         memset(AVI->track[j].audio_index, 0, (nai[j]+1)*(sizeof(audio_index_entry)));
         if(AVI->track[j].audio_index==0) ERR_EXIT(AVI_ERR_NO_MEM);
       }
     }   
     
     nvi = 0;
     for(j=0; j<AVI->anum; ++j) nai[j] = tot[j] = 0;
     
     ioff = idx_type == 1 ? 8 : AVI->movi_start+4;
     
     for(i=0;i<AVI->n_idx;i++) {
       /* Video */
       if(g_strncasecmp((char *)AVI->idx[i],AVI->video_tag,3) == 0) {
         AVI->video_index[nvi].key = str2ulong(AVI->idx[i]+ 4);
         AVI->video_index[nvi].pos = str2ulong(AVI->idx[i]+ 8)+ioff;
         AVI->video_index[nvi].len = str2ulong(AVI->idx[i]+12);
         nvi++;
       }
     
       /* Audio */
       for(j=0; j<AVI->anum; ++j) {
         
         if(g_strncasecmp((char *)AVI->idx[i],AVI->track[j].audio_tag,4) == 0) {
           AVI->track[j].audio_index[nai[j]].pos = str2ulong(AVI->idx[i]+ 8)+ioff;
           AVI->track[j].audio_index[nai[j]].len = str2ulong(AVI->idx[i]+12);
           AVI->track[j].audio_index[nai[j]].tot = tot[j];
           tot[j] += AVI->track[j].audio_index[nai[j]].len;
           nai[j]++;
         }
       }
     }
     
     for(j=0; j<AVI->anum; ++j) AVI->track[j].audio_bytes = tot[j];
   } 

   /* Reposition the file */   
   xio_lseek(AVI->fdes,AVI->movi_start,SEEK_SET);
   AVI->video_pos = 0;

   return(0);
}

static avi_t *
AVI_open_fd (gint fd, gboolean getIndex)
{
  avi_t *AVI=NULL;
  
  /* Create avi_t structure */
  AVI = (avi_t *) malloc(sizeof(avi_t));
  if(AVI==NULL)
  {
    AVI_errno = AVI_ERR_NO_MEM;
    return 0;
  }
  memset((void *)AVI,0,sizeof(avi_t));
  
  AVI->mode = AVI_MODE_READ; /* open for reading */
  
  /* file alread open */
  lseek (fd, 0, SEEK_SET);
  AVI->fdes = fd;
  
  AVI_errno = 0;
  avi_parse_input_file(AVI, getIndex);
  
  if (AVI != NULL && !AVI_errno) {
      AVI->aptr=0;
  }
  
  if (AVI_errno)
    return AVI = NULL;
  else
    return AVI;
}

static int
AVI_set_audio_track (avi_t *AVI, int track)
{
  
  if(track < 0 || track + 1 > AVI->anum) return(-1);

  //this info is not written to file anyway
  AVI->aptr=track;
  return 0;
}

static long
AVI_video_frames (avi_t *AVI)
{
   return AVI->video_frames;
}

static int
AVI_video_width (avi_t *AVI)
{
   return AVI->width;
}

static int
AVI_video_height (avi_t *AVI)
{
   return AVI->height;
}

static double
AVI_frame_rate (avi_t *AVI)
{
   return AVI->fps;
}

static int
AVI_audio_tracks (avi_t *AVI)
{
    return(AVI->anum);
}

static long
AVI_audio_bytes (avi_t *AVI)
{
   return AVI->track[AVI->aptr].audio_bytes;
}

gboolean
tnt_avi_extract_file_meta (gint fd, TntFileMeta *file_meta)
{
	gboolean     success = FALSE;
  TntStat      tstat;
  avi_t       *avi_file;        
  gint64       size;
  gint         duration, audio_tracks;
  gdouble      frame_rate;
  gint         i;

  g_return_val_if_fail (fd >= 0, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  fd = dup (fd);
  if (fd < 0)
    return FALSE;

  avi_file = AVI_open_fd (fd, TRUE);
  if (!avi_file)
  {
    close (fd);
    return FALSE;
  }

  tnt_file_meta_set_format (file_meta, TNT_FORMAT_AVI);

  /* Get the pixel dimensions and duration */
  frame_rate = AVI_frame_rate (avi_file);
  if (frame_rate == 0.0)
    frame_rate = 20.0;

  duration = AVI_video_frames (avi_file) / frame_rate;
  if (duration < 0)
    duration = 0;

  tnt_file_meta_set_duration (file_meta, duration);
  tnt_file_meta_set_width (file_meta, AVI_video_width (avi_file));
  tnt_file_meta_set_height (file_meta, AVI_video_height (avi_file));

  if (tnt_fstat (fd, &tstat))
    size = tstat.len;
  else
    size = 0;

  if (duration > 0 && size >= 0)
  {
    guint16 bitrate;

    /* Get the approximate video bitrate */
    /* FIXME: This could probably be tuned better */

    audio_tracks = AVI_audio_tracks (avi_file);    
    for (i = 0; i < audio_tracks; ++i)
    {
      AVI_set_audio_track (avi_file, i);
      size -= AVI_audio_bytes (avi_file);
    }

    size -= size / 50;  /* size *= 0.98 */

    if (size >= 0)
    {
      bitrate = (guint16) (((size / duration) * 8) / 1024);
      bitrate -= bitrate % 10;  /* Round down to multiple of 10 */
    }
    else
    {
      bitrate = 0;
    }

    tnt_file_meta_set_quality (file_meta, bitrate);
  }

  AVI_close (avi_file);
	return TRUE;
}
