/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-bandwidth.c - Describes a bandwidth goal and who belongs to it.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* Stats Collection
 *
 * 1. Set up a (signed) cps pool, based on the maximum cps.
 * 2. Count sockets in control/data categories and calculate ideal average cps,
 *    subtracting actual cps from the pool. Control sockets' ideal average is
 *    a specified fraction of data sockets' ideal average.
 *
 * Adjustment if Pool is Negative
 *
 * 1. Take a little bit from control sockets that are above average.
 * 2. If pool is still negative, take a little bit from data sockets that are
 *    above average.
 *
 * Adjustment if Pool is Positive
 *
 * 1. Give a little bit to all data sockets.
 * 2. Give a little bit to all control sockets. */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG(x)

#define BANDWIDTH_CHECK_TIMEOUT           1
#define BANDWIDTH_INCREASE_STEP           0.25
#define BANDWIDTH_DECREASE_STEP           0.25

#define DEVIANCE_FRACTION                 0.05
#define CONTROL_STREAM_FRACTION           0.20
#define CONTROL_STREAM_IDLE_FRACTION      0.90

#define QUALIFIED_DATA_FRACTION           0.95

#define UNCAPPED_FRACTION                 0.92
#define OVERSATURATED_GOAL_FRACTION       0.05
#define OVERSATURATED_EDGE_FRACTION       0.20

#define CPS_IS_CAPPED(cps, goal)     (cps > (goal * UNCAPPED_FRACTION) || cps + 75 > goal)

#define count_edges(bandwidth_group) (g_list_length (bandwidth_group->edge_list))
#define assert_types(o)              flux_object_check (FLUX_OBJECT (o), TNT_BANDWIDTH_GROUP_TYPE)

static void
collect_bandwidth_stats (TntBandwidthGroup *bandwidth_group, TntBandwidth *bandwidth)
{
  TntStreamType  stream_type;
  GList         *l;

  g_assert (bandwidth);
  memset (bandwidth, 0, sizeof (TntBandwidth));

  for (stream_type = 0; stream_type < TNT_STREAM_TYPE_NUM; stream_type++)
  {
    for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
    {
      TntEdge  *edge = l->data;
      TntTcpIO *tcp_io;

      tcp_io = tnt_edge_get_io (edge, stream_type);

      if (!tcp_io)
        continue;

      bandwidth->cps [stream_type] [TNT_TRANSFER_DIRECTION_IN] +=
        tnt_tcp_io_get_cps (tcp_io, TNT_TRANSFER_DIRECTION_IN);
      bandwidth->cps [stream_type] [TNT_TRANSFER_DIRECTION_OUT] +=
        tnt_tcp_io_get_cps (tcp_io, TNT_TRANSFER_DIRECTION_OUT);
    }
  }
}

static void
debug_log_bandwidth_stats (TntBandwidthGroup *bandwidth_group)
{
  GList *l;
  gint   bw_total_in = 0, bw_total_out = 0;

  g_print ("\033[H\033[2J");
  g_print ("Bandwidth group summary for '%s':\n", tnt_bandwidth_group_get_name (bandwidth_group));

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge       *edge = l->data;
    TntStreamType  stream_type;

    for (stream_type = 0; stream_type < TNT_STREAM_TYPE_NUM; stream_type++)
    {
      TntTcpIO *tcp_io = tnt_edge_get_io (edge, stream_type);

      if (tcp_io)
        g_print ("%016llx %s%s I=%6d/%6d O=%6d/%6d q=%3d/%3d\n", (guint64) tnt_edge_get_peer_fingerprint (edge), 
                 stream_type == TNT_STREAM_TYPE_CONTROL ? "C" : "D",
                 tnt_edge_get_qualified (edge) ? "*" : " ",
                 tnt_tcp_io_get_cps            (tcp_io, TNT_TRANSFER_DIRECTION_IN),
                 tnt_tcp_io_get_cps_goal       (tcp_io, TNT_TRANSFER_DIRECTION_IN),
                 tnt_tcp_io_get_cps            (tcp_io, TNT_TRANSFER_DIRECTION_OUT),
                 tnt_tcp_io_get_cps_goal       (tcp_io, TNT_TRANSFER_DIRECTION_OUT),
                 tnt_tcp_io_get_queue_out_len  (tcp_io, TNT_MESSAGE_ANY),
                 tnt_tcp_io_get_queue_out_max_len (tcp_io, TNT_MESSAGE_UNKNOWN));
      else
        g_print ("%016llx %s%s (no socket)\n", (guint64) tnt_edge_get_peer_fingerprint (edge),
                 stream_type == TNT_STREAM_TYPE_CONTROL ? "C" : "D",
                 tnt_edge_get_qualified (edge) ? "*" : " ");

      if (tcp_io)
      {
        bw_total_in  += tnt_tcp_io_get_cps (tcp_io, TNT_TRANSFER_DIRECTION_IN);
        bw_total_out += tnt_tcp_io_get_cps (tcp_io, TNT_TRANSFER_DIRECTION_OUT);
      }
    }
  }

  g_print ("\nTotal in: %d  Total out: %d\n", bw_total_in, bw_total_out);

  fflush (stdout);
}

static void
debug_log_routing_table (TntBandwidthGroup *bandwidth_group)
{
  GList *l;

  tnt_log_debug ("Routing tables for bandwidth group '%s':", tnt_bandwidth_group_get_name (bandwidth_group));

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;

    tnt_edge_debug_dump_routing_table (edge);
  }
}

static guint
get_edge_cps (TntEdge *edge, TntStreamType stream, TntTransferDirection dir)
{
  TntTcpIO *io;

  io = tnt_edge_get_io (edge, stream);
  if (io)
    return tnt_tcp_io_get_cps (io, dir);

  return 0;
}

static guint
get_edge_goal (TntEdge *edge, TntStreamType stream, TntTransferDirection dir)
{
  TntTcpIO *io;

  io = tnt_edge_get_io (edge, stream);
  if (io)
    return tnt_tcp_io_get_cps_goal (io, dir);

  return 0;
}

static void
set_edge_goal (TntEdge *edge, TntStreamType stream, TntTransferDirection dir, guint cps)
{
  TntTcpIO *io;

  io = tnt_edge_get_io (edge, stream);
  if (io)
    tnt_tcp_io_set_cps_goal (io, dir, cps);
}

static gint
count_edges_capped (TntBandwidthGroup *bandwidth_group, TntStreamType stream,
                    TntTransferDirection dir)
{
  GList *l;
  gint   num = 0;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps, goal;

    cps  = get_edge_cps  (edge, stream, dir);
    goal = get_edge_goal (edge, stream, dir);

    if (CPS_IS_CAPPED (cps, goal))
      num++;
  }

  return num;
}

static gint
count_edges_qualified (TntBandwidthGroup *bandwidth_group)
{
  GList *l;
  gint   num = 0;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;

    if (tnt_edge_get_qualified (edge))
      num++;
  }

  return num;
}

static gint
count_edges_qualified_capped (TntBandwidthGroup *bandwidth_group, TntStreamType stream,
                              TntTransferDirection dir)
{
  GList *l;
  gint   num = 0;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps, goal;

    if (!tnt_edge_get_qualified (edge))
      continue;

    cps  = get_edge_cps  (edge, stream, dir);
    goal = get_edge_goal (edge, stream, dir);

    if (CPS_IS_CAPPED (cps, goal))
      num++;
  }

  return num;
}

static gint
count_edges_unqualified_capped (TntBandwidthGroup *bandwidth_group, TntStreamType stream,
                                TntTransferDirection dir)
{
  GList *l;
  gint   num = 0;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps, goal;

    if (tnt_edge_get_qualified (edge))
      continue;

    cps  = get_edge_cps  (edge, stream, dir);
    goal = get_edge_goal (edge, stream, dir);

    if (CPS_IS_CAPPED (cps, goal))
      num++;
  }

  return num;
}

static void
increase_bandwidth (TntBandwidthGroup *bandwidth_group, gdouble bw_goal, gdouble bw_pool,
                    gint num_edges, TntTransferDirection dir)
{
  GList *l;
  gint   num_capped;

  /* Adjust control streams up */
  num_capped = count_edges_capped (bandwidth_group, TNT_STREAM_TYPE_CONTROL, dir);

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge  *edge = l->data;
    gdouble   cps, goal;

    cps  = get_edge_cps (edge, TNT_STREAM_TYPE_CONTROL, dir);
    goal = get_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir);

    if (CPS_IS_CAPPED (cps, goal))
    {
      gdouble new_cps = cps + (bw_pool / num_capped);
      if (new_cps > bw_goal)
        new_cps = bw_goal;
      bw_pool += cps - new_cps;
      set_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir, new_cps);
    }
  }

  /* Adjust data streams up */
  num_capped = count_edges_capped (bandwidth_group, TNT_STREAM_TYPE_DATA, dir);

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge  *edge = l->data;
    gdouble   cps, goal;

    cps  = get_edge_cps (edge, TNT_STREAM_TYPE_DATA, dir);
    goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);

    if (CPS_IS_CAPPED (cps, goal))
    {
      gdouble new_cps = cps + (bw_pool / num_capped);

      if (new_cps > bw_goal)
        new_cps = bw_goal;

      bw_pool += cps - new_cps;

      /* If the edge is handling user-requested blocks, make its bandwidth share
       * grow faster (data only). This applies to both incoming and outgoing. */

      /* FIXME: Kludge. */
      if (tnt_edge_get_qualified (edge))
        new_cps += 2000;

      set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, new_cps);
    }
  }
}

static gint
count_edges_above_cps (TntBandwidthGroup *bandwidth_group, TntStreamType stream,
                       TntTransferDirection dir, gdouble lim)
{
  GList *l;
  gint   num = 0;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps;

    cps = get_edge_cps (edge, stream, dir);
    if (cps > (lim * 1.01))
      num++;
  }

  return num;
}

static void
decrease_bandwidth (TntBandwidthGroup *bandwidth_group, gdouble bw_goal, gdouble bw_pool,
                    gint num_edges, TntTransferDirection dir)
{
  gdouble  control_avg, control_idle_avg, data_avg, data_idle_avg;
  gint     num_above_avg, num_idle_above_avg;
  GList   *l;

  /* Ideal averages */
  control_avg      = (bw_goal / num_edges) * CONTROL_STREAM_FRACTION;
  control_idle_avg = (bw_goal / num_edges) * CONTROL_STREAM_IDLE_FRACTION;
  data_avg         = (bw_goal / num_edges) * (1.0 - CONTROL_STREAM_FRACTION);
  data_idle_avg    = (bw_goal / num_edges) * (1.0 - CONTROL_STREAM_IDLE_FRACTION);

  /* Adjust control streams down towards average */
  num_idle_above_avg = count_edges_above_cps (bandwidth_group, TNT_STREAM_TYPE_CONTROL, dir, control_idle_avg);
  num_above_avg      = count_edges_above_cps (bandwidth_group, TNT_STREAM_TYPE_CONTROL, dir, control_avg) -
    num_idle_above_avg;

  g_assert (num_above_avg >= 0);
  g_assert (num_idle_above_avg >= 0);

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps  = get_edge_cps (edge, TNT_STREAM_TYPE_CONTROL, dir);

    if (tnt_edge_get_qualified (edge))
    {
      /* Servicing user */

      if (cps > control_avg * 1.01)
      {
        gdouble new_cps = cps + (bw_pool / num_above_avg);
        if (new_cps < control_avg)
          new_cps = control_avg;
        bw_pool += cps - new_cps;
        set_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir, new_cps);
      }
    }
    else
    {
      /* Idle */

      if (cps > control_idle_avg * 1.01)
      {
        gdouble new_cps = cps + (bw_pool / num_idle_above_avg);
        if (new_cps < control_idle_avg)
          new_cps = control_idle_avg;
        bw_pool += cps - new_cps;
        set_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir, new_cps);
      }
    }
  }

  if (bw_pool > -50)
    return;

  /* Adjust data streams down towards average */
  num_above_avg      = count_edges_above_cps (bandwidth_group, TNT_STREAM_TYPE_DATA, dir, data_avg);
  num_idle_above_avg = count_edges_above_cps (bandwidth_group, TNT_STREAM_TYPE_DATA, dir, data_idle_avg) -
    num_above_avg;

  g_assert (num_above_avg >= 0);
  g_assert (num_idle_above_avg >= 0);

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps  = get_edge_cps (edge, TNT_STREAM_TYPE_DATA, dir);;

    if (tnt_edge_get_qualified (edge))
    {
      /* Servicing user */

      if (cps > data_avg * 1.01)
      {
        gdouble new_cps = cps + (bw_pool / num_above_avg);
        if (new_cps < data_avg)
          new_cps = data_avg;
        bw_pool += cps - new_cps;
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, new_cps);
      }
    }
    else
    {
      /* Idle */

      if (cps > data_idle_avg * 1.01)
      {
        gdouble new_cps = cps + (bw_pool / num_idle_above_avg);
        if (new_cps < data_idle_avg)
          new_cps = data_idle_avg;
        bw_pool += cps - new_cps;
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, new_cps);
      }
    }
  }  
}

static void
adjust_bandwidth (TntBandwidthGroup *bandwidth_group, gdouble bw_goal,
                  gint num_edges, TntTransferDirection dir)
{
  gdouble  control_avg, control_idle_avg, data_avg, data_idle_avg, bw_pool, bw_share;
  gint     num_qualified, num_unqualified, num_qualified_capped, num_unqualified_capped;
  TntEdge *high_edge;
  GList   *l;

  /* --- 1. Make sure no streams are restricted to less bandwidth than they are entitled to --- */

  num_qualified   = count_edges_qualified (bandwidth_group);
  num_unqualified = num_edges - num_qualified;

  /* Don't divide by zero below */
  if (num_qualified == 0)
    num_qualified = 1;
  if (num_unqualified == 0)
    num_unqualified = 1;

  /* Ideal averages */
  control_avg      = (bw_goal / num_edges)       * CONTROL_STREAM_FRACTION;
  control_idle_avg = (bw_goal / num_edges)       * CONTROL_STREAM_IDLE_FRACTION;
  data_avg         = (bw_goal / num_qualified)   * (1.0 - CONTROL_STREAM_FRACTION) * QUALIFIED_DATA_FRACTION;
  data_idle_avg    = (bw_goal / num_unqualified) * (1.0 - CONTROL_STREAM_IDLE_FRACTION) * (1.0 - QUALIFIED_DATA_FRACTION);

  /* Control */

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  goal  = get_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir);

    if (tnt_edge_get_qualified (edge))
    {
      /* Servicing user */
      if (goal < control_avg)
        set_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir, control_avg);
    }
    else
    {
      /* Idle */
      if (goal < control_idle_avg)
        set_edge_goal (edge, TNT_STREAM_TYPE_CONTROL, dir, control_idle_avg);
    }
  }

  /* Data */

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);

    if (tnt_edge_get_qualified (edge))
    {
      /* Servicing user */
      if (goal < data_avg)
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, data_avg);
    }
    else
    {
      /* Idle */
      if (goal < data_idle_avg)
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, data_idle_avg);
    }
  }

  /* --- 2. If any priority data streams are capped at all, give them bandwidth from low-pri streams --- */

  num_qualified_capped = count_edges_qualified_capped (bandwidth_group, TNT_STREAM_TYPE_DATA, dir);
  if (num_qualified_capped > 0)
  {
    /* Take bw from unqualified data streams and the richest capped qualified data stream */

    bw_pool = 0.0;
    high_edge = NULL;

    for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
    {
      TntEdge *edge = l->data;
      gdouble  goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);
      gdouble  cps  = get_edge_cps  (edge, TNT_STREAM_TYPE_DATA, dir);

      if (tnt_edge_get_qualified (edge) && CPS_IS_CAPPED (cps, goal))
      {
        if (high_edge == NULL || goal > get_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir))
          high_edge = edge;
      }

      if (!tnt_edge_get_qualified (edge) && goal > data_idle_avg)
      {
        gdouble new_goal;

        new_goal = goal - (goal - data_idle_avg) * 0.1;
        if (new_goal < 100)
          new_goal = 100;

        if (new_goal < goal)
        {
          bw_pool += (goal - new_goal);
          set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, new_goal);
        }
      }
    }

    if (high_edge)
    {
      gdouble goal = get_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir);

      bw_pool += goal * 0.1;
      goal    *= 0.9;
      set_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir, goal);
    }

    /* Give bw to qualified data streams */

    bw_share = bw_pool / num_qualified_capped;

    for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
    {
      TntEdge *edge = l->data;
      gdouble  cps  = get_edge_cps  (edge, TNT_STREAM_TYPE_DATA, dir);
      gdouble  goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);

      if (tnt_edge_get_qualified (edge) && CPS_IS_CAPPED (cps, goal))
      {
        goal += bw_share;
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, goal);
      }
    }
  }

  /* --- 3. Take some bandwidth from the richest capped unqualified data stream and give to other uncapped --- */

  num_unqualified_capped = count_edges_unqualified_capped (bandwidth_group, TNT_STREAM_TYPE_DATA, dir);
  if (num_unqualified_capped > 0)
  {
    /* Take bw */

    bw_pool = 0.0;
    high_edge = NULL;

    for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
    {
      TntEdge *edge = l->data;
      gdouble  goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);
      gdouble  cps  = get_edge_cps  (edge, TNT_STREAM_TYPE_DATA, dir);

      if (!tnt_edge_get_qualified (edge) && CPS_IS_CAPPED (cps, goal))
      {
        if (high_edge == NULL || goal > get_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir))
          high_edge = edge;
      }
    }

    if (high_edge)
    {
      gdouble goal = get_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir);

      bw_pool += goal * 0.1;
      goal    *= 0.9;
      set_edge_goal (high_edge, TNT_STREAM_TYPE_DATA, dir, goal);
    }

    /* Give bw */

    bw_share = bw_pool / num_unqualified_capped;

    for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
    {
      TntEdge *edge = l->data;
      gdouble  cps  = get_edge_cps  (edge, TNT_STREAM_TYPE_DATA, dir);
      gdouble  goal = get_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir);

      if (!tnt_edge_get_qualified (edge) && CPS_IS_CAPPED (cps, goal))
      {
        goal += bw_share;
        set_edge_goal (edge, TNT_STREAM_TYPE_DATA, dir, goal);
      }
    }
  }
}

static void
adjust_direction_goals (TntBandwidthGroup *bandwidth_group, TntBandwidth *bandwidth_used,
                        gint num_edges, TntTransferDirection dir)
{
  gdouble bw_goal, bw_pool, bw_used;

  bw_goal =  bandwidth_group->goal_cps [dir];
  bw_used =  bandwidth_used->cps [TNT_STREAM_TYPE_CONTROL] [dir] +
             bandwidth_used->cps [TNT_STREAM_TYPE_DATA]    [dir];
  bw_pool =  bw_goal - bw_used;

  if (bw_pool < -(bw_goal * DEVIANCE_FRACTION))
  {
    /* We're using too much bandwidth */
    decrease_bandwidth (bandwidth_group, bw_goal, bw_pool, num_edges, dir);

    if (bandwidth_group->bandwidth_adjust_type [dir] != TNT_BANDWIDTH_ADJUSTMENT_DECREASE)
    {
      bandwidth_group->bandwidth_adjust_type [dir] = TNT_BANDWIDTH_ADJUSTMENT_DECREASE;
      bandwidth_group->bandwidth_adjust_changed_last [dir] = tnt_timestamp ();
    }
  }
  else if (bw_pool > bw_goal * DEVIANCE_FRACTION)
  {
    /* We're not using enough bandwidth */
    increase_bandwidth (bandwidth_group, bw_goal, bw_pool, num_edges, dir);

    if (bandwidth_group->bandwidth_adjust_type [dir] != TNT_BANDWIDTH_ADJUSTMENT_INCREASE)
    {
      bandwidth_group->bandwidth_adjust_type [dir] = TNT_BANDWIDTH_ADJUSTMENT_INCREASE;
      bandwidth_group->bandwidth_adjust_changed_last [dir] = tnt_timestamp ();
    }
  }
  else
  {
    if (bandwidth_group->bandwidth_adjust_type [dir] != TNT_BANDWIDTH_ADJUSTMENT_NONE)
    {
      bandwidth_group->bandwidth_adjust_type [dir] = TNT_BANDWIDTH_ADJUSTMENT_NONE;
      bandwidth_group->bandwidth_adjust_changed_last [dir] = tnt_timestamp ();
    }
  }

  /* Rectify any gross injustice */
  adjust_bandwidth (bandwidth_group, bw_goal, num_edges, dir);
}

static gboolean
oversaturated (TntBandwidthGroup *bandwidth_group, TntBandwidth *bandwidth_used, TntTransferDirection dir)
{
  gdouble  bw_goal, bw_used;
  GList   *l;

  bw_goal =  bandwidth_group->goal_cps [dir];
  bw_used =  bandwidth_used->cps [TNT_STREAM_TYPE_CONTROL] [dir] +
             bandwidth_used->cps [TNT_STREAM_TYPE_DATA]    [dir];

  if (bw_used < bw_goal * (1.0 - OVERSATURATED_GOAL_FRACTION))
    return FALSE;

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    gdouble  cps;

    cps = get_edge_cps (edge, TNT_STREAM_TYPE_CONTROL, dir) +
          get_edge_cps (edge, TNT_STREAM_TYPE_DATA, dir);

    /* If we have enough edges that all of them are under 20% of the bandwidth
     * goal, we're oversaturated. This will take at least 5 edges, and
     * probably a lot more. */
    if (cps > bw_goal * OVERSATURATED_EDGE_FRACTION)
      return FALSE;
  }

  tnt_log_debug ("Oversaturated!");
  return TRUE;
}

static void
maybe_trigger_action (TntBandwidthGroup *bandwidth_group, TntBandwidth *bw, TntTransferDirection dir)
{
  TntTimeStamp now;

  now = tnt_timestamp ();
  if (bandwidth_group->bandwidth_adjust_changed_last [dir] + 10 < now &&
      bandwidth_group->bandwidth_adjust_type [dir] == TNT_BANDWIDTH_ADJUSTMENT_INCREASE)
  {
    /* Get help in increasing transfer rate */
    if (bandwidth_group->action_func)
      bandwidth_group->action_func (bandwidth_group, dir, TNT_BANDWIDTH_ADJUSTMENT_INCREASE,
                                    bandwidth_group->action_func_data);
    bandwidth_group->bandwidth_adjust_changed_last [dir] = now;
  }
  else if (oversaturated (bandwidth_group, bw, dir))
  {
    /* Get help in decreasing transfer rate */
    if (bandwidth_group->action_func)
      bandwidth_group->action_func (bandwidth_group, dir, TNT_BANDWIDTH_ADJUSTMENT_DECREASE,
                                    bandwidth_group->action_func_data);
    bandwidth_group->bandwidth_adjust_changed_last [dir] = now;
  }
}

static void
adjust_io_goals (TntBandwidthGroup *bandwidth_group)
{
  TntBandwidth bw;
  gint         num_edges;

  collect_bandwidth_stats (bandwidth_group, &bw);
  num_edges = count_edges (bandwidth_group);

  adjust_direction_goals (bandwidth_group, &bw, num_edges, TNT_TRANSFER_DIRECTION_IN);
  adjust_direction_goals (bandwidth_group, &bw, num_edges, TNT_TRANSFER_DIRECTION_OUT);

  maybe_trigger_action (bandwidth_group, &bw, TNT_TRANSFER_DIRECTION_IN);
  maybe_trigger_action (bandwidth_group, &bw, TNT_TRANSFER_DIRECTION_OUT);
}

void
tnt_bandwidth_group_finalize (FluxObject *object)
{
  TntBandwidthGroup *bandwidth_group = TNT_BANDWIDTH_GROUP (object);
  GList             *l;

  assert_types (bandwidth_group);

  for (l = bandwidth_group->edge_list; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;

    flux_object_unref (FLUX_OBJECT (edge));
  }

  g_list_free (bandwidth_group->edge_list);
  if (bandwidth_group->name)
    g_free (bandwidth_group->name);
}

void
tnt_bandwidth_group_init (TntBandwidthGroup *bandwidth_group)
{
  flux_object_init              (FLUX_OBJECT (bandwidth_group));
  flux_object_set_type          (FLUX_OBJECT (bandwidth_group), TNT_BANDWIDTH_GROUP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (bandwidth_group), tnt_bandwidth_group_finalize);
}

TntBandwidthGroup *
tnt_bandwidth_group_new (void)
{
  TntBandwidthGroup *bandwidth_group;

  bandwidth_group = g_new0 (TntBandwidthGroup, 1);
  tnt_bandwidth_group_init (bandwidth_group);

  return bandwidth_group;
}

void
tnt_bandwidth_group_iterate (TntBandwidthGroup *bandwidth_group)
{
  assert_types (bandwidth_group);

  adjust_io_goals (bandwidth_group);

#if 0
  debug_log_bandwidth_stats (bandwidth_group);

  if (!(tnt_timestamp () % 240))
    debug_log_routing_table (bandwidth_group);
#endif
}

const gchar *
tnt_bandwidth_group_get_name (TntBandwidthGroup *bandwidth_group)
{
  assert_types (bandwidth_group);

  return bandwidth_group->name;
}

void
tnt_bandwidth_group_set_name (TntBandwidthGroup *bandwidth_group, const gchar *name)
{
  assert_types (bandwidth_group);
  g_return_if_fail (name != NULL);

  if (bandwidth_group->name)
    g_free (bandwidth_group->name);

  bandwidth_group->name = g_strdup (name);

  if (bandwidth_group->update_func)
    bandwidth_group->update_func (bandwidth_group, bandwidth_group->update_func_data);
}

void
tnt_bandwidth_group_get_accept (TntBandwidthGroup *bandwidth_group, gchar **ip, gchar **netmask)
{
  assert_types (bandwidth_group);

  if (ip)
    *ip      = tnt_ip_to_string (bandwidth_group->ip);
  if (netmask)
    *netmask = tnt_ip_to_string (bandwidth_group->netmask);
}

void
tnt_bandwidth_group_set_accept (TntBandwidthGroup *bandwidth_group, const gchar *ip, const gchar *netmask)
{
  assert_types (bandwidth_group);

  if (ip)
    bandwidth_group->ip = tnt_string_to_ip (ip);

  if (netmask)
    bandwidth_group->netmask = tnt_string_to_ip (netmask);

  if (bandwidth_group->update_func)
    bandwidth_group->update_func (bandwidth_group, bandwidth_group->update_func_data);
}

gint
tnt_bandwidth_group_get_goal (TntBandwidthGroup *bandwidth_group, TntTransferDirection dir)
{
  assert_types (bandwidth_group);

  return bandwidth_group->goal_cps [dir];
}

void
tnt_bandwidth_group_set_goal (TntBandwidthGroup *bandwidth_group, TntTransferDirection dir, gint cps)
{
  assert_types (bandwidth_group);
  g_return_if_fail (cps >= 0);

  bandwidth_group->goal_cps [dir] = cps;

  if (bandwidth_group->update_func)
    bandwidth_group->update_func (bandwidth_group, bandwidth_group->update_func_data);
}

gint
tnt_bandwidth_group_get_cps (TntBandwidthGroup *bandwidth_group, TntTransferDirection dir)
{
  TntBandwidth bw;

  assert_types (bandwidth_group);

  collect_bandwidth_stats (bandwidth_group, &bw);
  return bw.cps [TNT_STREAM_TYPE_CONTROL] [dir] +
         bw.cps [TNT_STREAM_TYPE_DATA]    [dir];
}

static gboolean
accept_ip (TntBandwidthGroup *bandwidth_group, TntNodeIP4 test_ip)
{
  DEBUG
  ({
    gchar *test_ip_str;
    gchar *group_ip_str;
    gchar *netmask_str;

    test_ip_str  = tnt_ip_to_string (test_ip);
    group_ip_str = tnt_ip_to_string (bandwidth_group->ip);
    netmask_str  = tnt_ip_to_string (bandwidth_group->netmask);
    g_print ("*** BWGroup %s: Comparing %s against %s (%s).\n",
           bandwidth_group->name, test_ip_str, group_ip_str, netmask_str);
    g_free (netmask_str);
    g_free (group_ip_str);
    g_free (test_ip_str);
  })

  if ((test_ip             & bandwidth_group->netmask) ==
      (bandwidth_group->ip & bandwidth_group->netmask))
  {
    DEBUG (g_print ("*** BWGroup %s: Matched.\n", bandwidth_group->name));
    return TRUE;
  }

  DEBUG (g_print ("*** BWGroup %s: No match.\n", bandwidth_group->name));
  return FALSE;
}

static gboolean
have_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  if (g_list_find (bandwidth_group->edge_list, edge))
    return TRUE;

  return FALSE;
}

static gboolean
accept_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  return accept_ip (bandwidth_group, tnt_edge_get_peer_ip4 (edge));
}

static gboolean
add_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  if (!have_edge (bandwidth_group, edge))
  {
    bandwidth_group->edge_list = g_list_prepend (bandwidth_group->edge_list, edge);
    flux_object_ref (FLUX_OBJECT (edge));
    return TRUE;
  }

  return FALSE;
}

gboolean
tnt_bandwidth_group_add_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  assert_types (bandwidth_group);
  g_return_val_if_fail (edge != NULL, FALSE);

  if (accept_edge (bandwidth_group, edge))
  {
    return add_edge (bandwidth_group, edge);
  }

  return FALSE;
}

gboolean
tnt_bandwidth_group_remove_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  GList *l;

  assert_types (bandwidth_group);
  g_return_val_if_fail (edge != NULL, FALSE);

  l = g_list_find (bandwidth_group->edge_list, edge);
  if (l)
  {
    flux_object_unref (FLUX_OBJECT (edge));
    bandwidth_group->edge_list = g_list_delete_link (bandwidth_group->edge_list, l);
    return TRUE;
  }

  return FALSE;
}

gboolean
tnt_bandwidth_group_have_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  assert_types (bandwidth_group);
  g_return_val_if_fail (edge != NULL, FALSE);

  return have_edge (bandwidth_group, edge);
}

gboolean
tnt_bandwidth_group_match_ip (TntBandwidthGroup *bandwidth_group, TntNodeIP4 ip)
{
  assert_types (bandwidth_group);

  return accept_ip (bandwidth_group, ip);
}

gboolean
tnt_bandwidth_group_match_edge (TntBandwidthGroup *bandwidth_group, TntEdge *edge)
{
  assert_types (bandwidth_group);
  g_return_val_if_fail (edge != NULL, FALSE);

  return accept_edge (bandwidth_group, edge);
}

void
tnt_bandwidth_group_set_action_func (TntBandwidthGroup *bandwidth_group, TntBandwidthActionFunc *func,
                                     gpointer data)
{
  assert_types (bandwidth_group);

  bandwidth_group->action_func      = func;
  bandwidth_group->action_func_data = data;
}

void
tnt_bandwidth_group_set_update_func (TntBandwidthGroup *bandwidth_group, TntBandwidthGroupFunc *func,
                                     gpointer data)
{
  assert_types (bandwidth_group);

  bandwidth_group->update_func      = func;
  bandwidth_group->update_func_data = data;
}

