/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-block-db.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <glib.h>
#include "tnt.h"

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
# include <direct.h>
#endif

#define DEBUG(x)

#define RANDOM_CULL_PROBES_MAX 64

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_BLOCK_DB_TYPE)

static void begin_state    (TntBlockDB *block_db, TntBlockDBState state);
static void end_state      (TntBlockDB *block_db);

/* Warning: This frees the path it's given */
static gchar *
full_path_to (TntBlockDB *block_db, gchar *path)
{
  gchar       *full_path;
  const gchar *prefix;

  prefix = tnt_storage_get_path (TNT_DATA_DB (block_db)->private_storage);
  full_path = g_build_path (G_DIR_SEPARATOR_S, prefix, path, NULL);

  g_free (path);
  return full_path;
}

static gboolean
ensure_file_path (TntBlockDB *block_db, TntFileID *file_id)
{
  gchar    *file_path;
  gboolean  result;

  file_path = full_path_to (block_db, tnt_file_id_to_string (file_id));
#ifndef G_PLATFORM_WIN32
  if (mkdir (file_path, S_IRWXU | S_IRWXG) == 0)
#else
  if (mkdir (file_path) == 0)
#endif
    result = TRUE;
  else
    result = FALSE;

  g_free (file_path);
  return result;
}

static TntFileBitmap *
get_bitmap_for_file (TntBlockDB *block_db, const TntFileID *file_id)
{
  TntFileBitmap *file_bitmap;
  gchar         *file_path;
  gchar         *bitmap_path;

  file_path = full_path_to (block_db, tnt_file_id_to_string (file_id));
  bitmap_path = g_build_path (G_DIR_SEPARATOR_S, file_path, TNT_BITMAP_FILE_NAME, NULL);
  g_free (file_path);

  file_bitmap = tnt_file_bitmap_load_from_file (bitmap_path, TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len));
  g_free (bitmap_path);
  return file_bitmap;
}

static void
put_bitmap_for_file (TntBlockDB *block_db, const TntFileID *file_id, TntFileBitmap *file_bitmap)
{
  gchar        *file_path;
  gchar        *bitmap_path;

  file_path = full_path_to (block_db, tnt_file_id_to_string (file_id));
  bitmap_path = g_build_path (G_DIR_SEPARATOR_S, file_path, TNT_BITMAP_FILE_NAME, NULL);
  g_free (file_path);

  tnt_file_bitmap_save_to_file (file_bitmap, bitmap_path);
  g_free (bitmap_path);
}

static void
assert_tallies (TntBlockDB *block_db)
{
  g_assert (block_db->file_count >= 0);
  g_assert (block_db->file_complete_count >= 0);
  g_assert (block_db->file_size >= 0);
  g_assert (block_db->file_complete_size >= 0);

  g_assert (block_db->file_count >= block_db->file_complete_count);
  g_assert (block_db->file_size >= block_db->file_complete_size);
}

static void
remove_file_overhead (TntBlockDB *block_db, const TntFileID *file_id)
{
  gchar *path;
  gint   i;

  /* Remove on-disk stuff */

  path = full_path_to (block_db, tnt_file_id_to_string (file_id));
  tnt_dir_delete (path);
  g_free (path);

  /* Remove from in-memory array */

  for (i = 0; i < block_db->known_files->len; i++)
  {
    TntFileID *this_id = &g_array_index (block_db->known_files, TntFileID, i);

    if (!memcmp (file_id, this_id, sizeof (TntFileID)))
    {
      g_array_remove_index_fast (block_db->known_files, i);
      break;
    }
  }
}

static const TntFileID *
pick_random_file_for_cull (TntBlockDB *block_db)
{
  const TntFileID *picked_id = NULL;
  gint             i;

  if (!block_db->known_files->len)
    return NULL;

  /* We don't want to copy the known_files array (so we could remove retrieval IDs
   * from it and select from the resulting set), as it can be several megabytes
   * for a database of 100GB. So, first try to pick an ID by doing random stabs and
   * asserting our luck. */

  for (i = 0; i < RANDOM_CULL_PROBES_MAX; i++)
  {
    gint          n         = g_random_int_range (0, block_db->known_files->len);
    TntFileID    *file_id   = &g_array_index (block_db->known_files, TntFileID, n);
    TntRetrieval *retrieval = NULL;
    gint          retrieval_n;

    retrieval_n = tnt_retrieval_list_find_retrieval (block_db->retrieval_list, file_id);
    if (retrieval_n >= 0)
      retrieval = tnt_retrieval_list_get_retrieval_n (block_db->retrieval_list, retrieval_n);

    /* We need a retrieval list if we're going to avoid deleting parts of ongoing
     * retrievals. */
    if (!block_db->retrieval_list || !retrieval || !tnt_retrieval_is_active (retrieval))
    {
      picked_id = file_id;
      break;
    }
  }

#if 0
  if (picked_id || !block_db->retrieval_list)
    return picked_id;

  /* We couldn't find a file that is not on the retrieval list; delete a part of the
   * lowest-priority retrieval we can find. Never delete parts of the highest-priority
   * retrieval, in case that's the only file we know about between known_files
   * and retrieval_list, and that file is bigger than our limit (we ignore the limit
   * in this case). */

  n = tnt_retrieval_list_get_n_incomplete_retrievals (block_db->retrieval_list);
  for (i = tnt_retrieval_list_get_n_retrievals (block_db->retrieval_list) - 1; i >= 0; i--)
  {
    TntRetrieval  *retrieval = tnt_retrieval_list_get_retrieval_n (block_db->retrieval_list, i);
    TntFileBitmap *bitmap    = tnt_retrieval_get_local_bitmap (retrieval);

    if (tnt_retrieval_is_active (retrieval))
    {
      if (!tnt_retrieval_is_complete (retrieval))
      {
        n--;
        if (n && !tnt_file_bitmap_is_empty (bitmap))
        {
          picked_id = tnt_retrieval_get_file_id (retrieval);
          break;
        }
      }
    }
    else if (!tnt_file_bitmap_is_empty (bitmap))
    {
      picked_id = tnt_retrieval_get_file_id (retrieval);
      break;
    }
  }
#endif

  return picked_id;
}

/* FIXME: If we retrieved a huge file and busted our limit good, we'll wind up blocking
 * while we remove blocks to get back below the limit. */
static void
cull_db (TntBlockDB *block_db)
{
  while (tnt_storage_get_free (TNT_DATA_DB (block_db)->private_storage) < TNT_DATA_BLOCK_SIZE * 2)
  {
    const TntFileID *file_id;
    TntFileBitmap   *file_bitmap;
    gchar           *block_path;
    gint             n;

    /* Get a file and its bitmap */

    file_id = pick_random_file_for_cull (block_db);
    if (!file_id)
    {
      DEBUG (g_print ("Could not find a file to cull.\n"));
      break;
    }

    file_bitmap = get_bitmap_for_file (block_db, file_id);
    if (!file_bitmap)
    {
      /* Shouldn't happen, but let's guard against it */
      remove_file_overhead (block_db, file_id);
      DEBUG (g_print ("Could not get bitmap for cull file.\n"));
      break;
    }

    /* Get a random block in the file */

    n = tnt_file_bitmap_count_set_bits (file_bitmap);
    if (!n)
    {
      /* Shouldn't happen, but let's guard against it */
      remove_file_overhead (block_db, file_id);
      flux_object_unref (FLUX_OBJECT (file_bitmap));
      DEBUG (g_print ("No set bits in cull bitmap.\n"));
      break;
    }

    n = g_random_int_range (0, n);
    n = tnt_file_bitmap_get_set_bit_n (file_bitmap, n);

    /* Remove that block */

    block_path = full_path_to (block_db, tnt_file_block_id_to_string (file_id, n));
    DEBUG (g_print ("Culling %s\n", block_path));
    remove (block_path);  /* FIXME: Error checking. */
    g_free (block_path);

    /* Update bitmaps, bookkeeping. Maybe clean up. */

    block_db->file_size -= TNT_BLOCK_IN_FILE_LEN (n, file_id->len);
    tnt_storage_set_used (TNT_DATA_DB (block_db)->private_storage, block_db->file_size);

    if (tnt_file_bitmap_is_complete (file_bitmap))
    {
      block_db->file_complete_count--;
      block_db->file_complete_size -= file_id->len;
    }

    tnt_file_bitmap_set_bit (file_bitmap, n, FALSE);

    if (block_db->retrieval_list &&
        (n = tnt_retrieval_list_find_retrieval (block_db->retrieval_list, file_id)) >= 0)
    {
      TntRetrieval *retrieval = tnt_retrieval_list_get_retrieval_n (block_db->retrieval_list, n);
      tnt_retrieval_set_local_bitmap (retrieval, file_bitmap);
    }

    if (tnt_file_bitmap_is_empty (file_bitmap))
    {
      /* WARNING: file_id may point into the known_files array, which may be
       * modified by this call. Don't rely on it after doing this. */
      remove_file_overhead (block_db, file_id);
      block_db->file_count--;
    }
    else
      put_bitmap_for_file (block_db, file_id, file_bitmap);

    assert_tallies (block_db);
    flux_object_unref (FLUX_OBJECT (file_bitmap));
  }
}

static gboolean
scan_step (TntBlockDB *block_db)
{
  TntStat        tstat;
  struct dirent *dent;
  const gchar   *top_path;
  gchar         *file_path;

  top_path = tnt_storage_get_path (TNT_DATA_DB (block_db)->private_storage);

  if (!block_db->top_dir_hd)
  {
    block_db->top_dir_hd = opendir (top_path);
  }

  if (!block_db->top_dir_hd)
  {
    /* This should not happen if we have a valid top-level path. The storage
     * should take care of it for us. */
    g_warning ("BlockDB could not open top-level path.");
    return TRUE;
  }

  dent = readdir (block_db->top_dir_hd);
  if (!dent)
  {
    /* No more file entries. */
    closedir (block_db->top_dir_hd);
    block_db->top_dir_hd = NULL;
    return TRUE;
  }

  if (!strcmp ("..", dent->d_name) || !strcmp (".", dent->d_name))
    return FALSE;

  if (!tnt_is_valid_file_id_string (dent->d_name))
    return FALSE;

  file_path = g_build_path (G_DIR_SEPARATOR_S, top_path, dent->d_name, NULL);

  if (!tnt_lstat (file_path, &tstat))
  {
    /* This won't happen a lot */
    g_free (file_path);
    return FALSE;
  }

  if (tstat.is_dir && tnt_dir_have_full_access (file_path))
  {
    gchar         *bitmap_path;
    guint          bitmap_len;
    TntFileID      file_id;
    TntFileBitmap *file_bitmap;
    guint64        data_size;

    /* Found a file's block directory */

    if (!tnt_string_to_file_id (dent->d_name, &file_id) || !file_id.len)
    {
      /* Invalid file ID in directory name */
      g_free (file_path);
      return FALSE;
    }

    bitmap_path = g_build_path (G_DIR_SEPARATOR_S, file_path, TNT_BITMAP_FILE_NAME, NULL);
    bitmap_len  = TNT_REAL_LEN_TO_BITMAP_LEN (file_id.len);

    file_bitmap = tnt_file_bitmap_load_from_file (bitmap_path, bitmap_len);
    g_free (bitmap_path);

    if (!file_bitmap)
    {
      remove_file_overhead (block_db, &file_id);  /* Remove bogus directory */
      g_free (file_path);
      return FALSE;
    }

    /* This shouldn't happen, but we need to guard against it */
    if (tnt_file_bitmap_is_empty (file_bitmap))
    {
      remove_file_overhead (block_db, &file_id);  /* Remove bogus directory */
      flux_object_unref (FLUX_OBJECT (file_bitmap));
      g_free (file_path);
      return FALSE;
    }

    /* We have the bitmap */

    data_size = tnt_file_bitmap_calc_data_size (file_bitmap, file_id.len);
    g_assert (data_size > 0);

    block_db->file_count++;
    block_db->file_size += data_size;

    if (tnt_file_bitmap_is_complete (file_bitmap))
    {
      g_assert (data_size == file_id.len);
      block_db->file_complete_count++;
      block_db->file_complete_size += data_size;
    }

    assert_tallies (block_db);
    g_array_append_val (block_db->known_files, file_id);
    flux_object_unref (FLUX_OBJECT (file_bitmap));
  }

  g_free (file_path);
  return FALSE;
}

/* For scanning and finding the files/blocks we have */
static gboolean
scan_cb (gpointer data)
{
  TntBlockDB *block_db = data;

  assert_types (block_db);

  if (scan_step (block_db))
  {
    end_state (block_db);
  }

  return TRUE;
}

/* --- File hashing --- */


/* --- State machine --- */

/* Callback from TntBlockAction */
static void
end_block_action (TntDataDB *data_db, gboolean success)
{
  TntBlockDB     *block_db = TNT_BLOCK_DB (data_db);
  TntBlockAction *block_action;

  assert_types (block_db);

  block_action = block_db->current_block_action;
  g_assert (block_action);

  /* Update bitmap if block written */

  if (success && tnt_block_action_get_type (block_action) == TNT_BLOCK_ACTION_PUT_BLOCK)
  {
    TntFileBitmap *file_bitmap;
    TntFileID      file_id;
    gint           block_id;
    gboolean       old_bit_state;
    gboolean       new_file = FALSE;

    /* FIXME: We should really check what the TntBlockAction says, e.g.
     * error value and such. For instance, it might've run out of disk space. */

    tnt_block_action_get_file_info (block_action, &file_id);
    block_id = tnt_block_action_get_block_id (block_action);

    file_bitmap = get_bitmap_for_file (block_db, &file_id);
    if (!file_bitmap)
    {
      file_bitmap = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (file_id.len));
      tnt_file_bitmap_fill_false (file_bitmap);
    }

    if (tnt_file_bitmap_is_empty (file_bitmap))
      new_file = TRUE;

    old_bit_state = tnt_file_bitmap_get_bit (file_bitmap, block_id);
    tnt_file_bitmap_set_bit (file_bitmap, block_id, TRUE);
    put_bitmap_for_file (block_db, &file_id, file_bitmap);

    if (!old_bit_state)
    {
      if (new_file)
      {
        g_array_append_val (block_db->known_files, file_id);
        block_db->file_count++;
      }

      block_db->file_size += TNT_BLOCK_IN_FILE_LEN (block_id, file_id.len);
      tnt_storage_set_used (TNT_DATA_DB (block_db)->private_storage, block_db->file_size);

      if (tnt_file_bitmap_is_complete (file_bitmap))
      {
        block_db->file_complete_count++;
        block_db->file_complete_size += file_id.len;
      }
    }

    flux_object_unref (FLUX_OBJECT (file_bitmap));
  }

  /* Dequeue */

  flux_object_unref (FLUX_OBJECT (block_action));
  block_db->current_block_action = NULL;

  end_state (block_db);
}

static void
begin_block_action (TntBlockDB *block_db, TntBlockAction *block_action)
{
  TntFileID   file_id;
  gint        block_id;
  gchar      *block_path;
  FluxFileIO *file_io;
  gint        fd;

  g_assert (block_action);
  g_assert (block_db->current_block_action == NULL);

  block_db->current_block_action = block_action;

  /* Create and position a FluxFileIO for tnt_block_action_run (). */

  switch (tnt_block_action_get_type (block_action))
  {
    case TNT_BLOCK_ACTION_GET_BLOCK_HASH:
    case TNT_BLOCK_ACTION_GET_BLOCK:
      tnt_block_action_get_file_info (block_action, &file_id);
      block_id = tnt_block_action_get_block_id (block_action);

      block_path = full_path_to (block_db, tnt_file_block_id_to_string (&file_id, block_id));

      fd = tnt_open (block_path, FALSE, FALSE, FALSE);
      g_free (block_path);
      if (fd < 0)
      {
        tnt_block_action_cancel (block_action, TNT_DATA_DB (block_db));
        return;
      }

      file_io = flux_file_io_new (fd);
      tnt_block_action_run (block_action, TNT_DATA_DB (block_db), file_io);
      flux_object_unref (FLUX_OBJECT (file_io));
      break;

    case TNT_BLOCK_ACTION_PUT_BLOCK:
      tnt_block_action_get_file_info (block_action, &file_id);
      block_id = tnt_block_action_get_block_id (block_action);

      if (tnt_storage_get_free (TNT_DATA_DB (block_db)->private_storage) < TNT_DATA_BLOCK_SIZE * 2)
        cull_db (block_db);

      ensure_file_path (block_db, &file_id);
      block_path = full_path_to (block_db, tnt_file_block_id_to_string (&file_id, block_id));

      fd = tnt_open (block_path, TRUE, TRUE, FALSE);
      g_free (block_path);
      if (fd < 0)
      {
        tnt_block_action_cancel (block_action, TNT_DATA_DB (block_db));
        return;
      }

      file_io = flux_file_io_new (fd);
      tnt_block_action_run (block_action, TNT_DATA_DB (block_db), file_io);
      flux_object_unref (FLUX_OBJECT (file_io));
      break;

    default:
      g_assert_not_reached ();
      break;
  }
}

/* Called on transition to new state */
static void
begin_state (TntBlockDB *block_db, TntBlockDBState state)
{
  TntBlockAction *action;

  block_db->state = state;

  switch (state)
  { 
    case TNT_BLOCK_DB_IDLE:
      DEBUG (g_print ("Begin idle.\n"));
      break;

    case TNT_BLOCK_DB_SCANNING:
      DEBUG (g_print ("Begin scan.\n"));
      block_db->callback_source_id = g_idle_add_full (G_PRIORITY_LOW + 4, scan_cb, block_db, NULL);
      break;

    case TNT_BLOCK_DB_ACTION:
      DEBUG (g_print ("Begin action.\n"));
      g_assert (block_db->block_action_queue);

      action = block_db->block_action_queue->data;
      block_db->block_action_queue = g_list_delete_link (block_db->block_action_queue,
                                                         block_db->block_action_queue);

      begin_block_action (block_db, action);
      break;

    default:
      g_assert_not_reached ();
      break;
  }
}

/* Called when a state ends. Decides what to do next, and performs the
 * transition to the appropriate next state. */
static void
end_state (TntBlockDB *block_db)
{
  TntBlockDBState next_state;

  /* Determine next state */

  if (block_db->block_action_queue)
    next_state = TNT_BLOCK_DB_ACTION;
  else
    next_state = TNT_BLOCK_DB_IDLE;

  /* If new state, end current state */

  if (next_state != block_db->state || next_state == TNT_BLOCK_DB_ACTION)
  {
    switch (block_db->state)
    {
      case TNT_BLOCK_DB_IDLE:
        break;

      case TNT_BLOCK_DB_SCANNING:
        g_source_remove (block_db->callback_source_id);
        block_db->callback_source_id = 0;
        tnt_storage_set_used (TNT_DATA_DB (block_db)->private_storage, block_db->file_size);
        break;

      case TNT_BLOCK_DB_ACTION:
        break;

      default:
        g_assert_not_reached ();
        break;
    }

    begin_state (block_db, next_state);
  }
}

/* --- *
 * API *
 * --- */

static gboolean
have_file_full (TntDataDB *block_db, const TntFileID *file_id)
{
  TntFileBitmap *file_bitmap;
  gboolean       result = FALSE;

  assert_types (block_db);

  file_bitmap = get_bitmap_for_file (TNT_BLOCK_DB (block_db), file_id);
  if (!file_bitmap)
    return FALSE;

  if (tnt_file_bitmap_is_complete (file_bitmap))
    result = TRUE;

  flux_object_unref (FLUX_OBJECT (file_bitmap));
  return result;
}

static gboolean
have_file_part (TntDataDB *block_db, const TntFileID *file_id)
{
  TntFileBitmap *file_bitmap;
  gboolean       result = FALSE;

  assert_types (block_db);

  file_bitmap = get_bitmap_for_file (TNT_BLOCK_DB (block_db), file_id);
  if (!file_bitmap)
    return FALSE;

  if (tnt_file_bitmap_count_set_bits (file_bitmap) > 0)
    result = TRUE;

  flux_object_unref (FLUX_OBJECT (file_bitmap));
  return result;
}

static gboolean
have_block (TntDataDB *block_db, const TntFileID *file_id, gint block_id)
{
  TntFileBitmap *file_bitmap;
  gboolean       result;

  assert_types (block_db);

  if ((guint64) block_id * (guint64) TNT_DATA_BLOCK_SIZE >= file_id->len)
    return FALSE;

  file_bitmap = get_bitmap_for_file (TNT_BLOCK_DB (block_db), file_id);
  if (!file_bitmap)
    return FALSE;

  result = tnt_file_bitmap_get_bit (file_bitmap, block_id);
  flux_object_unref (FLUX_OBJECT (file_bitmap));
  return result;
}

static TntFileBitmap *
get_file_bitmap (TntDataDB *block_db, const TntFileID *file_id)
{
  TntFileBitmap *file_bitmap;

  assert_types (block_db);

  file_bitmap = get_bitmap_for_file (TNT_BLOCK_DB (block_db), file_id);
  return file_bitmap;
}

static void
queue_block_action (TntBlockDB *block_db, TntPriority priority, TntBlockAction *block_action)
{
  if (priority < TNT_PRIORITY_HIGH)
    block_db->block_action_queue = g_list_append (block_db->block_action_queue, block_action);
  else
    block_db->block_action_queue = g_list_prepend (block_db->block_action_queue, block_action);

  if (block_db->state == TNT_BLOCK_DB_IDLE)
  {
    end_state (block_db);
  }
}

static gboolean
get_block_hash (TntDataDB *block_db, const TntFileID *file_id, gint block_id,
                TntPriority priority, TntGotBlockHashFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  assert_types (block_db);

  if (!have_block (block_db, file_id, block_id))
    return FALSE;

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_GET_BLOCK_HASH);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block_len (block_action, TNT_BLOCK_IN_FILE_LEN (block_id, file_id->len));
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_BLOCK_DB (block_db), priority, block_action);
  return TRUE;
}

static gboolean
get_block (TntDataDB *block_db, const TntFileID *file_id, gint block_id,
           TntPriority priority, TntGotBlockFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  assert_types (block_db);

  if (!have_block (block_db, file_id, block_id))
    return FALSE;

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_GET_BLOCK);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block_len (block_action, TNT_BLOCK_IN_FILE_LEN (block_id, file_id->len));
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_BLOCK_DB (block_db), priority, block_action);
  return TRUE;
}

static gboolean
put_block (TntDataDB *block_db, const TntFileID *file_id, gint block_id,
           gpointer block, guint block_len, TntPriority priority,
           TntWroteBlockFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  assert_types (block_db);

  if (have_block (block_db, file_id, block_id))
  {
    tnt_free_block (block);
    return TRUE;
  }

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_PUT_BLOCK);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block     (block_action, block);
  tnt_block_action_set_block_len (block_action, block_len);
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_BLOCK_DB (block_db), priority, block_action);
  return TRUE;
}

static void
cancel_action (TntDataDB *data_db, gpointer func, gpointer data)
{
  TntBlockDB *block_db = TNT_BLOCK_DB (data_db);
  GList      *l;

  assert_types (block_db);

  for (l = block_db->block_action_queue; l; )
  {
    TntBlockAction *action = l->data;
    gpointer action_func;
    gpointer action_data;

    tnt_block_action_get_post_func (action, &action_func, &action_data);

    if (action_func == func && action_data == data)
    {
      GList *unlink_l = l;

      l = g_list_next (l);
      block_db->block_action_queue = g_list_delete_link (block_db->block_action_queue, unlink_l);
      tnt_block_action_set_post_func (action, NULL, NULL);
      tnt_block_action_cancel (action, NULL);
      flux_object_unref (FLUX_OBJECT (action));
    }
    else
      l = g_list_next (l);
  }

  if (block_db->current_block_action)
  {
    gpointer action_func;
    gpointer action_data;

    tnt_block_action_get_post_func (block_db->current_block_action, &action_func, &action_data);

    if (action_func == func && action_data == data)
    {
      tnt_block_action_set_post_func (block_db->current_block_action, NULL, NULL);
      tnt_block_action_cancel (block_db->current_block_action, data_db);
      /* end_block_action () unrefs */
    }
  }
}

static gboolean
pick_random_file (TntDataDB *data_db, TntFileID *file_id)
{
  TntBlockDB *block_db = TNT_BLOCK_DB (data_db);
  gint        n;

  assert_types (block_db);

  if (block_db->known_files->len == 0)
    return FALSE;

  n = g_random_int_range (0, block_db->known_files->len);
  memcpy (file_id, &g_array_index (block_db->known_files, TntFileID, n), sizeof (TntFileID));
  return TRUE;
}

gboolean
tnt_block_db_init (TntBlockDB *block_db, TntStorage *private_storage)
{
  TntDataDB *data_db = TNT_DATA_DB (block_db);

  g_return_val_if_fail (private_storage != NULL, FALSE);

  tnt_data_db_init              (TNT_DATA_DB (block_db), private_storage);
  flux_object_set_type          (FLUX_OBJECT (block_db), TNT_BLOCK_DB_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (block_db), tnt_block_db_finalize);

  block_db->known_files = g_array_new (FALSE, FALSE, sizeof (TntFileID));

  data_db->have_file_full_func   = have_file_full;
  data_db->have_file_part_func   = have_file_part;
  data_db->have_block_func       = have_block;
  data_db->get_file_bitmap_func  = get_file_bitmap;
  data_db->get_block_hash_func   = get_block_hash;
  data_db->get_block_func        = get_block;
  data_db->put_block_func        = put_block;
  data_db->cancel_func           = cancel_action;
  data_db->pick_random_file_func = pick_random_file;
  data_db->action_done_func      = end_block_action;

  begin_state (block_db, TNT_BLOCK_DB_SCANNING);

  return TRUE;
}

TntBlockDB *
tnt_block_db_new (TntStorage *private_storage)
{
  TntBlockDB *block_db;

  block_db = g_new0 (TntBlockDB, 1);
  if (!tnt_block_db_init (block_db, private_storage))
  {
    g_free (block_db);
    return NULL;
  }

  return block_db;
}

void
tnt_block_db_finalize (FluxObject *object)
{
  TntBlockDB *block_db = TNT_BLOCK_DB (object);
  GList      *l;

  assert_types (block_db);

  if (block_db->callback_source_id)
    g_source_remove (block_db->callback_source_id);

  if (block_db->retrieval_list)
    flux_object_unref (FLUX_OBJECT (block_db->retrieval_list));

  g_array_free (TNT_BLOCK_DB (block_db)->known_files, TRUE);

  for (l = block_db->block_action_queue; l; l = g_list_next (l))
  {
    TntBlockAction *action = l->data;
    flux_object_unref (FLUX_OBJECT (action));
  }

  g_list_free (block_db->block_action_queue);

  if (block_db->current_block_action)
    flux_object_unref (FLUX_OBJECT (block_db->current_block_action));

  if (block_db->top_dir_hd)
  {
    closedir (block_db->top_dir_hd);
    block_db->top_dir_hd = NULL;
  }

  tnt_data_db_finalize (object);
}

guint
tnt_block_db_get_file_count (TntBlockDB *block_db, gboolean include_incomplete)
{
  assert_types (block_db);

  if (include_incomplete)
    return block_db->file_count;

  return block_db->file_complete_count;
}

guint64
tnt_block_db_get_file_size (TntBlockDB *block_db, gboolean include_incomplete)
{
  assert_types (block_db);

  if (include_incomplete)
    return block_db->file_size;

  return block_db->file_complete_size;
}

void
tnt_block_db_set_retrieval_list (TntBlockDB *block_db, TntRetrievalList *retrieval_list)
{
  assert_types (block_db);

  if (retrieval_list)
  {
    flux_object_check (FLUX_OBJECT (retrieval_list), TNT_RETRIEVAL_LIST_TYPE);
    flux_object_ref (FLUX_OBJECT (retrieval_list));
  }

  if (block_db->retrieval_list)
    flux_object_unref (FLUX_OBJECT (block_db->retrieval_list));

  block_db->retrieval_list = retrieval_list;
}
