/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-core-registry.c - Keeps track of initialized cores.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include <stdio.h>
#include <stdlib.h>
#include "tnt.h"

typedef struct
{
  gpointer         core;
  TntCallbackList *listeners;
}
CoreEntry;

static GHashTable *core_table = NULL;

static void
free_core_entry (CoreEntry *entry)
{
  flux_object_unref (FLUX_OBJECT (entry->listeners));
  g_free (entry);
}

static void
ensure_core_table (void)
{
  if (!core_table)
    core_table = g_hash_table_new_full (g_str_hash, g_str_equal,
                                        g_free, (GDestroyNotify) free_core_entry);
}

gpointer
tnt_core_registry_get_core (const gchar *core_id)
{
  CoreEntry *entry;

  ensure_core_table ();

  entry = g_hash_table_lookup (core_table, core_id);
  if (!entry)
    return NULL;

  return entry->core;
}

void
tnt_core_registry_register_core (const gchar *core_id, gpointer core)
{
  CoreEntry *entry;

  ensure_core_table ();

  entry = g_hash_table_lookup (core_table, core_id);
  if (!entry)
  {
    entry = g_new0 (CoreEntry, 1);
    entry->listeners = tnt_callback_list_new ();

    g_hash_table_insert (core_table, g_strdup (core_id), entry);
  }

  if (entry->core)
  {
    if (entry->core == core)
      g_warning ("Cannot register '%s' core twice.", core_id);
    else
      g_warning ("Cannot register multiple '%s' cores.", core_id);

    return;
  }

  entry->core = core;
  tnt_callback_list_run_while_2 (entry->listeners, tnt_test_always_true, core_id, core);
}

void
tnt_core_registry_unregister_core (const gchar *core_id, gpointer core)
{
  CoreEntry *entry;

  ensure_core_table ();

  entry = g_hash_table_lookup (core_table, core_id);

  if (!entry || !entry->core)
  {
    g_warning ("Cannot unregister unregistered '%s' core.", core_id);
    return;
  }

  entry->core = NULL;
  tnt_callback_list_run_while_2 (entry->listeners, tnt_test_always_true, core_id, NULL);
}

void
tnt_core_registry_attach_listener (const gchar *core_id, TntCoreListenerFunc *func, gpointer data)
{
  CoreEntry *entry;

  ensure_core_table ();

  entry = g_hash_table_lookup (core_table, core_id);
  if (!entry)
  {
    entry = g_new0 (CoreEntry, 1);
    entry->listeners = tnt_callback_list_new ();

    g_hash_table_insert (core_table, g_strdup (core_id), entry);
  }

  tnt_callback_list_add_callback (entry->listeners, func, data);
}

void
tnt_core_registry_detach_listener (const gchar *core_id, TntCoreListenerFunc *func, gpointer data)
{
  CoreEntry *entry;

  ensure_core_table ();

  entry = g_hash_table_lookup (core_table, core_id);
  if (!entry)
  {
    g_warning ("Cannot detach listener from unknown '%s' core.", core_id);
    return;
  }

  tnt_callback_list_remove_callback (entry->listeners, func, data);
}
