/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-disseminator.c - Maintains a queue of blocks to send.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include "tnt.h"
#include <string.h>

#define QUEUE_LEN_MAX 100

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_DISSEMINATOR_TYPE)

typedef struct
{
  TntNodeID node_id;
  TntFileID file_id;
  gint      block_id;
}
Request;

typedef struct
{
  TntNodeID node_id;
}
BlockedNode;

static GList *
find_request_list_by_node_id (GList *queue, TntNodeID node_id)
{
  GList *l;

  for (l = queue; l; l = g_list_next (l))
  {
    Request *request = l->data;

    if (request->node_id == node_id)
      return l;
  }

  return NULL;
}

static GList *
find_block_list_by_node_id (GList *list, TntNodeID node_id)
{
  GList *l;

  for (l = list; l; l = g_list_next (l))
  {
    BlockedNode *blocked_node = l->data;

    if (blocked_node->node_id == node_id)
      return l;
  }

  return NULL;
}

gint
tnt_disseminator_get_queue_len (TntDisseminator *disseminator)
{
  assert_types (disseminator);

  return disseminator->queue_len;
}

gboolean
tnt_disseminator_enqueue (TntDisseminator *disseminator, TntNodeID node_id,
                          const TntFileID *file_id, gint block_id)
{
  GList   *old_request_list;
  Request *request;

  assert_types (disseminator);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, FALSE);
  g_return_val_if_fail (file_id != NULL, FALSE);

  old_request_list = find_request_list_by_node_id (disseminator->queue, node_id);

  if (old_request_list)
  {
    request = old_request_list->data;
  }
  else
  {
    if (disseminator->queue_len >= QUEUE_LEN_MAX)
      return FALSE;

    request = g_new (Request, 1);

    disseminator->queue = g_list_prepend (disseminator->queue, request);
    if (!disseminator->last)
      disseminator->last = disseminator->queue;

    disseminator->queue_len++;
  }

  request->node_id = node_id;
  memcpy (&request->file_id, file_id, sizeof (TntFileID));
  request->block_id = block_id;

  return TRUE;
}

gboolean
tnt_disseminator_peek (TntDisseminator *disseminator, TntNodeID *node_id,
                       TntFileID *file_id, gint *block_id)
{
  Request *request;
  GList   *last;

  assert_types (disseminator);

  if (!disseminator->queue_len)
    return FALSE;

  /* Find the last request whose node id is not being blocked */
  for (last = disseminator->last; last; last = g_list_previous (last))
  {
    Request *r = last->data;
    if (!find_block_list_by_node_id (disseminator->blocked_nodes, r->node_id))
      break;
  }

  if (!last)
    return FALSE;

  request = last->data;
  if (node_id)
    *node_id = request->node_id;
  if (file_id)
    memcpy (file_id, &request->file_id, sizeof (TntFileID));
  if (block_id)
    *block_id = request->block_id;

  return TRUE;
}

gboolean
tnt_disseminator_peek_nth (TntDisseminator *disseminator, gint n,
                           TntNodeID *node_id, TntFileID *file_id, gint *block_id)
{
  Request *request;
  GList   *last;

  assert_types (disseminator);
  g_return_val_if_fail (n > -1, FALSE);

  if (!disseminator->queue_len)
    return FALSE;

  /* Find the nth request, skipping over requests whose node ids are being blocked */
  for (last = disseminator->last; last != NULL && n > -1; last = g_list_previous (last))
  {
    Request *r = last->data;

    if (find_block_list_by_node_id (disseminator->blocked_nodes, r->node_id))
      continue;

    n--;
  }

  if (!last)
    return FALSE;

  request = last->data;
  if (node_id)
    *node_id = request->node_id;
  if (file_id)
    memcpy (file_id, &request->file_id, sizeof (TntFileID));
  if (block_id)
    *block_id = request->block_id;

  return TRUE;
}

gboolean
tnt_disseminator_dequeue (TntDisseminator *disseminator, TntNodeID *node_id,
                          TntFileID *file_id, gint *block_id)
{
  Request *request;
  GList   *last;

  assert_types (disseminator);

  if (!disseminator->queue_len)
    return FALSE;

  /* Find the last request whose node id is not being blocked */
  for (last = disseminator->last; last; last = g_list_previous (last))
  {
    Request *r = last->data;
    if (!find_block_list_by_node_id (disseminator->blocked_nodes, r->node_id))
      break;
  }

  if (!last)
    return FALSE;

  request = last->data;
  if (node_id)
    *node_id = request->node_id;
  if (file_id)
    memcpy (file_id, &request->file_id, sizeof (TntFileID));
  if (block_id)
    *block_id = request->block_id;
  g_free (request);

  if (last == disseminator->last)
    disseminator->last = g_list_previous (last);
  disseminator->queue = g_list_delete_link (disseminator->queue, last);

  disseminator->queue_len--;
  return TRUE;
}

gboolean
tnt_disseminator_dequeue_nth (TntDisseminator *disseminator, gint n,
                              TntNodeID *node_id, TntFileID *file_id, gint *block_id)
{
  Request *request;
  GList   *last;

  assert_types (disseminator);
  g_return_val_if_fail (n > -1, FALSE);

  if (!disseminator->queue_len)
    return FALSE;

  /* Find the last request whose node id is not being blocked */
  for (last = disseminator->last; last != NULL && n > -1; last = g_list_previous (last))
  {
    Request *r = last->data;

    if (find_block_list_by_node_id (disseminator->blocked_nodes, r->node_id))
      continue;

    n--;
  }

  if (!last)
    return FALSE;

  request = last->data;
  if (node_id)
    *node_id = request->node_id;
  if (file_id)
    memcpy (file_id, &request->file_id, sizeof (TntFileID));
  if (block_id)
    *block_id = request->block_id;
  g_free (request);

  if (last == disseminator->last)
    disseminator->last = g_list_previous (last);
  disseminator->queue = g_list_delete_link (disseminator->queue, last);

  disseminator->queue_len--;
  return TRUE;
}

gboolean
tnt_disseminator_find_node (TntDisseminator *disseminator, TntNodeID node_id,
                            TntFileID *file_id, gint *block_id)
{
  GList   *request_list;
  Request *request;

  assert_types (disseminator);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, FALSE);

  request_list = find_request_list_by_node_id (disseminator->queue, node_id);
  if (!request_list)
    return FALSE;

  request = request_list->data;

  if (file_id)
    memcpy (file_id, &request->file_id, sizeof (TntFileID));
  if (block_id)
    *block_id = request->block_id;

  return TRUE;
}

void
tnt_disseminator_block_node (TntDisseminator *disseminator, TntNodeID node_id)
{
  BlockedNode *blocked_node;

  assert_types (disseminator);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  if (find_block_list_by_node_id (disseminator->blocked_nodes, node_id))
    return;

  blocked_node = g_new (BlockedNode, 1);
  blocked_node->node_id = node_id;
  disseminator->blocked_nodes = g_list_prepend (disseminator->blocked_nodes, blocked_node);
}

void
tnt_disseminator_unblock_node (TntDisseminator *disseminator, TntNodeID node_id)
{
  GList *l;

  assert_types (disseminator);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  l = find_block_list_by_node_id (disseminator->blocked_nodes, node_id);
  if (!l)
    return;

  g_free (l->data);
  disseminator->blocked_nodes = g_list_delete_link (disseminator->blocked_nodes, l);
}

void
tnt_disseminator_finalize (FluxObject *object)
{
  TntDisseminator *disseminator = TNT_DISSEMINATOR (object);
  GList           *l;

  assert_types (disseminator);

  for (l = disseminator->queue; l; l = g_list_next (l))
  {
    Request *request = l->data;
    g_free (request);
  }

  g_list_free (disseminator->queue);
}

void
tnt_disseminator_init (TntDisseminator *disseminator)
{
  flux_object_init              (FLUX_OBJECT (disseminator));
  flux_object_set_type          (FLUX_OBJECT (disseminator), TNT_DISSEMINATOR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (disseminator), tnt_disseminator_finalize);
}

TntDisseminator *
tnt_disseminator_new (void)
{
  TntDisseminator *disseminator;

  disseminator = g_new0 (TntDisseminator, 1);
  tnt_disseminator_init (disseminator);

  return disseminator;
}
