/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-assembler.c - Assembles files from blocks to a serial file.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

typedef struct
{
  TntFileAssemblyStatus  status;

  TntFileID              file_id;

  TntFileAssembler      *assembler;

  gchar                 *file_path;
  gchar                 *file_name;

  TntAssemblyFunc       *func;
  gpointer               func_data;
}
Assembly;

static void next_assembly (TntFileAssembler *assembler);
static void got_block_cb  (TntDataDB *data_db, TntFileID *file_id, gint block_id,
                           guint block_len, gpointer block, gpointer data);

static void
assembly_done (Assembly *assembly, TntFileAssemblyStatus new_status)
{
  TntFileAssembler *assembler = assembly->assembler;

  assembly->status = new_status;

  if (new_status == TNT_FILE_ASSEMBLY_COMPLETE)
  {
    tnt_file_db_end_serialization (assembler->dest_db, &assembly->file_id);
  }
  else
  {
    tnt_file_db_abort_serialization (assembler->dest_db, &assembly->file_id);
  }

  assembler->current_assembly = NULL;

  flux_object_ref (FLUX_OBJECT (assembler));

  if (assembly->func)
    assembly->func (assembler, &assembly->file_id, assembly->func_data);

  if (!assembler->current_assembly)
  {
    next_assembly (assembler);
  }

  flux_object_unref (FLUX_OBJECT (assembler));
}

static void
wrote_block_cb (TntDataDB *data_db, TntFileID *file_id, gint block_id,
                guint block_len, gpointer data)
{
  Assembly *assembly = data;

  g_assert (assembly != NULL);
  flux_object_check (FLUX_OBJECT (data_db), TNT_DATA_DB_TYPE);
  g_assert (file_id != NULL);

  if (block_len == 0)
  {
    /* Couldn't write the block! */
    assembly_done (assembly, TNT_FILE_ASSEMBLY_ERROR_OTHER_TARGET);
    return;
  }

  block_id++;
  if (block_id == TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len))
  {
    /* We've transferred all blocks */
    assembly_done (assembly, TNT_FILE_ASSEMBLY_COMPLETE);
    return;
  }

  if (!tnt_data_db_get_block (assembly->assembler->source_db, file_id, block_id, TNT_PRIORITY_HIGH,
                              got_block_cb, assembly))
    assembly_done (assembly, TNT_FILE_ASSEMBLY_ERROR_OTHER_SOURCE);
}

static void
got_block_cb (TntDataDB *data_db, TntFileID *file_id, gint block_id,
              guint block_len, gpointer block, gpointer data)
{
  Assembly *assembly = data;

  g_assert (assembly != NULL);
  flux_object_check (FLUX_OBJECT (data_db), TNT_DATA_DB_TYPE);
  g_assert (file_id != NULL);

  if (block == NULL || block_len == 0)
  {
    /* Couldn't read the block! */
    assembly_done (assembly, TNT_FILE_ASSEMBLY_ERROR_OTHER_SOURCE);
    return;
  }

  if (!tnt_data_db_put_block (TNT_DATA_DB (assembly->assembler->dest_db),
                              file_id, block_id, block, block_len,
                              TNT_PRIORITY_HIGH, wrote_block_cb, assembly))
    assembly_done (assembly, TNT_FILE_ASSEMBLY_ERROR_OTHER_TARGET);
}

static void
foreach_assembly_start_one (Assembly *assembly, gpointer dummy, TntFileAssembler *assembler)
{
  if (assembler->current_assembly)
    return;

  if (assembly->status != TNT_FILE_ASSEMBLY_WAITING)
    return;

  if (!tnt_data_db_have_file_full (assembler->source_db, &assembly->file_id))
  {
    assembly->status = TNT_FILE_ASSEMBLY_ERROR_INCOMPLETE;
    if (assembly->func)
      assembly->func (assembler, &assembly->file_id, assembly->func_data);
    return;
  }

  if (!tnt_file_db_begin_serialization (assembler->dest_db, &assembly->file_id,
                                        assembly->file_path, assembly->file_name))
  {
    assembly->status = TNT_FILE_ASSEMBLY_ERROR_OTHER;
    if (assembly->func)
      assembly->func (assembler, &assembly->file_id, assembly->func_data);
    return;
  }

  assembly->status = TNT_FILE_ASSEMBLY_RUNNING;
  assembler->current_assembly = assembly;

  if (tnt_data_db_get_block (assembler->source_db, &assembly->file_id, 0, TNT_PRIORITY_HIGH,
                             got_block_cb, assembly))
  {
    if (assembly->func)
      assembly->func (assembler, &assembly->file_id, assembly->func_data);
  }
  else
  {
    assembly_done (assembly, TNT_FILE_ASSEMBLY_ERROR_OTHER_SOURCE);
  }
}

static void
next_assembly (TntFileAssembler *assembler)
{
  g_assert (assembler->current_assembly == NULL);

  flux_object_ref (FLUX_OBJECT (assembler));
  g_hash_table_foreach (assembler->assembly_table, (GHFunc) foreach_assembly_start_one, assembler);
  flux_object_unref (FLUX_OBJECT (assembler));
}

gboolean
tnt_file_assembler_add_assembly (TntFileAssembler *assembler,
                                 const TntFileID *file_id,
                                 const gchar *file_path, const gchar *file_name,
                                 TntAssemblyFunc *func, gpointer data)
{
  Assembly *assembly;

  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (file_name != NULL, FALSE);
  g_return_val_if_fail (file_path != NULL, FALSE);

  assembly = g_new0 (Assembly, 1);
  memcpy (&assembly->file_id, file_id, sizeof (TntFileID));

  if (g_hash_table_lookup (assembler->assembly_table, assembly))
  {
    g_warning ("Attempted to add duplicate assembly.");
    g_free (assembly);
    return FALSE;
  }

  assembly->status      = TNT_FILE_ASSEMBLY_WAITING;
  assembly->assembler   = assembler;
  assembly->file_path   = g_strdup (file_path);
  assembly->file_name   = g_strdup (file_name);
  assembly->func        = func;
  assembly->func_data   = data;

  g_hash_table_insert (assembler->assembly_table, assembly, assembly);

  if (!assembler->current_assembly)
    next_assembly (assembler);

  return TRUE;
}

gboolean
tnt_file_assembler_remove_assembly (TntFileAssembler *assembler, const TntFileID *file_id)
{
  Assembly  cmp_assembly;
  Assembly *assembly;
  gboolean  result;

  g_return_val_if_fail (file_id != NULL, FALSE);

  memcpy (&cmp_assembly.file_id, file_id, sizeof (TntFileID));
  assembly = g_hash_table_lookup (assembler->assembly_table, &cmp_assembly);
  if (!assembly)
    return FALSE;

  if (assembly->status == TNT_FILE_ASSEMBLY_RUNNING)
  {
    tnt_data_db_cancel (TNT_DATA_DB (assembler->source_db), got_block_cb, assembly);
    tnt_data_db_cancel (TNT_DATA_DB (assembler->dest_db), wrote_block_cb, assembly);
  }

  result = g_hash_table_remove (assembler->assembly_table, assembly);

  if (assembler->current_assembly == assembly)
  {
    assembler->current_assembly = NULL;
    next_assembly (assembler);
  }

  return result;
}

TntFileAssemblyStatus
tnt_file_assembler_get_assembly_status (TntFileAssembler *assembler, const TntFileID *file_id)
{
  Assembly  cmp_assembly;
  Assembly *assembly;

  g_return_val_if_fail (file_id != NULL, FALSE);

  memcpy (&cmp_assembly.file_id, file_id, sizeof (TntFileID));
  assembly = g_hash_table_lookup (assembler->assembly_table, &cmp_assembly);
  if (!assembly)
    return TNT_FILE_ASSEMBLY_UNKNOWN;

  return assembly->status;
}

static guint
assembly_hash (const Assembly *assembly)
{
  guint32 val = 0;
  gint    i;

  for (i = 0; i < sizeof (assembly->file_id.hash); i++)
  {
    guint32 t;

    val ^= ((const guchar *) assembly->file_id.hash) [i];
    t = val & 0xf0000000;
    val <<= 4;
    val |= t >> 28;
  }

  for (i = 0; i < sizeof (assembly->file_id.len); i++)
  {
    guint32 t;

    val ^= ((const guchar *) &assembly->file_id.len) [i];
    t = val & 0xf0000000;
    val <<= 4;
    val |= t >> 28;
  }

  return val;
}

static gboolean
assembly_equal (const Assembly *assembly_1, const Assembly *assembly_2)
{
  if (!memcmp (&assembly_1->file_id, &assembly_2->file_id, sizeof (TntFileID)))
    return TRUE;

  return FALSE;
}

static void
assembly_free (Assembly *assembly)
{
  g_free (assembly->file_name);
  g_free (assembly->file_path);

  g_free (assembly);
}

TntFileAssembler *
tnt_file_assembler_new (TntDataDB *source_db, TntFileDB *dest_db)
{
  TntFileAssembler *assembler;

  assembler = g_new0 (TntFileAssembler, 1);
  if (!tnt_file_assembler_init (assembler, source_db, dest_db))
  {
    g_free (assembler);
    assembler = NULL;
  }

  return assembler;
}

gboolean
tnt_file_assembler_init (TntFileAssembler *assembler, TntDataDB *source_db, TntFileDB *dest_db)
{
  g_return_val_if_fail (source_db != NULL, FALSE);
  g_return_val_if_fail (dest_db != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (assembler));
  flux_object_set_type          (FLUX_OBJECT (assembler), TNT_FILE_ASSEMBLER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (assembler), tnt_file_assembler_finalize);

  flux_object_ref (FLUX_OBJECT (source_db));
  flux_object_ref (FLUX_OBJECT (dest_db));

  assembler->source_db = source_db;
  assembler->dest_db   = dest_db;

  assembler->assembly_table = g_hash_table_new_full ((GHashFunc) assembly_hash,
                                                     (GEqualFunc) assembly_equal,
                                                     NULL, (GDestroyNotify) assembly_free);

  return TRUE;
}

static void
foreach_assembly_cancel (Assembly *assembly, gpointer dummy, TntFileAssembler *assembler)
{
  if (assembly->status == TNT_FILE_ASSEMBLY_RUNNING)
  {
    tnt_data_db_cancel (TNT_DATA_DB (assembler->source_db), got_block_cb, assembly);
    tnt_data_db_cancel (TNT_DATA_DB (assembler->dest_db), wrote_block_cb, assembly);
  }
}

void
tnt_file_assembler_finalize (FluxObject *object)
{
  TntFileAssembler *assembler = TNT_FILE_ASSEMBLER (object);

  g_hash_table_foreach (assembler->assembly_table, (GHFunc) foreach_assembly_cancel, assembler);
  g_hash_table_destroy (assembler->assembly_table);
  flux_object_unref (FLUX_OBJECT (assembler->source_db));
  flux_object_unref (FLUX_OBJECT (assembler->dest_db));
}
