/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-core.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tnt.h"
#include "tnt-file-messages.h"

#define DEBUG(x)
#define DEBUG_DISSEMINATION(x)

#define REPEAT_QUERY_TIMEOUT        (1 * SECONDS_IN_MINUTE)
#define ASK_AVAIL_TIMEOUT           (3 * SECONDS_IN_MINUTE)
#define ASK_BLOCK_TIMEOUT           (5)
#define SUMMARIZE_TIMEOUT           (1)

#define REMOTE_QUERIES_MAX             100
#define RUNNING_DISSEMINATIONS_NOMINAL 4
#define ALLOC_BLOCKS_UNQUALIFIED_MAX   16
#define ALLOC_BLOCKS_QUALIFIED_MAX     64
#define QUEUE_LENGTH_PAIN_THRESHOLD    1000
#define QUEUE_LENGTH_REST_THRESHOLD    2000

/* These scores pertain to messages referring to files we are getting */
#define EDGE_SCORE_HAVE_FILE               1.0
#define EDGE_SCORE_HAVE_FILE_BITMAP        0.5
#define EDGE_SCORE_ON_QUEUE                4.0
#define EDGE_SCORE_OFF_QUEUE               4.0
#define EDGE_SCORE_DATA_BLOCK             10.0

/* Other score modifiers */
#define EDGE_SCORE_UNREQUESTED_DATA_BLOCK  0.5
#define EDGE_SCORE_QUERY                  -0.01

/* Number of matches to buffer for a remote query */
#define REMOTE_QUERY_BUFFERED_MATCHES_LOW  64
#define REMOTE_QUERY_BUFFERED_MATCHES_HIGH 256

/* Maximum runtime for a remote query, in seconds */
#define REMOTE_QUERY_DURATION_MAX          30

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_FILE_CORE_TYPE)

typedef struct
{
  TntNodeAddress4  address;
  TntEdge         *edge;
  TntTcpIO        *tcp_io;
  guint            outbound : 1;
}
PendingEdge;

/* --- Absolutely necessary forward declarations --- */

static gboolean          disseminate_next           (TntFileCore *file_core, TntBandwidthGroup *group);

static TntNodeID
my_node_id (TntFileCore *file_core)
{
  if (!file_core->network_core)
    return TNT_NODE_ID_INVALID;

  return tnt_network_core_get_node_id (file_core->network_core);
}

static guint16
generate_query_id (TntFileCore *file_core)
{
  guint16 id;

  id = file_core->next_local_query_id++;

  if (file_core->next_local_query_id == 65535)
    file_core->next_local_query_id = 0;

  return id;
}

static gboolean
get_incoming_file_id (TntEdge *edge, TntFileID *file_id)
{
  TntMessageType  type;
  TntTcpIO       *data_tcp_io;
  gpointer        payload, payload_copy;
  guint16         payload_size;
  TntMessage     *message;
  TntDataBlockM  *data_block_m;

  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);
  g_return_val_if_fail (file_id != NULL, FALSE);

  data_tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA);
  if (!data_tcp_io)
    return FALSE;

  if (!tnt_tcp_io_get_read_info (data_tcp_io, &type, &payload, &payload_size))
    return FALSE;

  if (type != TNT_MESSAGE_DATA_BLOCK)
    return FALSE;

  /* Set up message, assign payload, demarshal it, free message */

  payload_copy = g_malloc (payload_size);
  memcpy (payload_copy, payload, payload_size);
  message = tnt_message_new (type);
  tnt_message_set_payload (message, payload_copy, payload_size);
  data_block_m = tnt_message_get_struct (message);
  flux_object_unref (FLUX_OBJECT (message));

  /* Copy file ID from message struct, free message struct */

  if (!data_block_m)
    return FALSE;

  memcpy (file_id, &data_block_m->file_id, sizeof (TntFileID));
  g_free (data_block_m);
  return TRUE;
}

static guint
get_cps_in_for_file_id (TntFileCore *file_core, const TntFileID *file_id)
{
  const GPtrArray *edge_array;
  guint            cps = 0;
  gint             n;

  if (!file_core->network_core)
    return 0;

  edge_array = tnt_network_core_peek_edge_array (file_core->network_core);
  if (!edge_array)
    return 0;

  for (n = 0; n < edge_array->len; n++)
  {
    TntEdge   *edge = g_ptr_array_index (edge_array, n);
    TntFileID  edge_file_id;

    if (!get_incoming_file_id (edge, &edge_file_id))
      continue;

    if (!memcmp (file_id, &edge_file_id, sizeof (TntFileID)))
    {
      TntTcpIO *tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA);

      if (!tcp_io)
        continue;

      cps += tnt_tcp_io_get_cps (tcp_io, TNT_TRANSFER_DIRECTION_IN);
    }
  }

  return cps;
}

/* ---------------- *
 * Query Processing *
 * ---------------- */

static void
try_match_file (TntFileCore *file_core, const TntFileID *file_id, TntFileMeta *file_meta,
                TntNodeID node_id, gint remote_blocks)
{
  GList *l;

  for (l = file_core->local_queries; l; l = g_list_next (l))
  {
    TntQuery *query;

    query = l->data;
    tnt_query_try_match (query, file_id, file_meta, node_id, remote_blocks);
  }
}

static void
try_match_file_remote (TntFileCore *file_core, TntFileID *file_id, TntFileMeta *file_meta)
{
  GList *l;

  for (l = file_core->remote_queries; l; l = g_list_next (l))
  {
    TntQuery *query;

    query = l->data;
    tnt_query_try_match (query, file_id, file_meta, TNT_NODE_ID_INVALID, 0);
  }
}

static void
broadcast_local_query (TntFileCore *file_core, TntLocalQuery *query, TntEdge *focus_edge)
{
  TntQueryM   query_info;
  TntMessage *message;

  if (!file_core->network_core)
    return;

  query_info.id     = tnt_query_get_id (TNT_QUERY (query));
  query_info.string = (gchar *) tnt_query_get_string (TNT_QUERY (query));
  query_info.flags  = 0;

  if (tnt_query_get_match_partial (TNT_QUERY (query)))
    query_info.flags |= TNT_QUERY_M_MATCH_PARTIAL;

  message = TNT_MESSAGE (tnt_message_construct_routed (TNT_MESSAGE_QUERY, &query_info, NULL, 0, 10,
                                                       my_node_id (file_core), TNT_NODE_ID_INVALID));

  if (focus_edge)
    tnt_edge_send_message (focus_edge, message);
  else
    tnt_network_core_route_message (file_core->network_core, message);

  flux_object_unref (FLUX_OBJECT (message));
}

static void
broadcast_local_query_cancel (TntFileCore *file_core, TntLocalQuery *query)
{
  TntQueryCancelM  query_cancel_info;
  TntMessage      *message;

  if (!file_core->network_core)
    return;

  query_cancel_info.id = tnt_query_get_id (TNT_QUERY (query));

  message = TNT_MESSAGE (tnt_message_construct_routed (TNT_MESSAGE_QUERY_CANCEL, &query_cancel_info, NULL, 0, 10,
                                                       my_node_id (file_core), TNT_NODE_ID_INVALID));
  tnt_network_core_route_message (file_core->network_core, message);
  flux_object_unref (FLUX_OBJECT (message));
}

static void
local_query_match_pre_add_cb (TntLocalQuery *query, TntMatch *match, gpointer data)
{
  TntFileCore *file_core = data;

  assert_types (file_core);
  flux_object_check (FLUX_OBJECT (query), TNT_LOCAL_QUERY_TYPE);

  if (!tnt_match_is_file (match))
    return;

  if (file_core->file_db &&
      tnt_data_db_have_file_full (TNT_DATA_DB (file_core->file_db), tnt_match_get_file_id (match)))
  {
    tnt_match_set_local (match, TRUE);
  }
  else
  {
    tnt_match_set_local (match, FALSE);
  }
}

static gboolean
repeat_queries_cb (gpointer data)
{
  TntFileCore *file_core = data;
  GList       *l;

  for (l = file_core->local_queries; l; l = g_list_next (l))
  {
    TntLocalQuery *query = l->data;

    broadcast_local_query (file_core, query, NULL);
  }

  return TRUE;
}

static void
local_file_set_in_queries (TntFileCore *file_core, TntFileID *file_id, gboolean is_local)
{
  GList *l;

  for (l = file_core->local_queries; l; l = g_list_next (l))
  {
    TntLocalQuery *query;
    TntMatch      *match;

    query = l->data;

    match = tnt_local_query_find_match_by_file_id (query, file_id);
    if (match == NULL)
      continue;

    if (is_local)
      tnt_match_set_local (match, TRUE);
    else
      tnt_match_set_local (match, FALSE);

    tnt_local_query_match_updated (query, match);
  }
}

void
tnt_file_core_add_local_query (TntFileCore *file_core, TntLocalQuery *query)
{
  if (!g_list_find (file_core->local_queries, query))
  {
    flux_object_ref (FLUX_OBJECT (query));
    tnt_query_set_id (TNT_QUERY (query), generate_query_id (file_core));
    tnt_local_query_add_match_pre_add_func (query, local_query_match_pre_add_cb, file_core);
    file_core->local_queries = g_list_prepend (file_core->local_queries, query);

    if (file_core->meta_db)
      tnt_meta_db_query (file_core->meta_db, TNT_QUERY (query), TRUE, NULL, NULL);

    broadcast_local_query (file_core, query, NULL);
  }
}

void
tnt_file_core_remove_local_query (TntFileCore *file_core, TntLocalQuery *query)
{
  GList *node;

  node = g_list_find (file_core->local_queries, query);
  if (!node)
    return;

  tnt_local_query_remove_match_func (query, local_query_match_pre_add_cb, file_core);
  broadcast_local_query_cancel (file_core, query);

  if (file_core->meta_db)
    tnt_meta_db_cancel_query (file_core->meta_db, TNT_QUERY (query));

  flux_object_unref (FLUX_OBJECT (query));
  file_core->local_queries = g_list_delete_link (file_core->local_queries, node);
}

typedef struct
{
  TntFileCore    *file_core;
  TntRemoteQuery *remote_query;
  TntTimeStamp    start_stamp;
  gboolean        query_done;

  guint           callback_source;
}
RemoteQueryContext;

static void
remove_remote_query (TntFileCore *file_core, TntNodeID node_id, guint16 query_id)
{
  GList *l;

  for (l = file_core->remote_queries; l; l = g_list_next (l))
  {
    TntRemoteQuery *remote_query = l->data;

    if (tnt_remote_query_get_node_id (remote_query) == node_id &&
        tnt_query_get_id (TNT_QUERY (remote_query)) == query_id)
    {
      break;
    }
  }

  if (l)
  {
    TntRemoteQuery *remote_query = l->data;

    file_core->remote_queries = g_list_delete_link (file_core->remote_queries, l);
    flux_object_unref (FLUX_OBJECT (remote_query));  /* Remote query list reference */

    file_core->n_remote_queries--;
  }
}

static void
destroy_remote_query_context (RemoteQueryContext *context)
{
  if (context->file_core->meta_db)
    tnt_meta_db_cancel_query (context->file_core->meta_db, TNT_QUERY (context->remote_query));

  flux_object_unref (FLUX_OBJECT (context->remote_query));

  if (context->callback_source != 0)
    g_source_remove (context->callback_source);
}

static gint
get_queue_avail_for_message_type (TntEdge *edge, TntMessageType type)
{
  TntTcpIO *tcp_io;

  tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL);
  if (!tcp_io)
    return 0;

  return tnt_tcp_io_get_queue_out_max_len (tcp_io, type) -
         tnt_tcp_io_get_queue_out_len     (tcp_io, type);
}

static gint
get_queue_avail_for_send_to (TntNetworkCore *network_core, TntMessageType type, TntNodeID node_id)
{
  TntEdge *edge;

  edge = tnt_network_core_find_route (network_core, node_id);
  if (!edge)
    return -1;

  return get_queue_avail_for_message_type (edge, type);
}

static gboolean
send_remote_matches (gpointer data)
{
  RemoteQueryContext *context  = data;
  gboolean            hold_off = FALSE;
  gboolean            finished = FALSE;
  gint                i;

  if (!g_list_find (context->file_core->remote_queries, context->remote_query))
  {
    /* remove_remote_query() was called on this remote query */
    tnt_context_destroy (context);
    return FALSE;
  }

  /* See if query should be paused or restarted, to maintain optimal buffer fill */

  if (context->file_core->meta_db)
  {
    gint n_matches = tnt_query_get_n_matches (TNT_QUERY (context->remote_query));

    if (n_matches < REMOTE_QUERY_BUFFERED_MATCHES_LOW)
    {
      tnt_meta_db_query_set_paused (context->file_core->meta_db,
                                    TNT_QUERY (context->remote_query),
                                    FALSE);
    }
    else if (n_matches > REMOTE_QUERY_BUFFERED_MATCHES_HIGH)
    {
      tnt_meta_db_query_set_paused (context->file_core->meta_db,
                                    TNT_QUERY (context->remote_query),
                                    TRUE);
    }
  }

  /* Dispatch some matches */

  for (i = 0; i < 10; i++)
  {
    TntRoutedMessage  *rmessage;
    gboolean           have_match;
    TntFileID          file_id;
    TntFileMeta       *file_meta;
    TntFileInfoM       file_info;
    TntFileBitmap     *file_bitmap [2] = { NULL, NULL };
    gint               bitmap_len;

    if (context->file_core->network_core &&
        get_queue_avail_for_send_to (context->file_core->network_core, TNT_MESSAGE_FILE_INFO,
                                     tnt_remote_query_get_node_id (context->remote_query)) == 0)
    {
      hold_off = TRUE;
      DEBUG (g_print ("No queue space: Holdoff\n"));
      break;
    }

    have_match = tnt_remote_query_get_next_match (context->remote_query, &file_id, &file_meta);
    if (!have_match)
    {
      /* No more matches at the moment */

      if (context->query_done)
        finished = TRUE;
      else
        hold_off = TRUE;

      break;
    }

    memcpy (&file_info.id,   &file_id,  sizeof (TntFileID));
    memcpy (&file_info.meta, file_meta, sizeof (TntFileMeta));

    if (context->file_core->file_db)
      file_bitmap [0] = tnt_data_db_get_file_bitmap (TNT_DATA_DB (context->file_core->file_db), &file_id);
    if (context->file_core->block_db)
      file_bitmap [1] = tnt_data_db_get_file_bitmap (TNT_DATA_DB (context->file_core->block_db), &file_id);

    if (file_bitmap [0] && file_bitmap [1])
    {
      tnt_file_bitmap_or (file_bitmap [0], file_bitmap [1]);
      flux_object_unref (FLUX_OBJECT (file_bitmap [1]));
    }
    else if (!file_bitmap [0])
    {
      file_bitmap [0] = file_bitmap [1];
    }

    bitmap_len = TNT_REAL_LEN_TO_BITMAP_LEN (file_id.len);

    if (file_bitmap [0])
    {
      file_info.local_block_count = tnt_file_bitmap_count_set_bits (file_bitmap [0]);
      flux_object_unref (FLUX_OBJECT (file_bitmap [0]));
    }
    else
    {
      file_info.local_block_count = 0;
    }

    /* Screen for local availability */
    if (!file_bitmap [0] || file_info.local_block_count < bitmap_len)
    {
      if (!tnt_query_get_match_partial (TNT_QUERY (context->remote_query)))
      {
        tnt_file_meta_free (file_meta);
        continue;
      }
    }

    DEBUG (g_print ("Send match: %s %s\n", tnt_file_meta_get_best_title (&file_info.meta),
           context->pass == 0 ? "(avail)" : ""));

    if (context->file_core->network_core)
    {
      rmessage = tnt_message_construct_routed (TNT_MESSAGE_FILE_INFO, &file_info, NULL, 0, 4,
                                               my_node_id (context->file_core),
                                               tnt_remote_query_get_node_id (context->remote_query));

      tnt_network_core_route_message (context->file_core->network_core, TNT_MESSAGE (rmessage));

      flux_object_unref (FLUX_OBJECT (rmessage));
    }

    tnt_file_meta_free (file_meta);
  }

  if (finished || tnt_timestamp () - context->start_stamp > REMOTE_QUERY_DURATION_MAX)
  {
    DEBUG (g_print ("Query removed.\n"));
    /* Remove this query from list. Next callback will free context and disable callback. */
    remove_remote_query (context->file_core, tnt_remote_query_get_node_id (context->remote_query),
                         tnt_query_get_id (TNT_QUERY (context->remote_query)));
  }

  /* Hook up new callback source */
  if (hold_off)
  {
    context->callback_source = g_timeout_add (250, send_remote_matches, context);
  }
  else
  {
    context->callback_source = g_idle_add (send_remote_matches, context);
  }

  return FALSE;  /* Remove old callback source */
}

static void
remote_query_done (TntRemoteQuery *remote_query, RemoteQueryContext *context)
{
  context->query_done = TRUE;
}

static void
add_remote_query (TntFileCore *file_core, TntRemoteQuery *remote_query)
{
  RemoteQueryContext *context;

  if (!file_core->meta_db)
    return;

  flux_object_ref (FLUX_OBJECT (remote_query));  /* Remote query list reference */
  file_core->remote_queries = g_list_prepend (file_core->remote_queries, remote_query);

  context = tnt_context_factory_new_context (file_core->remote_query_context_factory);
  context->file_core       = file_core;
  context->start_stamp     = tnt_timestamp ();
  context->query_done      = FALSE;
  context->remote_query    = remote_query;
  context->callback_source = 0;
  flux_object_ref (FLUX_OBJECT (remote_query));  /* Callback context reference */

  tnt_meta_db_query (file_core->meta_db, TNT_QUERY (remote_query), TRUE,
                     (TntQueryDoneFunc *) remote_query_done, context);

  file_core->n_remote_queries++;

  /* Start recurrent match dispatch */
  send_remote_matches (context);
}

/* -------------------- *
 * Retrieval Processing *
 * -------------------- */

static void
try_set_retrieval_complete (TntFileCore *file_core, TntFileID *file_id)
{
  TntRetrieval  *retrieval;
  TntFileBitmap *file_bitmap;
  gint           n;

  if (!file_core->retrieval_list)
    return;

  n = tnt_retrieval_list_find_retrieval (file_core->retrieval_list, file_id);
  if (n == -1)
    return;

  retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, n);
  file_bitmap = tnt_retrieval_get_local_bitmap (retrieval);
  tnt_file_bitmap_fill_true (file_bitmap);
  tnt_retrieval_list_updated (file_core->retrieval_list, n);
}

static gboolean
maybe_update_retrieval_remote_bitmap (TntFileCore *file_core, TntFileID *file_id, TntNodeID source_id,
                                      TntFileBitmap *file_bitmap)
{
  TntRetrieval *retrieval;
  gint          n;

  if (!file_core->retrieval_list)
    return FALSE;

  n = tnt_retrieval_list_find_retrieval (file_core->retrieval_list, file_id);
  if (n == -1)
    return FALSE;

  retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, n);
  tnt_retrieval_update_remote_bitmap (retrieval, source_id, file_bitmap);
  return TRUE;
}

static void
ask_retrieval_availability (TntFileCore *file_core, TntRetrieval *retrieval, TntEdge *focus_edge)
{
  TntAskFileAvailM  ask_file_avail;
  TntRoutedMessage *rmessage;

#if 0
  /* FIXME: We can only do this if we're sure all blocks are available (and available
   * enough). */
  if (tnt_retrieval_get_source_count (retrieval) > 10)
    continue;
#endif

  memcpy (&ask_file_avail.file_id, tnt_retrieval_get_file_id (retrieval), sizeof (TntFileID));

  rmessage = tnt_message_construct_routed (TNT_MESSAGE_ASK_FILE_AVAIL, &ask_file_avail, NULL, 0, 10,
                                           my_node_id (file_core), TNT_NODE_ID_INVALID);

  if (focus_edge)
    tnt_edge_send_message (focus_edge, TNT_MESSAGE (rmessage));
  else if (file_core->network_core)
    tnt_network_core_route_message (file_core->network_core, TNT_MESSAGE (rmessage));

  flux_object_unref (FLUX_OBJECT (rmessage));
}

static gboolean
ask_retrieval_availability_cb (gpointer data)
{
  TntFileCore  *file_core    = data;
  TntRetrieval *retrieval    = NULL;
  gboolean      keep_handler = TRUE;
  gint          i;

  assert_types (file_core);

  if (!file_core->retrieval_list)
    return TRUE;

  for (i = file_core->next_retrieval_avail_ask;
       i < tnt_retrieval_list_get_n_retrievals (file_core->retrieval_list); i++)
  {
    retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, i);

    if (tnt_retrieval_is_active (retrieval))
      break;
  }

  if (retrieval)
  {
    ask_retrieval_availability (file_core, retrieval, NULL);

    if (file_core->next_retrieval_avail_ask == 0)
    {
      /* 100ms between individual queries */
      file_core->ask_avail_timeout_id =
        g_timeout_add (100, ask_retrieval_availability_cb, file_core);
      keep_handler = FALSE;
    }

    file_core->next_retrieval_avail_ask = i + 1;
  }
  else
  {
    if (file_core->next_retrieval_avail_ask != 0)
    {
      /* ASK_AVAIL_TIMEOUT seconds until start from top of list */
      file_core->ask_avail_timeout_id =
        g_timeout_add (1000 * ASK_AVAIL_TIMEOUT, ask_retrieval_availability_cb, file_core);
      keep_handler = FALSE;
    }

    file_core->next_retrieval_avail_ask = 0;
  }

  return keep_handler;
}

static void
ask_retrieval_block (TntFileCore *file_core)
{
  TntDataBlockRequestM  data_block_request;
  TntNodeID             node_id;
  gint                  block_id;

  if (!file_core->network_core)
    return;

  if (!file_core->retrieval_list)
    return;

  if (tnt_get_blocks_allocated () > ALLOC_BLOCKS_QUALIFIED_MAX)
    return;

  if (tnt_retrieval_list_ask_block (file_core->retrieval_list, &data_block_request.file_id,
                                    &node_id, &block_id))
  {
    TntRoutedMessage *rmessage;

    if (block_id < 0)
    {
      g_warning ("Asking for negative block_id!");
      return;
    }

    data_block_request.block_id = block_id;

    DEBUG
    (
      g_print ("Asking block %d of file sized %lld from node %08llx.\n",
               block_id, data_block_request.file_id.len, node_id);
    )

    rmessage = tnt_message_construct_routed (TNT_MESSAGE_DATA_BLOCK_REQUEST, &data_block_request, NULL, 0, 5,
                                             my_node_id (file_core), node_id);
    tnt_network_core_route_message (file_core->network_core, TNT_MESSAGE (rmessage));
    flux_object_unref (FLUX_OBJECT (rmessage));

    /* We want to hang on to this node. Try to establish a direct connection. */
    tnt_network_core_connect_to_node (file_core->network_core, node_id);
  }
}

static gboolean
ask_retrieval_block_cb (gpointer data)
{
  TntFileCore *file_core = data;

  assert_types (file_core);

  ask_retrieval_block (file_core);
  return TRUE;
}

static gboolean
summarize_cb (gpointer data)
{
  TntFileCore *file_core = data;
  gint           i, n;

  assert_types (file_core);

  if (!file_core->retrieval_list)
    return TRUE;

  n = tnt_retrieval_list_get_n_retrievals (file_core->retrieval_list);
  for (i = 0; i < n; i++)
  {
    TntRetrieval *retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, i);

    if (!tnt_retrieval_is_active (retrieval))
    {
      tnt_retrieval_set_cps_in (retrieval, 0);
      continue;
    }

    tnt_retrieval_set_cps_in (retrieval,
                              get_cps_in_for_file_id (file_core, tnt_retrieval_get_file_id (retrieval)));
  }
  
  return TRUE;
}

static TntFileBitmap *
get_local_file_bitmap (TntFileCore *file_core, TntFileID *file_id)
{
  TntFileBitmap *file_bitmap_a = NULL, *file_bitmap_b = NULL;

  if (file_core->file_db)
    file_bitmap_a = tnt_data_db_get_file_bitmap (TNT_DATA_DB (file_core->file_db),  file_id);
  if (file_core->block_db)
    file_bitmap_b = tnt_data_db_get_file_bitmap (TNT_DATA_DB (file_core->block_db), file_id);

  if (file_bitmap_a)
  {
    if (file_bitmap_b)
    {
      tnt_file_bitmap_or (file_bitmap_a, file_bitmap_b);
      flux_object_unref (FLUX_OBJECT (file_bitmap_b));
    }
  }
  else if (file_bitmap_b)
  {
    file_bitmap_a = file_bitmap_b;
  }
  else
  {
    file_bitmap_a = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len));
  }

  return file_bitmap_a;
}

static void
user_retrieval_added_cb (TntRetrievalList *retrieval_list, gint retrieval_n, gpointer data)
{
  TntFileCore   *file_core = data;
  TntRetrieval  *retrieval;
  TntFileBitmap *file_bitmap;

  assert_types (file_core);
  flux_object_check (FLUX_OBJECT (retrieval_list), TNT_RETRIEVAL_LIST_TYPE);

  retrieval = tnt_retrieval_list_get_retrieval_n (retrieval_list, retrieval_n);
  file_bitmap = get_local_file_bitmap (file_core, (TntFileID *) tnt_retrieval_get_file_id (retrieval));
  tnt_retrieval_set_local_bitmap (retrieval, file_bitmap);
  flux_object_unref (FLUX_OBJECT (file_bitmap));

  if (tnt_retrieval_is_active (retrieval))
    ask_retrieval_availability (file_core, retrieval, NULL);
}

static void
assembly_status_change_cb (TntFileAssembler *assembler, const TntFileID *file_id, gpointer data)
{
  TntFileCore           *file_core = data;
  gint                   retrieval_n;
  TntRetrieval          *retrieval;
  TntFileAssemblyStatus  status;

  assert_types (file_core);
  g_return_if_fail (file_id != NULL);

  status = tnt_file_assembler_get_assembly_status (assembler, file_id);

  retrieval_n = tnt_retrieval_list_find_retrieval (file_core->retrieval_list, file_id);
  if (retrieval_n < 0)
  {
    /* User may have cancelled the retrieval during assembly */
    tnt_file_assembler_remove_assembly (assembler, file_id);
    return;
  }

  retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, retrieval_n);

  if (status == TNT_FILE_ASSEMBLY_COMPLETE)
  {
    tnt_retrieval_set_active (retrieval, FALSE);
    tnt_file_assembler_remove_assembly (assembler, file_id);
  }
  else if (status >= TNT_FILE_ASSEMBLY_FIRST_ERROR &&
           status <= TNT_FILE_ASSEMBLY_LAST_ERROR)
  {
    /* FIXME: Need to handle errors and notify user */

    g_message ("Assembly failed with error %d!", status);
  }
}

static void
user_retrieval_completed_cb (TntRetrievalList *retrieval_list, gint retrieval_n, gpointer data)
{
  TntFileCore     *file_core = data;
  TntRetrieval    *retrieval;
  const TntFileID *file_id;
  TntFileMeta     *file_meta;
  const gchar     *path;
  gchar           *name;

  assert_types (file_core);

  if (!file_core->block_db || !file_core->file_db)
    return;

  retrieval = tnt_retrieval_list_get_retrieval_n (retrieval_list, retrieval_n);
  file_id   = tnt_retrieval_get_file_id (retrieval);
  file_meta = tnt_retrieval_get_file_meta (retrieval);
  path      = tnt_retrieval_get_path (retrieval);
  name      = tnt_file_meta_get_name (file_meta);

  tnt_retrieval_set_completed_func (retrieval, NULL, NULL);

  tnt_file_assembler_add_assembly (file_core->assembler, file_id, path, name,
                                   assembly_status_change_cb, file_core);

  if (name)
    g_free (name);
}

/* ------------------------ *
 * Dissemination Management *
 * ------------------------ */

typedef struct
{
  TntFileCore    *file_core;
  TntTimeStamp    start_stamp;
  TntNodeID       recipient_id;
  TntFileID       file_id;
  gint            block_id;

  gpointer        block_data;
  guint           block_len;

  TntMessage     *message;
  TntEdge        *through_edge;

  guint           from_disseminator : 1;
  guint           done              : 1;
}
DisseminationContext;

static void
dissemination_cleanup (DisseminationContext *context)
{
  if (context->from_disseminator)
  {
    tnt_disseminator_unblock_node (context->file_core->disseminator, context->recipient_id);

    if (context->file_core->network_core)
    {
      TntOffQueueM      off_queue;
      TntRoutedMessage *rmessage;

      memcpy (&off_queue.file_id, &context->file_id, sizeof (TntFileID));
      off_queue.block_id  = context->block_id;
      off_queue.timestamp = tnt_timestamp ();

      rmessage = tnt_message_construct_routed (TNT_MESSAGE_OFF_QUEUE, &off_queue, NULL, 0, 5,
                                               my_node_id (context->file_core), context->recipient_id);

      tnt_network_core_route_message (context->file_core->network_core, TNT_MESSAGE (rmessage));
      flux_object_unref (FLUX_OBJECT (rmessage));
    }
  }

  if (context->from_disseminator && context->through_edge && context->file_core->network_core)
  {
    /* FIXME: This is bad, and the only reason for having tnt_network_core_have_edge () */
    if (tnt_network_core_have_edge (context->file_core->network_core, context->through_edge))
      tnt_edge_qualification_unref (context->through_edge);
  }

  context->done = TRUE;
  tnt_context_destroy (context);
}

/* FIXME: We may get this callback after we're destroyed! */
static void
end_dissemination_send_cb (FluxObject *message, gpointer data)
{
  DisseminationContext *context = data;
  TntFileCore          *file_core;

  g_return_if_fail (context != NULL);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);
  assert_types (context->file_core);

  file_core = context->file_core;

  context->message = NULL;

  if (!context->from_disseminator)
    return;

  file_core->running_disseminations--;
  dissemination_cleanup (context);

  /* Maybe service next block immediately */
  if (file_core->running_disseminations < RUNNING_DISSEMINATIONS_NOMINAL)
    disseminate_next (file_core, NULL);
}

static void
start_dissemination_send (DisseminationContext *context)
{
  TntMessage *message = NULL;

  if (context->block_data && context->file_core->network_core)
  {
    TntDataBlockM data_block_m;

    memcpy (&data_block_m.file_id, &context->file_id, sizeof (TntFileID));
    data_block_m.block_id = context->block_id;

    if (context->recipient_id != TNT_NODE_ID_INVALID)
    {
      /* Have a node ID target; route */

      DEBUG_DISSEMINATION (g_print ("Dissemination start: routed.\n"));

      message = TNT_MESSAGE (
        tnt_message_construct_routed (TNT_MESSAGE_DATA_BLOCK, &data_block_m, context->block_data,
                                      context->block_len, 2, my_node_id (context->file_core),
                                      context->recipient_id));
      context->message = message;
      tnt_message_set_destroy_func (message, end_dissemination_send_cb, context);
      context->through_edge = tnt_network_core_route_message (context->file_core->network_core,
                                                              message);
    }
    else if (context->through_edge)
    {
      /* Have an edge target; direct to neighbour */

      DEBUG_DISSEMINATION (g_print ("Dissemination start: through edge.\n"));

      message = tnt_message_construct_direct (TNT_MESSAGE_DATA_BLOCK, &data_block_m, context->block_data,
                                              context->block_len);
      tnt_network_core_send_message (context->file_core->network_core, message,
                                     context->through_edge);
    }

    if (context->from_disseminator)
    {
      context->file_core->running_disseminations++;

      if (context->through_edge)
        tnt_edge_qualification_ref (context->through_edge);
    }

    if (message)
      flux_object_unref (FLUX_OBJECT (message));
  }

  if (!message)
  {
    DEBUG_DISSEMINATION (g_print ("Dissemination start: FAILED.\n"));
    dissemination_cleanup (context);
  }
}

static void
block_ready_for_dissemination (TntDataDB *data_db, TntFileID *file_id, gint block_id, guint block_len,
                               gpointer block, gpointer data)
{
  DisseminationContext *context = data;

  context->block_data = block;
  context->block_len  = block_len;
  start_dissemination_send (context);
}

static gboolean
dequeue_disseminator_block (TntFileCore *file_core, TntBandwidthGroup *group,
                            TntNodeID *node_id, TntFileID *file_id, gint *block_id)
{
  gint n;

  if (!file_core->network_core)
    return FALSE;

  if (group == NULL)
    return tnt_disseminator_dequeue (file_core->disseminator, node_id, file_id, block_id);

  /* Caller wants to disseminate to a specific bandwidth group. Iterate over the
   * queued requests and find the first that would be sent through an edge in that group. */

  for (n = 0; ; n++)
  {
    TntEdge *edge;

    /* We use the caller's fields to store temporary information. Probably unethical... */
    if (!tnt_disseminator_peek_nth (file_core->disseminator, n, node_id, file_id, block_id))
      break;     /* End of request list */

    edge = tnt_network_core_find_route (file_core->network_core, *node_id);
    if (edge == NULL)
      continue;  /* No known route to this destination */

    if (tnt_bandwidth_group_have_edge (group, edge))
      break;     /* Found it */
  }

  return tnt_disseminator_dequeue_nth (file_core->disseminator, n, node_id, file_id, block_id);
}

static void
destroy_dissemination_context (DisseminationContext *context)
{
  if (!context->done)
  {
    if (context->message)
      tnt_message_set_destroy_func (context->message, NULL, NULL);

    if (context->file_core->file_db)
      tnt_data_db_cancel (TNT_DATA_DB (context->file_core->file_db),
                          block_ready_for_dissemination, context);
    if (context->file_core->block_db)
      tnt_data_db_cancel (TNT_DATA_DB (context->file_core->block_db),
                          block_ready_for_dissemination, context);
  }
}

static gboolean
disseminate_next (TntFileCore *file_core, TntBandwidthGroup *group)
{
  DisseminationContext *context;
  TntNodeID             next_node_id;

  if (tnt_get_blocks_allocated () > ALLOC_BLOCKS_QUALIFIED_MAX)
  {
    DEBUG_DISSEMINATION (g_print ("Dissemination attempt: Too many qualified blocks allocated.\n"));
    return FALSE;
  }

  if (!file_core->network_core)
  {
    DEBUG_DISSEMINATION (g_print ("Dissemination attempt: No network core.\n"));
    return FALSE;
  }

  context = tnt_context_factory_new_context (file_core->dissemination_context_factory);

  if (!dequeue_disseminator_block (file_core, group, &context->recipient_id,
                                   &context->file_id, &context->block_id))
  {
    DEBUG_DISSEMINATION (g_print ("Dissemination attempt: Empty queue%s.\n", group ? " (group specific)" : ""));

    context->done = TRUE;
    tnt_context_destroy (context);
    return FALSE;
  }

  context->file_core = file_core;

  context->start_stamp = tnt_timestamp ();
  context->from_disseminator = TRUE;

  tnt_disseminator_block_node (file_core->disseminator, context->recipient_id);

  /* Ensure connection to this recipient. Note: This might arrive too late, or not at all.
   * In that case, the block is sent through the edge with the shortest known path to
   * destination. */
  tnt_network_core_connect_to_node (file_core->network_core, context->recipient_id);

  /* Ensure connection to next recipient */
  if (tnt_disseminator_peek (file_core->disseminator, &next_node_id, NULL, NULL))
    tnt_network_core_connect_to_node (file_core->network_core, next_node_id);

  /* Start reading the data block from disk. Note: Blocks may be removed from the local
   * cache while they're in the disseminator queue. While this isn't a serious problem,
   * maybe it could be avoided somehow. */
  if ((file_core->file_db != NULL &&
       tnt_file_db_get_state (file_core->file_db) != TNT_FILE_DB_SCANNING &&
       tnt_data_db_get_block (TNT_DATA_DB (file_core->file_db), &context->file_id, context->block_id,
                              TNT_PRIORITY_NORMAL, block_ready_for_dissemination, context)) ||
      (file_core->block_db != NULL &&
       tnt_data_db_get_block (TNT_DATA_DB (file_core->block_db), &context->file_id, context->block_id,
                              TNT_PRIORITY_NORMAL, block_ready_for_dissemination, context)))
  {
    DEBUG_DISSEMINATION (g_print ("Dissemination attempt: Loading block.\n"));
    return TRUE;
  }

  DEBUG_DISSEMINATION (g_print ("Dissemination attempt: Could not get block.\n"));

  dissemination_cleanup (context);
  return FALSE;
}

static gint
pick_random_block (TntFileBitmap *file_bitmap)
{
  gint total, n;

  total = tnt_file_bitmap_count_set_bits (file_bitmap);
  if (!total)
    return -1;

  n = tnt_file_bitmap_get_set_bit_n (file_bitmap, g_random_int_range (0, total));
  if (n == -1)
    g_warning ("tnt_file_bitmap_get_set_bit_n() returned -1!");

  return n;
}

/* Returns TRUE if a random dissemination was started, otherwise FALSE. */
static gboolean
disseminate_random (TntFileCore *file_core, TntBandwidthGroup *group)
{
  DisseminationContext *context;
  const GPtrArray      *edge_array;
  guint                 i, n;
  GList                *eligible_edges = NULL;
  guint                 rough_file_db_size, rough_block_db_size;
  TntDataDB            *pick_db;
  TntFileBitmap        *file_bitmap;

  /* FIXME: This can be made to work even if only one of the databases is present.
   * That doesn't happen in real life, though. */
  if (!file_core->file_db || !file_core->block_db)
    return FALSE;

  if (tnt_get_blocks_allocated () > ALLOC_BLOCKS_UNQUALIFIED_MAX)
    return FALSE;

  if (!file_core->network_core)
    return FALSE;

  edge_array = tnt_network_core_peek_edge_array (file_core->network_core);
  if (!edge_array)
    return FALSE;

  /* Determine eligible edges */
  for (i = 0, n = 0; i < edge_array->len; i++)
  {
    TntEdge  *edge = g_ptr_array_index (edge_array, i);
    TntTcpIO *tcp_io;

    /* If a bandwidth group was requested, limit selected edges to that */
    if (group && !tnt_bandwidth_group_have_edge (group, edge))
      continue;

    tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA);
    if (!tcp_io)
      continue;

    /* Have empty outgoing data queue -> eligible */
    if (!tnt_tcp_io_get_queue_out_len (tcp_io, TNT_MESSAGE_ANY))
    {
      eligible_edges = g_list_prepend (eligible_edges, edge);
      n++;
    }
  }

  if (n == 0)
    return FALSE;

  context = tnt_context_factory_new_context (file_core->dissemination_context_factory);
  context->file_core = file_core;
  context->start_stamp = tnt_timestamp ();
  context->from_disseminator = FALSE;
  context->recipient_id = TNT_NODE_ID_INVALID;

  i = g_random_int_range (0, n);
  context->through_edge = g_list_nth (eligible_edges, i)->data;
  g_list_free (eligible_edges);

  /* Decide if we're picking a block from the file DB or the block DB, based on their
   * relative sizes. The bigger DB has a higher chance of being picked. The file DB will
   * not be picked if it's scanning or hashing its contents.
   *
   * NOTE: On 32-bit archs, this will only make a sane decision if the user is sharing
   * less than 2 terabytes total. Not a huge problem, I guess. */
  rough_file_db_size = (tnt_file_db_get_file_size (file_core->file_db, FALSE) >> 10);
  rough_block_db_size = (tnt_block_db_get_file_size (file_core->block_db, TRUE) >> 10);

  if (rough_file_db_size + rough_block_db_size == 0)
    goto error;

  i = g_random_int_range (0, rough_file_db_size + rough_block_db_size);

  if (tnt_file_db_get_state (file_core->file_db) == TNT_FILE_DB_SCANNING ||
      i >= rough_file_db_size)
  {
    pick_db = TNT_DATA_DB (file_core->block_db);
  }
  else
  {
    pick_db = TNT_DATA_DB (file_core->file_db);
  }

  if (!tnt_data_db_pick_random_file (pick_db, &context->file_id))
    goto error;

  file_bitmap = tnt_data_db_get_file_bitmap (pick_db, &context->file_id);
  if (!file_bitmap)
    goto error;

  context->block_id = pick_random_block (file_bitmap);
  flux_object_unref (FLUX_OBJECT (file_bitmap));

  if (context->block_id >= 0 &&
      tnt_data_db_get_block (TNT_DATA_DB (pick_db), &context->file_id, context->block_id,
                             TNT_PRIORITY_LOW, block_ready_for_dissemination, context))
  {
    DEBUG_DISSEMINATION (g_print ("Random dissemination successful.\n"));
    return TRUE;
  }

error:
  dissemination_cleanup (context);
  return FALSE;
}

static gboolean
starvation_handler (TntNetworkCore *network_core, TntBandwidthGroup *bandwidth_group,
                    TntTransferDirection dir, TntFileCore *file_core)
{
  gboolean handled = FALSE;

  if (dir == TNT_TRANSFER_DIRECTION_OUT)
  {
    DEBUG_DISSEMINATION (g_print ("Starvation: Attempting dissemination.\n"));
    handled = disseminate_next (file_core, bandwidth_group);
    if (!handled)
      handled = disseminate_random (file_core, bandwidth_group);
  }

  return handled;
}

static void
new_edge_handler (TntNetworkCore *network_core, TntEdge *edge, TntFileCore *file_core)
{
  GList *l;
  gint   i;

  /* Ask availability for all files we're retrieving */
  if (file_core->retrieval_list)
  {
    for (i = 0; i < tnt_retrieval_list_get_n_retrievals (file_core->retrieval_list); i++)
    {
      TntRetrieval *retrieval = tnt_retrieval_list_get_retrieval_n (file_core->retrieval_list, i);

      if (tnt_retrieval_is_active (retrieval))
        ask_retrieval_availability (file_core, retrieval, edge);
    }
  }

  /* Send running queries through edge */
  for (l = file_core->local_queries; l; l = g_list_next (l))
  {
    TntLocalQuery *query = l->data;
    broadcast_local_query (file_core, query, edge);
  }
}

static void
pre_data_block_handler (TntNetworkCore *network_core, TntEdge *edge,
                        TntMessage *message, TntFileCore *file_core)
{
  TntDataBlockM *data_block_m;

  if (file_core->retrieval_list && tnt_message_has_payload (message) &&
      (data_block_m = tnt_message_get_struct (message)))
  {
    if (tnt_retrieval_list_find_retrieval (file_core->retrieval_list, &data_block_m->file_id) >= 0)
    {
      tnt_edge_qualification_ref (edge);
    }

    g_free (data_block_m);
  }
}

/* ---------------- *
 * Message Handling *
 * ---------------- */

/* --- Handlers --- */

typedef struct
{
  TntFileCore *file_core;
  TntNodeID    source_id;
  guint        local_block_count;
  guint        finished : 1;
}
IncomingFileInfoContext;

static void
file_meta_merged_cb (const TntFileID *file_id, TntFileMeta *file_meta, IncomingFileInfoContext *context)
{
  try_match_file (context->file_core, file_id, file_meta, context->source_id,
                  context->local_block_count);

  tnt_file_meta_free (file_meta);

  context->finished = TRUE;
  tnt_context_destroy (context);
}

static void
destroy_incoming_file_info_context (IncomingFileInfoContext *context)
{
  if (!context->finished)
  {
    if (context->file_core->meta_db)
      tnt_meta_db_cancel_callback (context->file_core->meta_db, file_meta_merged_cb, context);
  }
}

static gboolean
handle_incoming_file_info (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntFileInfoM    *file_info;
  TntNodeID        source_id = TNT_NODE_ID_INVALID;

  if (!tnt_message_has_payload (message))
    return FALSE;

  file_info = tnt_message_get_struct (message);
  if (!file_info)
    return TRUE;

  if (tnt_message_is_routed (message))
    source_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));

  tnt_file_meta_strip_foreign (&file_info->meta);

  if (file_core->meta_db)
  {
    gint     queue_len;
    gboolean store_meta;

    /* FIXME: What if local_block_count is higher than real len? */

    queue_len = tnt_meta_db_get_action_queue_len (file_core->meta_db);

    if (file_info->local_block_count > 0 && queue_len < QUEUE_LENGTH_PAIN_THRESHOLD)
      store_meta = TRUE;
    else
      store_meta = FALSE;

    if (queue_len < QUEUE_LENGTH_REST_THRESHOLD)
    {
      IncomingFileInfoContext *context;

      context = tnt_context_factory_new_context (file_core->incoming_file_info_context_factory);
      context->file_core = file_core;
      context->source_id = source_id;
      context->local_block_count = file_info->local_block_count;

      tnt_meta_db_merge_file_meta (file_core->meta_db, &file_info->id, &file_info->meta,
                                   store_meta, (TntFileMetaFunc *) file_meta_merged_cb, context);
    }
    else
    {
      try_match_file (file_core, &file_info->id, &file_info->meta, source_id,
                      file_info->local_block_count);
    }
  }

  tnt_message_free_struct (message, file_info);
  return TRUE;
}

static gboolean
handle_incoming_query (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntQueryM      *query_info;
  TntRemoteQuery *remote_query;

  if (!tnt_message_has_payload (message))
    return FALSE;

  /* Must have source */
  if (!tnt_message_is_routed (message) ||
      tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)) == TNT_NODE_ID_INVALID)
    return FALSE;

  query_info = tnt_message_get_struct (message);
  if (!query_info)
    return TRUE;

  if (query_info->string && g_utf8_validate (query_info->string, -1, NULL))
  {
    TntNodeID node_id;

    node_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));

    if (source_edge)
      tnt_edge_change_score (source_edge, EDGE_SCORE_QUERY);

    remote_query = tnt_remote_query_new (node_id);
    tnt_query_set_id (TNT_QUERY (remote_query), query_info->id);
    tnt_query_set_string (TNT_QUERY (remote_query), query_info->string);

    DEBUG (g_print ("Query: %s\n", query_info->string));

    if (query_info->flags & TNT_QUERY_M_MATCH_PARTIAL)
    {
      tnt_query_set_match_partial (TNT_QUERY (remote_query), TRUE);
      DEBUG (g_print ("Match partials.\n"));
    }
    else
    {
      tnt_query_set_match_partial (TNT_QUERY (remote_query), FALSE);
      DEBUG (g_print ("Do not match partials.\n"));
    }

    DEBUG
    (
      g_print ("n_remote_queries == %d, list length == %d\n",
               file_core->n_remote_queries, g_list_length (file_core->remote_queries));
      {
        GList *l;

        for (l = file_core->remote_queries; l; l = g_list_next (l))
        {
          TntQuery *query = l->data;

          g_print ("Still running: %s\n", tnt_query_get_string (query));
        }
      }
    )

    if (file_core->n_remote_queries < REMOTE_QUERIES_MAX)
      add_remote_query (file_core, remote_query);

    flux_object_unref (FLUX_OBJECT (remote_query));  /* Initial reference */
  }

  if (query_info->string)
    g_free (query_info->string);
  g_free (query_info);
  return TRUE;
}

static gboolean
handle_incoming_query_cancel (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntNodeID        node_id = TNT_NODE_ID_INVALID;
  TntQueryCancelM *query_cancel_info;

  if (!tnt_message_has_payload (message))
    return FALSE;

  /* Must have source */
  if (!tnt_message_is_routed (message) ||
      tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)) == TNT_NODE_ID_INVALID)
    return FALSE;

  query_cancel_info = tnt_message_get_struct (message);
  if (!query_cancel_info)
    return TRUE;

  node_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));

  remove_remote_query (file_core, node_id, query_cancel_info->id);
  g_free (query_cancel_info);
  return TRUE;
}

static gboolean
handle_incoming_data_block_request (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntNodeID             node_id;
  TntDataBlockRequestM *data_block_request;
  gboolean              have_block = FALSE;

  if (!tnt_message_has_payload (message))
    return FALSE;

  if (!tnt_message_is_routed (message))
    return FALSE;

  if (!tnt_message_is_for_me (message, my_node_id (file_core)))
    return TRUE;

  if (!file_core->network_core)
    return TRUE;

  node_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));
  if (node_id == TNT_NODE_ID_INVALID)
    return FALSE;

  data_block_request = tnt_message_get_struct (message);
  if (!data_block_request)
    return TRUE;

  if (file_core->file_db &&
      tnt_data_db_have_block (TNT_DATA_DB (file_core->file_db),
                              &data_block_request->file_id,
                              data_block_request->block_id))
  {
    DEBUG_DISSEMINATION (g_print ("Data block requested: File DB has block %d.\n", data_block_request->block_id));
    have_block = TRUE;
  }
  else if (file_core->block_db &&
           tnt_data_db_have_block (TNT_DATA_DB (file_core->block_db),
                                   &data_block_request->file_id,
                                   data_block_request->block_id))
  {
    DEBUG_DISSEMINATION (g_print ("Data block requested: Block DB has block %d.\n", data_block_request->block_id));
    have_block = TRUE;
  }
  else
  {
    DEBUG_DISSEMINATION (g_print ("Data block requested: Block %d not found.\n", data_block_request->block_id));
  }

  if (have_block &&
      tnt_disseminator_enqueue (file_core->disseminator, node_id, &data_block_request->file_id,
                                data_block_request->block_id))
  {
    TntOnQueueM       on_queue;
    TntRoutedMessage *rmessage;

    DEBUG_DISSEMINATION (g_print ("Data block requested: Queued successfully.\n"));

    memcpy (&on_queue.file_id, &data_block_request->file_id, sizeof (TntFileID));
    on_queue.block_id  = data_block_request->block_id;
    on_queue.timestamp = tnt_timestamp ();

    rmessage = tnt_message_construct_routed (TNT_MESSAGE_ON_QUEUE, &on_queue, NULL, 0, 5,
                                             my_node_id (file_core), node_id);
    tnt_network_core_route_message (file_core->network_core, TNT_MESSAGE (rmessage));
    flux_object_unref (FLUX_OBJECT (rmessage));

    /* Give it a chance to start sending immediately */
    /* FIXME: Should find the bandwidth group corresponding to the edge, and pass that */
    if (file_core->running_disseminations < RUNNING_DISSEMINATIONS_NOMINAL)
      disseminate_next (file_core, NULL);
  }
  else
  {
    DEBUG_DISSEMINATION (g_print ("Data block requested: Not queued.\n"));
  }

  g_free (data_block_request);
  return TRUE;
}

static void
data_block_stored_cb (TntDataDB *data_db, TntFileID *file_id, gint block_id,
                      guint block_len, gpointer data)
{
  TntFileCore *file_core = data;

  assert_types (file_core);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (block_id >= 0);
  g_return_if_fail (block_len > 0);

  if (file_core->retrieval_list)
    tnt_retrieval_list_got_block (file_core->retrieval_list, file_id, block_id);

  /* FIXME: We should call network_decision () here, and ask_retrieval_block () from that */
  ask_retrieval_block (file_core);
}

static void
store_block (TntFileCore *file_core, const TntFileID *file_id, gint block_id, gpointer data, guint len)
{
  gpointer block_data_copy;

  /* FIXME: Should have refcounted data blocks... */
  block_data_copy = tnt_alloc_block (len);
  memcpy (block_data_copy, data, len);

  tnt_data_db_put_block (TNT_DATA_DB (file_core->block_db), file_id, block_id,
                         block_data_copy, len, TNT_PRIORITY_NORMAL,
                         data_block_stored_cb, file_core);
}

static gboolean
handle_incoming_data_block (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntDataBlockM *data_block_m;
  gpointer       block_data;
  guint          block_len;

  if (!tnt_message_has_payload (message))
    return FALSE;

  data_block_m = tnt_message_get_struct (message);
  if (!data_block_m)
    return TRUE;

  tnt_message_get_data_block (message, &block_data, &block_len);

  if (block_data && block_len && block_len <= TNT_DATA_BLOCK_SIZE && file_core->block_db)
  {
    if (file_core->retrieval_list &&
        tnt_retrieval_list_find_retrieval (file_core->retrieval_list, &data_block_m->file_id) >= 0)
    {
      if (tnt_get_blocks_allocated () < ALLOC_BLOCKS_QUALIFIED_MAX)
        store_block (file_core, &data_block_m->file_id, data_block_m->block_id, block_data, block_len);

      if (source_edge)
      {
        tnt_edge_change_score (source_edge, EDGE_SCORE_DATA_BLOCK);

        /* Undo what we did in receiving_data_block_from_edge (). Note that (rarely), we can
         * receive a data block in a file we're getting without first getting that callback, since
         * the user can start getting a file after we've started receiving a data block for it.
         * tnt_edge_qualification_unref () is tolerant of this condition, but it's sort of nasty. */
        tnt_edge_qualification_unref (source_edge);
      }
    }
    else
    {
      if (tnt_get_blocks_allocated () < ALLOC_BLOCKS_UNQUALIFIED_MAX)
        store_block (file_core, &data_block_m->file_id, data_block_m->block_id, block_data, block_len);

      /* FIXME: If the user removed the retrieval while we were getting a block for it,
       * the edge can stay qualified forever. The right fix for this is to add a callback for
       * removal from the retrieval list, and make that callback and user_retrieval_added_cb()
       * (see above) check all edges, using tnt_edge_get_incoming_file_id(), and ref/unref
       * qualification. */

      if (source_edge)
        tnt_edge_change_score (source_edge, EDGE_SCORE_UNREQUESTED_DATA_BLOCK);
    }
  }

  g_free (data_block_m);
  return TRUE;
}

static gboolean
handle_incoming_ask_file_avail (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntAskFileAvailM *ask_file_avail;

  /* We only want to do this if we can know who to reply to */
  if (!tnt_message_is_routed (message) || !tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    return FALSE;

  if (!tnt_message_has_payload (message))
    return FALSE;

  if (!file_core->network_core)
    return TRUE;

  ask_file_avail = tnt_message_get_struct (message);
  if (!ask_file_avail)
    return TRUE;

  if ((file_core->file_db &&
       tnt_data_db_have_file_full (TNT_DATA_DB (file_core->file_db), &ask_file_avail->file_id)) ||
      (file_core->block_db &&
       tnt_data_db_have_file_full (TNT_DATA_DB (file_core->block_db), &ask_file_avail->file_id)))
  {
    TntRoutedMessage *rmessage;
    TntHaveFileM      have_file_info;

    memcpy (&have_file_info.file_id, &ask_file_avail->file_id, sizeof (TntFileID));

    rmessage = tnt_message_construct_routed (TNT_MESSAGE_HAVE_FILE, &have_file_info, NULL, 0, 5,
                                             my_node_id (file_core),
                                             tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)));
    tnt_network_core_route_message (file_core->network_core, TNT_MESSAGE (rmessage));
    flux_object_unref (FLUX_OBJECT (rmessage));
  }
  else if (file_core->block_db &&
           tnt_data_db_have_file_part (TNT_DATA_DB (file_core->block_db), &ask_file_avail->file_id))
  {
    TntFileBitmap *file_bitmap = NULL;

    if (file_core->block_db)
      file_bitmap = tnt_data_db_get_file_bitmap (TNT_DATA_DB (file_core->block_db),
                                                 &ask_file_avail->file_id);

    if (file_bitmap)
    {
      TntRoutedMessage   *rmessage;
      TntHaveFileBitmapM  have_file_bitmap;

      memcpy (&have_file_bitmap.file_id, &ask_file_avail->file_id, sizeof (TntFileID));
      have_file_bitmap.bitmap          = tnt_file_bitmap_get_bitmap (file_bitmap);
      have_file_bitmap.bitmap_len_bits = tnt_file_bitmap_get_len_bits (file_bitmap);
      have_file_bitmap.bitmap_len      = TNT_BITS_TO_BYTES (have_file_bitmap.bitmap_len_bits);
      flux_object_unref (FLUX_OBJECT (file_bitmap));

      rmessage = tnt_message_construct_routed (TNT_MESSAGE_HAVE_FILE_BITMAP, &have_file_bitmap, NULL, 0, 5,
                                               my_node_id (file_core),
                                               tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)));
      tnt_network_core_route_message (file_core->network_core, TNT_MESSAGE (rmessage));
      flux_object_unref (FLUX_OBJECT (rmessage));
      g_free (have_file_bitmap.bitmap);
    }
    else
    {
      g_warning ("Block DB says it has part of file, but returns NULL bitmap.");
    }
  }

  g_free (ask_file_avail);
  return TRUE;
}

static gboolean
handle_incoming_have_file (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntHaveFileM  *have_file;
  TntNodeID      source_id;
  TntFileBitmap *file_bitmap;

  /* We only want to do this if we can know who it's from */
  if (!tnt_message_is_routed (message) || !tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    return FALSE;

  if (!tnt_message_has_payload (message))
    return FALSE;

  have_file = tnt_message_get_struct (message);
  if (!have_file)
    return TRUE;

  source_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));
  file_bitmap = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (have_file->file_id.len));
  tnt_file_bitmap_fill_true (file_bitmap);

  if (maybe_update_retrieval_remote_bitmap (file_core, &have_file->file_id, source_id, file_bitmap) && source_edge)
    tnt_edge_change_score (source_edge, EDGE_SCORE_HAVE_FILE);

  flux_object_unref (FLUX_OBJECT (file_bitmap));
  g_free (have_file);
  return TRUE;
}

static gboolean
handle_incoming_have_file_bitmap (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntHaveFileBitmapM *have_file_bitmap;

  /* We only want to do this if we can know who it's from */
  if (!tnt_message_is_routed (message) || !tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    return FALSE;

  if (!tnt_message_has_payload (message))
    return FALSE;

  have_file_bitmap = tnt_message_get_struct (message);
  if (!have_file_bitmap)
    return TRUE;

  if (have_file_bitmap->bitmap_len_bits == TNT_REAL_LEN_TO_BITMAP_LEN (have_file_bitmap->file_id.len) &&
      TNT_BITS_TO_BYTES (have_file_bitmap->bitmap_len_bits) == have_file_bitmap->bitmap_len &&
      have_file_bitmap->bitmap)
  {
    TntNodeID      source_id;
    TntFileBitmap *file_bitmap;

    source_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));
    file_bitmap = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (have_file_bitmap->file_id.len));
    tnt_file_bitmap_set_bitmap (file_bitmap, have_file_bitmap->bitmap);

    if (maybe_update_retrieval_remote_bitmap (file_core, &have_file_bitmap->file_id, source_id, file_bitmap) && source_edge)
      tnt_edge_change_score (source_edge, EDGE_SCORE_HAVE_FILE_BITMAP);

    flux_object_unref (FLUX_OBJECT (file_bitmap));
  }
  else if (have_file_bitmap->bitmap)
  {
    g_free (have_file_bitmap->bitmap);
  }

  g_free (have_file_bitmap);
  return TRUE;
}

static gboolean
handle_incoming_on_queue (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntOnQueueM *on_queue;

  /* We only want to do this if we can know who it's from */
  if (!tnt_message_is_routed (message) || !tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    return FALSE;

  if (!tnt_message_has_payload (message))
    return TRUE;

  if (!tnt_message_is_for_me (message, my_node_id (file_core)))
    return TRUE;

  on_queue = tnt_message_get_struct (message);
  if (!on_queue)
    return TRUE;

  if (!file_core->retrieval_list)
    return TRUE;

  DEBUG
  (
    g_print ("On queue for block %d of file sized %lld from node %08llx.\n",
             on_queue->block_id, on_queue->file_id.len,
             tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)));
  )

  tnt_retrieval_list_on_queue (file_core->retrieval_list, &on_queue->file_id,
                               tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)),
                               on_queue->block_id, on_queue->timestamp);

  if (source_edge)
    tnt_edge_change_score (source_edge, EDGE_SCORE_ON_QUEUE);

  g_free (on_queue);
  return TRUE;
}

static gboolean
handle_incoming_off_queue (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message, TntFileCore *file_core)
{
  TntOffQueueM *off_queue;

  /* We only want to do this if we can know who it's from */
  if (!tnt_message_is_routed (message) || !tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    return FALSE;

  if (!tnt_message_has_payload (message))
    return TRUE;

  if (!tnt_message_is_for_me (message, my_node_id (file_core)))
    return TRUE;

  off_queue = tnt_message_get_struct (message);
  if (!off_queue)
    return TRUE;

  if (!file_core->retrieval_list)
    return TRUE;

  DEBUG
  (
    g_print ("Off queue for block %d of file sized %lld from node %08llx.\n",
             off_queue->block_id, off_queue->file_id.len,
             tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)));
  )

  tnt_retrieval_list_off_queue (file_core->retrieval_list,
                                tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)),
                                off_queue->block_id, off_queue->timestamp);

  if (source_edge)
    tnt_edge_change_score (source_edge, EDGE_SCORE_OFF_QUEUE);

  g_free (off_queue);
  return TRUE;
}

/* --- Dispatcher --- */

typedef struct
{
  TntMessageType message_type;
  gpointer       func;
  gboolean       may_generate_replies;
}
MessageHandlerMap;

static MessageHandlerMap message_handler_map [] =
{
  { TNT_MESSAGE_FILE_INFO,          handle_incoming_file_info,          FALSE },
  { TNT_MESSAGE_QUERY,              handle_incoming_query,              TRUE  },
  { TNT_MESSAGE_QUERY_CANCEL,       handle_incoming_query_cancel,       FALSE },
  { TNT_MESSAGE_DATA_BLOCK_REQUEST, handle_incoming_data_block_request, TRUE  },
  { TNT_MESSAGE_DATA_BLOCK,         handle_incoming_data_block,         FALSE },
  { TNT_MESSAGE_ASK_FILE_AVAIL,     handle_incoming_ask_file_avail,     TRUE  },
  { TNT_MESSAGE_HAVE_FILE,          handle_incoming_have_file,          FALSE },
  { TNT_MESSAGE_HAVE_FILE_BITMAP,   handle_incoming_have_file_bitmap,   FALSE },
  { TNT_MESSAGE_ON_QUEUE,           handle_incoming_on_queue,           FALSE },
  { TNT_MESSAGE_OFF_QUEUE,          handle_incoming_off_queue,          TRUE  },

  { TNT_MESSAGE_UNKNOWN, NULL, FALSE }
};

static void
attach_network_handlers (TntFileCore *file_core)
{
  gint i;

  g_assert (file_core->network_core);

  for (i = 0; message_handler_map [i].func; i++)
  {
    tnt_network_core_attach_message_handler (file_core->network_core,
                                             message_handler_map [i].message_type,
                                             message_handler_map [i].may_generate_replies,
                                             (TntMessageHandler *) message_handler_map [i].func,
                                             file_core);
  }

  tnt_network_core_attach_new_edge_handler (file_core->network_core,
                                            (TntNewEdgeHandler *) new_edge_handler,
                                            file_core);
  tnt_network_core_attach_starvation_handler (file_core->network_core,
                                              (TntStarvationHandler *) starvation_handler,
                                              file_core);
  tnt_network_core_attach_pre_data_handler (file_core->network_core,
                                            TNT_MESSAGE_DATA_BLOCK,
                                            (TntPreDataHandler *) pre_data_block_handler,
                                            file_core);
}

static void
detach_network_handlers (TntFileCore *file_core)
{
  gint i;

  g_assert (file_core->network_core);

  for (i = 0; message_handler_map [i].func; i++)
  {
    tnt_network_core_detach_message_handler (file_core->network_core,
                                             message_handler_map [i].message_type,
                                             (TntMessageHandler *) message_handler_map [i].func,
                                             file_core);
  }

  tnt_network_core_detach_new_edge_handler (file_core->network_core,
                                            (TntNewEdgeHandler *) new_edge_handler,
                                            file_core);
  tnt_network_core_detach_starvation_handler (file_core->network_core,
                                              (TntStarvationHandler *) starvation_handler,
                                              file_core);
  tnt_network_core_detach_pre_data_handler (file_core->network_core,
                                            TNT_MESSAGE_DATA_BLOCK,
                                            (TntPreDataHandler *) pre_data_block_handler,
                                            file_core);
}

/* ------------------------- *
 * Local Database Operations *
 * ------------------------- */

static void
local_meta_extracted_cb (TntFileID *file_id, TntFileMeta *file_meta, TntFileCore *file_core)
{
  if (file_meta)
  {
    try_match_file        (file_core, file_id, file_meta, TNT_NODE_ID_INVALID, 0);
    try_match_file_remote (file_core, file_id, file_meta);
    tnt_file_meta_free (file_meta);

    try_set_retrieval_complete (file_core, file_id);
    local_file_set_in_queries (file_core, file_id, TRUE);
  }
}

/* Called whenever file_db gets the hash for a file. Makes sure meta_db is up-to-date
 * on the local file name, and adds some keywords based on its path if the keyword
 * field is empty. */
static void
file_update_cb (TntFileDB *file_db, TntFileID *file_id, gpointer data)
{
  TntFileCore *file_core = data;

  assert_types (file_core);
  flux_object_check (FLUX_OBJECT (file_db), TNT_FILE_DB_TYPE);
  g_return_if_fail (file_id != NULL);

  if (file_core->meta_db && file_core->file_db)
  {
    gchar *name;
    gchar *path;
    gchar *relative_path;

    name          = tnt_file_db_get_file_name (file_core->file_db, file_id);
    path          = tnt_file_db_get_file_path (file_core->file_db, file_id);
    relative_path = tnt_file_db_get_file_relative_path (file_core->file_db, file_id);

    tnt_meta_db_extract_file_meta (file_core->meta_db, file_id, name, path, relative_path,
                                   (TntFileMetaFunc *) local_meta_extracted_cb, file_core);

    g_free (relative_path);
    g_free (path);
    g_free (name);
  }
}

/* Called whenever file_db needs the hash for a file. The meta_db might have it cached. */
static void
file_id_hint_cb (TntFileDB *file_db, const gchar *path, const gchar *name, guint64 len,
                 TntTimeStamp timestamp, TntFileIDFunc *func, gpointer func_data, gpointer data)
{
  TntFileCore *file_core = data;

  assert_types (file_core);
  flux_object_check (FLUX_OBJECT (file_db), TNT_FILE_DB_TYPE);
  g_return_if_fail (name != NULL);
  g_return_if_fail (func != NULL);

  if (file_core->meta_db)
  {
    gchar *name_utf8;

    name_utf8 = tnt_filename_to_utf8_lossy (name);
    tnt_meta_db_get_local_file_id (file_core->meta_db, name_utf8, len, timestamp,
                                   func, func_data);
    g_free (name_utf8);
  }
  else
  {
    func (NULL, func_data);
  }
}

/* ------------------------ *
 * Configuration Management *
 * ------------------------ */

static void
configure_db (TntFileCore *file_core, TntConfig *config)
{
  gchar      *base_path;
  gchar      *sub_path;
  TntStorage *storage;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);

  if (file_core->block_db)
  {
    flux_object_unref (FLUX_OBJECT (file_core->block_db));
    file_core->block_db = NULL;
  }

  if (file_core->meta_db)
  {
    flux_object_unref (FLUX_OBJECT (file_core->meta_db));
    file_core->meta_db = NULL;
  }

  if (file_core->assembler)
  {
    flux_object_unref (FLUX_OBJECT (file_core->assembler));
    file_core->assembler = NULL;
  }

  if (!base_path)
    return;

  sub_path = g_build_path (G_DIR_SEPARATOR_S, base_path, "data", NULL);
  storage = tnt_storage_init (sub_path);
  g_free (sub_path);

  if (storage)
  {
    tnt_storage_set_limit (storage, tnt_config_get_uint64 (config, "filing", "data-cache-size", NULL));
    file_core->block_db = tnt_block_db_new (storage);
  }

  sub_path = g_build_path (G_DIR_SEPARATOR_S, base_path, "meta", NULL);
  storage = tnt_storage_init (sub_path);
  g_free (sub_path);

  if (storage)
  {
    tnt_storage_set_limit (storage, tnt_config_get_uint64 (config, "filing", "meta-cache-size", NULL));
    file_core->meta_db = tnt_meta_db_new (storage);
  }

  if (file_core->block_db && file_core->file_db)
  {
    file_core->assembler = tnt_file_assembler_new (TNT_DATA_DB (file_core->block_db),
                                                   file_core->file_db);
  }

  /* FIXME: We can't move the retrieval list with the current API, so we create it
   * once and keep using that for the entire session. */

  if (!file_core->retrieval_list)
  {
    storage = tnt_storage_open (base_path);

    if (storage)
    {
      file_core->retrieval_list = tnt_retrieval_list_new (storage);
      tnt_retrieval_list_add_retrieval_add_func (file_core->retrieval_list,
                                                 user_retrieval_added_cb, file_core);
      tnt_retrieval_list_add_retrieval_completed_func (file_core->retrieval_list,
                                                       user_retrieval_completed_cb, file_core);
    }
  }

  if (file_core->block_db && file_core->retrieval_list)
  {
    tnt_block_db_set_retrieval_list (file_core->block_db, file_core->retrieval_list);
    tnt_retrieval_list_load (file_core->retrieval_list, file_core->block_db);
  }

  g_free (base_path);  
}

static void
validate_db (TntFileCore *file_core, TntConfig *config)
{
  gchar *base_path;
  gchar *nice_path;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  nice_path = tnt_nice_path (base_path);

  if (!base_path || !strlen (base_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_FILE_ERROR_DATA_DIR_UNSET,
                                 "The database directory is not set.");
  }
  else if (!nice_path)
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_FILE_ERROR_DATA_DIR_RELATIVE,
                                 "The database directory is not absolute.");
  }
  else if (!tnt_path_exists (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_FILE_ERROR_DATA_DIR_NONEXISTENT,
                                 "The database directory doesn't exist.");
  }
  else if (!tnt_path_is_dir (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_FILE_ERROR_DATA_DIR_NONEXISTENT,
                                 "The database path is not a directory.");
  }
  else if (!tnt_dir_have_full_access (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_FILE_ERROR_DATA_DIR_PERMS,
                                 "The database path does not provide read/write access.");
  }

  g_free (base_path);
  g_free (nice_path);
}

static void
configure_shares (TntFileCore *file_core, TntConfig *config)
{
  FluxTree *shares_tree, *node;
  GList    *configured_dirs, *l;

  shares_tree     = tnt_config_peek_node (config, "filing", "shares", NULL);
  configured_dirs = tnt_file_db_get_user_dir_list (file_core->file_db);

  /* Check that TntFileDB's dirs exist in config */

  for (l = configured_dirs; l; l = g_list_next (l))
  {
    gchar    *configured_dir      = l->data;
    gchar    *configured_dir_nice = tnt_nice_path (configured_dir);
    FluxTree *requested_dir_node  = NULL;

    if (!configured_dir_nice)
      continue;

    if (shares_tree)
    {
      /* Can't use flux_tree_find_first_child (), since we have to trim paths */
      for (node = flux_tree_get_first_child (shares_tree); node; node = flux_tree_get_next (node))
      {
        gchar *found_dir      = flux_tree_data_get_str (node);
        gchar *found_dir_nice = tnt_nice_path (found_dir);
        gint   result;

        g_free (found_dir);

        if (!found_dir_nice)
          continue;

        result = strcmp (found_dir_nice, configured_dir_nice);
        g_free (found_dir_nice);

        if (!result)
        {
          requested_dir_node = node;
          break;
        }
      }
    }

    if (!requested_dir_node)
      tnt_file_db_remove_dir (file_core->file_db, configured_dir_nice);

    g_free (configured_dir_nice);
  }

  /* Check that config's dirs exist in TntFileDB */

  if (!shares_tree)
    return;

  for (node = flux_tree_get_first_child (shares_tree); node; node = flux_tree_get_next (node))
  {
    gchar *requested_dir      = flux_tree_data_get_str (node);
    gchar *requested_dir_nice = tnt_nice_path (requested_dir);

    g_free (requested_dir);

    if (!requested_dir_nice)
      continue;

    for (l = configured_dirs; l; l = g_list_next (l))
    {
      gchar *configured_dir      = l->data;
      gchar *configured_dir_nice = tnt_nice_path (configured_dir);
      gint   result;

      if (!configured_dir_nice)
        continue;

      result = strcmp (configured_dir_nice, requested_dir_nice);
      g_free (configured_dir_nice);

      if (!result)
        break;
    }

    if (!l)
      tnt_file_db_add_dir (file_core->file_db, requested_dir_nice, TRUE);

    g_free (requested_dir_nice);
  }

  tnt_list_free_with_data (configured_dirs);
}

static void
validate_shares (TntFileCore *file_core, TntConfig *config)
{
  FluxTree *shares_tree, *node;

  shares_tree = tnt_config_peek_node (config, "filing", "shares", NULL);
  if (!shares_tree)
    return;

  for (node = flux_tree_get_first_child (shares_tree); node; node = flux_tree_get_next (node))
  {
    gchar *requested_dir      = flux_tree_data_get_str (node);
    gchar *requested_dir_nice = tnt_nice_path (requested_dir);

    if (!requested_dir_nice)
    {
      tnt_config_validation_error (config, "file", node,
                                   TNT_FILE_ERROR_SHARE_RELATIVE,
                                   "The share '%s' is not absolute.", requested_dir);
    }
    else if (!tnt_path_exists (requested_dir_nice))
    {
      tnt_config_validation_error (config, "file", node,
                                   TNT_FILE_ERROR_SHARE_NONEXISTENT,
                                   "The share '%s' doesn't exist.", requested_dir);
    }
    else if (!tnt_path_is_dir (requested_dir_nice))
    {
      tnt_config_validation_error (config, "file", node,
                                   TNT_FILE_ERROR_SHARE_NONEXISTENT,
                                   "The share '%s' is not a directory.", requested_dir);
    }
    else if (!tnt_dir_have_read_access (requested_dir_nice))
    {
      tnt_config_validation_error (config, "file", node,
                                   TNT_FILE_ERROR_SHARE_PERMS,
                                   "The share '%s' is not readable.", requested_dir);
    }

    g_free (requested_dir);
    g_free (requested_dir_nice);
  }
}

static void
configure_incoming_dir (TntFileCore *file_core, TntConfig *config)
{
  gchar      *incoming_dir;
  TntStorage *storage;

  incoming_dir = tnt_config_get_str (config, "filing", "incoming-dir", NULL);
  if (!incoming_dir)
    return;

  storage = tnt_storage_open (incoming_dir);
  g_free (incoming_dir);

  if (storage && file_core->retrieval_list)
    tnt_retrieval_list_set_incoming_storage (file_core->retrieval_list, storage);
}

static void
validate_incoming_dir (TntFileCore *file_core, TntConfig *config)
{
  gchar *base_path;
  gchar *nice_path;

  base_path = tnt_config_get_str (config, "filing", "incoming-dir", NULL);
  nice_path = tnt_nice_path (base_path);

  if (!base_path || !strlen (base_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "filing", "incoming-dir", NULL),
                                 TNT_FILE_ERROR_INCOMING_DIR_UNSET,
                                 "The incoming directory is not set.");
  }
  else if (!nice_path)
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "filing", "incoming-dir", NULL),
                                 TNT_FILE_ERROR_INCOMING_DIR_RELATIVE,
                                 "The incoming directory is not absolute.");
  }
  else if (!tnt_path_exists (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "filing", "incoming-dir", NULL),
                                 TNT_FILE_ERROR_INCOMING_DIR_NONEXISTENT,
                                 "The incoming directory doesn't exist.");
  }
  else if (!tnt_path_is_dir (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "filing", "incoming-dir", NULL),
                                 TNT_FILE_ERROR_INCOMING_DIR_NONEXISTENT,
                                 "The incoming path is not a directory.");
  }
  else if (!tnt_dir_have_full_access (nice_path))
  {
    tnt_config_validation_error (config, "file",
                                 tnt_config_peek_node (config, "filing", "incoming-dir", NULL),
                                 TNT_FILE_ERROR_INCOMING_DIR_PERMS,
                                 "The incoming path does not provide read/write access.");
  }

  g_free (base_path);
  g_free (nice_path);
}

static void
configure_cache_sizes (TntFileCore *file_core, TntConfig *config)
{
  TntStorage *storage;
  guint64     cache_size;

  if (file_core->block_db)
  {
    cache_size = tnt_config_get_uint64 (config, "filing", "data-cache-size", NULL);
    storage = tnt_data_db_get_storage (TNT_DATA_DB (file_core->block_db));
    tnt_storage_set_limit (storage, cache_size);
  }

  if (file_core->meta_db)
  {
    cache_size = tnt_config_get_uint64 (config, "filing", "meta-cache-size", NULL);
    storage = tnt_meta_db_get_storage (file_core->meta_db);
    tnt_storage_set_limit (storage, cache_size);
  }
}

static void
config_set (TntFileCore *file_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (file_core));

  configure_db           (file_core, config);
  configure_shares       (file_core, config);
  configure_incoming_dir (file_core, config);
  configure_cache_sizes  (file_core, config);
}

static void
config_changed (TntFileCore *file_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (file_core));

  if (tnt_config_is_changed (config, "common", "data-dir", NULL))
    configure_db (file_core, config);
  if (tnt_config_is_changed (config, "filing", "shares", NULL))
    configure_shares (file_core, config);
  if (tnt_config_is_changed (config, "filing", "incoming-dir", NULL))
    configure_incoming_dir (file_core, config);
  if (tnt_config_is_changed (config, "filing", "data-cache-size", NULL) ||
      tnt_config_is_changed (config, "filing", "meta-cache-size", NULL))
    configure_cache_sizes (file_core, config);
}

static void
config_changed_cb (TntConfig *config, TntFileCore *file_core)
{
  config_changed (file_core);
}

static void
config_validate (TntFileCore *file_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (file_core));

  validate_db           (file_core, config);
  validate_shares       (file_core, config);
  validate_incoming_dir (file_core, config);
}

static void
config_validate_cb (TntConfig *config, TntFileCore *file_core)
{
  config_validate (file_core);
}

/* --------------------------- *
 * Initialization and Shutdown *
 * --------------------------- */

static void
network_core_set (TntFileCore *file_core)
{
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_QUERY,              100);
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_FILE_INFO,          400);
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_ASK_FILE_AVAIL,     100);
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_HAVE_FILE,          100);
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_HAVE_FILE_BITMAP,   100);
  tnt_network_core_set_outgoing_cap (file_core->network_core, TNT_MESSAGE_DATA_BLOCK_REQUEST, 100);

  attach_network_handlers (file_core);
}

static void
network_core_listen_cb (const gchar *core_id, TntNetworkCore *network_core, TntFileCore *file_core)
{
  file_core->network_core = network_core;

  if (file_core->network_core)
    network_core_set (file_core);
}

static void
init_periodic_callbacks (TntFileCore *file_core)
{
  file_core->repeat_queries_timeout_id =
    g_timeout_add (1000 * REPEAT_QUERY_TIMEOUT, repeat_queries_cb, file_core);
  file_core->ask_avail_timeout_id =
    g_timeout_add (1000 * ASK_AVAIL_TIMEOUT, ask_retrieval_availability_cb, file_core);
  file_core->ask_block_timeout_id =
    g_timeout_add (1000 * ASK_BLOCK_TIMEOUT, ask_retrieval_block_cb, file_core);
  file_core->summarize_timeout_id =
    g_timeout_add (1000 * SUMMARIZE_TIMEOUT, summarize_cb, file_core);
}

static void
stop_periodic_callbacks (TntFileCore *file_core)
{
  g_source_remove (file_core->repeat_queries_timeout_id);
  g_source_remove (file_core->ask_avail_timeout_id);
  g_source_remove (file_core->ask_block_timeout_id);
  g_source_remove (file_core->summarize_timeout_id);
}

void
tnt_file_core_finalize (FluxObject *object)
{
  TntFileCore *file_core = TNT_FILE_CORE (object);
  TntConfig   *config;

  assert_types (file_core);

  config = tnt_core_peek_config (TNT_CORE (file_core));

  /* TODO: Free array elements */
  /* TODO: Disconnect database callbacks */

  stop_periodic_callbacks (file_core);

  flux_object_unref (FLUX_OBJECT (file_core->remote_query_context_factory));
  flux_object_unref (FLUX_OBJECT (file_core->dissemination_context_factory));
  flux_object_unref (FLUX_OBJECT (file_core->incoming_file_info_context_factory));

  if (file_core->network_core)
    detach_network_handlers (file_core);

  flux_object_unref (FLUX_OBJECT (file_core->file_db));
  flux_object_unref (FLUX_OBJECT (file_core->disseminator));

  if (file_core->block_db)
    flux_object_unref (FLUX_OBJECT (file_core->block_db));
  if (file_core->meta_db)
    flux_object_unref (FLUX_OBJECT (file_core->meta_db));

  if (file_core->assembler)
    flux_object_unref (FLUX_OBJECT (file_core->assembler));

  tnt_config_detach_listener  (config, (TntConfigListenerFunc *) config_changed_cb, file_core);
  tnt_config_detach_validator (config, (TntConfigValidatorFunc *) config_validate_cb, file_core);

  tnt_core_registry_detach_listener ("network", (TntCoreListenerFunc *) network_core_listen_cb, file_core);
  tnt_core_registry_unregister_core ("file", file_core);
}

gboolean
tnt_file_core_init (TntFileCore *file_core, TntConfig *config)
{
  g_return_val_if_fail (config != NULL, FALSE);
  g_return_val_if_fail (tnt_core_registry_get_core ("file") == NULL, FALSE);

  if (!tnt_core_init (TNT_CORE (file_core), config))
    return FALSE;

  flux_object_set_type          (FLUX_OBJECT (file_core), TNT_FILE_CORE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (file_core), tnt_file_core_finalize);

  /* Messages */

  tnt_file_messages_register ();

  /* Config */

  tnt_config_attach_listener  (config, (TntConfigListenerFunc *)  config_changed_cb,  file_core);
  tnt_config_attach_validator (config, (TntConfigValidatorFunc *) config_validate_cb, file_core);

  /* Databases */

  file_core->file_db      = tnt_file_db_new ();
  file_core->disseminator = tnt_disseminator_new ();

  tnt_file_db_set_file_update_func (file_core->file_db, file_update_cb, file_core);
  tnt_file_db_set_file_id_hint_func (file_core->file_db, (TntFileIDHintFunc *) file_id_hint_cb, file_core);

  /* Misc */

  init_periodic_callbacks (file_core);
  config_set (file_core);

  file_core->remote_query_context_factory =
    tnt_context_factory_new (RemoteQueryContext, destroy_remote_query_context);
  file_core->dissemination_context_factory =
    tnt_context_factory_new (DisseminationContext, destroy_dissemination_context);
  file_core->incoming_file_info_context_factory =
    tnt_context_factory_new (IncomingFileInfoContext, destroy_incoming_file_info_context);

  /* Monitor other cores */

  file_core->network_core = tnt_core_registry_get_core ("network");
  if (file_core->network_core)
    network_core_set (file_core);

  tnt_core_registry_attach_listener ("network", (TntCoreListenerFunc *) network_core_listen_cb, file_core);

  /* Register self */

  tnt_core_registry_register_core ("file", file_core);

  return TRUE;
}

TntFileCore *
tnt_file_core_new (TntConfig *config)
{
  TntFileCore *file_core;

  file_core = g_new0 (TntFileCore, 1);
  if (!tnt_file_core_init (file_core, config))
  {
    g_free (file_core);
    file_core = NULL;
  }

  return file_core;
}

TntFileDB *
tnt_file_core_peek_file_db (TntFileCore *file_core)
{
  assert_types (file_core);

  return file_core->file_db;
}

TntBlockDB *
tnt_file_core_peek_block_db (TntFileCore *file_core)
{
  assert_types (file_core);

  return file_core->block_db;
}

TntMetaDB *
tnt_file_core_peek_meta_db (TntFileCore *file_core)
{
  assert_types (file_core);

  return file_core->meta_db;
}

TntRetrievalList *
tnt_file_core_peek_retrieval_list (TntFileCore *file_core)
{
  assert_types (file_core);

  return file_core->retrieval_list;
}

GList *
tnt_file_core_peek_local_queries (TntFileCore *file_core)
{
  assert_types (file_core);

  return file_core->local_queries;
}
