/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-db.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* This code is pretty unwieldy. Probably needs a rewrite... */

#include "config.h"

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>
#include "tnt.h"

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#define DEBUG(x)

#define IDLE_STAT_INTERVAL_MS 100
#define STABLE_FILE_TIMEOUT_S 10

#define DETERMINE_CHANGE_TIME(tstat) MIN (MAX ((tstat)->mtime, (tstat)->ctime), tnt_timestamp ())

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_FILE_DB_TYPE)

typedef struct
{
  gchar   *name;

  /* If the dir disappears, we keep it around but inactive. It might
   * be that it's simply been unmounted, and could reappear at a
   * later time. */
  guint    active      : 1;

  /* If we should add dirs found within this dir to our list of dirs. */
  guint    shared      : 1;

  /* If we finished scanning this dir. */
  guint    scanned     : 1;

  /* If we finished scanning this dir and its descendants. */
  guint    scanned_all : 1;

  /* If we statted this dir in the current sweep. */
  guint    statted     : 1;

  /* Cumulative stats for all files and subdirs in this dir. */
  gint     files_hashed_count;
  gint     files_count;
  gint64   files_size;

  /* Last modification time of the dir reported by stat (). */
  guint    last_modification_time;
}
DirStat;

typedef struct
{
  TntFileID     file_id;
  gchar        *name;
  GNode        *dir;  /* Points to a node in file_db->dir_tree */

  TntTimeStamp  change_time;

  /* If the file disappears, we keep it around in an inactive state */
  guint         active  : 1;

  /* If the file was validated by a scan */
  guint         scanned : 1;

  /* If the id field is valid */
  guint         hashed  : 1;
}
FileStat;

typedef struct
{
  GNode *found_node;
}
TreeSearchContext;

typedef struct
{
  TntFileDB *file_db;

  TntFileID  file_id;

  gchar     *name;
  gchar     *path;
}
SerializationContext;

static void     begin_state    (TntFileDB *file_db, TntFileDBState state);
static void     end_state      (TntFileDB *file_db);
static void     hash_next_file (TntFileDB *file_db);
static gboolean scan_cb        (gpointer data);

/* Directories and files we want to skip */
static gchar *skip_names [] =
{
  "resource.frk",
  "thumbs.db",
  "trans.tbl",
  NULL
};

/* --- DirStat/FileStat list building and maintenance --- */

static gboolean
should_skip (const gchar *name)
{
  gint i;

  for (i = 0; skip_names [i]; i++)
  {
    if (!strcasecmp (name, skip_names [i]))
      return TRUE;
  }

  return FALSE;
}

static SerializationContext *
find_serialization_context_by_file_id (TntFileDB *file_db, const TntFileID *file_id)
{
  GList *l;

  for (l = file_db->serialization_list; l; l = g_list_next (l))
  {
    SerializationContext *ser_context = l->data;

    if (!memcmp (file_id, &ser_context->file_id, sizeof (TntFileID)))
      return ser_context;
  }

  return NULL;
}

static SerializationContext *
find_serialization_context_by_path_and_name (TntFileDB *file_db, const gchar *path, const gchar *name)
{
  GList *l;

  for (l = file_db->serialization_list; l; l = g_list_next (l))
  {
    SerializationContext *ser_context = l->data;

    if (!tnt_compare_paths (path, ser_context->path) &&
        !tnt_compare_paths (name, ser_context->name))
      return ser_context;
  }

  return NULL;
}

static void
free_serialization_context (SerializationContext *ser_context)
{
  g_free (ser_context->name);
  g_free (ser_context->path);
  g_free (ser_context);
}

static gchar *
make_serialization_path_to_temp_file (SerializationContext *ser_context)
{
  gchar *file_id_string;
  gchar *full_path;

  file_id_string = tnt_file_id_to_string (&ser_context->file_id);
  full_path = g_strconcat (ser_context->path, G_DIR_SEPARATOR_S, 
#ifndef G_PLATFORM_WIN32
                           ".",
#endif
                           file_id_string, NULL);

  g_free (file_id_string);
  return full_path;
}

static gchar *
make_serialization_path_to_file (SerializationContext *ser_context)
{
  return g_strjoin (G_DIR_SEPARATOR_S, ser_context->path, ser_context->name, NULL);
}

static DirStat *
make_dir_stat (const gchar *name)
{
  DirStat *dir_stat;

  dir_stat = g_new0 (DirStat, 1);
  dir_stat->name      = g_strdup (name);
  dir_stat->active    = TRUE;
  dir_stat->shared    = FALSE;
  dir_stat->scanned   = FALSE;
  dir_stat->statted   = TRUE;

  return dir_stat;
}

static GNode *
find_dir_child_by_name (GNode *parent, const gchar *name)
{
  GNode *c;

  for (c = g_node_first_child (parent); c; c = g_node_next_sibling (c))
  {
    DirStat *dir_stat = c->data;

    if (!strcmp (dir_stat->name, name))
      break;
  }

  return c;
}

static GNode *
create_dir_named_child (GNode *parent, const gchar *name)
{
  DirStat *dir_stat;

  dir_stat = make_dir_stat (name);
  return g_node_append_data (parent, dir_stat);
}

static GNode *
ensure_dir_node (GNode *root, const gchar *path)
{
  GSList *pelem, *l;
  GNode  *node = root;

  pelem = tnt_path_split (path);
  for (l = pelem; l; l = g_slist_next (l))
  {
    gchar *name = l->data;
    GNode *child;

    child = find_dir_child_by_name (node, name);
    if (!child)
      child = create_dir_named_child (node, name);

    node = child;
  }

  tnt_slist_free_with_data (pelem);
  return node;
}

static gboolean
path_is_shared (GNode *root, const gchar *path)
{
  GSList   *pelem, *l;
  GNode    *node   = root;
  gboolean  result = FALSE;

  pelem = tnt_path_split (path);
  for (l = pelem; l; l = g_slist_next (l))
  {
    gchar   *name = l->data;
    DirStat *dir_stat;
    GNode   *child;

    DEBUG (g_print ("pelem: %s\n", name));

    child = find_dir_child_by_name (node, name);
    if (!child)
    {
      DEBUG (g_print ("Not found.\n"));
      break;
    }

    node = child;
    dir_stat = node->data;

    if (dir_stat && dir_stat->shared)
    {
      result = TRUE;
      DEBUG (g_print ("Success!\n"));
      break;
    }
  }

  tnt_slist_free_with_data (pelem);
  return result;
}

static GNode *
find_dir_node (GNode *root, const gchar *path)
{
  GSList *pelem, *l;
  GNode  *node = root;

  pelem = tnt_path_split (path);
  for (l = pelem; l; l = g_slist_next (l))
  {
    gchar *name = l->data;

    node = find_dir_child_by_name (node, name);
    if (!node)
      break;
  }

  tnt_slist_free_with_data (pelem);
  return node;
}

static gchar *
build_dir_path (GNode *dir_node)
{
  GSList *pelem = NULL;
  gchar  *path;

  for (; dir_node; )
  {
    DirStat *dir_stat = dir_node->data;

#ifndef G_PLATFORM_WIN32
    if (dir_stat && dir_stat->name [0] && dir_stat->name [0] != '/')
#else
    if (dir_stat && dir_stat->name [0])
#endif
      pelem = g_slist_prepend (pelem, dir_stat->name);

    dir_node = dir_node->parent;
  }

#ifndef G_PLATFORM_WIN32
  /* To get an absolute path */
  pelem = g_slist_prepend (pelem, "");

  /* To get a single '/' if we're expressing the root path */
  if (!g_slist_next (pelem))
    pelem = g_slist_prepend (pelem, "");
#else
  /* To make C: -> C:\ */
  if (!g_slist_next (pelem))
    pelem = g_slist_append (pelem, "");
#endif

  path = tnt_path_join (pelem);
  g_slist_free (pelem);
  return path;
}

static gchar *
build_relative_path (GNode *dir_node)
{
  GSList *pelem = NULL;
  gchar  *path;

  for (; dir_node && dir_node->parent; )
  {
    DirStat *dir_stat = dir_node->data;

    if (!dir_stat || !dir_stat->shared)
      break;

#ifndef G_PLATFORM_WIN32
    if (dir_stat->name [0] && dir_stat->name [0] != '/')
#else
    /* Don't include the root device name on Win32 */
    if (!dir_node->parent->parent)
      break;

    if (dir_stat->name [0])
#endif
      pelem = g_slist_prepend (pelem, dir_stat->name);

    dir_node = dir_node->parent;
  }

  path = tnt_path_join (pelem);
  g_slist_free (pelem);
  return path;
}

static void
free_dir_stat (DirStat *dir_stat)
{
  g_free (dir_stat->name);
  g_free (dir_stat);
}

/* This is used in conjunction with a GHashTable _only_, and therefore returns
 * TRUE/FALSE, not -1/0/1 */
static gint
compare_files_id (const FileStat *file_0, const FileStat *file_1)
{
  if (file_0->file_id.len != file_1->file_id.len)
    return FALSE;

  if (memcmp (file_0->file_id.hash, file_1->file_id.hash, TNT_FILE_ID_LEN))
    return FALSE;

  return TRUE;
}

/* This is used in conjunction with a GHashTable _only_, and therefore returns
 * TRUE/FALSE, not -1/0/1 */
static gint
compare_files_dir_and_name (const FileStat *file_0, const FileStat *file_1)
{
  if (file_0->dir != file_1->dir)
    return FALSE;  /* No need to differentiate < and > */

  if (strcmp (file_0->name, file_1->name))
    return FALSE;

  return TRUE;
}

static DirStat *
find_dir_stat (TntFileDB *file_db, const gchar *path)
{
  GNode *dir_node;

  dir_node = find_dir_node (file_db->dir_tree, path);
  if (dir_node)
    return dir_node->data;

  return NULL;
}

static gboolean
unscanned_dir_traverse (GNode *node, gpointer data)
{
  TreeSearchContext *context  = data;
  DirStat           *dir_stat = node->data;

  if (dir_stat && dir_stat->shared && !dir_stat->scanned)
  {
    context->found_node = node;
    return TRUE;
  }

  return FALSE;
}

static GNode *
find_first_unscanned_dir_node (TntFileDB *file_db)
{
  TreeSearchContext context;

  memset (&context, 0, sizeof (context));
  g_node_traverse (file_db->dir_tree, G_LEVEL_ORDER, G_TRAVERSE_ALL,
                   -1, unscanned_dir_traverse, &context);

  return context.found_node;
}

static gboolean
unstatted_dir_traverse (GNode *node, gpointer data)
{
  TreeSearchContext *context  = data;
  DirStat           *dir_stat = node->data;

  if (dir_stat && dir_stat->shared && !dir_stat->statted)
  {
    context->found_node = node;
    return TRUE;
  }

  return FALSE;
}

static GNode *
find_first_unstatted_dir_node (TntFileDB *file_db)
{
  TreeSearchContext context;

  memset (&context, 0, sizeof (context));
  g_node_traverse (file_db->dir_tree, G_LEVEL_ORDER, G_TRAVERSE_ALL,
                   -1, unstatted_dir_traverse, &context);

  return context.found_node;
}

static gboolean
unstat_dir_traverse (GNode *node, gpointer data)
{
  DirStat *dir_stat = node->data;

  if (dir_stat)
    dir_stat->statted = FALSE;

  return FALSE;
}

static void
unstat_all_dirs (TntFileDB *file_db)
{
  g_node_traverse (file_db->dir_tree, G_IN_ORDER, G_TRAVERSE_ALL,
                   -1, unstat_dir_traverse, NULL);
}

static void
invalidate_files_in_dir (TntFileDB *file_db, DirStat *dir_stat)
{
  GList *l;

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
    FileStat *file_stat = l->data;

    if (file_stat->dir->data == dir_stat)
      file_stat->scanned = FALSE;
  }
}

static FileStat *
find_file_stat_by_id (TntFileDB *file_db, const TntFileID *file_id)
{
  GList    *result_node;
  FileStat  file_stat;

  memset (&file_stat, 0, sizeof (file_stat));
  memcpy (&file_stat.file_id, file_id, sizeof (TntFileID));

  result_node = g_hash_table_lookup (file_db->files_by_id, &file_stat);
  if (result_node)
    return result_node->data;

  return NULL;
}

static FileStat *
find_file_stat_by_dir_and_name (TntFileDB *file_db, GNode *dir_node, gchar *name)
{
  GList    *result_node;
  FileStat  file_stat;

  memset (&file_stat, 0, sizeof (file_stat));
  file_stat.name = name;
  file_stat.dir  = dir_node;

  result_node = g_hash_table_lookup (file_db->files_by_location, &file_stat);
  if (result_node)
    return result_node->data;

  return NULL;
}

static void
propagate_stats_upwards (TntFileDB *file_db, GNode *dir_node,
                         gint hashed_file_count_change, gint64 hashed_file_size_change,
                         gint file_count_change, gint64 file_size_change)
{
  file_db->file_hashed_count += hashed_file_count_change;
  file_db->file_hashed_size  += hashed_file_size_change;
  file_db->file_count        += file_count_change;
  file_db->file_size         += file_size_change;

  g_assert (file_db->file_hashed_count >= 0);
  g_assert (file_db->file_hashed_count <= file_db->file_count);
  g_assert (file_db->file_hashed_size >= 0);
  g_assert (file_db->file_hashed_size <= file_db->file_size);

  for ( ; dir_node; dir_node = dir_node->parent)
  {
    DirStat *dir_stat = dir_node->data;

    if (!dir_stat)
      continue;

    dir_stat->files_hashed_count += hashed_file_count_change;
    dir_stat->files_count        += file_count_change;
    dir_stat->files_size         += file_size_change;

    g_assert (dir_stat->files_hashed_count >= 0);
    g_assert (dir_stat->files_hashed_count <= dir_stat->files_count);

    if (file_db->handle_change_func)
      file_db->handle_change_func (file_db, (TntFileDBHandle) dir_node, file_db->handle_change_func_data);
  }
}

static GList *
find_first_unhashed_file_node (TntFileDB *file_db)
{
  GList *l;

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
    FileStat             *file_stat = l->data;
    SerializationContext *ser_context;
    gchar                *path;
    gchar                *full_path;
    guint64               file_len;
    TntStat               tstat;

    if (file_stat->active != TRUE || file_stat->hashed != FALSE)
      continue;

    /* Only hash files that have been unchanged for N seconds */
    if (file_stat->change_time + STABLE_FILE_TIMEOUT_S >= tnt_timestamp ())
    {
      file_db->need_stabilization = TRUE;
      continue;
    }

    g_assert (file_stat->dir);

    /* Don't hash files that are being serialized. I'm not sure this check is really
     * needed, since the scanner skips files that look like serialization targets. */

    path = build_dir_path (file_stat->dir);
    full_path = g_build_path (G_DIR_SEPARATOR_S, path, file_stat->name, NULL);

    ser_context = find_serialization_context_by_path_and_name (file_db, path, file_stat->name);

    g_free (path);

    if (ser_context)
    {
      g_free (full_path);
      continue;
    }

    /* Check that file still exists */
    if (!tnt_stat (full_path, &tstat))
    {
      file_stat->active = FALSE;
      g_free (full_path);
      continue;
    }

    g_free (full_path);

    /* Make sure it didn't change */
    if (file_stat->file_id.len == tstat.len)
      return l;

    /* If it did, update known timestamp & size */
    file_stat->change_time = DETERMINE_CHANGE_TIME (&tstat);

    propagate_stats_upwards (file_db, file_stat->dir, 0, 0, 0, (gint64) tstat.len - (gint64) file_stat->file_id.len);
    file_stat->file_id.len = tstat.len;

    file_db->need_stabilization = TRUE;
  }

  return NULL;
}

static FileStat *
make_file_stat (GNode *parent_dir, gchar *name, guint64 len)
{
  FileStat *file_stat;

  file_stat = g_new0 (FileStat, 1);
  file_stat->name        = g_strdup (name);
  file_stat->file_id.len = len;
  file_stat->dir         = parent_dir;
  file_stat->active      = TRUE;
  file_stat->scanned     = TRUE;
  file_stat->hashed      = FALSE;

  return file_stat;
}

static void
free_file_stat (FileStat *file_stat)
{
  g_free (file_stat->name);
  g_free (file_stat);
}

static GList *
build_top_dir_node_list (GNode *root)
{
  GList *top_list = NULL;
  GNode *node;

  if (((DirStat *) root->data)->shared)
    return g_list_append (top_list, root);

  for (node = g_node_first_child (root); node; node = g_node_next_sibling (node))
  {
    DirStat *dir_stat = node->data;

    if (dir_stat->shared)
    {
      top_list = g_list_append (top_list, node);
    }
    else
    {
      top_list = g_list_concat (top_list, build_top_dir_node_list (node));
    }
  }

  return top_list;
}

static gboolean
add_dir_tree_traverse (GNode *node, gpointer data)
{
  TntFileDB *file_db = data;
  DirStat   *dir_stat = node->data;

  if (dir_stat)
  {
    dir_stat->shared = TRUE;

    if (file_db->handle_add_func)
      file_db->handle_add_func (file_db, (TntFileDBHandle) node, file_db->handle_add_func_data);
  }

  return FALSE;
}

static gboolean
unset_scanned_all_traverse (GNode *node, gpointer data)
{
  DirStat *dir_stat = node->data;

  if (dir_stat && !dir_stat->shared)
    dir_stat->scanned_all = FALSE;

  return FALSE;
}

static void
add_dir_to_user_list (TntFileDB *file_db, const gchar *path)
{
  GNode   *node;
  DirStat *dir_stat;

  node     = ensure_dir_node (file_db->dir_tree, path);
  dir_stat = node->data;

  /* If it's not shared at this point, it was just created */
  if (!dir_stat->shared)
  {
    GList *old_top_dirs, *l;

    /* Remove old top dirs encompassed by this share */
    old_top_dirs = build_top_dir_node_list (node);
    for (l = g_list_last (old_top_dirs); l; l = g_list_previous (l))
    {
      if (file_db->handle_remove_func)
        file_db->handle_remove_func (file_db, (TntFileDBHandle) l->data, file_db->handle_remove_func_data);
    }
    g_list_free (old_top_dirs);

    /* Unset scanned_all in subdirs. Subdirs can exist because a deeper share could have been
     * unshared, and our share is added towards the top of the lingering path */
    g_node_traverse (node, G_IN_ORDER, G_TRAVERSE_ALL, -1, unset_scanned_all_traverse, NULL);

    /* Re-add directory tree */
    g_node_traverse (node, G_PRE_ORDER, G_TRAVERSE_ALL, -1, add_dir_tree_traverse, file_db);

    /* Unset scanned_all from here on upwards. Note that this destroys node pointer */
    for ( ; node; node = node->parent)
    {
      DirStat *here_stat = node->data;
      if (!here_stat)
        break;

      here_stat->scanned_all = FALSE;
    }
  }

  DEBUG (g_print ("Added dir has %d hashed files and %d regular%s.\n",
                  dir_stat->files_hashed_count, dir_stat->files_count,
                  dir_stat->scanned ? ", and is scanned" : "");)
}

static guint
hash_file_stat_location (FileStat *file_stat)
{
  GNode  *dir_node;
  gchar  *path;
  guint   val;

  dir_node = file_stat->dir;

  path = build_dir_path (dir_node);
  val  = g_str_hash (file_stat->name);
  val ^= g_str_hash (path);
  g_free (path);

  return val;
}

static guint
hash_file_stat_id (FileStat *file_stat)
{
  guint32 val = 0;
  gint    i;

  for (i = 0; i < sizeof (file_stat->file_id.hash); i++)
  {
    guint32 t;

    val ^= ((const guchar *) file_stat->file_id.hash) [i];
    t = val & 0xf0000000;
    val <<= 4;
    val |= t >> 28;
  }

  for (i = 0; i < sizeof (file_stat->file_id.len); i++)
  {
    guint32 t;

    val ^= ((const guchar *) &file_stat->file_id.len) [i];
    t = val & 0xf0000000;
    val <<= 4;
    val |= t >> 28;
  }

  return val;
}

static void
emit_handle_change_upwards (TntFileDB *file_db, GNode *dir_node)
{
  if (!file_db->handle_change_func)
    return;

  for ( ; dir_node; dir_node = dir_node->parent)
  {
    file_db->handle_change_func (file_db, (TntFileDBHandle) dir_node, file_db->handle_change_func_data);
  }
}

static void
add_file_to_list (TntFileDB *file_db, FileStat *file_stat)
{
  DEBUG (g_print ("Adding file (size %lld)\n", file_stat->file_id.len));

  file_db->file_list = g_list_prepend (file_db->file_list, file_stat);
  /* Using file_db->file_list here depends on g_list_prepend() above */
  g_hash_table_insert (file_db->files_by_location, file_stat, file_db->file_list);

  if (!file_db->file_list_last)
    file_db->file_list_last = file_db->file_list;

  propagate_stats_upwards (file_db, file_stat->dir, 0, 0, 1, file_stat->file_id.len);
}

static void
remove_file_node_from_list (TntFileDB *file_db, GList *node)
{
  FileStat *file_stat;

  file_stat = node->data;

  DEBUG (g_print ("Removing file (size %lld)", file_stat->file_id.len));

  g_hash_table_remove (file_db->files_by_location, file_stat);

  /* We may be doing something with this file - say, hashing it. The callbacks will look
   * for (file_db->stat_node == NULL) and take appropriate action. It's ok. Really! */
  if (node == file_db->stat_node)
  {
    DEBUG (g_print (" (running)"));
    if (file_db->state == TNT_FILE_DB_HASHING && file_db->file_io)
    {
      guint64 pos;

      pos = flux_file_io_get_position (file_db->file_io);
      propagate_stats_upwards (file_db, file_stat->dir, 0, -pos, 0, 0);
    }
    file_db->stat_node = NULL;
  }
  else if (file_stat->hashed)
  {
    DEBUG (g_print (" (hashed)"));
    g_hash_table_remove (file_db->files_by_id, file_stat);
    propagate_stats_upwards (file_db, file_stat->dir, -1, -file_stat->file_id.len, 0, 0);
  }

  DEBUG (g_print ("\n"));

  propagate_stats_upwards (file_db, file_stat->dir, 0, 0, -1, -file_stat->file_id.len);
  free_file_stat (file_stat);

  if (file_db->file_list_last == node)
    file_db->file_list_last = g_list_previous (node);

  file_db->file_list = g_list_remove_link (file_db->file_list, node);
  g_list_free_1 (node);
}

static void
end_dir_read (TntFileDB *file_db)
{
  if (file_db->dir_hd)
  {
    closedir (file_db->dir_hd);
    file_db->dir_hd    = NULL;
  }
}

static gboolean
remove_dir_pre_traverse (GNode *node, gpointer data)
{
  TntFileDB *file_db = data;

  if (file_db->handle_remove_func)
    file_db->handle_remove_func (file_db, (TntFileDBHandle) node, file_db->handle_remove_func_data);

  return FALSE;
}

static gboolean
remove_dir_traverse (GNode *node, gpointer data)
{
  TntFileDB *file_db  = data;
  DirStat   *dir_stat = node->data;

  /* We may be doing something with this dir - say, scanning it. The callbacks
   * will look for (file_db->stat_node == NULL) and take appropriate action.
   * It's ok. Really! */
  if (node == (GNode *) file_db->stat_node)
  {
    end_dir_read (file_db);
    file_db->stat_node = NULL;
  }

  file_db->dir_count--;
  free_dir_stat (dir_stat);
  return FALSE;
}

static void
remove_dir_node (TntFileDB *file_db, GNode *top_dir_node)
{
  GList   *l, *next;

  /* Remove files in tree */
  for (l = file_db->file_list; l; l = next)
  {
    FileStat *file_stat = l->data;
    GNode    *dir_node  = file_stat->dir;

    next = g_list_next (l);

    if (top_dir_node == dir_node || g_node_is_ancestor (top_dir_node, dir_node))
      remove_file_node_from_list (file_db, l);
  }

  /* Emit removal callbacks */
  g_node_traverse (top_dir_node, G_IN_ORDER, G_TRAVERSE_ALL,
                   -1, remove_dir_pre_traverse, file_db);

  /* Free dir stats */
  g_node_traverse (top_dir_node, G_IN_ORDER, G_TRAVERSE_ALL,
                   -1, remove_dir_traverse, file_db);

  /* And unlink */
  g_node_destroy (top_dir_node);

  if (top_dir_node == file_db->dir_tree)
  {
    file_db->dir_tree       = g_node_new (NULL);
    file_db->dir_tree->data = make_dir_stat ("/");
  }
}

static gchar *
build_path_to_file (FileStat *file_stat)
{
  gchar *dir_path;
  gchar *full_path;

  dir_path = build_dir_path (file_stat->dir);
  full_path = g_strjoin (G_DIR_SEPARATOR_S, dir_path, file_stat->name, NULL);
  g_free (dir_path);

  return full_path;
}

static void
remove_invalid_files (TntFileDB *file_db)
{
  GList *l, *next;

  for (l = file_db->file_list; l; l = next)
  {
    FileStat *file_stat = l->data;

    next = g_list_next (l);

    if (!file_stat->scanned)
      remove_file_node_from_list (file_db, l);
  }
}

static void
remove_dir_path (TntFileDB *file_db, const gchar *top_path)
{
  GNode *top_dir_node;

  top_dir_node = find_dir_node (file_db->dir_tree, top_path);
  if (!top_dir_node)
  {
    g_warning ("Tried to remove non-existent path.");
    return;
  }

  remove_dir_node (file_db, top_dir_node);
}

static GList *
build_top_dir_list (GNode *root)
{
  GList *top_list, *l;

  top_list = build_top_dir_node_list (root);
  for (l = top_list; l; l = g_list_next (l))
  {
    GNode *node = l->data;
    l->data = build_dir_path (node);
  }

  return top_list;
}

/* --- Idle activities --- */

static gboolean
idle_timeout_cb (gpointer data)
{
  TntFileDB    *file_db = data;
  GNode        *dir_node;
  DirStat      *dir_stat;
  gchar        *dir_path;
  TntStat       tstat;
  TntTimeStamp  current_modification_time;

  assert_types (file_db);

  if (file_db->stabilization_time != 0 &&
      file_db->stabilization_time <= tnt_timestamp ())
  {
    file_db->need_hash          = TRUE;
    file_db->need_stabilization = FALSE;
    end_state (file_db);
    return TRUE;
  }

  dir_node = find_first_unstatted_dir_node (file_db);
  if (!dir_node)
  {
    unstat_all_dirs (file_db);
    dir_node = find_first_unstatted_dir_node (file_db);
  }

  if (!dir_node)
    return TRUE;

  dir_stat = dir_node->data;
  dir_stat->statted = TRUE;
  dir_path = build_dir_path (dir_node);

  if (!tnt_lstat (dir_path, &tstat))
  {
    /* Directory could not be found - remove it and its files. */
    remove_dir_node (file_db, dir_node);
  }
  else if (tstat.mtime != dir_stat->last_modification_time)
  {
    /* Children were added or removed */

    dir_stat->scanned  = FALSE;
    file_db->need_scan = TRUE;
    invalidate_files_in_dir (file_db, dir_stat);
    dir_stat->last_modification_time = tstat.mtime;

    /* Do a scan to see what changed. Files that are still there will be
     * re-validated, new files will be added, invalid files will be removed
     * when the scan ends. */
    end_state (file_db);
  }

  g_free (dir_path);
  return TRUE;
}

/* --- Filesystem scanner --- */

static void
id_hint_cb (TntFileID *file_id, TntFileDB *file_db)
{
  GList    *file_node = file_db->file_list;
  FileStat *file_stat;

  DEBUG (g_print ("id_hint_cb. File %sfound.\n", file_id ? "" : "not "));

  /* Reconnect scan cb */
  file_db->callback_source_id = g_idle_add_full (G_PRIORITY_LOW + 5, scan_cb, file_db, NULL);

  if (!file_node)
  {
    DEBUG (g_print ("No file node.\n"));
    flux_object_unref (FLUX_OBJECT (file_db));
    return;
  }

  file_stat = file_node->data;

  if (!file_id || file_stat != file_db->id_hint_file_stat || file_stat->file_id.len != file_id->len)
  {
    DEBUG (g_print ("Need hash.\n"));
    file_db->need_hash = TRUE;
    flux_object_unref (FLUX_OBJECT (file_db));
    return;
  }

  DEBUG (g_print ("Adding pre-hashed file.\n"));

  memcpy (&file_stat->file_id, file_id, sizeof (TntFileID));

  propagate_stats_upwards (file_db, file_stat->dir, 1, file_stat->file_id.len, 0, 0);
  file_stat->hashed = TRUE;
  g_hash_table_insert (file_db->files_by_id, file_stat, file_node);

  /* Reattach at end of file list */
  file_db->file_list = g_list_remove_link (file_db->file_list, file_db->file_list);
  if (file_db->file_list_last != file_node)
  {
    g_list_concat (file_db->file_list_last, file_node);
    file_db->file_list_last = file_node;
  }

  if (!file_db->file_list)
    file_db->file_list = file_db->file_list_last;

  /* Notify user of new file */
  if (file_db->file_update_func)
    file_db->file_update_func (file_db, &file_stat->file_id, file_db->file_update_func_data);

  flux_object_unref (FLUX_OBJECT (file_db));
}

/* Returns -1 if all dirs have been scanned and there is nothing left to do
 * in this state, 0 if there are more dirs to scan, and 1 if the idle callback
 * should be disconnected (pausing the scan). */
static gint
scan_step (TntFileDB *file_db)
{
  TntStat        tstat;
  struct dirent *dent;
  GNode         *dir_node;
  DirStat       *dir_stat;
  gchar         *dir_path;
  gchar         *full_path;

  if (!file_db->stat_node)
    file_db->stat_node = (gpointer) find_first_unscanned_dir_node (file_db);

  if (!file_db->stat_node)
  {
    /* Couldn't find an unscanned dir. We're done! */
    return -1;
  }

  dir_node = ((GNode *) file_db->stat_node);
  dir_stat = dir_node->data;
  dir_path = build_dir_path (dir_node);

  if (!file_db->dir_hd)
  {
    gchar *dir_path_utf8;

    dir_path_utf8 = tnt_filename_to_utf8_lossy (dir_path);
    tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, 0, "Scanning '%s'.", dir_path_utf8);
    g_free (dir_path_utf8);

    file_db->dir_hd = opendir (dir_path);
  }

  if (!file_db->dir_hd)
  {
    /* Couldn't open this dir. Mark it inactive. */
    file_db->stat_node = NULL;
    dir_stat->scanned  = TRUE;
    dir_stat->active   = FALSE;
    g_free (dir_path);
    return 0;
  }

  dent = readdir (file_db->dir_hd);
  if (!dent)
  {
    GNode *this_node;

    /* End-of-dir reached. This dir completed, but there could be others. */
    end_dir_read (file_db);
    file_db->stat_node = NULL;
    dir_stat->scanned  = TRUE;
    dir_stat->active   = TRUE;
    g_free (dir_path);

    /* If this dir is a leaf, propagate scanned_all flag upwards */
    for (this_node = dir_node; this_node; this_node = this_node->parent)
    {
      DirStat *this_stat = this_node->data;
      GNode   *child_node;

      if (!this_stat || this_stat->scanned_all || !this_stat->scanned)
        break;

      for (child_node = g_node_first_child (this_node);
           child_node;
           child_node = g_node_next_sibling (child_node))
      {
        DirStat *child_stat = child_node->data;

        if (!child_stat || !child_stat->scanned_all)
          break;
      }

      if (child_node)
        break;

      this_stat->scanned_all = TRUE;

      if (file_db->handle_change_func)
        file_db->handle_change_func (file_db, (TntFileDBHandle) this_node, file_db->handle_change_func_data);
    }

    return 0;
  }

  /* Don't include entries that start in a dot. This takes care of ".", "..",
   * files that are currently being assembled, and other temporary/hidden entries
   * on Unix. Also skip entries that are on our global black list, like
   * "resource.frk" and "thumbs.db". */
  if (dent->d_name [0] == '.' || should_skip (dent->d_name))
  {
    g_free (dir_path);
    return 0;
  }

#ifdef G_PLATFORM_WIN32
  /* Windows doesn't like filenames with a leading dot, so files that are
   * being assembled don't have it. Look for a valid file id string instead. */
  if (tnt_is_valid_file_id_string (dent->d_name))
  {
    g_free (dir_path);
    return 0;
  }
#endif

  full_path = g_strjoin (G_DIR_SEPARATOR_S, dir_path, dent->d_name, NULL);

  if (!tnt_lstat (full_path, &tstat))
  {
    DEBUG (g_print ("Couldn't stat '%s'.\n", full_path));

    /* Couldn't stat this entry. Throw it away. */
    g_free (dir_path);
    g_free (full_path);
    return 0;
  }

  if (tstat.is_file && tstat.len > 4 && tnt_file_have_read_access (full_path))
  {
    /* Found a file. Hashing will eliminate differently-named duplicates later. */
    FileStat *file_stat;

    file_stat = find_file_stat_by_dir_and_name (file_db, dir_node, dent->d_name);
    if (file_stat)
    {
      DEBUG (g_print ("Found known file.\n"));

      /* We've seen this file (path) before. */
      if (file_stat->file_id.len != tstat.len)
      {
        DEBUG (g_print ("  Changed!\n"));

        /* File size changed! We need to rehash this. */
        if (file_stat->hashed)
          propagate_stats_upwards (file_db, file_stat->dir,
                                   -1, -file_stat->file_id.len,
                                   0, (gint64) tstat.len - (gint64) file_stat->file_id.len);
        else
          propagate_stats_upwards (file_db, file_stat->dir,
                                   0, 0,
                                   0, (gint64) tstat.len - (gint64) file_stat->file_id.len);

        file_stat->file_id.len = tstat.len;
        file_stat->change_time = DETERMINE_CHANGE_TIME (&tstat);
        file_stat->hashed      = FALSE;
        file_db->need_hash     = TRUE;
      }
      file_stat->scanned = TRUE;
    }
    else
    {
      DEBUG (g_print ("Found '%s' is %lld bytes long.\n", full_path, tstat.len));

      /* NOTE: Some duplication of code from hash_eof_cb () */

      /* This is new to us. */
      file_stat = make_file_stat (dir_node, dent->d_name, tstat.len);
      file_stat->change_time = DETERMINE_CHANGE_TIME (&tstat);
      add_file_to_list (file_db, file_stat);

      /* Ask user for cached ID. */
      if (file_db->file_id_hint_func)
      {
        /* Unreffed in id_hint_cb () */
        flux_object_ref (FLUX_OBJECT (file_db));

        file_db->id_hint_file_stat = file_stat;

        file_db->file_id_hint_func (file_db, dir_path, file_stat->name, tstat.len,
                                    tstat.mtime, (TntFileIDFunc *) id_hint_cb, file_db,
                                    file_db->file_id_hint_func_data);

        g_free (dir_path);
        g_free (full_path);

        /* Disconnect scan handler; reconnected in file ID callback */
        return 1;
      }

      file_db->need_hash = TRUE;
    }
  }
  else if (tstat.is_dir && dir_stat->shared && tnt_dir_have_read_access (full_path) &&
           strlen (full_path) < TNT_PATH_NAME_LEN && !find_dir_stat (file_db, full_path))
  {
    /* Found a subdir, and we are adding subdirs. */
    GNode   *sub_dir_node;
    DirStat *sub_dir_stat;

    sub_dir_node = ensure_dir_node (file_db->dir_tree, full_path);
    sub_dir_stat = sub_dir_node->data;
    sub_dir_stat->shared = TRUE;
    sub_dir_stat->last_modification_time = tstat.mtime;

    if (file_db->handle_add_func)
      file_db->handle_add_func (file_db, (TntFileDBHandle) sub_dir_node, file_db->handle_add_func_data);
  }

  g_free (dir_path);
  g_free (full_path);
  return 0;
}

static gboolean
scan_cb (gpointer data)
{
  TntFileDB *file_db = data;
  gint       result;

  assert_types (file_db);

  result = scan_step (file_db);

  if (result == 1)
  {
    /* Pause */
    return FALSE;
  }
  else if (result == -1)
  {
    /* End */
    file_db->need_scan = FALSE;
    end_state (file_db);
  }

  /* Continue */
  return TRUE;
}

/* --- File hashing --- */

static void
hash_eof_cb (FluxFileIO *file_io, guint len_left, gpointer data)
{
  TntFileDB *file_db = data;
  gchar     *buf;
  guchar     hash [20];
  FileStat  *file_stat;

  assert_types (file_db);

  DEBUG (g_print ("Got %d bytes (EOF)\n", len_left));

#ifdef G_PLATFORM_WIN32
  /* Hack to work around GLib brokenness on win32 */
  flux_io_set_read_func (FLUX_IO (file_io), NULL, NULL);
  flux_dual_fifo_set_read_hook (FLUX_DUAL_FIFO (file_io), NULL, NULL);
#endif

  /* As misfortune would have it, the file can be removed from the list as
   * it's being hashed. In that case, file_db->stat_node will be set to NULL
   * and the file and its length will be subtracted from the hashed-file tally.
   * This means we have to add the last numbers to the tally to get it to
   * match up, although storing the information we have so laboriously obtained
   * is pointless, and indeed, impossible. */

  if (file_db->stat_node)
  {
    guint64 pos;

    file_stat = file_db->stat_node->data;

    buf = g_malloc (len_left);
    flux_dual_fifo_read (FLUX_DUAL_FIFO (file_io), buf, len_left);
    flux_rmd160_add (TNT_DATA_DB (file_db)->rmd_160, buf, len_left);
    g_free (buf);

    flux_rmd160_end      (TNT_DATA_DB (file_db)->rmd_160);
    flux_rmd160_get_hash (TNT_DATA_DB (file_db)->rmd_160, hash);

    tnt_fold_hash (hash, file_stat->file_id.hash, TNT_FILE_ID_LEN);
    file_stat->hashed = TRUE;

    /* Reattach at end of list, so future searches for unhashed files will not have
     * to traverse over hashed files */
    file_db->file_list = g_list_remove_link (file_db->file_list, file_db->stat_node);
    if (file_db->file_list_last != file_db->stat_node)
    {
      g_list_concat (file_db->file_list_last, file_db->stat_node);
      file_db->file_list_last = file_db->stat_node;
    }

    if (!file_db->file_list)
      file_db->file_list = file_db->file_list_last;

    /* Count the file as hashed */
    pos = flux_file_io_get_position (file_io);
    if (pos > file_stat->file_id.len)
    {
      propagate_stats_upwards (file_db, file_stat->dir, 1, len_left, 0, pos - file_stat->file_id.len);
      file_stat->file_id.len = pos;
    }
    else
      propagate_stats_upwards (file_db, file_stat->dir, 1, len_left, 0, 0);

    g_hash_table_insert (file_db->files_by_id, file_stat, file_db->stat_node);

    /* Notify master of new files as they show up */
    if (file_db->file_update_func)
    {
      file_db->file_update_func (file_db, &file_stat->file_id, file_db->file_update_func_data);
    }
  }

  flux_object_unref (FLUX_OBJECT (file_io));
  g_source_remove (file_db->report_timeout_id);

  file_db->stat_node           = NULL;
  file_db->file_io             = NULL;
  file_db->report_timeout_id   = 0;

  end_state (file_db);
}

static void
hash_error_cb (FluxObject *object, GError *error, gpointer data)
{
  FluxFileIO *file_io = FLUX_FILE_IO (object);
  TntFileDB  *file_db = data;

  /* We got a read error while hashing the file. It may be locked, or there
   * was a hardware failure. */

  assert_types (file_db);

  DEBUG (g_print ("Got I/O error during file hash.\n"));

#ifdef G_PLATFORM_WIN32
  /* Hack to work around GLib brokenness on win32 */
  flux_io_set_read_func (FLUX_IO (file_io), NULL, NULL);
  flux_dual_fifo_set_read_hook (FLUX_DUAL_FIFO (file_io), NULL, NULL);
#endif

  if (file_db->stat_node)
  {
    FileStat *file_stat = file_db->stat_node->data;
    guint64   pos;

    /* Reattach at end of list, so we can hash files that don't have errors */
    file_db->file_list = g_list_remove_link (file_db->file_list, file_db->stat_node);
    if (file_db->file_list_last != file_db->stat_node)
    {
      g_list_concat (file_db->file_list_last, file_db->stat_node);
      file_db->file_list_last = file_db->stat_node;
    }

    if (!file_db->file_list)
      file_db->file_list = file_db->file_list_last;

    /* Uncount the length we've hashed */
    pos = flux_file_io_get_position (file_io);
    DEBUG (g_print ("Got I/O error at position %lld.\n", pos));
    propagate_stats_upwards (file_db, file_stat->dir, 0, -pos, 0, 0);
  }

  flux_object_unref (FLUX_OBJECT (file_io));
  g_source_remove (file_db->report_timeout_id);

  file_db->stat_node         = NULL;
  file_db->file_io           = NULL;
  file_db->report_timeout_id = 0;

  g_error_free (error);

  end_state (file_db);
}

/* Timer callback */
static gboolean
hash_report_progress_cb (gpointer data)
{
  TntFileDB *file_db = data;
  FileStat  *file_stat;
  gchar     *full_path_utf8;

  assert_types (file_db);

  if (!file_db->stat_node || file_db->state != TNT_FILE_DB_HASHING)
    return TRUE;

  file_stat = file_db->stat_node->data;
  full_path_utf8 = tnt_filename_to_utf8_replace (build_path_to_file (file_stat));

  tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, TNT_LOG_FOLLOWUP,
           "Hashing '%s' (%d%%).", full_path_utf8,
           (flux_file_io_get_position (file_db->file_io) * 100) / file_stat->file_id.len);

  g_free (full_path_utf8);
  return TRUE;
}

static void
hash_read_cb (FluxIO *io, guint min_len, gpointer data)
{
  GList     *file_node = NULL;
  FileStat  *file_stat = NULL;
  TntFileDB *file_db = data;
  gchar     *buf;
  guint64    pos;

  buf = g_malloc (min_len);
  flux_dual_fifo_read (FLUX_DUAL_FIFO (io), buf, min_len);
  flux_rmd160_add (TNT_DATA_DB (file_db)->rmd_160, buf, min_len);
  g_free (buf);

  file_node = file_db->stat_node;
  if (file_node)
    file_stat = file_node->data;

  if (file_stat)
  {
    pos = flux_file_io_get_position (FLUX_FILE_IO (io));
    if (pos > file_stat->file_id.len)
    {
      propagate_stats_upwards (file_db, file_stat->dir, 0, min_len, 0, pos - file_stat->file_id.len);
      file_stat->file_id.len = pos;
    }
    else
      propagate_stats_upwards (file_db, file_stat->dir, 0, min_len, 0, 0);
  }
  else
  {
    /* The file was removed while hashing */

#ifdef G_PLATFORM_WIN32
    /* Hack to work around GLib brokenness on win32 */
    flux_io_set_read_func (io, NULL, NULL);
    flux_dual_fifo_set_read_hook (FLUX_DUAL_FIFO (io), NULL, NULL);
#endif

    flux_object_unref (FLUX_OBJECT (io));
    file_db->file_io = NULL;

    g_source_remove (file_db->report_timeout_id);

    end_state (file_db);
  }
}

/* Returns TRUE when nothing left to do. */
static gboolean
hash_step (TntFileDB *file_db)
{
  FileStat *file_stat;

  DEBUG (g_print ("In hash_step().\n"));

  if (!file_db->stat_node)
    file_db->stat_node = find_first_unhashed_file_node (file_db);

  if (!file_db->stat_node)
  {
    /* Couldn't find an unhashed file. We're done! */
    DEBUG (g_print ("No unhashed nodes.\n"));
    return TRUE;
  }

  file_stat = file_db->stat_node->data;
  emit_handle_change_upwards (file_db, file_stat->dir);

  if (!file_db->file_io)
  {
    gchar  *full_path, *full_path_utf8;
    gint    fd;

    full_path = build_path_to_file (file_stat);
    full_path_utf8 = tnt_filename_to_utf8_lossy (full_path);

    DEBUG (g_print ("Starting hash for '%s'.\n", full_path));

    tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, 0, "Hashing '%s'.",
             full_path_utf8);
    g_free (full_path_utf8);

    fd = tnt_open (full_path, FALSE, FALSE, TRUE);
    g_free (full_path);

    if (fd < 0)
    {
      DEBUG (g_print ("Failed!\n"));

      /* remove_file_node_from_list () sets file_db->stat_node to NULL */
      remove_file_node_from_list (file_db, file_db->stat_node);
      file_db->need_scan = TRUE;
      end_state (file_db);
      return FALSE;
    }

    DEBUG (g_print ("Ok.\n"));

    flux_rmd160_begin (TNT_DATA_DB (file_db)->rmd_160);

    file_db->file_io = flux_file_io_new (fd);
    flux_io_set_read_threshold  (FLUX_IO (file_db->file_io), 1024);
    flux_io_set_read_func       (FLUX_IO (file_db->file_io), hash_read_cb, file_db);
    flux_file_io_set_eof_func   (file_db->file_io, hash_eof_cb, file_db);
    flux_file_io_set_error_func (file_db->file_io, hash_error_cb, file_db);

    file_db->report_timeout_id = g_timeout_add (1000, hash_report_progress_cb, file_db);
  }

  return FALSE;
}

static void
hash_next_file (TntFileDB *file_db)
{
  assert_types (file_db);

  if (hash_step (file_db))
  {
    file_db->need_hash = FALSE;
    end_state (file_db);
  }
}

/* --- State machine --- */

/* Callback from TntBlockAction */
static void
end_block_action (TntDataDB *data_db, gboolean success)
{
  TntFileDB      *file_db = TNT_FILE_DB (data_db);
  TntBlockAction *block_action;

  assert_types (file_db);

  block_action = file_db->current_block_action;
  g_assert (block_action);

  flux_object_unref (FLUX_OBJECT (block_action));
  file_db->current_block_action = NULL;
  end_state (file_db);
}

static void
begin_block_action (TntFileDB *file_db, TntBlockAction *block_action)
{
  FileStat             *file_stat;
  TntFileID             file_id;
  gchar                *file_path;
  gchar                *file_name_utf8;
  FluxFileIO           *file_io;
  gint                  fd;
  SerializationContext *ser_context;

  g_assert (block_action);
  g_assert (file_db->current_block_action == NULL);

  file_db->current_block_action = block_action;

  /* Create and position a FluxFileIO for tnt_block_action_run (). */

  switch (tnt_block_action_get_type (block_action))
  {
    case TNT_BLOCK_ACTION_GET_BLOCK_HASH:
    case TNT_BLOCK_ACTION_GET_BLOCK:
      tnt_block_action_get_file_info (block_action, &file_id);
      file_stat = find_file_stat_by_id (file_db, &file_id);
      if (!file_stat)
      {
        tnt_block_action_cancel (block_action, TNT_DATA_DB (file_db));
        return;
      }

      file_name_utf8 = tnt_filename_to_utf8_lossy (file_stat->name);
      tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, 0,
               tnt_block_action_get_type (block_action) == TNT_BLOCK_ACTION_GET_BLOCK_HASH ?
               "Hashing block %d of file '%s'." : "Reading block %d of file '%s'.",
               tnt_block_action_get_block_id (block_action), file_name_utf8);
      g_free (file_name_utf8);

      file_path = build_path_to_file (file_stat);
      fd = tnt_open (file_path, FALSE, FALSE, TRUE);
      g_free (file_path);
      if (fd < 0)
      {
        DEBUG (g_print ("Block read action failed!\n"));
        tnt_block_action_cancel (block_action, TNT_DATA_DB (file_db));
        return;
      }

      file_io = flux_file_io_new (fd);
      flux_file_io_seek_to (file_io, (guint64) tnt_block_action_get_block_id (block_action) * (guint64) TNT_DATA_BLOCK_SIZE);
      tnt_block_action_run (block_action, TNT_DATA_DB (file_db), file_io);
      flux_object_unref (FLUX_OBJECT (file_io));
      break;

    case TNT_BLOCK_ACTION_PUT_BLOCK:
      tnt_block_action_get_file_info (block_action, &file_id);

      ser_context = find_serialization_context_by_file_id (file_db, &file_id);
      if (ser_context)
      {
        if (!tnt_path_create (ser_context->path))
        {
          /* FIXME: More user friendly alert */
          g_warning ("Can't create serialization target path.");
          return;
        }

        file_path = make_serialization_path_to_temp_file (ser_context);
        fd = tnt_open (file_path, TRUE, FALSE, FALSE);
        if (fd < 0)
          g_warning ("Couldn't write to '%s'!", file_path);

        g_free (file_path);
        if (fd < 0)
        {
          tnt_block_action_cancel (block_action, TNT_DATA_DB (file_db));
          return;
        }

        file_io = flux_file_io_new (fd);
        flux_file_io_seek_to (file_io, (guint64) tnt_block_action_get_block_id (block_action) * (guint64) TNT_DATA_BLOCK_SIZE);
        tnt_block_action_run (block_action, TNT_DATA_DB (file_db), file_io);
        flux_object_unref (FLUX_OBJECT (file_io));
      }
      else
      {
        g_warning ("Tried to put block to file not being serialized.");
        tnt_block_action_cancel (block_action, TNT_DATA_DB (file_db));
        return;
      }
      break;

    default:
      g_assert_not_reached ();
      break;
  }
}

/* Called on transition to new state */
static void
begin_state (TntFileDB *file_db, TntFileDBState state)
{
  TntBlockAction *action;

  file_db->state = state;

  switch (state)
  {
    case TNT_FILE_DB_IDLE:
      if (file_db->need_stabilization == TRUE)
        file_db->stabilization_time = tnt_timestamp () + STABLE_FILE_TIMEOUT_S / 2;
      else
        file_db->stabilization_time = 0;

      file_db->callback_source_id = g_timeout_add (IDLE_STAT_INTERVAL_MS, idle_timeout_cb, file_db);
      break;

    case TNT_FILE_DB_SCANNING:
      tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_LOW, TNT_LOG_MOOD_NEUTRAL, 0, "Scanning shares.");
      file_db->callback_source_id = g_idle_add_full (G_PRIORITY_LOW + 5, scan_cb, file_db, NULL);
      break;

    case TNT_FILE_DB_HASHING:
      hash_next_file (file_db);
      break;

    case TNT_FILE_DB_ACTION:
      g_assert (file_db->block_action_queue);

      action = file_db->block_action_queue->data;
      file_db->block_action_queue = g_list_delete_link (file_db->block_action_queue,
                                                        file_db->block_action_queue);

      begin_block_action (file_db, action);
      break;

    default:
      g_assert_not_reached ();
      break;
  }
}

/* Called when a state ends. Decides what to do next, and performs the
 * transition to the appropriate next state. */
static void
end_state (TntFileDB *file_db)
{
  TntFileDBState  next_state;

  DEBUG (g_print ("In end_state ().\n"));

  /* Determine next state */

  if (file_db->block_action_queue)
    next_state = TNT_FILE_DB_ACTION;
  else if (file_db->need_scan)
    next_state = TNT_FILE_DB_SCANNING;
  else if (file_db->need_hash)
    next_state = TNT_FILE_DB_HASHING;
  else
    next_state = TNT_FILE_DB_IDLE;

  /* Shut down current state and go to next state */

  switch (file_db->state)
  {
    case TNT_FILE_DB_IDLE:
      g_source_remove (file_db->callback_source_id);
      break;

    case TNT_FILE_DB_SCANNING:
      tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, 0,
               "Local file scan done.");

      g_source_remove (file_db->callback_source_id);
      remove_invalid_files (file_db);
      break;

    case TNT_FILE_DB_HASHING:
      tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_NEUTRAL, 0,
               "Done hashing files.");

      /* Hashing only keeps one FluxFileIO at a time, and when the last one's
       * destroyed (in hash_eof_cb), there is no async handler for this
       * state. Thus we do nothing. */
      break;

    case TNT_FILE_DB_ACTION:
      break;

    default:
      g_assert_not_reached ();
      break;
  }

  begin_state (file_db, next_state);
}

/* --- *
 * API *
 * --- */

static gboolean
have_file_full (TntDataDB *file_db, const TntFileID *file_id)
{
  if (find_file_stat_by_id (TNT_FILE_DB (file_db), file_id))
    return TRUE;

  return FALSE;
}

static gboolean
have_file_part (TntDataDB *file_db, const TntFileID *file_id)
{
  if (find_file_stat_by_id (TNT_FILE_DB (file_db), file_id))
    return TRUE;

  return FALSE;
}

static gboolean
have_block (TntDataDB *file_db, const TntFileID *file_id, gint block_id)
{
  if (find_file_stat_by_id (TNT_FILE_DB (file_db), file_id) &&
      (guint64) block_id * (guint64) TNT_DATA_BLOCK_SIZE < file_id->len)
    return TRUE;

  return FALSE;
}

static TntFileBitmap *
get_file_bitmap (TntDataDB *file_db, const TntFileID *file_id)
{
  if (have_file_full (file_db, file_id))
  {
    TntFileBitmap *file_bitmap = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len));
    tnt_file_bitmap_fill_true (file_bitmap);
    return file_bitmap;
  }

  return NULL;
}

static void
queue_block_action (TntFileDB *file_db, TntPriority priority, TntBlockAction *block_action)
{
  if (priority < TNT_PRIORITY_HIGH)
    file_db->block_action_queue = g_list_append (file_db->block_action_queue, block_action);
  else
    file_db->block_action_queue = g_list_prepend (file_db->block_action_queue, block_action);
  
  if (file_db->state == TNT_FILE_DB_IDLE)
  {
    end_state (file_db);
  }
}

static gboolean
get_block_hash (TntDataDB *file_db, const TntFileID *file_id, gint block_id,
                TntPriority priority,
                TntGotBlockHashFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  if (!have_block (file_db, file_id, block_id))
    return FALSE;

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_GET_BLOCK_HASH);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block_len (block_action, TNT_BLOCK_IN_FILE_LEN (block_id, file_id->len));
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_FILE_DB (file_db), priority, block_action);
  return TRUE;
}

static gboolean
get_block (TntDataDB *file_db, const TntFileID *file_id, gint block_id,
           TntPriority priority,
           TntGotBlockFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  if (!have_block (file_db, file_id, block_id))
    return FALSE;

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_GET_BLOCK);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block_len (block_action, TNT_BLOCK_IN_FILE_LEN (block_id, file_id->len));
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_FILE_DB (file_db), priority, block_action);
  return TRUE;
}

static gboolean
put_block (TntDataDB *file_db, const TntFileID *file_id, gint block_id,
           gpointer block, guint block_len, TntPriority priority,
           TntWroteBlockFunc *func, gpointer data)
{
  TntBlockAction *block_action;

  if (have_block (file_db, file_id, block_id))
  {
    tnt_free_block (block);
    return TRUE;
  }

  block_action = tnt_block_action_new (TNT_BLOCK_ACTION_PUT_BLOCK);
  tnt_block_action_set_file_info (block_action, file_id);
  tnt_block_action_set_block_id  (block_action, block_id);
  tnt_block_action_set_block     (block_action, block);
  tnt_block_action_set_block_len (block_action, block_len);
  tnt_block_action_set_post_func (block_action, func, data);

  queue_block_action (TNT_FILE_DB (file_db), priority, block_action);
  return TRUE;
}

static void
cancel_action (TntDataDB *data_db, gpointer func, gpointer data)
{
  TntFileDB *file_db = TNT_FILE_DB (data_db);
  GList     *l;

  assert_types (file_db);

  for (l = file_db->block_action_queue; l; )
  {
    TntBlockAction *action = l->data;
    gpointer action_func;
    gpointer action_data;

    tnt_block_action_get_post_func (action, &action_func, &action_data);

    if (action_func == func && action_data == data)
    {
      GList *unlink_l = l;

      l = g_list_next (l);
      file_db->block_action_queue = g_list_delete_link (file_db->block_action_queue, unlink_l);
      tnt_block_action_set_post_func (action, NULL, NULL);
      tnt_block_action_cancel (action, NULL);
      flux_object_unref (FLUX_OBJECT (action));
    }
    else
      l = g_list_next (l);
  }

  if (file_db->current_block_action)
  {
    gpointer action_func;
    gpointer action_data;

    tnt_block_action_get_post_func (file_db->current_block_action, &action_func, &action_data);

    if (action_func == func && action_data == data)
    {
      tnt_block_action_set_post_func (file_db->current_block_action, NULL, NULL);
      tnt_block_action_cancel (file_db->current_block_action, data_db);
      /* end_block_action () unrefs */
    }
  }
}

static gboolean
pick_random_file (TntDataDB *data_db, TntFileID *file_id)
{
  TntFileDB *file_db = TNT_FILE_DB (data_db);
  FileStat  *stat = NULL;
  GList     *l;
  gint       n;

  if (file_db->file_hashed_count == 0)
    return FALSE;

  n = g_random_int_range (0, file_db->file_hashed_count);

  for (l = file_db->file_list_last; l; l = g_list_previous (l))
  {
    stat = l->data;
    if (!stat->hashed)
      continue;

    if (n == 0)
      break;

    n--;
  }

  /* Shouldn't really happen... */
  if (!l)
    return FALSE;

  memcpy (file_id, &stat->file_id, sizeof (TntFileID));
  return TRUE;
}

void
tnt_file_db_init (TntFileDB *file_db)
{
  TntDataDB *data_db = TNT_DATA_DB (file_db);

  tnt_data_db_init              (TNT_DATA_DB (file_db), NULL);
  flux_object_set_type          (FLUX_OBJECT (file_db), TNT_FILE_DB_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (file_db), tnt_file_db_finalize);

  file_db->dir_tree             = g_node_new (NULL);
#ifndef G_PLATFORM_WIN32
  file_db->dir_tree->data       = make_dir_stat (G_DIR_SEPARATOR_S);
#else
  file_db->dir_tree->data       = make_dir_stat ("");
#endif

  file_db->files_by_location    = g_hash_table_new ((GHashFunc) hash_file_stat_location,
                                                    (GCompareFunc) compare_files_dir_and_name);
  file_db->files_by_id          = g_hash_table_new ((GHashFunc) hash_file_stat_id,
                                                    (GCompareFunc) compare_files_id);

  data_db->have_file_full_func   = have_file_full;
  data_db->have_file_part_func   = have_file_part;
  data_db->have_block_func       = have_block;
  data_db->get_file_bitmap_func  = get_file_bitmap;
  data_db->get_block_hash_func   = get_block_hash;
  data_db->get_block_func        = get_block;
  data_db->put_block_func        = put_block;
  data_db->cancel_func           = cancel_action;
  data_db->pick_random_file_func = pick_random_file;
  data_db->action_done_func      = end_block_action;

  begin_state (file_db, TNT_FILE_DB_IDLE);
}

TntFileDB *
tnt_file_db_new (void)
{
  TntFileDB *file_db;

  file_db = g_new0 (TntFileDB, 1);
  tnt_file_db_init (file_db);
  return file_db;
}

static gboolean
free_dir_tree (GNode *node, gpointer data)
{
  DirStat *dir_stat = node->data;

  free_dir_stat (dir_stat);
  return FALSE;
}

void
tnt_file_db_finalize (FluxObject *object)
{
  TntFileDB *file_db = TNT_FILE_DB (object);
  GList     *l;

  assert_types (file_db);

  if (file_db->state == TNT_FILE_DB_IDLE ||
      file_db->state == TNT_FILE_DB_SCANNING)
    g_source_remove (file_db->callback_source_id);

  if (file_db->file_io)
  {
    flux_io_set_read_func (FLUX_IO (file_db->file_io), NULL, NULL);
    flux_file_io_set_eof_func (file_db->file_io, NULL, NULL);
    flux_file_io_set_error_func (file_db->file_io, NULL, NULL);

#ifdef G_PLATFORM_WIN32
    /* Hack to work around GLib brokenness on win32 */
    flux_dual_fifo_set_read_hook (FLUX_DUAL_FIFO (file_db->file_io), NULL, NULL);
#endif

    flux_object_unref (FLUX_OBJECT (file_db->file_io));
  }

  if (file_db->report_timeout_id)
    g_source_remove (file_db->report_timeout_id);

  g_node_traverse (file_db->dir_tree, G_IN_ORDER, G_TRAVERSE_ALL,
                   -1, free_dir_tree, NULL);
  g_node_destroy (file_db->dir_tree);

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
    FileStat *file_stat = l->data;
    free_file_stat (file_stat);
  }

  for (l = file_db->block_action_queue; l; l = g_list_next (l))
  {
    TntBlockAction *action = l->data;
    flux_object_unref (FLUX_OBJECT (action));
  }
  g_list_free (file_db->block_action_queue);

  if (file_db->current_block_action)
    flux_object_unref (FLUX_OBJECT (file_db->current_block_action));

  end_dir_read (file_db);

  g_hash_table_destroy (file_db->files_by_location);
  g_hash_table_destroy (file_db->files_by_id);

  tnt_data_db_finalize (object);
}

TntFileDBState
tnt_file_db_get_state (TntFileDB *file_db)
{
  return file_db->state;
}

/* Callable by application programmer only. Do not use internally. */
void
tnt_file_db_add_dir (TntFileDB *file_db, const gchar *path, gboolean recursive)
{
  gchar   *canon_path;
  DirStat *dir_stat;

  assert_types (file_db);
  g_return_if_fail (path != NULL);

  /* Canonicalize and verify the path */

  canon_path = tnt_nice_path (path);
  if (!canon_path || !tnt_path_is_dir (canon_path))
  {
    /* Doesn't exist, not accessible, or not a directory */
    g_free (canon_path);
    return;
  }

  if (!tnt_dir_have_read_access (canon_path))
  {
    /* Not readable for us; fail silently */
    g_free (canon_path);
    return;
  }

  if ((dir_stat = find_dir_stat (file_db, canon_path)) && dir_stat->shared)
  {
    /* Already have that dir */
    g_free (canon_path);
    return;
  }

  add_dir_to_user_list (file_db, canon_path);
  file_db->need_scan = TRUE;

  canon_path = tnt_filename_to_utf8_replace (canon_path);
  tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_LOW, TNT_LOG_MOOD_PLEASING, 0, "Added share '%s'.",
           canon_path);
  g_free (canon_path);

  if (file_db->state == TNT_FILE_DB_IDLE)
  {
    end_state (file_db);
  }
}

void
tnt_file_db_remove_dir (TntFileDB *file_db, const gchar *path)
{
  assert_types (file_db);
  g_return_if_fail (path != NULL);

  remove_dir_path (file_db, path);
}

/* User must do a deep free on returned list */
GList *
tnt_file_db_get_user_dir_list (TntFileDB *file_db)
{
  assert_types (file_db);

  return build_top_dir_list (file_db->dir_tree);
}

gboolean
tnt_file_db_get_dir_stats (TntFileDB *file_db, gchar *path, guint *files, guint64 *size)
{
  gchar    *canon_path;
  gboolean  result = FALSE;
  GNode    *top_node;
  GList    *l;

  assert_types (file_db);
  g_return_val_if_fail (path != NULL, FALSE);

  canon_path = tnt_nice_path (path);
  if (!canon_path)
    return FALSE;

  top_node = find_dir_node (file_db->dir_tree, canon_path);

  if (files)
    *files = 0;
  if (size)
    *size  = 0;

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
    FileStat *file_stat = l->data;
    GNode    *dir_node  = file_stat->dir;

    if (top_node == dir_node || g_node_is_ancestor (top_node, dir_node))
    {
      /* We found an exact match or a subdir */

      if (files)
        (*files)++;
      if (size)
        (*size) += file_stat->file_id.len;

      result = TRUE;
    }
  }

  g_free (canon_path);
  return result;
}

gboolean
tnt_file_db_dir_is_scanned (TntFileDB *file_db, gchar *path)
{
  DirStat *dir_stat;

  assert_types (file_db);
  g_return_val_if_fail (path != NULL, FALSE);

  dir_stat = find_dir_stat (file_db, path);
  if (!dir_stat)
    return FALSE;

  return dir_stat->scanned;
}

guint
tnt_file_db_get_dir_count (TntFileDB *file_db)
{
  assert_types (file_db);

  return file_db->dir_count;
}

guint
tnt_file_db_get_file_count (TntFileDB *file_db, gboolean include_unhashed)
{
  assert_types (file_db);

  if (include_unhashed)
    return file_db->file_count;

  return file_db->file_hashed_count;
}

guint64
tnt_file_db_get_file_size (TntFileDB *file_db, gboolean include_unhashed)
{
  assert_types (file_db);

  if (include_unhashed)
    return file_db->file_size;

  return file_db->file_hashed_size;
}

gboolean
tnt_file_db_pick_random_file (TntFileDB *file_db, TntFileID *file_id)
{
  GList *l;
  guint  num, found = 0;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, FALSE);

  if (!file_db->file_hashed_count)
    return FALSE;

  num = g_random_int_range (0, file_db->file_hashed_count);

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
     FileStat *file_stat = l->data;

    if (file_stat->hashed)
    {
      if (found == num)
      {
        memcpy (file_id, &file_stat->file_id, sizeof (TntFileID));
        break;
      }
      found++;
    }
  }

  return TRUE;
}

void
tnt_file_db_set_file_update_func (TntFileDB *file_db, TntFileUpdateFunc *func, gpointer data)
{
  assert_types (file_db);

  file_db->file_update_func      = func;
  file_db->file_update_func_data = data;
}

void
tnt_file_db_set_file_id_hint_func (TntFileDB *file_db, TntFileIDHintFunc *func, gpointer data)
{
  assert_types (file_db);

  file_db->file_id_hint_func      = func;
  file_db->file_id_hint_func_data = data;
}

gchar *
tnt_file_db_get_file_name (TntFileDB *file_db, const TntFileID *file_id)
{
  FileStat *file_stat;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, NULL);

  file_stat = find_file_stat_by_id (file_db, file_id);
  if (!file_stat)
    return NULL;

  if (!file_stat->name)
  {
    g_warning ("Found nameless file!");
    return NULL;
  }

  return g_strdup (file_stat->name);
}

gchar *
tnt_file_db_get_file_path (TntFileDB *file_db, const TntFileID *file_id)
{
  FileStat *file_stat;
  gchar    *path;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, NULL);

  file_stat = find_file_stat_by_id (file_db, file_id);
  if (!file_stat)
    return NULL;

  path = build_dir_path (file_stat->dir);
  return path;
}

gchar *
tnt_file_db_get_file_relative_path (TntFileDB *file_db, const TntFileID *file_id)
{
  FileStat *file_stat;
  gchar    *path;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, NULL);

  file_stat = find_file_stat_by_id (file_db, file_id);
  if (!file_stat)
    return NULL;

  path = build_relative_path (file_stat->dir);
  return path;
}

TntTimeStamp
tnt_file_db_get_file_timestamp (TntFileDB *file_db, const TntFileID *file_id)
{
  TntTimeStamp timestamp;
  TntStat      tstat;
  FileStat    *file_stat;
  gchar       *full_path;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, 0);

  file_stat = find_file_stat_by_id (file_db, file_id);
  if (!file_stat)
    return 0;

  full_path = build_path_to_file (file_stat);
  if (!full_path)
    return 0;

  if (tnt_stat (full_path, &tstat))
    timestamp = tstat.mtime;
  else
    timestamp = 0;

  g_free (full_path);
  return timestamp;
}

gint
tnt_file_db_n_user_dirs (TntFileDB *file_db)
{
  GList *top_dirs;
  gint   n;

  assert_types (file_db);

  top_dirs = build_top_dir_node_list (file_db->dir_tree);
  n = g_list_length (top_dirs);
  g_list_free (top_dirs);
  return n;
}

TntFileDBHandle
tnt_file_db_nth_user_dir (TntFileDB *file_db, gint n)
{
  GList           *top_dirs;
  TntFileDBHandle  handle;

  assert_types (file_db);
  g_return_val_if_fail (n >= 0, NULL);

  top_dirs = build_top_dir_node_list (file_db->dir_tree);
  handle = g_list_nth_data (top_dirs, n);
  g_list_free (top_dirs);
  return handle;
}

gboolean
tnt_file_db_begin_serialization (TntFileDB *file_db, const TntFileID *file_id,
                                 const gchar *path, const gchar *name)
{
  SerializationContext *ser_context;

  assert_types (file_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (path != NULL, FALSE);
  g_return_val_if_fail (name != NULL, FALSE);

  if (find_serialization_context_by_file_id (file_db, file_id))
  {
    g_warning ("Tried to start serialization for file already serializing.");
    return FALSE;
  }

  ser_context = g_new0 (SerializationContext, 1);
  memcpy (&ser_context->file_id, file_id, sizeof (TntFileID));
  ser_context->file_db = file_db;
  ser_context->name = g_strdup (name);
  ser_context->path = g_strdup (path);

  file_db->serialization_list = g_list_prepend (file_db->serialization_list, ser_context);
  return TRUE;
}

static void
serialization_rename_cb (const gchar *old_path, const gchar *new_path,
                         SerializationContext *ser_context)
{
  TntFileDB *file_db = ser_context->file_db;

  if (!new_path)
  {
    unlink (old_path);
    goto out;
  }

  /* If serialization and rename went well, and the target path is shared,
   * insert the file into our database (remember hash table) */

  DEBUG (g_print ("Serialized to path '%s'.\n", ser_context->path));

  if (path_is_shared (file_db->dir_tree, ser_context->path))
  {
    GNode    *dir_node;
    FileStat *file_stat;
    GList    *file_node;

    DEBUG (g_print ("Serialized file in shared path!\n"));

    dir_node = ensure_dir_node (file_db->dir_tree, ser_context->path);
    file_stat = find_file_stat_by_dir_and_name (file_db, dir_node, ser_context->name);

    if (file_stat)
    {
      DEBUG (g_print ("A file stat exists for it.\n"));

      /* This shouldn't happen, unless we're unlink()ing new_path before renaming */
      if (file_stat->hashed)
        propagate_stats_upwards (file_db, dir_node,
                                 -1, -file_stat->file_id.len,
                                 -1, -file_stat->file_id.len);
      else
        propagate_stats_upwards (file_db, dir_node,
                                 0, 0,
                                 -1, -file_stat->file_id.len);

      g_hash_table_remove (file_db->files_by_id, file_stat);

      file_stat->active  = TRUE;
      file_stat->scanned = TRUE;

      file_node = g_list_find (file_db->file_list, file_stat);  /* Slow, sort of bad */

      propagate_stats_upwards (file_db, dir_node, 0, 0, 1, ser_context->file_id.len);
    }
    else
    {
      DEBUG (g_print ("Creating new file stat.\n"));

      file_stat = make_file_stat (dir_node, ser_context->name, ser_context->file_id.len);
      file_stat->change_time = tnt_timestamp ();
      add_file_to_list (file_db, file_stat);

      file_node = file_db->file_list;  /* Relies on add_file_to_list() behaviour... */
    }

    memcpy (&file_stat->file_id, &ser_context->file_id, sizeof (TntFileID));

    file_stat->hashed = TRUE;
    propagate_stats_upwards (file_db, dir_node, 1, file_stat->file_id.len, 0, 0);
    g_hash_table_insert (file_db->files_by_id, file_stat, file_node);

    /* Notify user of new file */
    if (file_db->file_update_func)
      file_db->file_update_func (file_db, &file_stat->file_id, file_db->file_update_func_data);
  }

out:
  flux_object_unref (FLUX_OBJECT (file_db));
  free_serialization_context (ser_context);
}

void
tnt_file_db_end_serialization (TntFileDB *file_db, const TntFileID *file_id)
{
  SerializationContext *ser_context;
  gchar                *old_path;
  gchar                *new_path;

  assert_types (file_db);
  g_return_if_fail (file_id != NULL);

  ser_context = find_serialization_context_by_file_id (file_db, file_id);
  if (!ser_context)
  {
    g_warning ("Tried to end serialization of file not being serialized.");
    return;
  }

  old_path = make_serialization_path_to_temp_file (ser_context);
  new_path = make_serialization_path_to_file (ser_context);

  flux_object_ref (FLUX_OBJECT (file_db));
  tnt_persistent_rename (old_path, new_path,
                         (TntPersistentRenameFunc *) serialization_rename_cb, ser_context);

  g_free (new_path);
  g_free (old_path);

  file_db->serialization_list = g_list_remove (file_db->serialization_list, ser_context);
}

void
tnt_file_db_abort_serialization (TntFileDB *file_db, const TntFileID *file_id)
{
  SerializationContext *ser_context;
  gchar                *file_path;
  gint                  result;

  assert_types (file_db);
  g_return_if_fail (file_id != NULL);

  ser_context = find_serialization_context_by_file_id (file_db, file_id);
  if (!ser_context)
  {
    g_warning ("Tried to abort serialization of file not being serialized.");
    return;
  }

  file_path = make_serialization_path_to_temp_file (ser_context);
  result = unlink (file_path);
  g_free (file_path);

  file_db->serialization_list = g_list_remove (file_db->serialization_list, ser_context);
  free_serialization_context (ser_context);
}

/* --- Operations on handles --- */

gint
tnt_file_db_handle_n_children (TntFileDB *file_db, TntFileDBHandle handle)
{
  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, 0);

  return g_node_n_children ((GNode *) handle);
}

TntFileDBHandle
tnt_file_db_handle_nth_child (TntFileDB *file_db, TntFileDBHandle handle, gint n)
{
  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, NULL);
  g_return_val_if_fail (n >= 0, NULL);

  return g_node_nth_child ((GNode *) handle, n);
}

TntFileDBHandle
tnt_file_db_handle_parent (TntFileDB *file_db, TntFileDBHandle handle)
{
  GNode *parent;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, NULL);

  parent = ((GNode *) handle)->parent;
  if (!parent || !parent->data || !((DirStat *) parent->data)->shared)
    return NULL;

  return parent;
}

gboolean
tnt_file_db_handle_is_user_dir (TntFileDB *file_db, TntFileDBHandle handle)
{
  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, FALSE);

  if (!tnt_file_db_handle_parent (file_db, handle))
    return TRUE;

  return FALSE;
}

gint
tnt_file_db_handle_get_position (TntFileDB *file_db, TntFileDBHandle handle)
{
  GNode *parent;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, -1);

  parent = tnt_file_db_handle_parent (file_db, handle);

  if (!parent)
  {
    GList *top_dirs;
    gint   n;

    top_dirs = build_top_dir_node_list (file_db->dir_tree);
    n = g_list_index (top_dirs, handle);
    g_list_free (top_dirs);
    return n;
  }

  return g_node_child_position (parent, (GNode *) handle);
}

gchar *
tnt_file_db_handle_get_name (TntFileDB *file_db, TntFileDBHandle handle)
{
  DirStat *dir_stat;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, NULL);

  if (tnt_file_db_handle_is_user_dir (file_db, handle))
    return build_dir_path ((GNode *) handle);

  dir_stat = ((GNode *) handle)->data;
  if (dir_stat && dir_stat->name)
    return g_strdup (dir_stat->name);

  return NULL;
}

gchar *
tnt_file_db_handle_get_path (TntFileDB *file_db, TntFileDBHandle handle)
{
  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, NULL);

  return build_dir_path ((GNode *) handle);
}

guint
tnt_file_db_handle_get_file_count (TntFileDB *file_db, TntFileDBHandle handle)
{
  DirStat *dir_stat;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, 0);

  dir_stat = ((GNode *) handle)->data;
  return dir_stat->files_count;
}

guint64
tnt_file_db_handle_get_file_size (TntFileDB *file_db, TntFileDBHandle handle)
{
  DirStat *dir_stat;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, 0);

  dir_stat = ((GNode *) handle)->data;
  return dir_stat->files_size;
}

TntFileDBHandleState
tnt_file_db_handle_get_state (TntFileDB *file_db, TntFileDBHandle handle)
{
  TntFileDBHandleState  state;
  DirStat              *dir_stat;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, TNT_FILE_DB_HANDLE_ONLINE);

  dir_stat = ((GNode *) handle)->data;

  if (!dir_stat->scanned_all)
  {
    state = TNT_FILE_DB_HANDLE_SCANNING;
  }
  else if (dir_stat->files_hashed_count < dir_stat->files_count)
  {
    state = TNT_FILE_DB_HANDLE_TALLIED;

    /* FIXME: Sort of annoying structuring here */
    if (file_db->state == TNT_FILE_DB_HASHING)
    {
      if (dir_stat->files_hashed_count)
      {
        state = TNT_FILE_DB_HANDLE_HASHING;
      }
      else if (file_db->stat_node)
      {
        FileStat *file_stat = ((GList *) file_db->stat_node)->data;

        if (g_node_is_ancestor ((GNode *) handle, file_stat->dir))
          state = TNT_FILE_DB_HANDLE_HASHING;
      }
    }
  }
  else
  {
    state = TNT_FILE_DB_HANDLE_ONLINE;
  }

  return state;
}

GList *
tnt_file_db_handle_get_files (TntFileDB *file_db, TntFileDBHandle handle)
{
  GNode *top_node  = handle;
  GList *file_list = NULL;
  GList *l;

  assert_types (file_db);
  g_return_val_if_fail (handle != NULL, NULL);

  for (l = file_db->file_list; l; l = g_list_next (l))
  {
    FileStat *file_stat = l->data;
    GNode    *dir_node  = file_stat->dir;

    DEBUG (g_print ("File (size %lld) is in list.\n", file_stat->file_id.len));

    if (!file_stat->hashed)
      continue;

    if (top_node == dir_node || g_node_is_ancestor (top_node, dir_node))
    {
      /* File is in path */
      TntFileID *file_id;

      file_id = g_new (TntFileID, 1);
      memcpy (file_id, &file_stat->file_id, sizeof (TntFileID));
      file_list = g_list_append (file_list, file_id);
    }
  }

  return file_list;
}

void
tnt_file_db_set_handle_add_func (TntFileDB *file_db, TntHandleFunc *func, gpointer data)
{
  assert_types (file_db);

  file_db->handle_add_func      = func;
  file_db->handle_add_func_data = data;
}

void
tnt_file_db_set_handle_remove_func (TntFileDB *file_db, TntHandleFunc *func, gpointer data)
{
  assert_types (file_db);

  file_db->handle_remove_func      = func;
  file_db->handle_remove_func_data = data;
}

void
tnt_file_db_set_handle_change_func (TntFileDB *file_db, TntHandleFunc *func, gpointer data)
{
  assert_types (file_db);

  file_db->handle_change_func      = func;
  file_db->handle_change_func_data = data;
}
