/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-meta.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG_FILE_META_MEMORY 1

#define SET_SIMPLE_FUNC(prop, field, type) \
  void tnt_file_meta_set_ ## prop (TntFileMeta *meta, type prop) \
  { \
    g_return_if_fail (meta != NULL); \
    meta->field = prop; \
  }

#define GET_SIMPLE_FUNC(prop, field, type) \
  type tnt_file_meta_get_ ## prop (TntFileMeta *meta) \
  { \
    g_return_val_if_fail (meta != NULL, 0);  /* 0 should be pretty universal... */  \
    return meta->field; \
  }

#define SET_ENUMERATED_FUNC(prop, field, type, len) \
  void tnt_file_meta_set_ ## prop (TntFileMeta *meta, gint n, type prop) \
  { \
    g_return_if_fail (meta != NULL); \
    g_return_if_fail (n < len); \
    meta->field [n] = prop; \
  }

#define GET_ENUMERATED_FUNC(prop, field, type, len) \
  type tnt_file_meta_get_ ## prop (TntFileMeta *meta, gint n) \
  { \
    g_return_val_if_fail (meta != NULL, 0); \
    g_return_val_if_fail (n < len, 0); \
    return meta->field [n]; \
  }

#define SET_STRING_FUNC(prop, field) \
  void tnt_file_meta_set_ ## prop (TntFileMeta *meta, const gchar *prop) \
  { \
    g_return_if_fail (prop != NULL); \
    if (meta->field) \
      g_free (meta->field); \
    if (!FLUX_STRING_IS_EMPTY (prop)) \
    { \
      meta->field = g_strdup (prop); \
      cleanse_string (meta->field); \
    } \
    else \
      meta->field = NULL; \
  }

#define GET_STRING_FUNC(prop, field) \
  gchar *tnt_file_meta_get_ ## prop (TntFileMeta *meta) \
  { \
    if (meta->field == NULL) \
      return NULL; \
    return g_strdup (meta->field); \
  }

#ifndef DEBUG_FILE_META_MEMORY
static FluxMemChunk *file_meta_mem_chunk = NULL;
#endif

/* Marshalling info for TntFileMeta struct */
FluxMarshalledStructEntry tnt_file_meta_m_entries [] =
{
  FLUX_MARSHAL_STRING_DYNAMIC ("name",            TntFileMeta, name, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("title",           TntFileMeta, title, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("collection",      TntFileMeta, collection, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("performer",       TntFileMeta, performer, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("author",          TntFileMeta, author, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("keywords",        TntFileMeta, keywords, TRUE, NULL),
  FLUX_MARSHAL_STRING_DYNAMIC ("seq",             TntFileMeta, seq, TRUE, NULL),
  FLUX_MARSHAL_UINT16         ("fragment",        TntFileMeta, fragment, TRUE, 0),
  FLUX_MARSHAL_UINT16         ("max_fragment",    TntFileMeta, max_fragment, TRUE, 0),
  FLUX_MARSHAL_UINT16         ("quality",         TntFileMeta, quality, TRUE, 0),
  FLUX_MARSHAL_UINT16         ("format",          TntFileMeta, format, TRUE, 0),
  FLUX_MARSHAL_UINT8          ("media",           TntFileMeta, media, TRUE, 0),

  FLUX_MARSHAL_BRANCH_BEGIN   ("date"),
  FLUX_MARSHAL_INT16          ("year",            TntFileMeta, year, TRUE, 0),
  FLUX_MARSHAL_UINT8          ("month",           TntFileMeta, month, TRUE, 0),
  FLUX_MARSHAL_UINT8          ("day",             TntFileMeta, day, TRUE, 0),
  FLUX_MARSHAL_BRANCH_END     (),

  FLUX_MARSHAL_BRANCH_BEGIN   ("language"),
  FLUX_MARSHAL_UINT8          ("primary",         TntFileMeta, lang [0], TRUE, 0),
  FLUX_MARSHAL_UINT8          ("secondary",       TntFileMeta, lang [1], TRUE, 0),
  FLUX_MARSHAL_BRANCH_END     (),

  FLUX_MARSHAL_BRANCH_BEGIN   ("category"),
  FLUX_MARSHAL_UINT8          ("primary",         TntFileMeta, cat [0], TRUE, 0),
  FLUX_MARSHAL_UINT8          ("secondary",       TntFileMeta, cat [1], TRUE, 0),
  FLUX_MARSHAL_BRANCH_END     (),

  FLUX_MARSHAL_BRANCH_BEGIN   ("dimensions"),
  FLUX_MARSHAL_UINT32         ("duration",        TntFileMeta, duration, TRUE, 0),
  FLUX_MARSHAL_UINT16         ("width",           TntFileMeta, width, TRUE, 0),
  FLUX_MARSHAL_UINT16         ("height",          TntFileMeta, height, TRUE, 0),
  FLUX_MARSHAL_BRANCH_END     (),

  FLUX_MARSHAL_UINT8          ("flags",           TntFileMeta, flags, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

/* UI labels by TntMediaType */
TntMediaLabels media_labels [] =
{
  { "Title", "Collection", "Publisher",    "Author",       "Keywords", "Width", "Height", "Duration", "Installment", "Year", "Part", "Category", "Languages" },
  { "Title", "Collection", "Publisher",    "Author",       "Keywords", "Width", "Height", "Duration", "Installment", "Year", "Part", "Category", "Languages" },
  { "Title", "Series",     "Performer",    "Creator",      "Keywords", "Width", "Height", "Duration", "Installment", "Year", "Part", "Category", "Languages" },
  { "Title", "Album",      "Track Artist", "Album Artist", "Keywords", "Width", "Height", "Duration", "Track",       "Year", "Part", "Category", "Languages" },
  { "Title", "Series",     "Performer",    "Director",     "Keywords", "Width", "Height", "Duration", "Installment", "Year", "Part", "Category", "Languages" },
  { "Title", "Series",     "Publisher",    "Author",       "Keywords", "Width", "Height", "Duration", "Version",     "Year", "Part", "Category", "Languages" },
  { "Title", "Collection", "Publisher",    "Author",       "Keywords", "Width", "Height", "Duration", "Installment", "Year", "Part", "Category", "Languages" }
};

/* Note: Must be in sync with TntMediaType enum */
static gchar *media_names [] =
{
  "Unknown",
  "Text",
  "Picture",
  "Audio",
  "Video",
  "Software",
  "Other"
};

const gchar *
_tnt_get_media_label (gint type, const gchar *name)
{
  /* Bleh. This was a bad idea to begin with, and now we have
   * to work around Woe32 DLL brokenness. */

  if (!strcmp (name, "title"))
    return media_labels [type].title;
  if (!strcmp (name, "collection"))
    return media_labels [type].collection;
  if (!strcmp (name, "performer"))
    return media_labels [type].performer;
  if (!strcmp (name, "author"))
    return media_labels [type].author;
  if (!strcmp (name, "keywords"))
    return media_labels [type].keywords;
  if (!strcmp (name, "width"))
    return media_labels [type].width;
  if (!strcmp (name, "height"))
    return media_labels [type].height;
  if (!strcmp (name, "duration"))
    return media_labels [type].duration;
  if (!strcmp (name, "seq"))
    return media_labels [type].seq;
  if (!strcmp (name, "year"))
    return media_labels [type].year;
  if (!strcmp (name, "fragment"))
    return media_labels [type].fragment;
  if (!strcmp (name, "category"))
    return media_labels [type].category;
  if (!strcmp (name, "lang"))
    return media_labels [type].lang;

  return NULL;
}

const gchar *
tnt_media_name (TntMediaType media)
{
  return media_names [media];
}

TntFileMeta *
tnt_file_meta_new (void)
{
#ifndef DEBUG_FILE_META_MEMORY
  if (!file_meta_mem_chunk)
    file_meta_mem_chunk = flux_mem_chunk_new (32, sizeof (TntFileMeta));

  return flux_mem_chunk_alloc0 (file_meta_mem_chunk);
#else
  return g_new0 (TntFileMeta, 1);
#endif
}

TntFileMeta *
tnt_file_meta_dup (TntFileMeta *file_meta)
{
  TntFileMeta *new_meta;

  g_return_val_if_fail (file_meta != NULL, NULL);

  /* Having a file_meta to dup indicates we also have the memchunk,
   * but a crazy user (say, me) might want to dup a struct that was
   * not allocated here. That would be bad, and should be avoided.
   * I'll keep this check around anyhow - to be removed in final
   * stages of debugging. */

#ifndef DEBUG_FILE_META_MEMORY
  if (!file_meta_mem_chunk)
    file_meta_mem_chunk = flux_mem_chunk_new (32, sizeof (TntFileMeta));

  new_meta = flux_mem_chunk_alloc (file_meta_mem_chunk);
#else
  new_meta = g_new (TntFileMeta, 1);
#endif

  memcpy (new_meta, file_meta, sizeof (TntFileMeta));

  if (new_meta->name)
    new_meta->name = g_strdup (new_meta->name);
  if (new_meta->title)
    new_meta->title = g_strdup (new_meta->title);
  if (new_meta->collection)
    new_meta->collection = g_strdup (new_meta->collection);
  if (new_meta->performer)
    new_meta->performer = g_strdup (new_meta->performer);
  if (new_meta->author)
    new_meta->author = g_strdup (new_meta->author);
  if (new_meta->keywords)
    new_meta->keywords = g_strdup (new_meta->keywords);
  if (new_meta->seq)
    new_meta->seq = g_strdup (new_meta->seq);

  return new_meta;
}

void
tnt_file_meta_free (TntFileMeta *file_meta)
{
  g_return_if_fail (file_meta != NULL);

  g_free (file_meta->name);
  g_free (file_meta->title);
  g_free (file_meta->collection);
  g_free (file_meta->performer);
  g_free (file_meta->author);
  g_free (file_meta->keywords);
  g_free (file_meta->seq);

#ifndef DEBUG_FILE_META_MEMORY
  flux_mem_chunk_free (file_meta_mem_chunk, file_meta);
#else
  g_free (file_meta);
#endif
}

void
tnt_file_meta_reclaim_memory (void)
{
#ifndef DEBUG_FILE_META_MEMORY
  if (file_meta_mem_chunk)
    flux_mem_chunk_clean (file_meta_mem_chunk);
#endif
}

/* Returns TRUE if changed */
static gboolean
cleanse_string (gchar *string)
{
  gboolean  valid;
  gint      string_len;
  gchar    *p;
  gunichar *string_ucs4;
  gunichar *pu;
  gchar    *normalized_utf8;
  gint      normalized_utf8_len;
  gboolean  changed = FALSE, ucs4_changed = FALSE;

  if (!string)
    return FALSE;

  string_len = strlen (string);

  /* Terminate string at the first invalid sequence */

  valid = g_utf8_validate (string, string_len, (const gchar **) &p);
  if (!valid)
  {
    *p = '\0';
    string_len = p - string;
    changed = TRUE;
  }

  if (string_len == 0)
    return changed;

  /* Replace all undisplayable chars with spaces (get rid of newlines etc) */

  string_ucs4 = g_utf8_to_ucs4_fast (string, string_len, NULL);

  for (pu = string_ucs4; *pu; pu++)
  {
    if (!g_unichar_isgraph (*pu) && *pu != ' ')
    {
      *pu = ' ';
      ucs4_changed = TRUE;
      changed = TRUE;
    }
  }

  /* Normalize the UTF-8 to maximally composed form */

  if (ucs4_changed)
  {
    gchar *temp_utf8;

    temp_utf8 = g_ucs4_to_utf8 (string_ucs4, -1, NULL, NULL, NULL);
    normalized_utf8 = g_utf8_normalize (temp_utf8, -1, G_NORMALIZE_DEFAULT_COMPOSE);
    g_free (temp_utf8);
  }
  else
  {
    normalized_utf8 = g_utf8_normalize (string, -1, G_NORMALIZE_DEFAULT_COMPOSE);
  }

  g_free (string_ucs4);
  normalized_utf8_len = strlen (normalized_utf8);

  if (ucs4_changed || normalized_utf8_len < string_len)
  {
    memcpy (string, normalized_utf8, MIN (normalized_utf8_len, string_len));
    if (normalized_utf8_len < string_len)
      *(string + normalized_utf8_len) = '\0';
  }

  g_free (normalized_utf8);
  return changed;
}

GET_SIMPLE_FUNC (timestamp, timestamp, TntTimeStamp);
SET_SIMPLE_FUNC (timestamp, timestamp, TntTimeStamp);

GET_SIMPLE_FUNC (format, format, TntFormatType);
SET_SIMPLE_FUNC (format, format, TntFormatType);

GET_SIMPLE_FUNC (width, width, guint16);
SET_SIMPLE_FUNC (width, width, guint16);

GET_SIMPLE_FUNC (height, height, guint16);
SET_SIMPLE_FUNC (height, height, guint16);

GET_SIMPLE_FUNC (duration, duration, guint32);
SET_SIMPLE_FUNC (duration, duration, guint32);

GET_SIMPLE_FUNC (year, year, gint16);
SET_SIMPLE_FUNC (year, year, gint16);

GET_SIMPLE_FUNC (month, month, guint8);
SET_SIMPLE_FUNC (month, month, guint8);

GET_SIMPLE_FUNC (day, day, guint8);
SET_SIMPLE_FUNC (day, day, guint8);

GET_SIMPLE_FUNC (media, media, guint8);
SET_SIMPLE_FUNC (media, media, guint8);

GET_SIMPLE_FUNC (quality, quality, guint16);
SET_SIMPLE_FUNC (quality, quality, guint16);

GET_SIMPLE_FUNC (fragment, fragment, guint16);
SET_SIMPLE_FUNC (fragment, fragment, guint16);

GET_SIMPLE_FUNC (max_fragment, max_fragment, guint16);
SET_SIMPLE_FUNC (max_fragment, max_fragment, guint16);

GET_ENUMERATED_FUNC (lang, lang, guint8, 2);
SET_ENUMERATED_FUNC (lang, lang, guint8, 2);

GET_ENUMERATED_FUNC (category, cat, guint8, 2);
SET_ENUMERATED_FUNC (category, cat, guint8, 2);

GET_STRING_FUNC (name, name);
SET_STRING_FUNC (name, name);

GET_STRING_FUNC (title, title);
SET_STRING_FUNC (title, title);

GET_STRING_FUNC (collection, collection);
SET_STRING_FUNC (collection, collection);

GET_STRING_FUNC (performer, performer);
SET_STRING_FUNC (performer, performer);

GET_STRING_FUNC (author, author);
SET_STRING_FUNC (author, author);

GET_STRING_FUNC (keywords, keywords);
SET_STRING_FUNC (keywords, keywords);

GET_STRING_FUNC (seq, seq);
SET_STRING_FUNC (seq, seq);

gchar *
tnt_file_meta_get_best_title (TntFileMeta *meta)
{
  g_return_val_if_fail (meta != NULL, NULL);

  if (!FLUX_STRING_IS_EMPTY (meta->title))
    return tnt_file_meta_get_title (meta);

  return tnt_file_meta_get_name (meta);
}

gchar *
tnt_file_meta_get_best_description (TntFileMeta *meta)
{
  g_return_val_if_fail (meta != NULL, NULL);

  if (!FLUX_STRING_IS_EMPTY (meta->collection))
    return tnt_file_meta_get_collection (meta);

  if (!FLUX_STRING_IS_EMPTY (meta->author))
    return tnt_file_meta_get_author (meta);

  if (!FLUX_STRING_IS_EMPTY (meta->performer))
    return tnt_file_meta_get_performer (meta);

  return tnt_file_meta_get_keywords (meta);
}

gboolean
tnt_file_meta_is_local_user_input (TntFileMeta *meta)
{
  g_return_val_if_fail (meta != NULL, FALSE);

  if (meta->flags & TNT_FILE_META_LOCAL_USER_INPUT)
    return TRUE;

  return FALSE;
}

void
tnt_file_meta_set_local_user_input (TntFileMeta *meta, gboolean value)
{
  g_return_if_fail (meta != NULL);

  if (value)
    meta->flags |= (TNT_FILE_META_LOCAL_USER_INPUT | TNT_FILE_META_USER_INPUT);
  else
    meta->flags &= ~TNT_FILE_META_LOCAL_USER_INPUT;
}

gboolean
tnt_file_meta_is_user_input (TntFileMeta *meta)
{
  g_return_val_if_fail (meta != NULL, FALSE);

  if (meta->flags & TNT_FILE_META_USER_INPUT)
    return TRUE;

  return FALSE;
}

void
tnt_file_meta_set_user_input (TntFileMeta *meta, gboolean value)
{
  g_return_if_fail (meta != NULL);

  if (value)
    meta->flags |= TNT_FILE_META_USER_INPUT;
  else
    meta->flags &= ~(TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT);
}

gboolean
tnt_file_meta_is_local_file (TntFileMeta *meta)
{
  g_return_val_if_fail (meta != NULL, FALSE);

  if (meta->flags & TNT_FILE_META_LOCAL_FILE)
    return TRUE;

  return FALSE;
}

void
tnt_file_meta_set_local_file (TntFileMeta *meta, gboolean value)
{
  g_return_if_fail (meta != NULL);

  if (value)
    meta->flags |= TNT_FILE_META_LOCAL_FILE;
  else
    meta->flags &= ~TNT_FILE_META_LOCAL_FILE;
}

#define MATCH_STRING_PROP(prop) \
  if (!match && (str = tnt_file_meta_get_ ## prop (meta))) \
  { \
    gchar *temp = tnt_utf8_strip_accents (str); \
    gchar *str_low = g_utf8_strdown (temp, -1); \
    if (strstr (str_low, substr)) \
      match = TRUE; \
    g_free (str_low); \
    g_free (temp); \
    g_free (str); \
  }

gboolean
tnt_file_meta_matches_string (TntFileMeta *meta, TntMatchField field, const gchar *query_str)
{
  gint          i, seg_len;
  gchar        *query_str_low, *temp_str;
  gboolean      result = TRUE;
  const guchar *break_chars = " \n\t()[].-_:;+/&!@#=?<>\"\\";

  g_return_val_if_fail (meta != NULL, FALSE);
  g_return_val_if_fail (query_str != NULL, FALSE);

  temp_str = tnt_utf8_strip_accents (query_str);
  query_str_low = g_utf8_strdown (temp_str, -1);
  g_free (temp_str);

  for (i = 0; query_str_low [i] && result; )
  {
    seg_len = strcspn (query_str_low + i, break_chars);
    if (seg_len)
    {
      gchar    *str, *substr;
      gboolean  match = FALSE;

      substr = g_strndup (query_str_low + i, seg_len);

      if (field == TNT_MATCH_FIELD_ANY)
      {
        MATCH_STRING_PROP (name);
        MATCH_STRING_PROP (title);
        MATCH_STRING_PROP (collection);
        MATCH_STRING_PROP (performer);
        MATCH_STRING_PROP (author);
        MATCH_STRING_PROP (keywords);
        MATCH_STRING_PROP (seq);
      }
      else if (field == TNT_MATCH_FIELD_FILENAME)
      {
        MATCH_STRING_PROP (name);
      }
      else if (field == TNT_MATCH_FIELD_TITLE)
      {
        MATCH_STRING_PROP (title);
      }
      else if (field == TNT_MATCH_FIELD_COLLECTION)
      {
        MATCH_STRING_PROP (collection);
      }
      else if (field == TNT_MATCH_FIELD_AUTHOR)
      {
        MATCH_STRING_PROP (author);
        if (meta->media == TNT_MEDIA_AUDIO)
        {
          /* Music author/performer is fuzzy */
          MATCH_STRING_PROP (performer);
        }
      }
      else if (field == TNT_MATCH_FIELD_PERFORMER)
      {
        MATCH_STRING_PROP (performer);
        if (meta->media == TNT_MEDIA_AUDIO)
        {
          /* Music author/performer is fuzzy */
          MATCH_STRING_PROP (author);
        }
      }
      else if (field == TNT_MATCH_FIELD_KEYWORDS)
      {
        MATCH_STRING_PROP (keywords);
      }

      if (!match)
        result = FALSE;

      g_free (substr);
    }

    i += seg_len;
    i += strspn (query_str_low + i, break_chars);
  }

  g_free (query_str_low);
  return result;
}

static void
pick_string (TntFileMeta *a, TntFileMeta *b, TntFileMeta *out, guint offset)
{
  gchar *a_str, *b_str, **out_str;

  a_str   = G_STRUCT_MEMBER   (gchar *, a, offset);
  b_str   = G_STRUCT_MEMBER   (gchar *, b, offset);
  out_str = G_STRUCT_MEMBER_P (out, offset);

  /* If one of the in strings is empty, pick the other one */
  if (FLUX_STRING_IS_EMPTY (a_str))
  {
    if (!FLUX_STRING_IS_EMPTY (b_str))
      *out_str = g_strdup (b_str);
  }
  else if (FLUX_STRING_IS_EMPTY (b_str))
  {
    /* We know that *a_str != '\0' here */
    *out_str = g_strdup (a_str);
  }
  /* If one string has higher user priority than the other, pick that one */
  else if ((a->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)) >
           (b->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)))
  {
    *out_str = g_strdup (a_str);
  }
  else if ((a->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)) <
           (b->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)))
  {
    *out_str = g_strdup (b_str);
  }
  /* If all else fails, pick the longest string */
  else if (strlen (b_str) > strlen (a_str))
  {
    *out_str = g_strdup (b_str);
  }
  else
  {
    *out_str = g_strdup (a_str);
  }
}

#define PICK_STRING(a, b, out, member, token) \
  pick_string (a, b, out, G_STRUCT_OFFSET (TntFileMeta, member))

#define PICK_OPTION(member, other_val, unspecified_val) \
  G_STMT_START{                                         \
  if (a->member == unspecified_val)                     \
    out->member = b->member;                            \
  else if (b->member == unspecified_val)                \
    out->member = a->member;                            \
  else if (a->member == other_val)                      \
    out->member = b->member;                            \
  else if (b->member == other_val)                      \
    out->member = a->member;                            \
  else if (best)                                        \
    out->member = best->member;                         \
  else                                                  \
    out->member = a->member;                            \
  }G_STMT_END

TntFileMeta *
tnt_file_meta_merge (TntFileMeta *a, TntFileMeta *b)
{
  TntFileMeta *out, *best = NULL;

  g_return_val_if_fail (a != NULL, NULL);
  g_return_val_if_fail (b != NULL, NULL);

  out = tnt_file_meta_new ();

  if      ((a->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)) >
           (b->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)))
    best = a;
  else if ((a->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)) <
           (b->flags & (TNT_FILE_META_USER_INPUT | TNT_FILE_META_LOCAL_USER_INPUT)))
    best = b;

  /* Preserve local file names and timestamps */
  if (a->flags & TNT_FILE_META_LOCAL_FILE)
  {
    out->name = g_strdup (a->name);
    out->timestamp = a->timestamp;
  }
  else if (b->flags & TNT_FILE_META_LOCAL_FILE)
  {
    out->name = g_strdup (b->name);
    out->timestamp = b->timestamp;
  }
  else
  {
    PICK_STRING (a, b, out, name, NAME);
    out->timestamp = best ? best->timestamp : a->timestamp;
  }

  PICK_STRING (a, b, out, title, TITLE);
  PICK_STRING (a, b, out, collection, COLLECTION);
  PICK_STRING (a, b, out, performer, PERFORMER);
  PICK_STRING (a, b, out, author, AUTHOR);
  PICK_STRING (a, b, out, keywords, KEYWORDS);
  PICK_STRING (a, b, out, seq, SEQ);

  PICK_OPTION (lang [0], TNT_LANG_OTHER, TNT_LANG_UNSPECIFIED);
  PICK_OPTION (lang [1], TNT_LANG_OTHER, TNT_LANG_UNSPECIFIED);

  PICK_OPTION (cat [0], TNT_CATEGORY_OTHER, TNT_CATEGORY_UNSPECIFIED);
  PICK_OPTION (cat [1], TNT_CATEGORY_OTHER, TNT_CATEGORY_UNSPECIFIED);

  if (!best)
    best = a;

  out->format       = best->format;
  out->year         = best->year;
  out->month        = best->month;
  out->day          = best->day;
  out->media        = best->media;
  out->width        = best->width;
  out->height       = best->height;
  out->duration     = best->duration;
  out->fragment     = best->fragment;
  out->max_fragment = best->max_fragment;
  out->quality      = best->quality;

  out->flags  = best ? best->flags  : a->flags;
  out->flags |= (a->flags | b->flags) & TNT_FILE_META_LOCAL_FILE;

  return out;
}

#define CLEANSE_STRING(meta, field, token) \
  cleanse_string (meta->field)

static gboolean
cleanse_all (TntFileMeta *file_meta)
{
  return
    CLEANSE_STRING (file_meta, name,       NAME)       |
    CLEANSE_STRING (file_meta, title,      TITLE)      |
    CLEANSE_STRING (file_meta, collection, COLLECTION) |
    CLEANSE_STRING (file_meta, performer,  PERFORMER)  |
    CLEANSE_STRING (file_meta, author,     AUTHOR)     |
    CLEANSE_STRING (file_meta, keywords,   KEYWORDS)   |
    CLEANSE_STRING (file_meta, seq,        SEQ);
}

/* Returns TRUE if changed */
gboolean
tnt_file_meta_strip_foreign (TntFileMeta *file_meta)
{
  gboolean result = FALSE;

  g_return_val_if_fail (file_meta != NULL, FALSE);

  /* Strip all locality flags */
  if (file_meta->flags & TNT_FILE_META_LOCAL_USER_INPUT)
  {
    file_meta->flags &= ~TNT_FILE_META_LOCAL_USER_INPUT;
    result = TRUE;
  }

  if (file_meta->flags & TNT_FILE_META_LOCAL_FILE)
  {
    file_meta->flags &= ~TNT_FILE_META_LOCAL_FILE;
    result = TRUE;
  }

  /* Strip strings */
  result |= cleanse_all (file_meta);

  return result;
}
