/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-local-query.c - Describes a metadata query for local presentation.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_LOCAL_QUERY_TYPE)

#define MATCH_LOCAL_FLAG   (1 << 0)
#define MATCH_PARTIAL_FLAG (1 << 1)

typedef struct
{
  TntMatchFunc *func;
  gpointer      data;
}
MatchFunc;

static TntMatch *
find_match_by_file_id (TntLocalQuery *query, const TntFileID *file_id)
{
  return g_hash_table_lookup (query->file_id_hash, file_id);
}

static void
run_callbacks (TntLocalQuery *query, GList *callbacks, TntMatch *match)
{
  GList *l;

  for (l = callbacks; l; l = g_list_next (l))
  {
    MatchFunc *match_func;

    match_func = l->data;
    g_assert (match_func);

    match_func->func (query, match, match_func->data);
  }
}

static void
remove_callback (GList **callbacks, TntMatchFunc *func, gpointer data)
{
  GList *l;
  GList *next;

  for (l = *callbacks; l; l = next)
  {
    MatchFunc *match_func;

    next = g_list_next (l);

    match_func = l->data;
    g_assert (match_func);

    if (match_func->func == func && match_func->data == data)
    {
      g_free (match_func);
      *callbacks = g_list_delete_link (*callbacks, l);
    }
  }
}

static gboolean
grace_timeout_cb (TntMatch *match, gpointer data)
{
  TntLocalQuery *query = data;

  if (!tnt_match_get_parent (match))
    return FALSE;

  if (!(match->flags & TNT_MATCH_OUT_OF_GRACE))
  {
    if (tnt_timestamp () >= tnt_match_determine_grace_timeout (match->first_seen))
    {
      match->flags |= TNT_MATCH_OUT_OF_GRACE;
      run_callbacks (query, query->match_update_funcs, match);
    }
    else
      query->have_grace_timeout = TRUE;
  }

  return FALSE;
}

static void maybe_start_grace_timeout (TntLocalQuery *query);

static gboolean
grace_timeout (gpointer data)
{
  TntLocalQuery *query = data;

  assert_types (query);

  query->have_grace_timeout = FALSE;
  tnt_match_for_all (query->root_match, grace_timeout_cb, query);

  if (query->have_grace_timeout)
  {
    query->have_grace_timeout = FALSE;
    maybe_start_grace_timeout (query);
  }

  return FALSE;
}

static void
maybe_start_grace_timeout (TntLocalQuery *query)
{
  if (!query->have_grace_timeout)
  {
    TntTimeStamp now_time, update_time;

    now_time    = tnt_timestamp ();
    update_time = now_time - (now_time % TNT_MATCH_GRACE_INTERVAL) + TNT_MATCH_GRACE_INTERVAL;

    query->grace_timeout      = g_timeout_add ((update_time - now_time) * 1000, grace_timeout, query);
    query->have_grace_timeout = TRUE;
  }
}

static gboolean
query_matches_meta (TntLocalQuery *query, TntFileMeta *file_meta)
{
  if (tnt_file_meta_matches_string (file_meta, query->query.match_field, query->query.string) &&
      (query->query.media    == TNT_MEDIA_UNKNOWN ||
       query->query.media    == tnt_file_meta_get_media (file_meta)) &&
      (query->query.category == 0 ||
       query->query.category == tnt_file_meta_get_category (file_meta, 0)))
    return TRUE;

  return FALSE;
}

static gboolean
query_matches_match (TntLocalQuery *query, TntMatch *match)
{
  if ((!tnt_query_get_match_local (TNT_QUERY (query)) && tnt_match_is_local (match)) ||
      (!tnt_query_get_match_partial (TNT_QUERY (query)) && !tnt_match_is_local (match) &&
       tnt_match_get_completeness (match) != TNT_COMPLETENESS_COMPLETE))
    return FALSE;

  return query_matches_meta (query, tnt_match_get_file_meta (match));
}

static void
add_to_group_index (TntLocalQuery *query, TntMatch *match)
{
  gchar *str;

  if (!match->file_meta)
    return;

  str = tnt_file_meta_get_collection (match->file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    g_hash_table_insert (query->collection_hash, str_down, match);

    g_free (str);
    return;
  }

  str = tnt_file_meta_get_keywords (match->file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    g_hash_table_insert (query->keyword_hash, str_down, match);

    g_free (str);
    return;
  }
}

static void
remove_from_group_index (TntLocalQuery *query, TntMatch *match)
{
  gchar *str;

  if (!match->file_meta)
    return;

  str = tnt_file_meta_get_collection (match->file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    g_hash_table_remove (query->collection_hash, str_down);

    g_free (str_down);
    g_free (str);
    return;
  }

  str = tnt_file_meta_get_keywords (match->file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    g_hash_table_remove (query->keyword_hash, str_down);

    g_free (str_down);
    g_free (str);
    return;
  }
}

static gboolean
filter_matches_cb (TntMatch *match, gpointer data)
{
  TntLocalQuery *query  = data;
  gboolean       result = FALSE;

  if (!tnt_match_get_parent (match))
    return FALSE;

  if (tnt_match_is_group (match))
  {
    if (tnt_match_get_n_children (match) == 0)
    {
      result = TRUE;
    }
    else if (tnt_match_get_n_children (match) == 1)
    {
      TntMatch *child = g_ptr_array_index (tnt_match_get_children (match), 0);

      /* FIXME: Is this right? Shouldn't the child become toplevel? */
      g_hash_table_remove (query->file_id_hash, (gpointer) tnt_match_get_file_id (child));
      run_callbacks (query, query->match_remove_funcs, child);
      result = TRUE;
    }
  }
  else
  {
    if (!query_matches_match (query, match))
    {
      g_hash_table_remove (query->file_id_hash, (gpointer) tnt_match_get_file_id (match));
      result = TRUE;
    }
  }

  if (result == TRUE)
  {
    /* Remove from group index only if we're a top-level match (other matches
     * can't generate groups, and if their index strings are removed, it will
     * look like their containing group went away) */
    if (!tnt_match_get_parent (tnt_match_get_parent (match)))
      remove_from_group_index (query, match);

    run_callbacks (query, query->match_remove_funcs, match);
  }

  return result;
}

static void
filter_matches (TntQuery *query)
{
  TntLocalQuery *local_query = TNT_LOCAL_QUERY (query);

  tnt_match_for_all_remove (local_query->root_match, filter_matches_cb, local_query);
}

static void
add_to_group (TntLocalQuery *query, TntMatch *group, TntMatch *match)
{
  TntMatch    *old_match;
  TntFileMeta *group_meta;
  TntFileMeta *match_meta;

  if (!tnt_match_is_group (group))
  {
    old_match = tnt_match_dup (group);

    /* Update the TntFileID-based index to point to the duplicate of the match
     * that's being converted to a group. This must be done before it's converted,
     * since on conversion, its TntFileID will be overwritten by other data. */
    g_hash_table_replace (query->file_id_hash, (gpointer) tnt_match_get_file_id (old_match), old_match);

    tnt_match_convert_to_group (group);
    tnt_match_add_child (group, old_match);

    group_meta = tnt_match_get_file_meta (group);
    if (group_meta)
    {
      tnt_file_meta_set_name         (group_meta, "");
      tnt_file_meta_set_title        (group_meta, "");
      tnt_file_meta_set_seq          (group_meta, "");
      tnt_file_meta_set_fragment     (group_meta, 0);
      tnt_file_meta_set_max_fragment (group_meta, 0);

      if (tnt_file_meta_get_media (group_meta) == TNT_MEDIA_AUDIO)
        tnt_file_meta_set_performer (group_meta, "");  /* Track artist, usually */
    }
  }

  group_meta = tnt_match_get_file_meta (group);
  match_meta = tnt_match_get_file_meta (match);

  if (group_meta && match_meta)
  {
    guint16 group_width,  match_width;
    guint16 group_height, match_height;

    /* Accumulate duration */
    tnt_file_meta_set_duration (group_meta, tnt_file_meta_get_duration (group_meta) +
                                tnt_file_meta_get_duration (match_meta));

    /* Average quality */
    if (tnt_file_meta_get_quality (match_meta))
    {
      guint quality;

      quality = tnt_file_meta_get_quality (group_meta);
      quality *= tnt_match_get_n_quality_children (group);
      quality += tnt_file_meta_get_quality (match_meta);
      tnt_match_inc_quality_children (group);
      quality /= tnt_match_get_n_quality_children (group);
      tnt_file_meta_set_quality (group_meta, quality);
    }

    /* Maintain dimensions only if all children with dimensions are the same */
    group_width  = tnt_file_meta_get_width  (group_meta);
    group_height = tnt_file_meta_get_height (group_meta);
    match_width  = tnt_file_meta_get_width  (match_meta);
    match_height = tnt_file_meta_get_height (match_meta);

    if (!group_width && !group_height && !tnt_match_get_dimension_ban (group))
    {
      tnt_file_meta_set_width (group_meta, match_width);
      tnt_file_meta_set_height (group_meta, match_height);
    }
    else if (group_width && match_width &&
             (group_width != match_width || group_height != match_height))
    {
      tnt_file_meta_set_width (group_meta, 0);
      tnt_file_meta_set_height (group_meta, 0);
      tnt_match_set_dimension_ban (group, TRUE);
    }
  }

  tnt_match_add_child (group, match);
}

static TntMatch *
find_parent_group (TntLocalQuery *query, TntFileMeta *file_meta)
{
  TntMatch *parent;
  gchar    *str;

  str = tnt_file_meta_get_collection (file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    parent = g_hash_table_lookup (query->collection_hash, str_down);

    g_free (str_down);
    g_free (str);

    return parent;
  }

  str = tnt_file_meta_get_keywords (file_meta);
  if (str)
  {
    gchar *str_down = g_utf8_casefold (str, -1);

    parent = g_hash_table_lookup (query->keyword_hash, str_down);

    g_free (str_down);
    g_free (str);

    return parent;
  }

  return NULL;
}

static TntMatch *
new_match (TntLocalQuery *query, const TntFileID *file_id, TntFileMeta *file_meta)
{
  TntMatch *match;

  match = tnt_match_new ();
  tnt_match_set_file_id (match, file_id);
  match->file_meta = tnt_file_meta_dup (file_meta);

  return match;
}

static void
insert_match (TntLocalQuery *query, TntMatch *match)
{
  TntMatch *parent_match;

  /* Try to add it to a collection */
  parent_match = find_parent_group (query, match->file_meta);

  if (parent_match)
  {
    add_to_group (query, parent_match, match);
  }
  else
  {
    tnt_match_add_child (query->root_match, match);
    add_to_group_index (query, match);
  }

  g_hash_table_insert (query->file_id_hash, (gpointer) tnt_match_get_file_id (match), match);

  run_callbacks (query, query->match_add_funcs, match);

  maybe_start_grace_timeout (query);
}

static gboolean
destructive_string_equal (gchar *a, gchar *b)
{
  gint result;

  if (!a)
  {
    if (!b)
      return TRUE;

    g_free (b);
    return FALSE;
  }
  else if (!b)
  {
    g_free (a);
    return FALSE;
  }

  /* a and b are both non-null */

  result = strcmp (a, b);

  g_free (a);
  g_free (b);

  return result ? FALSE : TRUE;
}

static gboolean
same_group (TntFileMeta *a, TntFileMeta *b)
{
  gchar *str_a, *str_b;

  /* Collection */

  str_a = tnt_file_meta_get_collection (a);
  str_b = tnt_file_meta_get_collection (b);

  /* If both str_a and str_b are NULL, we proceed to keywords. Else, collection equality
   * determines the outcome */

  if (str_a || str_b)
  {
    return destructive_string_equal (str_a, str_b);
  }

  /* Keywords */

  str_a = tnt_file_meta_get_keywords (a);
  str_b = tnt_file_meta_get_keywords (b);
  return destructive_string_equal (str_a, str_b);
}

static gboolean
same_meta (TntFileMeta *a, TntFileMeta *b)
{
  if (destructive_string_equal (tnt_file_meta_get_name      (a), tnt_file_meta_get_name      (b)) &&
      destructive_string_equal (tnt_file_meta_get_title     (a), tnt_file_meta_get_title     (b)) &&
      destructive_string_equal (tnt_file_meta_get_performer (a), tnt_file_meta_get_performer (b)) &&
      destructive_string_equal (tnt_file_meta_get_author    (a), tnt_file_meta_get_author    (b)))
    return TRUE;

  return FALSE;
}

static gboolean
try_match (TntQuery *query, const TntFileID *file_id, TntFileMeta *file_meta,
           TntNodeID node_id, gint remote_blocks)
{
  TntLocalQuery *local_query = TNT_LOCAL_QUERY (query);
  TntMatch      *match;

  assert_types (query);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  match = find_match_by_file_id (local_query, file_id);

  if (match == NULL)
  {
    if (query_matches_meta (local_query, file_meta))
    {
      match = new_match (local_query, file_id, file_meta);
      run_callbacks (local_query, local_query->match_pre_add_funcs, match);

      if (node_id != TNT_NODE_ID_INVALID && remote_blocks > 0)
        tnt_match_add_hit (match, node_id, remote_blocks);

      if (query_matches_match (local_query, match))
      {
        insert_match (local_query, match);

        DEBUG (g_print ("New match: %s %s %s\n", tnt_file_meta_get_best_title (file_meta),
                        completeness == TNT_COMPLETENESS_COMPLETE || completeness == TNT_COMPLETENESS_PARTIAL ?
                        "[data]" : "", node_id != TNT_NODE_ID_INVALID ? "[node]" : ""));
      }
      else
      {
        tnt_match_free (match);
        match = NULL;
      }
    }
  }
  else
  {
    gboolean     changed = FALSE;
    TntFileMeta *old_meta;
    TntFileMeta *merged_meta;

    g_assert (tnt_match_is_file (match));

    if (node_id != TNT_NODE_ID_INVALID && remote_blocks > 0)
      changed = tnt_match_add_hit (match, node_id, remote_blocks);

    old_meta = tnt_match_get_file_meta (match);
    merged_meta = tnt_file_meta_merge (old_meta, file_meta);

    if (query_matches_meta (local_query, merged_meta))
    {
      if (same_group (old_meta, merged_meta))
      {
        /* Check if the metadata changed at all */

        if (!changed && !same_meta (old_meta, merged_meta))
          changed = TRUE;

        match->file_meta = merged_meta;
      }
      else
      {
        TntMatch *parent;

        /* Remove from old group */

        g_hash_table_remove (local_query->file_id_hash, (gpointer) tnt_match_get_file_id (match));
        run_callbacks (local_query, local_query->match_remove_funcs, match);

        parent = tnt_match_get_parent (match);

        if (tnt_match_get_parent (parent))
        {
          TntFileMeta *group_meta;
          TntFileMeta *match_meta;

          /* Parent is a group */

          group_meta = tnt_match_get_file_meta (parent);
          match_meta = tnt_match_get_file_meta (match);

          /* Accumulate duration */
          tnt_file_meta_set_duration (group_meta, tnt_file_meta_get_duration (group_meta) -
                                      tnt_file_meta_get_duration (match_meta));

          /* Average quality */
          if (tnt_file_meta_get_quality (match_meta))
          {
            guint quality;
            guint n_quality_children;

            quality = tnt_file_meta_get_quality (group_meta);
            quality *= tnt_match_get_n_quality_children (parent);
            quality -= tnt_file_meta_get_quality (match_meta);
            tnt_match_dec_quality_children (parent);
            n_quality_children = tnt_match_get_n_quality_children (parent);
            if (n_quality_children)
            {
              quality /= n_quality_children;
              tnt_file_meta_set_quality (group_meta, quality);
            }
            else
            {
              tnt_file_meta_set_quality (group_meta, 0);
            }
          }

          /* FIXME: No way to reset dimensions if all dimensioned children go away? */
        }
        else
        {
          /* Parent is the top-level node, remove from group index */
          remove_from_group_index (local_query, match);
        }

        tnt_match_unlink (match);

        /* If containing group has just one element left, move the element up to top level
         * and nuke the group */
        if (tnt_match_get_parent (parent) && tnt_match_get_n_children (parent) == 1)
        {
          TntMatch *child;

          remove_from_group_index (local_query, parent);

          child = g_ptr_array_index (tnt_match_get_children (parent), 0);

          run_callbacks (local_query, local_query->match_remove_funcs, child);
          g_hash_table_remove (local_query->file_id_hash, (gpointer) tnt_match_get_file_id (child));
          tnt_match_unlink (child);

          run_callbacks (local_query, local_query->match_remove_funcs, parent);
          tnt_match_unlink (parent);
          tnt_match_free (parent);

          insert_match (local_query, child);
        }

        /* Reinsert */

        match->file_meta = merged_meta;
        insert_match (local_query, match);

        /* Don't send any other updates */
        changed = FALSE;
      }

      tnt_file_meta_free (old_meta);
    }
    else
    {
      tnt_file_meta_free (merged_meta);
    }

    if (changed)
    {
      TntMatch *parent;

      parent = tnt_match_get_parent (match);

      run_callbacks (local_query, local_query->match_update_funcs, match);
      if (parent && tnt_match_get_parent (parent))
        run_callbacks (local_query, local_query->match_update_funcs, parent);
    }
  }

  /* May have been changed in callbacks, triggering a removal here */
  if (match && tnt_match_for_all_remove (match, filter_matches_cb, query))
    match = NULL;

  return match ? TRUE : FALSE;
}

static gint
get_n_matches (TntQuery *query)
{
  TntLocalQuery *local_query = TNT_LOCAL_QUERY (query);

  assert_types (local_query);

  return g_hash_table_size (local_query->file_id_hash);
}

void
tnt_local_query_finalize (FluxObject *object)
{
  TntLocalQuery *query = TNT_LOCAL_QUERY (object);

  assert_types (query);

  tnt_match_free (query->root_match);
  tnt_match_reclaim_memory ();
  tnt_file_meta_reclaim_memory ();

  if (query->have_grace_timeout)
    g_source_remove (query->grace_timeout);

  g_hash_table_destroy (query->file_id_hash);
  g_hash_table_destroy (query->collection_hash);
  g_hash_table_destroy (query->keyword_hash);

  tnt_query_finalize (object);
}

void
tnt_local_query_init (TntLocalQuery *local_query)
{
  TntQuery *query;

  tnt_query_init                (TNT_QUERY (local_query));
  flux_object_set_type          (FLUX_OBJECT (local_query), TNT_LOCAL_QUERY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (local_query), tnt_local_query_finalize);

  query = TNT_QUERY (local_query);

  query->filter_func        = filter_matches;
  query->get_n_matches_func = get_n_matches;
  query->try_match_func     = try_match;

  local_query->root_match = tnt_match_new ();
  tnt_match_convert_to_group (local_query->root_match);

  local_query->file_id_hash    = g_hash_table_new      ((GHashFunc)      tnt_file_id_hash,
                                                        (GCompareFunc)   tnt_file_id_equal);

  local_query->collection_hash = g_hash_table_new_full ((GHashFunc)      g_str_hash,
                                                        (GCompareFunc)   g_str_equal,
                                                        (GDestroyNotify) g_free,
                                                        (GDestroyNotify) NULL);

  local_query->keyword_hash    = g_hash_table_new_full ((GHashFunc)      g_str_hash,
                                                        (GCompareFunc)   g_str_equal,
                                                        (GDestroyNotify) g_free,
                                                        (GDestroyNotify) NULL);
}

TntLocalQuery *
tnt_local_query_new (void)
{
  TntLocalQuery *query;

  query = g_new0 (TntLocalQuery, 1);
  tnt_local_query_init (query);

  return query;
}

void
tnt_local_query_add_match_pre_add_func (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  MatchFunc *match_func;

  assert_types (query);
  g_return_if_fail (func != NULL);

  match_func = g_new (MatchFunc, 1);
  match_func->func = func;
  match_func->data = data;

  query->match_pre_add_funcs = g_list_append (query->match_pre_add_funcs, match_func);
}

void
tnt_local_query_add_match_add_func (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  MatchFunc *match_func;

  assert_types (query);
  g_return_if_fail (func != NULL);

  match_func = g_new (MatchFunc, 1);
  match_func->func = func;
  match_func->data = data;

  query->match_add_funcs = g_list_append (query->match_add_funcs, match_func);
}

void
tnt_local_query_add_match_update_func (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  MatchFunc *match_func;

  assert_types (query);
  g_return_if_fail (func != NULL);

  match_func = g_new (MatchFunc, 1);
  match_func->func = func;
  match_func->data = data;

  query->match_update_funcs = g_list_append (query->match_update_funcs, match_func);
}

void
tnt_local_query_add_match_remove_func (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  MatchFunc *match_func;

  assert_types (query);
  g_return_if_fail (func != NULL);

  match_func = g_new (MatchFunc, 1);
  match_func->func = func;
  match_func->data = data;

  query->match_remove_funcs = g_list_append (query->match_remove_funcs, match_func);
}

void
tnt_local_query_remove_match_func (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  assert_types (query);
  g_return_if_fail (func != NULL);

  remove_callback (&query->match_pre_add_funcs, func, data);
  remove_callback (&query->match_add_funcs,     func, data);
  remove_callback (&query->match_update_funcs,  func, data);
  remove_callback (&query->match_remove_funcs,  func, data);
}

TntMatch *
tnt_local_query_get_root_match (TntLocalQuery *query)
{
  assert_types (query);

  return query->root_match;
}

TntMatch *
tnt_local_query_get_match_n (TntLocalQuery *query, gint match_n)
{
  const GPtrArray *children;

  assert_types (query);
  g_return_val_if_fail (match_n < tnt_match_get_n_children (query->root_match), NULL);

  children = tnt_match_get_children (query->root_match);
  return g_ptr_array_index (children, match_n);
}

TntMatch *
tnt_local_query_find_match_by_file_id (TntLocalQuery *query, TntFileID *file_id)
{
  assert_types (query);
  g_return_val_if_fail (file_id != NULL, NULL);

  return find_match_by_file_id (query, file_id);
}

void
tnt_local_query_match_updated (TntLocalQuery *query, TntMatch *match)
{
  assert_types (query);
  g_return_if_fail (match != NULL);

  if ((!tnt_query_get_match_local (TNT_QUERY (query)) && tnt_match_is_local (match)) ||
      (!tnt_query_get_match_partial (TNT_QUERY (query)) && !tnt_match_is_local (match) &&
       tnt_match_get_completeness (match) != TNT_COMPLETENESS_COMPLETE))
  {
    TntMatch *parent = match->parent;

    tnt_match_for_all_remove (match, filter_matches_cb, query);

    /* If parent group is left with 0 children, and it's not the root,
     * remove that too */
    if (parent && parent->parent && tnt_match_get_n_children (parent) == 0)
    {
      remove_from_group_index (query, parent);
      run_callbacks (query, query->match_remove_funcs, parent);
      tnt_match_free (parent);
    }
  }
  else
    run_callbacks (query, query->match_update_funcs, match);
}

typedef struct
{
  TntLocalQuery *query;
  TntMatchFunc  *func;
  gpointer       func_data;
}
TraversalContext;

static gboolean
traverse_all_cb (TntMatch *match, gpointer data)
{
  TraversalContext *context = data;

  /* Don't call back for the root node */
  if (!tnt_match_get_parent (match))
    return FALSE;

  context->func (context->query, match, context->func_data);
  return FALSE;
}

void
tnt_local_query_for_all_matches (TntLocalQuery *query, TntMatchFunc *func, gpointer data)
{
  TraversalContext *context;

  context = g_new0 (TraversalContext, 1);
  context->query     = query;
  context->func      = func;
  context->func_data = data;

  tnt_match_for_all (query->root_match, traverse_all_cb, context);

  g_free (context);
}
