/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-log.c - Logger.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <time.h>
#include "tnt.h"

#ifndef G_PLATFORM_WIN32
# define TENTACLES_DIR ".tentacles"
#else
# define TENTACLES_DIR "tentacles"
#endif

typedef struct
{
  TntLogLevel  min_level;
  TntLogFlags  mask;
  TntLogFunc  *func;
  gpointer     func_data;
}
Recipient;

static GList    *recipients_list  = NULL;
static FILE     *debug_log_fh     = NULL;
static gboolean  debug_log_opened = FALSE;

void
tnt_log_add_recipient (TntLogLevel min_level, TntLogFlags mask, TntLogFunc *func, gpointer data)
{
  Recipient *recipient;

  g_return_if_fail (func != NULL);

  recipient = g_new0 (Recipient, 1);
  recipient->min_level = min_level;
  recipient->mask      = mask;
  recipient->func      = func;
  recipient->func_data = data;

  recipients_list = g_list_append (recipients_list, recipient);
}

void
tnt_log_remove_recipient (TntLogFunc *func)
{
  GList *l;

  g_return_if_fail (func != NULL);

  for (l = recipients_list; l; l = g_list_next (l))
  {
    Recipient *recipient = l->data;

    if (recipient->func == func)
    {
      recipients_list = g_list_remove_link (recipients_list, l);
      g_free (recipient);
      g_list_free_1 (l);
      break;
    }
  }
}

void
tnt_log_v (TntLogSubsystem subsys, TntLogLevel level, TntLogMood mood,
           TntLogFlags flags, const gchar *format, va_list args)
{
  gchar *message;
  GList *l;

  g_return_if_fail (format != NULL);

  message = g_strdup_vprintf (format, args);

  for (l = recipients_list; l; l = g_list_next (l))
  {
    Recipient *recipient = l->data;

    if (recipient->min_level <= level && (flags & recipient->mask) == flags)
      recipient->func (subsys, level, mood, flags, message, recipient->func_data);
  }

  g_free (message);
}

void
tnt_log (TntLogSubsystem subsys, TntLogLevel level, TntLogMood mood,
         TntLogFlags flags, const gchar *format, ...)
{
  va_list args;

  g_return_if_fail (format != NULL);

  va_start (args, format);
  tnt_log_v (subsys, level, mood, flags, format, args);
  va_end (args);
}

static gchar *
generate_timestamp (glong sec)
{
  time_t     tt;
  struct tm *tm;

  tt = sec;
  tm = localtime (&tt);

  return g_strdup_printf ("[%02d.%02d %02d:%02d:%02d]",
                          tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
}

void
tnt_log_debug_v (const gchar *format, va_list args)
{
  GTimeVal  gtv;
  gchar    *message;
  gchar    *timestamp;
  gchar    *log_entry;

  if (!debug_log_opened)
  {
    const gchar *home_path;
    gchar       *full_path;

    home_path = g_get_home_dir ();
    full_path = g_strjoin ("/", home_path, TENTACLES_DIR "/log", NULL);

    debug_log_fh = fopen (full_path, "at");
    debug_log_opened = TRUE;

    g_free (full_path);

    if (!debug_log_fh)
      g_warning ("Couldn't open log file!");
  }

  if (!debug_log_fh)
    return;

  g_get_current_time (&gtv);
  timestamp = generate_timestamp (gtv.tv_sec);
  message   = g_strdup_vprintf (format, args);
  log_entry = g_strjoin (" ", timestamp, message, NULL);

  fputs (log_entry, debug_log_fh);
  fputc ('\n', debug_log_fh);
  fflush (debug_log_fh);

  g_free (log_entry);
  g_free (message);
  g_free (timestamp);
}

void
tnt_log_debug (const gchar *format, ...)
{
  va_list args;

  g_return_if_fail (format != NULL);

  va_start (args, format);
  tnt_log_debug_v (format, args);
  va_end (args);
}
