/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-mpeg.c - MPEG1/2 metadata extraction.
 * Uses code from mpeg2dec.
 *
 * Copyright (C) 2005      Joakim Ziegler
 * Copyright (C) 2000-2003 Michel Lespinasse
 * Copyright (C) 1999-2000 Aaron Holtzman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Joakim Ziegler    <joakim@avmaria.com>
 *          Michel Lespinasse <walken@zoy.org>
 *          Aaron Holtzman    <aholtzma@ess.engr.uvic.ca>
 */

#include "config.h"
#include <string.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#else
# include <inttypes.h>
#endif

#include "tnt.h"

#ifdef ATTRIBUTE_ALIGNED_MAX
#define ATTR_ALIGN(align) __attribute__ ((__aligned__ ((ATTRIBUTE_ALIGNED_MAX < align) ? ATTRIBUTE_ALIGNED_MAX : align)))
#else
#define ATTR_ALIGN(align)
#endif

#ifdef HAVE_BUILTIN_EXPECT
#define likely(x) __builtin_expect ((x) != 0, 1)
#define unlikely(x) __builtin_expect ((x) != 0, 0)
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif

#define BUFFER_SIZE (1194 * 1024)

#define SEQ_FLAG_MPEG2 1
#define SEQ_FLAG_CONSTRAINED_PARAMETERS 2
#define SEQ_FLAG_PROGRESSIVE_SEQUENCE 4
#define SEQ_FLAG_LOW_DELAY 8
#define SEQ_FLAG_COLOUR_DESCRIPTION 16

#define SEQ_MASK_VIDEO_FORMAT 0xe0
#define SEQ_VIDEO_FORMAT_COMPONENT 0
#define SEQ_VIDEO_FORMAT_PAL 0x20
#define SEQ_VIDEO_FORMAT_NTSC 0x40
#define SEQ_VIDEO_FORMAT_SECAM 0x60
#define SEQ_VIDEO_FORMAT_MAC 0x80
#define SEQ_VIDEO_FORMAT_UNSPECIFIED 0xa0

#define SEQ_EXT 2
#define SEQ_DISPLAY_EXT 4
#define QUANT_MATRIX_EXT 8
#define COPYRIGHT_EXT 0x10
#define PIC_DISPLAY_EXT 0x80
#define PIC_CODING_EXT 0x100

/* macroblock modes */
#define MACROBLOCK_INTRA 1
#define MACROBLOCK_PATTERN 2
#define MACROBLOCK_MOTION_BACKWARD 4
#define MACROBLOCK_MOTION_FORWARD 8
#define MACROBLOCK_QUANT 16
#define DCT_TYPE_INTERLACED 32
/* motion_type */
#define MOTION_TYPE_SHIFT 6
#define MC_FIELD 1
#define MC_FRAME 2
#define MC_16X8 2
#define MC_DMV 3

/* picture structure */
#define TOP_FIELD 1
#define BOTTOM_FIELD 2
#define FRAME_PICTURE 3

/* picture coding type */
#define I_TYPE 1
#define P_TYPE 2
#define B_TYPE 3
#define D_TYPE 4

#define GOP_FLAG_DROP_FRAME 1
#define GOP_FLAG_BROKEN_LINK 2
#define GOP_FLAG_CLOSED_GOP 4

#define PIC_MASK_CODING_TYPE 7
#define PIC_FLAG_CODING_TYPE_I 1
#define PIC_FLAG_CODING_TYPE_P 2
#define PIC_FLAG_CODING_TYPE_B 3
#define PIC_FLAG_CODING_TYPE_D 4

#define PIC_FLAG_TOP_FIELD_FIRST 8
#define PIC_FLAG_PROGRESSIVE_FRAME 16
#define PIC_FLAG_COMPOSITE_DISPLAY 32
#define PIC_FLAG_SKIP 64
#define PIC_FLAG_TAGS 128
#define PIC_MASK_COMPOSITE_DISPLAY 0xfffff000

#define MPEG2_ACCEL_X86_MMX 1
#define MPEG2_ACCEL_X86_3DNOW 2
#define MPEG2_ACCEL_X86_MMXEXT 4
#define MPEG2_ACCEL_PPC_ALTIVEC 1
#define MPEG2_ACCEL_ALPHA 1
#define MPEG2_ACCEL_ALPHA_MVI 2
#define MPEG2_ACCEL_SPARC_VIS 1
#define MPEG2_ACCEL_SPARC_VIS2 2
#define MPEG2_ACCEL_DETECT 0x80000000

#define RECEIVED(code,state) (((state) << 8) + (code))
#define META_BUFFER_SIZE 1024

typedef enum {
    STATE_BUFFER = 0,
    STATE_SEQUENCE = 1,
    STATE_SEQUENCE_REPEATED = 2,
    STATE_GOP = 3,
    STATE_PICTURE = 4,
    STATE_SLICE_1ST = 5,
    STATE_PICTURE_2ND = 6,
    STATE_SLICE = 7,
    STATE_END = 8,
    STATE_INVALID = 9,
    STATE_INVALID_END = 10
} mpeg2_state_t;

typedef enum {
    MPEG2_CONVERT_SET = 0,
    MPEG2_CONVERT_STRIDE = 1,
    MPEG2_CONVERT_START = 2
} mpeg2_convert_stage_t;

typedef enum {
    MPEG2_ALLOC_MPEG2DEC = 0,
    MPEG2_ALLOC_CHUNK = 1,
    MPEG2_ALLOC_YUV = 2,
    MPEG2_ALLOC_CONVERT_ID = 3,
    MPEG2_ALLOC_CONVERTED = 4
} mpeg2_alloc_t;


static const guint8 default_intra_quantizer_matrix[64] ATTR_ALIGN(16) = {
    8,
    16, 16,
    19, 16, 19,
    22, 22, 22, 22,
    22, 22, 26, 24, 26,
    27, 27, 27, 26, 26, 26,
    26, 27, 27, 27, 29, 29, 29,
    34, 34, 34, 29, 29, 29, 27, 27,
    29, 29, 32, 32, 34, 34, 37,
    38, 37, 35, 35, 34, 35,
    38, 38, 40, 40, 40,
    48, 48, 46, 46,
    56, 56, 58,
    69, 69,
    83
};

guint8 mpeg2_scan_norm[64] ATTR_ALIGN(16) = {
    /* Zig-Zag scan pattern */
     0,  1,  8, 16,  9,  2,  3, 10, 17, 24, 32, 25, 18, 11,  4,  5,
    12, 19, 26, 33, 40, 48, 41, 34, 27, 20, 13,  6,  7, 14, 21, 28,
    35, 42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
    58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63
};

guint8 mpeg2_scan_alt[64] ATTR_ALIGN(16) = {
    /* Alternate scan pattern */
     0, 8,  16, 24,  1,  9,  2, 10, 17, 25, 32, 40, 48, 56, 57, 49,
    41, 33, 26, 18,  3, 11,  4, 12, 19, 27, 34, 42, 50, 58, 35, 43,
    51, 59, 20, 28,  5, 13,  6, 14, 21, 29, 36, 44, 52, 60, 37, 45,
    53, 61, 22, 30,  7, 15, 23, 31, 38, 46, 54, 62, 39, 47, 55, 63
};

typedef struct mpeg2dec_s mpeg2dec_t;
typedef struct mpeg2_decoder_s mpeg2_decoder_t;
typedef struct mpeg2_info_s mpeg2_info_t;
typedef struct mpeg2_convert_init_s mpeg2_convert_init_t;
typedef void mpeg2_mc_fct (guint8 *, const guint8 *, int, int);

typedef struct mpeg2_fbuf_s {
    guint8 * buf[3];
    void * id;
} mpeg2_fbuf_t;

typedef struct {
    guint8 * ref[2][3];
    guint8 ** ref2[2];
    int pmv[2][2];
    int f_code[2];
} motion_t;

typedef void motion_parser_t (mpeg2_decoder_t * decoder,
			      motion_t * motion,
			      mpeg2_mc_fct * const * table);

typedef struct mpeg2_gop_s {
  guint8 hours;
  guint8 minutes;
  guint8 seconds;
  guint8 pictures;
  guint32 flags;
} mpeg2_gop_t;

typedef struct mpeg2_picture_s {
    unsigned int temporal_reference;
    unsigned int nb_fields;
    guint32 tag, tag2;
    guint32 flags;
    struct {
	int x, y;
    } display_offset[3];
} mpeg2_picture_t;

struct mpeg2_decoder_s {
  /* first, state that carries information from one macroblock to the */
  /* next inside a slice, and is never used outside of mpeg2_slice() */
  
  /* bit parsing stuff */
  guint32 bitstream_buf;		/* current 32 bit working set */
  int bitstream_bits;			/* used bits in working set */
  const guint8 * bitstream_ptr;	/* buffer with stream data */
  
  guint8 * dest[3];
  
  int offset;
  int stride;
  int uv_stride;
  int slice_stride;
  int slice_uv_stride;
  int stride_frame;
  unsigned int limit_x;
  unsigned int limit_y_16;
  unsigned int limit_y_8;
  unsigned int limit_y;
  
  /* Motion vectors */
  /* The f_ and b_ correspond to the forward and backward motion */
  /* predictors */
  motion_t b_motion;
  motion_t f_motion;
  motion_parser_t * motion_parser[5];
  
  /* predictor for DC coefficients in intra blocks */
  gint16 dc_dct_pred[3];
  
  /* DCT coefficients */
  gint16 DCTblock[64] ATTR_ALIGN(64);
  
  guint8 * picture_dest[3];
  void (* convert) (void * convert_id, guint8 * const * src,
                    unsigned int v_offset);
  void * convert_id;
  
  int dmv_offset;
  unsigned int v_offset;
  
  /* now non-slice-specific information */
  
  /* sequence header stuff */
  guint16 * quantizer_matrix[4];
  guint16 (* chroma_quantizer[2])[64];
  guint16 quantizer_prescale[4][32][64];
  
  /* The width and height of the picture snapped to macroblock units */
  int width;
  int height;
  int vertical_position_extension;
  int chroma_format;
  
  /* picture header stuff */
  
  /* what type of picture this is (I, P, B, D) */
  int coding_type;
  
  /* picture coding extension stuff */
  
  /* quantization factor for intra dc coefficients */
  int intra_dc_precision;
  /* top/bottom/both fields */
  int picture_structure;
  /* bool to indicate all predictions are frame based */
  int frame_pred_frame_dct;
  /* bool to indicate whether intra blocks have motion vectors */
  /* (for concealment) */
  int concealment_motion_vectors;
  /* bool to use different vlc tables */
  int intra_vlc_format;
  /* used for DMV MC */
  int top_field_first;
  
  /* stuff derived from bitstream */
  
  /* pointer to the zigzag scan we're supposed to be using */
  const guint8 * scan;
  
  int second_field;
  
  int mpeg1;
};

typedef struct {
  mpeg2_fbuf_t fbuf;
} fbuf_alloc_t;

typedef struct mpeg2_sequence_s {
  unsigned int width, height;
  unsigned int chroma_width, chroma_height;
  unsigned int byte_rate;
  unsigned int vbv_buffer_size;
  guint32 flags;
  
  unsigned int picture_width, picture_height;
  unsigned int display_width, display_height;
  unsigned int pixel_width, pixel_height;
  unsigned int frame_period;
  
  guint8 profile_level_id;
  guint8 colour_primaries;
  guint8 transfer_characteristics;
  guint8 matrix_coefficients;
} mpeg2_sequence_t;

struct mpeg2_info_s {
    const mpeg2_sequence_t * sequence;
    const mpeg2_gop_t * gop;
    const mpeg2_picture_t * current_picture;
    const mpeg2_picture_t * current_picture_2nd;
    const mpeg2_fbuf_t * current_fbuf;
    const mpeg2_picture_t * display_picture;
    const mpeg2_picture_t * display_picture_2nd;
    const mpeg2_fbuf_t * display_fbuf;
    const mpeg2_fbuf_t * discard_fbuf;
    const guint8 * user_data;
    unsigned int user_data_len;
};

typedef int mpeg2_convert_t (int stage, void * id,
			     const mpeg2_sequence_t * sequence, int stride,
			     guint32 accel, void * arg,
			     mpeg2_convert_init_t * result);

struct mpeg2dec_s {
  mpeg2_decoder_t decoder;
  
  mpeg2_info_t info;
  
  guint32 shift;
  int is_display_initialized;
  mpeg2_state_t (* action) (struct mpeg2dec_s * mpeg2dec);
  mpeg2_state_t state;
  guint32 ext_state;
  
  /* allocated in init - gcc has problems allocating such big structures */
  guint8 * chunk_buffer;
  /* pointer to start of the current chunk */
  guint8 * chunk_start;
  /* pointer to current position in chunk_buffer */
  guint8 * chunk_ptr;
  /* last start code ? */
  guint8 code;
  
  /* picture tags */
  guint32 tag_current, tag2_current, tag_previous, tag2_previous;
  int num_tags;
  int bytes_since_tag;
  
  int first;
  int alloc_index_user;
  int alloc_index;
  guint8 first_decode_slice;
  guint8 nb_decode_slices;
  
  unsigned int user_data_len;
  
  mpeg2_sequence_t new_sequence;
  mpeg2_sequence_t sequence;
  mpeg2_gop_t new_gop;
  mpeg2_gop_t gop;
  mpeg2_picture_t new_picture;
  mpeg2_picture_t pictures[4];
  mpeg2_picture_t * picture;
  /*const*/ mpeg2_fbuf_t * fbuf[3];	/* 0: current fbuf, 1-2: prediction fbufs */
  
  fbuf_alloc_t fbuf_alloc[3];
  int custom_fbuf;
  
  guint8 * yuv_buf[3][3];
  int yuv_index;
  mpeg2_convert_t * convert;
  void * convert_arg;
  unsigned int convert_id_size;
  int convert_stride;
  void (* convert_start) (void * id, const mpeg2_fbuf_t * fbuf,
                          const mpeg2_picture_t * picture,
                          const mpeg2_gop_t * gop);
  
  guint8 * buf_start;
  guint8 * buf_end;
  
  gint16 display_offset_x, display_offset_y;
  
  int copy_matrix;
  gint8 q_scale_type, scaled[4];
  guint8 quantizer_matrix[4][64];
  guint8 new_quantizer_matrix[4][64];
};

struct mpeg2_convert_init_s {
    unsigned int id_size;
    unsigned int buf_size[3];
    void (* start) (void * id, const mpeg2_fbuf_t * fbuf,
		    const mpeg2_picture_t * picture, const mpeg2_gop_t * gop);
    void (* copy) (void * id, guint8 * const * src, unsigned int v_offset);
};


static inline mpeg2_state_t seek_chunk (mpeg2dec_t *mpeg2dec);
mpeg2_state_t mpeg2_parse_header (mpeg2dec_t *mpeg2dec);
mpeg2_state_t mpeg2_header_picture_start (mpeg2dec_t *mpeg2dec);
void mpeg2_reset_info (mpeg2_info_t *info);
void mpeg2_header_state_init (mpeg2dec_t * mpeg2dec);


static int mpeg2_accels = 0;


const mpeg2_info_t * mpeg2_info (mpeg2dec_t * mpeg2dec)
{
    return &(mpeg2dec->info);
}

mpeg2_state_t mpeg2_seek_header (mpeg2dec_t * mpeg2dec)
{
  while (mpeg2dec->code != 0xb3 &&
         ((mpeg2dec->code != 0xb7 && mpeg2dec->code != 0xb8 &&
           mpeg2dec->code) || mpeg2dec->sequence.width == (unsigned)-1))
    if (seek_chunk (mpeg2dec) == STATE_BUFFER)
	    return STATE_BUFFER;
  mpeg2dec->chunk_start = mpeg2dec->chunk_ptr = mpeg2dec->chunk_buffer;
  mpeg2dec->user_data_len = 0;
  return (mpeg2dec->code ? mpeg2_parse_header (mpeg2dec) :
          mpeg2_header_picture_start (mpeg2dec));
}

void mpeg2_reset (mpeg2dec_t * mpeg2dec, int full_reset)
{
  mpeg2dec->buf_start = mpeg2dec->buf_end = NULL;
  mpeg2dec->num_tags = 0;
  mpeg2dec->shift = 0xffffff00;
  mpeg2dec->code = 0xb4;
  mpeg2dec->action = mpeg2_seek_header;
  mpeg2dec->state = STATE_INVALID;
  mpeg2dec->first = 1;
  
  mpeg2_reset_info(&(mpeg2dec->info));
  mpeg2dec->info.gop = NULL;
  mpeg2dec->info.user_data = NULL;
  mpeg2dec->info.user_data_len = 0;
  if (full_reset) {
    mpeg2dec->info.sequence = NULL;
    mpeg2_header_state_init (mpeg2dec);
  }
}

static inline int skip_chunk (mpeg2dec_t * mpeg2dec, int bytes)
{
  guint8 * current;
  guint32 shift;
  guint8 * limit;
  guint8 byte;

  if (!bytes)
    return 0;

  current = mpeg2dec->buf_start;
  shift = mpeg2dec->shift;
  limit = current + bytes;
  
  do
  {
    byte = *current++;
    if (shift == 0x00000100)
    {
	    int skipped;
      
	    mpeg2dec->shift = 0xffffff00;
	    skipped = current - mpeg2dec->buf_start;
	    mpeg2dec->buf_start = current;
	    return skipped;
    }
    shift = (shift | byte) << 8;
  }
  while (current < limit);
  
  mpeg2dec->shift = shift;
  mpeg2dec->buf_start = current;
  return 0;
}

void mpeg2_buffer (mpeg2dec_t * mpeg2dec, guint8 * start, guint8 * end)
{
  mpeg2dec->buf_start = start;
  mpeg2dec->buf_end = end;
}

static inline mpeg2_state_t seek_chunk (mpeg2dec_t * mpeg2dec)
{
  int size, skipped;
  
  size = mpeg2dec->buf_end - mpeg2dec->buf_start;
  skipped = skip_chunk (mpeg2dec, size);
  if (!skipped)
  {
    mpeg2dec->bytes_since_tag += size;
    return STATE_BUFFER;
  }
  mpeg2dec->bytes_since_tag += skipped;
  mpeg2dec->code = mpeg2dec->buf_start[-1];
  return (mpeg2_state_t)-1;
}

int mpeg2_header_picture (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_picture_t * picture = &(mpeg2dec->new_picture);
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  int type;
  
  type = (buffer [1] >> 3) & 7;
  mpeg2dec->ext_state = PIC_CODING_EXT;
  
  picture->temporal_reference = (buffer[0] << 2) | (buffer[1] >> 6);
  
  picture->flags |= type;

  if (type == PIC_FLAG_CODING_TYPE_P || type == PIC_FLAG_CODING_TYPE_B)
  {
    /* forward_f_code and backward_f_code - used in mpeg1 only */
    decoder->f_motion.f_code[1] = (buffer[3] >> 2) & 1;
    decoder->f_motion.f_code[0] =
	    (((buffer[3] << 1) | (buffer[4] >> 7)) & 7) - 1;
    decoder->b_motion.f_code[1] = (buffer[4] >> 6) & 1;
    decoder->b_motion.f_code[0] = ((buffer[4] >> 3) & 7) - 1;
  }
  
  /* XXXXXX decode extra_information_picture as well */
  
  picture->nb_fields = 2;
  
  mpeg2dec->q_scale_type = 0;
  decoder->intra_dc_precision = 7;
  decoder->frame_pred_frame_dct = 1;
  decoder->concealment_motion_vectors = 0;
  decoder->scan = mpeg2_scan_norm;
  decoder->picture_structure = FRAME_PICTURE;
  mpeg2dec->copy_matrix = 0;
  
  return 0;
}

static int sequence_ext (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_sequence_t * sequence = &(mpeg2dec->new_sequence);
  guint32 flags;
  
  if (!(buffer[3] & 1))
  return 1;
  
  sequence->profile_level_id = (buffer[0] << 4) | (buffer[1] >> 4);
  
  sequence->display_width = sequence->picture_width +=
    ((buffer[1] << 13) | (buffer[2] << 5)) & 0x3000;
  sequence->display_height = sequence->picture_height +=
    (buffer[2] << 7) & 0x3000;
  sequence->width = (sequence->picture_width + 15) & ~15;
  sequence->height = (sequence->picture_height + 15) & ~15;
  flags = sequence->flags | SEQ_FLAG_MPEG2;
  if (!(buffer[1] & 8))
  {
    flags &= ~SEQ_FLAG_PROGRESSIVE_SEQUENCE;
    sequence->height = (sequence->height + 31) & ~31;
  }
  if (buffer[5] & 0x80)
    flags |= SEQ_FLAG_LOW_DELAY;
  sequence->flags = flags;
  sequence->chroma_width = sequence->width;
  sequence->chroma_height = sequence->height;
  switch (buffer[1] & 6)
  {
  case 0:	/* invalid */
    return 1;
  case 2:	/* 4:2:0 */
    sequence->chroma_height >>= 1;
  case 4:	/* 4:2:2 */
    sequence->chroma_width >>= 1;
  }
  
  sequence->byte_rate += ((buffer[2]<<25) | (buffer[3]<<17)) & 0x3ffc0000;
  sequence->vbv_buffer_size |= buffer[4] << 21;
  sequence->frame_period =
    sequence->frame_period * ((buffer[5]&31)+1) / (((buffer[5]>>2)&3)+1);
  
  mpeg2dec->ext_state = SEQ_DISPLAY_EXT;
  
  return 0;
}

static int sequence_display_ext (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_sequence_t * sequence = &(mpeg2dec->new_sequence);
  guint32 flags;
  
  flags = ((sequence->flags & ~SEQ_MASK_VIDEO_FORMAT) |
           ((buffer[0]<<4) & SEQ_MASK_VIDEO_FORMAT));
  if (buffer[0] & 1)
  {
    flags |= SEQ_FLAG_COLOUR_DESCRIPTION;
    sequence->colour_primaries = buffer[1];
    sequence->transfer_characteristics = buffer[2];
    sequence->matrix_coefficients = buffer[3];
    buffer += 3;
  }
  
  if (!(buffer[2] & 2))	/* missing marker_bit */
    return 1;
  
  sequence->display_width = (buffer[1] << 6) | (buffer[2] >> 2);
  sequence->display_height =
    ((buffer[2]& 1 ) << 13) | (buffer[3] << 5) | (buffer[4] >> 3);
  
  return 0;
}

static int quant_matrix_ext (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  int i, j;
  
  for (i = 0; i < 4; i++)
  {
    if (buffer[0] & (8 >> i))
    {
	    for (j = 0; j < 64; j++)
      {
        mpeg2dec->new_quantizer_matrix[i][mpeg2_scan_norm[j]] =
          (buffer[j] << (i+5)) | (buffer[j+1] >> (3-i));
      }
	    mpeg2dec->copy_matrix |= 1 << i;
	    buffer += 64;
    }
  }
  return 0;
}

static int copyright_ext (mpeg2dec_t * mpeg2dec)
{
    return 0;
}

static int picture_display_ext (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_picture_t * picture = &(mpeg2dec->new_picture);
  int i, nb_pos;
  
  nb_pos = picture->nb_fields;
  if (mpeg2dec->sequence.flags & SEQ_FLAG_PROGRESSIVE_SEQUENCE)
    nb_pos >>= 1;
  
  for (i = 0; i < nb_pos; i++)
  {
    int x, y;
    
    x = ((buffer[4*i] << 24) | (buffer[4*i+1] << 16) |
         (buffer[4*i+2] << 8) | buffer[4*i+3]) >> (11-2*i);
    y = ((buffer[4*i+2] << 24) | (buffer[4*i+3] << 16) |
         (buffer[4*i+4] << 8) | buffer[4*i+5]) >> (10-2*i);
    if (! (x & y & 1))
	    return 1;
    picture->display_offset[i].x = mpeg2dec->display_offset_x = x >> 1;
    picture->display_offset[i].y = mpeg2dec->display_offset_y = y >> 1;
  }
  for (; i < 3; i++)
  {
    picture->display_offset[i].x = mpeg2dec->display_offset_x;
    picture->display_offset[i].y = mpeg2dec->display_offset_y;
  }
  return 0;
}

static int picture_coding_ext (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_picture_t * picture = &(mpeg2dec->new_picture);
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  guint32 flags;
  
  /* pre subtract 1 for use later in compute_motion_vector */
  decoder->f_motion.f_code[0] = (buffer[0] & 15) - 1;
  decoder->f_motion.f_code[1] = (buffer[1] >> 4) - 1;
  decoder->b_motion.f_code[0] = (buffer[1] & 15) - 1;
  decoder->b_motion.f_code[1] = (buffer[2] >> 4) - 1;
  
  flags = picture->flags;
  decoder->intra_dc_precision = 7 - ((buffer[2] >> 2) & 3);
  decoder->picture_structure = buffer[2] & 3;
  switch (decoder->picture_structure)
  {
  case TOP_FIELD:
    flags |= PIC_FLAG_TOP_FIELD_FIRST;
  case BOTTOM_FIELD:
    picture->nb_fields = 1;
    break;
  case FRAME_PICTURE:
    if (!(mpeg2dec->sequence.flags & SEQ_FLAG_PROGRESSIVE_SEQUENCE))
    {
	    picture->nb_fields = (buffer[3] & 2) ? 3 : 2;
	    flags |= (buffer[3] & 128) ? PIC_FLAG_TOP_FIELD_FIRST : 0;
    } else
	    picture->nb_fields = (buffer[3]&2) ? ((buffer[3]&128) ? 6 : 4) : 2;
    break;
  default:
    return 1;
  }
  decoder->top_field_first = buffer[3] >> 7;
  decoder->frame_pred_frame_dct = (buffer[3] >> 6) & 1;
  decoder->concealment_motion_vectors = (buffer[3] >> 5) & 1;
  mpeg2dec->q_scale_type = buffer[3] & 16;
  decoder->intra_vlc_format = (buffer[3] >> 3) & 1;
  decoder->scan = (buffer[3] & 4) ? mpeg2_scan_alt : mpeg2_scan_norm;
  flags |= (buffer[4] & 0x80) ? PIC_FLAG_PROGRESSIVE_FRAME : 0;
  if (buffer[4] & 0x40)
    flags |= (((buffer[4]<<26) | (buffer[5]<<18) | (buffer[6]<<10)) &
              PIC_MASK_COMPOSITE_DISPLAY) | PIC_FLAG_COMPOSITE_DISPLAY;
  picture->flags = flags;
  
  mpeg2dec->ext_state = PIC_DISPLAY_EXT | COPYRIGHT_EXT | QUANT_MATRIX_EXT;
  
  return 0;
}

int mpeg2_header_extension (mpeg2dec_t * mpeg2dec)
{
  static int (* parser[]) (mpeg2dec_t *) = {
    0, sequence_ext, sequence_display_ext, quant_matrix_ext,
    copyright_ext, 0, 0, picture_display_ext, picture_coding_ext
  };
  int ext, ext_bit;
  
  ext = mpeg2dec->chunk_start[0] >> 4;
  ext_bit = 1 << ext;
  
  if (!(mpeg2dec->ext_state & ext_bit))
    return 0;	/* ignore illegal extensions */
  mpeg2dec->ext_state &= ~ext_bit;
  return parser[ext] (mpeg2dec);
}

int mpeg2_header_user_data (mpeg2dec_t * mpeg2dec)
{
  mpeg2dec->user_data_len += mpeg2dec->chunk_ptr - 1 - mpeg2dec->chunk_start;
  mpeg2dec->chunk_start = mpeg2dec->chunk_ptr - 1;
    
  return 0;
}

int mpeg2_header_sequence (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_sequence_t * sequence = &(mpeg2dec->new_sequence);
  static unsigned int frame_period[16] = {
    0, 1126125, 1125000, 1080000, 900900, 900000, 540000, 450450, 450000,
    /* unofficial: xing 15 fps */
    1800000,
    /* unofficial: libmpeg3 "Unofficial economy rates" 5/10/12/15 fps */
    5400000, 2700000, 2250000, 1800000, 0, 0
  };
  int i;
  
  if ((buffer[6] & 0x20) != 0x20)	/* missing marker_bit */
    return 1;

  i = (buffer[0] << 16) | (buffer[1] << 8) | buffer[2];
  if (! (sequence->display_width = sequence->picture_width = i >> 12))
    return 1;
  if (! (sequence->display_height = sequence->picture_height = i & 0xfff))
    return 1;
  sequence->width = (sequence->picture_width + 15) & ~15;
  sequence->height = (sequence->picture_height + 15) & ~15;
  sequence->chroma_width = sequence->width >> 1;
  sequence->chroma_height = sequence->height >> 1;
  
  sequence->flags = (SEQ_FLAG_PROGRESSIVE_SEQUENCE |
                     SEQ_VIDEO_FORMAT_UNSPECIFIED);
  
  sequence->pixel_width = buffer[3] >> 4;	/* aspect ratio */
  sequence->frame_period = frame_period[buffer[3] & 15];
  
  sequence->byte_rate = (buffer[4]<<10) | (buffer[5]<<2) | (buffer[6]>>6);
  
  sequence->vbv_buffer_size = ((buffer[6]<<16)|(buffer[7]<<8))&0x1ff800;
  
  if (buffer[7] & 4)
    sequence->flags |= SEQ_FLAG_CONSTRAINED_PARAMETERS;
  
  mpeg2dec->copy_matrix = 3;
  if (buffer[7] & 2) {
    for (i = 0; i < 64; i++)
	    mpeg2dec->new_quantizer_matrix[0][mpeg2_scan_norm[i]] =
        (buffer[i+7] << 7) | (buffer[i+8] >> 1);
    buffer += 64;
  } else
    for (i = 0; i < 64; i++)
	    mpeg2dec->new_quantizer_matrix[0][mpeg2_scan_norm[i]] =
        default_intra_quantizer_matrix[i];
  
  if (buffer[7] & 1)
    for (i = 0; i < 64; i++)
	    mpeg2dec->new_quantizer_matrix[1][mpeg2_scan_norm[i]] =
        buffer[i+8];
  else
    memset (mpeg2dec->new_quantizer_matrix[1], 16, 64);
  
  sequence->profile_level_id = 0x80;
  sequence->colour_primaries = 0;
  sequence->transfer_characteristics = 0;
  sequence->matrix_coefficients = 0;
  
  mpeg2dec->ext_state = SEQ_EXT;
  mpeg2dec->state = STATE_SEQUENCE;
  mpeg2dec->display_offset_x = mpeg2dec->display_offset_y = 0;
  
  return 0;
}

int mpeg2_header_gop (mpeg2dec_t * mpeg2dec)
{
  guint8 * buffer = mpeg2dec->chunk_start;
  mpeg2_gop_t * gop = &(mpeg2dec->new_gop);
  
  if (! (buffer[1] & 8))
    return 1;
  gop->hours = (buffer[0] >> 2) & 31;
  gop->minutes = ((buffer[0] << 4) | (buffer[1] >> 4)) & 63;
  gop->seconds = ((buffer[1] << 3) | (buffer[2] >> 5)) & 63;
  gop->pictures = ((buffer[2] << 1) | (buffer[3] >> 7)) & 63;
  gop->flags = (buffer[0] >> 7) | ((buffer[3] >> 4) & 6);
  mpeg2dec->state = STATE_GOP;
  return 0;
}

static inline int copy_chunk (mpeg2dec_t * mpeg2dec, int bytes)
{
  guint8 * current;
  guint32 shift;
  guint8 * chunk_ptr;
  guint8 * limit;
  guint8 byte;
  
  if (!bytes)
    return 0;
  
  current = mpeg2dec->buf_start;
  shift = mpeg2dec->shift;
  chunk_ptr = mpeg2dec->chunk_ptr;
  limit = current + bytes;
  
  do
  {
    byte = *current++;
    if (shift == 0x00000100)
    {
      int copied;
      
	    mpeg2dec->shift = 0xffffff00;
	    mpeg2dec->chunk_ptr = chunk_ptr + 1;
	    copied = current - mpeg2dec->buf_start;
	    mpeg2dec->buf_start = current;
	    return copied;
    }
    shift = (shift | byte) << 8;
    *chunk_ptr++ = byte;
  }
  while (current < limit);
  
  mpeg2dec->shift = shift;
  mpeg2dec->buf_start = current;
  return 0;
}

mpeg2_state_t mpeg2_header_picture_start (mpeg2dec_t * mpeg2dec)
{
  mpeg2_picture_t * picture = &(mpeg2dec->new_picture);

  mpeg2dec->state = ((mpeg2dec->state != STATE_SLICE_1ST) ?
                     STATE_PICTURE : STATE_PICTURE_2ND);
  picture->flags = 0;
  picture->tag = picture->tag2 = 0;
  if (mpeg2dec->num_tags)
  {
    if (mpeg2dec->bytes_since_tag >= 4)
    {
	    mpeg2dec->num_tags = 0;
	    picture->tag = mpeg2dec->tag_current;
	    picture->tag2 = mpeg2dec->tag2_current;
	    picture->flags = PIC_FLAG_TAGS;
    }
    else if (mpeg2dec->num_tags > 1)
    {
	    mpeg2dec->num_tags = 1;
	    picture->tag = mpeg2dec->tag_previous;
	    picture->tag2 = mpeg2dec->tag2_previous;
	    picture->flags = PIC_FLAG_TAGS;
    }
  }
  picture->display_offset[0].x = picture->display_offset[1].x =
    picture->display_offset[2].x = mpeg2dec->display_offset_x;
  picture->display_offset[0].y = picture->display_offset[1].y =
    picture->display_offset[2].y = mpeg2dec->display_offset_y;
  return mpeg2_parse_header (mpeg2dec);
}

static inline void finalize_sequence (mpeg2_sequence_t * sequence)
{
  int width;
  int height;
  
  sequence->byte_rate *= 50;
  
  if (sequence->flags & SEQ_FLAG_MPEG2)
  {
    switch (sequence->pixel_width)
    {
    case 1:		/* square pixels */
	    sequence->pixel_width = sequence->pixel_height = 1;	return;
    case 2:		/* 4:3 aspect ratio */
	    width = 4; height = 3;	break;
    case 3:		/* 16:9 aspect ratio */
	    width = 16; height = 9;	break;
    case 4:		/* 2.21:1 aspect ratio */
	    width = 221; height = 100;	break;
    default:	/* illegal */
	    sequence->pixel_width = sequence->pixel_height = 0;	return;
    }
    width *= sequence->display_height;
    height *= sequence->display_width;
    
  }
  else
  {
    if (sequence->byte_rate == 50 * 0x3ffff) 
	    sequence->byte_rate = 0;        /* mpeg-1 VBR */ 
    
    switch (sequence->pixel_width) {
    case 0:	case 15:	/* illegal */
	    sequence->pixel_width = sequence->pixel_height = 0;		return;
    case 1:	/* square pixels */
	    sequence->pixel_width = sequence->pixel_height = 1;		return;
    case 3:	/* 720x576 16:9 */
	    sequence->pixel_width = 64;	sequence->pixel_height = 45;	return;
    case 6:	/* 720x480 16:9 */
	    sequence->pixel_width = 32;	sequence->pixel_height = 27;	return;
    case 12:	/* 720*480 4:3 */
	    sequence->pixel_width = 8;	sequence->pixel_height = 9;	return;
    default:
	    height = 88 * sequence->pixel_width + 1171;
	    width = 2000;
    }
  }

  sequence->pixel_width = width;
  sequence->pixel_height = height;
  while (width) {	/* find greatest common divisor */
    int tmp = width;
    width = height % tmp;
    height = tmp;
  }
  sequence->pixel_width /= height;
  sequence->pixel_height /= height;
}

static void copy_matrix (mpeg2dec_t * mpeg2dec, int index)
{
  if (memcmp (mpeg2dec->quantizer_matrix[index],
              mpeg2dec->new_quantizer_matrix[index], 64))
  {
    memcpy (mpeg2dec->quantizer_matrix[index],
            mpeg2dec->new_quantizer_matrix[index], 64);
    mpeg2dec->scaled[index] = -1;
  }
}

static void finalize_matrix (mpeg2dec_t * mpeg2dec)
{
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  int i;
  
  for (i = 0; i < 2; i++) {
    if (mpeg2dec->copy_matrix & (1 << i))
	    copy_matrix (mpeg2dec, i);
    if ((mpeg2dec->copy_matrix & (4 << i)) &&
        memcmp (mpeg2dec->quantizer_matrix[i],
                mpeg2dec->new_quantizer_matrix[i+2], 64))
    {
	    copy_matrix (mpeg2dec, i + 2);
	    decoder->chroma_quantizer[i] = decoder->quantizer_prescale[i+2];
    } 
    else if (mpeg2dec->copy_matrix & (5 << i))
	    decoder->chroma_quantizer[i] = decoder->quantizer_prescale[i];
  }
}

void mpeg2_reset_info (mpeg2_info_t * info)
{
  info->current_picture = info->current_picture_2nd = NULL;
  info->display_picture = info->display_picture_2nd = NULL;
  info->current_fbuf = info->display_fbuf = info->discard_fbuf = NULL;
}

static int (* free_hook) (void * buf) = NULL;

void mpeg2_free (void * buf)
{
  if (free_hook && free_hook (buf))
    return;
  
  if (buf)
    free (*(((void **)buf) - 1));
}

void mpeg2_header_state_init (mpeg2dec_t * mpeg2dec)
{
  if (mpeg2dec->sequence.width != (unsigned)-1)
  {
    int i;
    
    mpeg2dec->sequence.width = (unsigned)-1;
    if (!mpeg2dec->custom_fbuf)
	    for (i = mpeg2dec->alloc_index_user;
           i < mpeg2dec->alloc_index; i++)
      {
        mpeg2_free (mpeg2dec->fbuf_alloc[i].fbuf.buf[0]);
        mpeg2_free (mpeg2dec->fbuf_alloc[i].fbuf.buf[1]);
        mpeg2_free (mpeg2dec->fbuf_alloc[i].fbuf.buf[2]);
	    }
    if (mpeg2dec->convert_start)
	    for (i = 0; i < 3; i++) {
        mpeg2_free (mpeg2dec->yuv_buf[i][0]);
        mpeg2_free (mpeg2dec->yuv_buf[i][1]);
        mpeg2_free (mpeg2dec->yuv_buf[i][2]);
	    }
    if (mpeg2dec->decoder.convert_id)
	    mpeg2_free (mpeg2dec->decoder.convert_id);
  }
  mpeg2dec->decoder.coding_type = I_TYPE;
  mpeg2dec->decoder.convert = NULL;
  mpeg2dec->decoder.convert_id = NULL;
  mpeg2dec->picture = mpeg2dec->pictures;
  mpeg2dec->fbuf[0] = &mpeg2dec->fbuf_alloc[0].fbuf;
  mpeg2dec->fbuf[1] = &mpeg2dec->fbuf_alloc[1].fbuf;
  mpeg2dec->fbuf[2] = &mpeg2dec->fbuf_alloc[2].fbuf;
  mpeg2dec->first = 1;
  mpeg2dec->alloc_index = 0;
  mpeg2dec->alloc_index_user = 0;
  mpeg2dec->first_decode_slice = 1;
  mpeg2dec->nb_decode_slices = 0;
  mpeg2dec->convert = NULL;
  mpeg2dec->convert_start = NULL;
  mpeg2dec->custom_fbuf = 0;
  mpeg2dec->yuv_index = 0;
}

static mpeg2_state_t seek_sequence (mpeg2dec_t * mpeg2dec)
{
  mpeg2_reset_info (&(mpeg2dec->info));
  mpeg2dec->info.sequence = NULL;
  mpeg2dec->info.gop = NULL;
  mpeg2_header_state_init (mpeg2dec);
  mpeg2dec->action = mpeg2_seek_header;
  return mpeg2_seek_header (mpeg2dec);
}

mpeg2_state_t mpeg2_header_end (mpeg2dec_t * mpeg2dec)
{
  mpeg2_picture_t * picture;
  int b_type;
  
  b_type = (mpeg2dec->decoder.coding_type == B_TYPE);
  picture = mpeg2dec->pictures;
  if ((mpeg2dec->picture >= picture + 2) ^ b_type)
    picture = mpeg2dec->pictures + 2;
  
  mpeg2_reset_info (&(mpeg2dec->info));
  if (!(mpeg2dec->sequence.flags & SEQ_FLAG_LOW_DELAY)) {
    mpeg2dec->info.display_picture = picture;
    if (picture->nb_fields == 1)
	    mpeg2dec->info.display_picture_2nd = picture + 1;
    mpeg2dec->info.display_fbuf = mpeg2dec->fbuf[b_type];
    if (!mpeg2dec->convert)
	    mpeg2dec->info.discard_fbuf = mpeg2dec->fbuf[b_type + 1];
  } else if (!mpeg2dec->convert)
    mpeg2dec->info.discard_fbuf = mpeg2dec->fbuf[b_type];
  mpeg2dec->action = seek_sequence;
  return STATE_END;
}

static void info_user_data (mpeg2dec_t * mpeg2dec)
{
  if (mpeg2dec->user_data_len) {
    mpeg2dec->info.user_data = mpeg2dec->chunk_buffer;
    mpeg2dec->info.user_data_len = mpeg2dec->user_data_len - 3;
  }
}

static mpeg2_state_t invalid_end_action (mpeg2dec_t * mpeg2dec)
{
  mpeg2_reset_info (&(mpeg2dec->info));
  mpeg2dec->info.gop = NULL;
  info_user_data (mpeg2dec);
  mpeg2_header_state_init (mpeg2dec);
  mpeg2dec->sequence = mpeg2dec->new_sequence;
  mpeg2dec->action = mpeg2_seek_header;
  mpeg2dec->state = STATE_SEQUENCE;
  return STATE_SEQUENCE;
}

void mpeg2_header_sequence_finalize (mpeg2dec_t * mpeg2dec)
{
  mpeg2_sequence_t * sequence = &(mpeg2dec->new_sequence);
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  
  finalize_sequence (sequence);
  finalize_matrix (mpeg2dec);
  
  decoder->mpeg1 = !(sequence->flags & SEQ_FLAG_MPEG2);
  decoder->width = sequence->width;
  decoder->height = sequence->height;
  decoder->vertical_position_extension = (sequence->picture_height > 2800);
  decoder->chroma_format = ((sequence->chroma_width == sequence->width) +
                            (sequence->chroma_height == sequence->height));

  if (mpeg2dec->sequence.width != (unsigned)-1)
  {
    unsigned int new_byte_rate;
    
    /*
     * According to 6.1.1.6, repeat sequence headers should be
     * identical to the original. However some DVDs dont respect
     * that and have different bitrates in the repeat sequence
     * headers. So we'll ignore that in the comparison and still
     * consider these as repeat sequence headers.
     *
     * However, be careful not to alter the current sequence when
     * returning STATE_INVALID_END.
     */
    new_byte_rate = sequence->byte_rate;
    sequence->byte_rate = mpeg2dec->sequence.byte_rate;
    if (memcmp (&(mpeg2dec->sequence), sequence,
                sizeof (mpeg2_sequence_t)))
    {
	    decoder->stride_frame = sequence->width;
	    sequence->byte_rate = new_byte_rate;
	    mpeg2_header_end (mpeg2dec);
	    mpeg2dec->action = invalid_end_action;
	    mpeg2dec->state = STATE_INVALID_END;
	    return;
    }
    sequence->byte_rate = new_byte_rate;
    mpeg2dec->state = STATE_SEQUENCE_REPEATED;
  }
  else
    decoder->stride_frame = sequence->width;
  mpeg2dec->sequence = *sequence;
  mpeg2_reset_info (&(mpeg2dec->info));
  mpeg2dec->info.sequence = &(mpeg2dec->sequence);
  mpeg2dec->info.gop = NULL;
  info_user_data (mpeg2dec);
}

void mpeg2_header_gop_finalize (mpeg2dec_t * mpeg2dec)
{
  mpeg2dec->gop = mpeg2dec->new_gop;
  mpeg2_reset_info (&(mpeg2dec->info));
  mpeg2dec->info.gop = &(mpeg2dec->gop);
  info_user_data (mpeg2dec);
}

static void * (* malloc_hook) (unsigned size, mpeg2_alloc_t reason) = NULL;

void * mpeg2_malloc (unsigned size, mpeg2_alloc_t reason)
{
  char * buf;
  
  if (malloc_hook)
  {
    buf = (char *) malloc_hook (size, reason);
    if (buf)
	    return buf;
  }
  
  if (size)
  {
    buf = (char *) malloc (size + 63 + sizeof (void **));
    if (buf)
    {
	    char * align_buf;
      
	    align_buf = buf + 63 + sizeof (void **);
	    align_buf -= (long)align_buf & 63;
	    *(((void **)align_buf) - 1) = buf;
	    return align_buf;
    }
  }
  return NULL;
}

void mpeg2_set_fbuf (mpeg2dec_t * mpeg2dec, int b_type)
{
  int i;
  
  for (i = 0; i < 3; i++)
    if (mpeg2dec->fbuf[1] != &mpeg2dec->fbuf_alloc[i].fbuf &&
        mpeg2dec->fbuf[2] != &mpeg2dec->fbuf_alloc[i].fbuf)
    {
	    mpeg2dec->fbuf[0] = &mpeg2dec->fbuf_alloc[i].fbuf;
	    mpeg2dec->info.current_fbuf = mpeg2dec->fbuf[0];
	    if (b_type || (mpeg2dec->sequence.flags & SEQ_FLAG_LOW_DELAY))
      {
        if (b_type || mpeg2dec->convert)
          mpeg2dec->info.discard_fbuf = mpeg2dec->fbuf[0];
        mpeg2dec->info.display_fbuf = mpeg2dec->fbuf[0];
	    }
	    break;
    }
}

void mpeg2_header_picture_finalize (mpeg2dec_t * mpeg2dec, guint32 accels)
{
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  int old_type_b = (decoder->coding_type == B_TYPE);
  int low_delay = mpeg2dec->sequence.flags & SEQ_FLAG_LOW_DELAY;
  
  finalize_matrix (mpeg2dec);
  decoder->coding_type = mpeg2dec->new_picture.flags & PIC_MASK_CODING_TYPE;
  
  if (mpeg2dec->state == STATE_PICTURE) {
    mpeg2_picture_t * picture;
    mpeg2_picture_t * other;
    
    decoder->second_field = 0;
    
    picture = other = mpeg2dec->pictures;
    if (old_type_b ^ (mpeg2dec->picture < mpeg2dec->pictures + 2))
	    picture += 2;
    else
	    other += 2;
    mpeg2dec->picture = picture;
    *picture = mpeg2dec->new_picture;
    
    if (!old_type_b)
    {
	    mpeg2dec->fbuf[2] = mpeg2dec->fbuf[1];
	    mpeg2dec->fbuf[1] = mpeg2dec->fbuf[0];
    }
    mpeg2dec->fbuf[0] = NULL;
    mpeg2_reset_info (&(mpeg2dec->info));
    mpeg2dec->info.current_picture = picture;
    mpeg2dec->info.display_picture = picture;
    if (decoder->coding_type != B_TYPE)
    {
	    if (!low_delay)
      {
        if (mpeg2dec->first)
        {
          mpeg2dec->info.display_picture = NULL;
          mpeg2dec->first = 0;
        } 
        else
        {
          mpeg2dec->info.display_picture = other;
          if (other->nb_fields == 1)
            mpeg2dec->info.display_picture_2nd = other + 1;
          mpeg2dec->info.display_fbuf = mpeg2dec->fbuf[1];
        }
	    }
	    if (!low_delay + !mpeg2dec->convert)
        mpeg2dec->info.discard_fbuf =
          mpeg2dec->fbuf[!low_delay + !mpeg2dec->convert];
    }
    if (mpeg2dec->convert)
    {
	    mpeg2_convert_init_t convert_init;
	    if (!mpeg2dec->convert_start)
      {
        int y_size, uv_size;
        
        mpeg2dec->decoder.convert_id =
          mpeg2_malloc (mpeg2dec->convert_id_size,
                        MPEG2_ALLOC_CONVERT_ID);
        mpeg2dec->convert (MPEG2_CONVERT_START,
                           mpeg2dec->decoder.convert_id,
                           &(mpeg2dec->sequence),
                           mpeg2dec->convert_stride, accels,
                           mpeg2dec->convert_arg, &convert_init);
        mpeg2dec->convert_start = convert_init.start;
        mpeg2dec->decoder.convert = convert_init.copy;
        
        y_size = decoder->stride_frame * mpeg2dec->sequence.height;
        uv_size = y_size >> (2 - mpeg2dec->decoder.chroma_format);
        mpeg2dec->yuv_buf[0][0] =
          (guint8 *) mpeg2_malloc (y_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[0][1] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[0][2] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[1][0] =
          (guint8 *) mpeg2_malloc (y_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[1][1] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[1][2] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
        y_size = decoder->stride_frame * 32;
        uv_size = y_size >> (2 - mpeg2dec->decoder.chroma_format);
        mpeg2dec->yuv_buf[2][0] =
          (guint8 *) mpeg2_malloc (y_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[2][1] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
        mpeg2dec->yuv_buf[2][2] =
          (guint8 *) mpeg2_malloc (uv_size, MPEG2_ALLOC_YUV);
	    }
	    if (!mpeg2dec->custom_fbuf)
      {
        while (mpeg2dec->alloc_index < 3)
        {
          mpeg2_fbuf_t * fbuf;
          
          fbuf = &mpeg2dec->fbuf_alloc[mpeg2dec->alloc_index++].fbuf;
          fbuf->id = NULL;
          fbuf->buf[0] =
            (guint8 *) mpeg2_malloc (convert_init.buf_size[0],
                                      MPEG2_ALLOC_CONVERTED);
          fbuf->buf[1] =
            (guint8 *) mpeg2_malloc (convert_init.buf_size[1],
                                      MPEG2_ALLOC_CONVERTED);
          fbuf->buf[2] =
            (guint8 *) mpeg2_malloc (convert_init.buf_size[2],
                                      MPEG2_ALLOC_CONVERTED);
        }
        mpeg2_set_fbuf (mpeg2dec, (decoder->coding_type == B_TYPE));
	    }
    }
    else if (!mpeg2dec->custom_fbuf)
    {
	    while (mpeg2dec->alloc_index < 3)
      {
        mpeg2_fbuf_t * fbuf;
        int y_size, uv_size;
        
        fbuf = &(mpeg2dec->fbuf_alloc[mpeg2dec->alloc_index++].fbuf);
        fbuf->id = NULL;
        y_size = decoder->stride_frame * mpeg2dec->sequence.height;
        uv_size = y_size >> (2 - decoder->chroma_format);
        fbuf->buf[0] = (guint8 *) mpeg2_malloc (y_size,
                                                 MPEG2_ALLOC_YUV);
        fbuf->buf[1] = (guint8 *) mpeg2_malloc (uv_size,
                                                 MPEG2_ALLOC_YUV);
        fbuf->buf[2] = (guint8 *) mpeg2_malloc (uv_size,
                                                 MPEG2_ALLOC_YUV);
	    }
	    mpeg2_set_fbuf (mpeg2dec, (decoder->coding_type == B_TYPE));
    }
  }
  else
  {
    decoder->second_field = 1;
    mpeg2dec->picture++;	/* second field picture */
    *(mpeg2dec->picture) = mpeg2dec->new_picture;
    mpeg2dec->info.current_picture_2nd = mpeg2dec->picture;
    if (low_delay || decoder->coding_type == B_TYPE)
	    mpeg2dec->info.display_picture_2nd = mpeg2dec->picture;
  }

  info_user_data (mpeg2dec);
}

static void prescale (mpeg2dec_t * mpeg2dec, int index)
{
  static int non_linear_scale [] = {
    0,  1,  2,  3,  4,  5,   6,   7,
    8, 10, 12, 14, 16, 18,  20,  22,
    24, 28, 32, 36, 40, 44,  48,  52,
    56, 64, 72, 80, 88, 96, 104, 112
  };
  int i, j, k;
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  
  if (mpeg2dec->scaled[index] != mpeg2dec->q_scale_type)
  {
    mpeg2dec->scaled[index] = mpeg2dec->q_scale_type;
    for (i = 0; i < 32; i++)
    {
	    k = mpeg2dec->q_scale_type ? non_linear_scale[i] : (i << 1);
	    for (j = 0; j < 64; j++)
        decoder->quantizer_prescale[index][i][j] =
          k * mpeg2dec->quantizer_matrix[index][j];
    }
  }
}

void mpeg2_init_fbuf (mpeg2_decoder_t * decoder, guint8 * current_fbuf[3],
		      guint8 * forward_fbuf[3], guint8 * backward_fbuf[3])
{
    int offset, stride, height, bottom_field;

    stride = decoder->stride_frame;
    bottom_field = (decoder->picture_structure == BOTTOM_FIELD);
    offset = bottom_field ? stride : 0;
    height = decoder->height;
    
    decoder->picture_dest[0] = current_fbuf[0] + offset;
    decoder->picture_dest[1] = current_fbuf[1] + (offset >> 1);
    decoder->picture_dest[2] = current_fbuf[2] + (offset >> 1);
    
    decoder->f_motion.ref[0][0] = forward_fbuf[0] + offset;
    decoder->f_motion.ref[0][1] = forward_fbuf[1] + (offset >> 1);
    decoder->f_motion.ref[0][2] = forward_fbuf[2] + (offset >> 1);
    
    decoder->b_motion.ref[0][0] = backward_fbuf[0] + offset;
    decoder->b_motion.ref[0][1] = backward_fbuf[1] + (offset >> 1);
    decoder->b_motion.ref[0][2] = backward_fbuf[2] + (offset >> 1);
    
    if (decoder->picture_structure != FRAME_PICTURE)
    {
      decoder->dmv_offset = bottom_field ? 1 : -1;
      decoder->f_motion.ref2[0] = decoder->f_motion.ref[bottom_field];
      decoder->f_motion.ref2[1] = decoder->f_motion.ref[!bottom_field];
      decoder->b_motion.ref2[0] = decoder->b_motion.ref[bottom_field];
      decoder->b_motion.ref2[1] = decoder->b_motion.ref[!bottom_field];
      offset = stride - offset;
      
      if (decoder->second_field && (decoder->coding_type != B_TYPE))
        forward_fbuf = current_fbuf;
      
      decoder->f_motion.ref[1][0] = forward_fbuf[0] + offset;
      decoder->f_motion.ref[1][1] = forward_fbuf[1] + (offset >> 1);
      decoder->f_motion.ref[1][2] = forward_fbuf[2] + (offset >> 1);
      
      decoder->b_motion.ref[1][0] = backward_fbuf[0] + offset;
      decoder->b_motion.ref[1][1] = backward_fbuf[1] + (offset >> 1);
      decoder->b_motion.ref[1][2] = backward_fbuf[2] + (offset >> 1);
      
      stride <<= 1;
      height >>= 1;
    }
    
    decoder->stride = stride;
    decoder->uv_stride = stride >> 1;
    decoder->slice_stride = 16 * stride;
    decoder->slice_uv_stride =
      decoder->slice_stride >> (2 - decoder->chroma_format);
    decoder->limit_x = 2 * decoder->width - 32;
    decoder->limit_y_16 = 2 * height - 32;
    decoder->limit_y_8 = 2 * height - 16;
    decoder->limit_y = height - 16;
}

mpeg2_state_t mpeg2_header_slice_start (mpeg2dec_t * mpeg2dec)
{
  mpeg2_decoder_t * decoder = &(mpeg2dec->decoder);
  
  mpeg2dec->info.user_data = NULL;	mpeg2dec->info.user_data_len = 0;
  mpeg2dec->state = ((mpeg2dec->picture->nb_fields > 1 ||
                      mpeg2dec->state == STATE_PICTURE_2ND) ?
                     STATE_SLICE : STATE_SLICE_1ST);
  
  if (mpeg2dec->decoder.coding_type != D_TYPE)
  {
    prescale (mpeg2dec, 0);
    if (decoder->chroma_quantizer[0] == decoder->quantizer_prescale[2])
	    prescale (mpeg2dec, 2);
    if (mpeg2dec->decoder.coding_type != I_TYPE)
    {
	    prescale (mpeg2dec, 1);
	    if (decoder->chroma_quantizer[1] == decoder->quantizer_prescale[3])
        prescale (mpeg2dec, 3);
    }
  }
  
  if (!(mpeg2dec->nb_decode_slices))
    mpeg2dec->picture->flags |= PIC_FLAG_SKIP;
  else if (mpeg2dec->convert_start) {
    mpeg2dec->convert_start (decoder->convert_id, mpeg2dec->fbuf[0],
                             mpeg2dec->picture, mpeg2dec->info.gop);
    
    if (mpeg2dec->decoder.coding_type == B_TYPE)
    {
      mpeg2_init_fbuf (&(mpeg2dec->decoder), mpeg2dec->yuv_buf[2],
                       mpeg2dec->yuv_buf[mpeg2dec->yuv_index ^ 1],
                       mpeg2dec->yuv_buf[mpeg2dec->yuv_index]);
    }
    else
    {
      mpeg2_init_fbuf (&(mpeg2dec->decoder),
                       mpeg2dec->yuv_buf[mpeg2dec->yuv_index ^ 1],
                       mpeg2dec->yuv_buf[mpeg2dec->yuv_index],
                       mpeg2dec->yuv_buf[mpeg2dec->yuv_index]);
      if (mpeg2dec->state == STATE_SLICE)
        mpeg2dec->yuv_index ^= 1;
    }
  }
  else
  {
    int b_type;
    
    b_type = (mpeg2dec->decoder.coding_type == B_TYPE);
    mpeg2_init_fbuf (&(mpeg2dec->decoder), mpeg2dec->fbuf[0]->buf,
                     mpeg2dec->fbuf[b_type + 1]->buf,
                     mpeg2dec->fbuf[b_type]->buf);
  }
  mpeg2dec->action = NULL;
  return (mpeg2_state_t)-1;
}

guint32 mpeg2_accel (guint32 accel)
{
  return 0;
}

mpeg2_state_t mpeg2_parse_header (mpeg2dec_t * mpeg2dec)
{
  static int (* process_header[]) (mpeg2dec_t * mpeg2dec) =
  {
    mpeg2_header_picture, mpeg2_header_extension, mpeg2_header_user_data,
    mpeg2_header_sequence, NULL, NULL, NULL, NULL, mpeg2_header_gop
  };
  int size_buffer, size_chunk, copied;
  
  mpeg2dec->action = mpeg2_parse_header;
  mpeg2dec->info.user_data = NULL;	mpeg2dec->info.user_data_len = 0;
  while (1)
  {
    size_buffer = mpeg2dec->buf_end - mpeg2dec->buf_start;
    size_chunk = (mpeg2dec->chunk_buffer + BUFFER_SIZE -
                  mpeg2dec->chunk_ptr);
    if (size_buffer <= size_chunk)
    {
      copied = copy_chunk (mpeg2dec, size_buffer);
      if (!copied)
      {
        mpeg2dec->bytes_since_tag += size_buffer;
        mpeg2dec->chunk_ptr += size_buffer;
        return STATE_BUFFER;
      }
    }
    else
    {
      copied = copy_chunk (mpeg2dec, size_chunk);
	    if (!copied)
      {
        /* filled the chunk buffer without finding a start code */
        mpeg2dec->bytes_since_tag += size_chunk;
        mpeg2dec->code = 0xb4;
        mpeg2dec->action = mpeg2_seek_header;
        return STATE_INVALID;
      }
    }
    mpeg2dec->bytes_since_tag += copied;
    
    if (process_header[mpeg2dec->code & 0x0b] (mpeg2dec))
    {
      mpeg2dec->code = mpeg2dec->buf_start[-1];
      mpeg2dec->action = mpeg2_seek_header;
      return STATE_INVALID;
    }
    
    mpeg2dec->code = mpeg2dec->buf_start[-1];
    switch (RECEIVED (mpeg2dec->code, mpeg2dec->state))
    {      
      /* state transition after a sequence header */
    case RECEIVED (0x00, STATE_SEQUENCE):
      mpeg2dec->action = mpeg2_header_picture_start;
    case RECEIVED (0xb8, STATE_SEQUENCE):
      mpeg2_header_sequence_finalize (mpeg2dec);
      break;
      
      /* other legal state transitions */
    case RECEIVED (0x00, STATE_GOP):
      mpeg2_header_gop_finalize (mpeg2dec);
      mpeg2dec->action = mpeg2_header_picture_start;
      break;
    case RECEIVED (0x01, STATE_PICTURE):
    case RECEIVED (0x01, STATE_PICTURE_2ND):
      mpeg2_header_picture_finalize (mpeg2dec, mpeg2_accels);
      mpeg2dec->action = mpeg2_header_slice_start;
      break;
      
      /* legal headers within a given state */
    case RECEIVED (0xb2, STATE_SEQUENCE):
    case RECEIVED (0xb2, STATE_GOP):
    case RECEIVED (0xb2, STATE_PICTURE):
    case RECEIVED (0xb2, STATE_PICTURE_2ND):
    case RECEIVED (0xb5, STATE_SEQUENCE):
    case RECEIVED (0xb5, STATE_PICTURE):
    case RECEIVED (0xb5, STATE_PICTURE_2ND):
      mpeg2dec->chunk_ptr = mpeg2dec->chunk_start;
      continue;
        
    default:
	    mpeg2dec->action = mpeg2_seek_header;
	    return STATE_INVALID;
    }
    
    mpeg2dec->chunk_start = mpeg2dec->chunk_ptr = mpeg2dec->chunk_buffer;
    mpeg2dec->user_data_len = 0;
    return mpeg2dec->state;
  }
}

mpeg2_state_t mpeg2_parse (mpeg2dec_t * mpeg2dec)
{
  if (mpeg2dec->action)
  {
    mpeg2_state_t state;
    
    state = mpeg2dec->action (mpeg2dec);
    if ((int)state >= 0)
      return state;
  }
  
  /* Possible to remove this while? Test when the rest works. */
  while (1)
  {
    if ((unsigned) (mpeg2dec->code - 1) >= 0xb0 - 1)
	    break;
    if (seek_chunk (mpeg2dec) == STATE_BUFFER)
	    return STATE_BUFFER;
  }
  
  switch (mpeg2dec->code) {
  case 0x00:
    mpeg2dec->action = mpeg2_header_picture_start;
    return mpeg2dec->state;
  case 0xb7:
    mpeg2dec->action = mpeg2_header_end;
    break;
  case 0xb3:
  case 0xb8:
    mpeg2dec->action = mpeg2_parse_header;
    break;
  default:
    mpeg2dec->action = seek_chunk;
    return STATE_INVALID;
  }
  return (mpeg2dec->state == STATE_SLICE) ? STATE_SLICE : STATE_INVALID;
}

mpeg2dec_t * mpeg2_init (void)
{
  mpeg2dec_t * mpeg2dec;
  
  mpeg2_accel (MPEG2_ACCEL_DETECT);
  
  mpeg2dec = (mpeg2dec_t *) mpeg2_malloc (sizeof (mpeg2dec_t),
                                          MPEG2_ALLOC_MPEG2DEC);
  if (mpeg2dec == NULL)
    return NULL;

  memset (mpeg2dec->decoder.DCTblock, 0, 64 * sizeof (gint16));
  memset (mpeg2dec->quantizer_matrix, 0, 4 * 64 * sizeof (guint8));
  
  mpeg2dec->chunk_buffer = (guint8 *) mpeg2_malloc (BUFFER_SIZE + 4,
                                                     MPEG2_ALLOC_CHUNK);
  
  mpeg2dec->sequence.width = (unsigned)-1;
  mpeg2_reset (mpeg2dec, 1);
  
  return mpeg2dec;
}

void mpeg2_close (mpeg2dec_t * mpeg2dec)
{
  mpeg2_header_state_init (mpeg2dec);
  mpeg2_free (mpeg2dec->chunk_buffer);
  mpeg2_free (mpeg2dec);
}


gboolean
tnt_mpg_extract_file_meta (gint fd, TntFileMeta *file_meta)
{
  guint8                  buffer [META_BUFFER_SIZE];
  mpeg2dec_t             *mpg_decoder;
  const mpeg2_info_t     *mpg_info;
  const mpeg2_sequence_t *mpg_seq;
  mpeg2_state_t           mpg_state;
  size_t                  mpg_size;
  guint32                 frame = 0;
  TntStat                 tstat;
  gint64                  file_size;
  guint32                 header_size = 0;
  gdouble                 frame_period = 0.0;
  guint32                 estimate = 0;

  g_return_val_if_fail (fd >= 0, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  mpg_decoder = mpeg2_init ();

  if (!mpg_decoder)
  {
    return FALSE;
  }

  mpg_info = mpeg2_info (mpg_decoder);
  mpg_size = (size_t)-1;

  do
  {
    mpg_state = mpeg2_parse (mpg_decoder);
    mpg_seq   = mpg_info->sequence;
    switch (mpg_state)
    {
    case STATE_BUFFER:
      mpg_size = read (fd, buffer, META_BUFFER_SIZE);
      mpeg2_buffer (mpg_decoder, buffer, buffer + mpg_size);
      break;
    case STATE_SLICE:
    case STATE_END:
    case STATE_INVALID_END:
      if (mpg_info->display_fbuf)
      {
        if (frame == 5)
        {
          tnt_file_meta_set_width    (file_meta, mpg_seq->width);
          tnt_file_meta_set_height   (file_meta, mpg_seq->height);
          if (mpg_seq->byte_rate)
          {
            tnt_file_meta_set_quality  (file_meta, (guint16) ((mpg_seq->byte_rate * 8) / 1000));            
          }
          frame_period = ((gdouble)mpg_seq->frame_period) / 27000000.0;
        }
        if (frame == 800)
        {
          tnt_fstat (fd, &tstat);
          file_size = tstat.len;
          estimate  = ((guint32) (frame_period * (( ((gdouble)(file_size - header_size) / (gdouble)lseek (fd, 0, SEEK_CUR)) * (frame + 1) ))));
          if (estimate < 12000 && estimate > 0)
          {
            tnt_file_meta_set_duration (file_meta, estimate);
          }
          mpeg2_close (mpg_decoder);
          return TRUE;
        }
        frame++;
      }
      break;
    default:
      break;
    }
    if (frame == 0) {
      header_size += mpg_size;
    }
  }
  while (mpg_size);

  estimate = frame_period * frame;
  if (estimate < 12000 && estimate > 0)
  {
    tnt_file_meta_set_duration (file_meta, (guint32) (frame_period * frame));
  }
  mpeg2_close (mpg_decoder);
  return TRUE;
}
