/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-net-db.c - Keeps information on nodes in the network.
 *
 * Copyright (C) 2002-2005 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
# include <io.h>
#endif

#include "config.h"
#include "tnt.h"

#define NODE_LIST_FILENAME    "node.list"
#define ADDRESS_LIST_FILENAME "address.list"
#define DEFAULT_PEER_ADDRESS  "soledad.copyleft.no"
#define DEFAULT_PEER_PORT     5112
#define ADDRESS_MAX           1024
#define DEFAULT_HOLDOFF       30  /* Magic number; seconds-ish */

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_NET_DB_TYPE)

typedef struct
{
  TntNodeID   node_id;
  TntNodeMeta node_meta;
}
NodeFileRecord;

typedef struct
{
  TntNodeID    node_id;
  TntNodeMeta *node_meta;
}
NodeRecord;

typedef struct
{
  TntNodeAddress4 address;
  TntTimeStamp    attempt_time;
  guint16         attempt_success : 1;
  guint16         successive_attempts;
}
AddressRecord;

typedef struct
{
  TntNodeIP4  match_ip;
  TntNodeIP4  match_mask;
  GTree      *address_by_score;
}
AddressSubset;

static FluxMemChunk *node_record_mem_chunk = NULL;

static gchar *
full_node_list_path (TntNetDB *net_db)
{
  return g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (net_db->storage),
                    NODE_LIST_FILENAME, NULL);
}

static gchar *
full_address_list_path (TntNetDB *net_db)
{
  return g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (net_db->storage),
                    ADDRESS_LIST_FILENAME, NULL);
}

static guint
hash_node_id (gconstpointer data)
{
  const TntNodeID *node_id = data;

  g_assert (node_id != NULL);

  return (*node_id & 0xffffffff) ^ (*node_id >> 32);
}

static gboolean
compare_node_ids (gconstpointer a, gconstpointer b)
{
  const TntNodeID *node_id_a = a, *node_id_b = b;

  g_assert (a != NULL);
  g_assert (b != NULL);

  if (*node_id_a == *node_id_b)
    return TRUE;

  return FALSE;
}

static GList *
find_node_meta_list (TntNetDB *net_db, TntNodeID node_id)
{
  GList *node_meta_list;

  node_meta_list = g_hash_table_lookup (net_db->node_meta_table, &node_id);
  return node_meta_list;
}

static NodeRecord *
find_node_record (TntNetDB *net_db, TntNodeID node_id)
{
  GList *node_meta_list;

  node_meta_list = find_node_meta_list (net_db, node_id);
  if (!node_meta_list)
    return NULL;

  return node_meta_list->data;
}

static TntNodeMeta *
find_node_meta (TntNetDB *net_db, TntNodeID node_id)
{
  NodeRecord *record;

  record = find_node_record (net_db, node_id);
  if (!record)
    return NULL;

  return record->node_meta;
}

static void
remove_node (TntNetDB *net_db, TntNodeID node_id)
{
  GList      *node_meta_list;
  NodeRecord *record;

  node_meta_list = find_node_meta_list (net_db, node_id);
  if (!node_meta_list)
    return;

  record = node_meta_list->data;

  g_hash_table_remove (net_db->node_meta_table, &node_id);
  tnt_node_meta_free (record->node_meta);
  flux_mem_chunk_free (node_record_mem_chunk, record);
  net_db->node_meta_list = g_list_delete_link (net_db->node_meta_list, node_meta_list);
}

/* Assumes we don't already have the node_id */
static void
add_node (TntNetDB *net_db, TntNodeID node_id, TntNodeMeta *node_meta)
{
  NodeRecord      *record;
  TntNodeAddress4  address;

  record = flux_mem_chunk_alloc (node_record_mem_chunk);
  record->node_id   = node_id;
  record->node_meta = node_meta;

  net_db->node_meta_list = g_list_prepend (net_db->node_meta_list, record);
  g_hash_table_insert (net_db->node_meta_table, &record->node_id, net_db->node_meta_list);

  /* Add IPs to anonymous address list, if present */

  tnt_node_meta_get_primary_address (node_meta, &address);
  if (address.ip != TNT_NODE_IP_INVALID && address.port != TNT_NODE_PORT_INVALID)
    tnt_net_db_add_address (net_db, &address);

  tnt_node_meta_get_secondary_address (node_meta, &address);
  if (address.ip != TNT_NODE_IP_INVALID && address.port != TNT_NODE_PORT_INVALID)
    tnt_net_db_add_address (net_db, &address);
}

static gint
n_nodes_in_table (TntNetDB *net_db)
{
  return g_hash_table_size (net_db->node_meta_table);
}

static guint
address_hash_address (const AddressRecord *record)
{
  return (guint) record->address.ip ^ (guint) record->address.port;
}

static gboolean
address_equal_address (const AddressRecord *record_a, const AddressRecord *record_b)
{
  if (record_a->address.ip == record_b->address.ip &&
      record_a->address.port == record_b->address.port)
    return TRUE;

  return FALSE;
}

static gint
address_calc_score (const AddressRecord *record)
{
  gint score;

  if (record->successive_attempts)
    score = (DEFAULT_HOLDOFF << MIN (record->successive_attempts, 24)) +
      record->successive_attempts * 5;
  else
    score = 0;

  if (!record->attempt_success)
    score = -score;

  if (record->attempt_time)
  {
    TntTimeStamp now = tnt_timestamp ();

    /* Protect against clock skew */
    if (now > record->attempt_time)
      score += now - record->attempt_time;
  }

  return score;
}

/* GCompareDataFunc */
static gint
address_compare_address (const AddressRecord *record_a, const AddressRecord *record_b)
{
  if (record_a->address.ip < record_b->address.ip)
    return -1;
  if (record_a->address.ip > record_b->address.ip)
    return 1;
  if (record_a->address.port < record_b->address.port)
    return -1;
  if (record_a->address.port > record_b->address.port)
    return 1;

  return 0;
}

/* GCompareDataFunc */
static gint
address_compare_score (const AddressRecord *record_a, const AddressRecord *record_b)
{
  gint score_a;
  gint score_b;
  gint address_result;
  gint result;

  /* Different addresses never indicate that they have the same score, but the
   * same address will always have the same score. */

  address_result = address_compare_address (record_a, record_b);
  if (!address_result)
    return 0;

  score_a = address_calc_score (record_a);
  score_b = address_calc_score (record_b);

  result = score_a - score_b;
  if (result == 0)
    result = address_result;

  return result;
}

static gint
address_search_best (const AddressRecord *record, gconstpointer *best_record)
{
  *best_record = record;
  return 1;
}

static gint
address_search_worst (const AddressRecord *record, gconstpointer *worst_record)
{
  *worst_record = record;
  return -1;
}

static void
address_destroy (AddressRecord *record)
{
  g_free (record);
}

static gboolean
ip_match (TntNodeIP4 ip, TntNodeIP4 match_ip, TntNodeIP4 match_mask)
{
  if ((ip & match_mask) == (match_ip & match_mask))
    return TRUE;

  return FALSE;
}

static gboolean
find_address_record (TntNetDB *net_db, const TntNodeAddress4 *address, AddressRecord **record_dest)
{
  AddressRecord  lookup_record;
  AddressRecord *record;

  memset (&lookup_record, 0, sizeof (AddressRecord));
  memcpy (&lookup_record.address, address, sizeof (TntNodeAddress4));

  record = g_hash_table_lookup (net_db->unblocked_addresses, &lookup_record);
  if (!record)
    record = g_hash_table_lookup (net_db->blocked_addresses, &lookup_record);

  if (record)
  {
    if (record_dest)
      *record_dest = record;

    return TRUE;
  }

  return FALSE;
}

static void
address_subsets_insert_address (TntNetDB *net_db, AddressRecord *record)
{
  GList *l;

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    AddressSubset *subset = l->data;

    if (ip_match (record->address.ip, subset->match_ip, subset->match_mask))
      g_tree_insert (subset->address_by_score, record, record);
  }
}

static void
address_subsets_remove_address (TntNetDB *net_db, const AddressRecord *record)
{
  GList *l;

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    AddressSubset *subset = l->data;

    if (ip_match (record->address.ip, subset->match_ip, subset->match_mask))
      g_tree_remove (subset->address_by_score, record);
  }
}

static gint
n_addresses_in_table (TntNetDB *net_db)
{
  return g_hash_table_size (net_db->blocked_addresses) +
         g_hash_table_size (net_db->unblocked_addresses);
}

typedef struct
{
  TntNetDB   *net_db;
  FluxFileIO *file_io;
  gint        fd;
}
IOContext;

static void
load_node_list (TntNetDB *net_db)
{
  gchar          *path;
  gint            node_file;
  NodeFileRecord  file_record;

  path = full_node_list_path (net_db);
  node_file = tnt_open (path, FALSE, FALSE, FALSE);
  g_free (path);
  if (node_file < 0)
    return;

  /* FIXME: Release any previously existing nodes */

  /* FIXME: Not interrupt-safe */
  while (read (node_file, &file_record, sizeof (NodeFileRecord)) == sizeof (NodeFileRecord))
  {
    add_node (net_db, file_record.node_id, tnt_node_meta_dup (&file_record.node_meta));
  }

  tnt_close (node_file);
}

static void
save_node_entry (gpointer key, gpointer value, gpointer data)
{
  IOContext      *context   = data;
  NodeRecord     *record    = ((GList *) value)->data;
  NodeFileRecord  file_record;

  file_record.node_id = record->node_id;
  memcpy (&file_record.node_meta, record->node_meta, sizeof (TntNodeMeta));
  write (context->fd, &file_record, sizeof (NodeFileRecord));
}

static void
save_node_list (TntNetDB *net_db)
{
  IOContext *context;
  gchar     *path;
  gint       node_file;

  path = full_node_list_path (net_db);
  node_file = tnt_open (path, TRUE, TRUE, FALSE);
  g_free (path);
  if (node_file < 0)
    return;

  context = g_new0 (IOContext, 1);
  context->net_db = net_db;
  context->fd     = node_file;
  flux_object_ref (FLUX_OBJECT (net_db));

  g_hash_table_foreach (net_db->node_meta_table, save_node_entry, context);

  flux_object_unref (FLUX_OBJECT (context->net_db));
  tnt_close (context->fd);
  g_free (context);
}

static void
save_address_entry (gpointer key, gpointer value, gpointer data)
{
  IOContext     *context = data;
  AddressRecord *record  = value;

  flux_dual_fifo_write (FLUX_DUAL_FIFO (context->file_io), record, sizeof (AddressRecord));
}

static void
save_address_list_done (FluxFileIO *file_io, gpointer data)
{
  TntNetDB *net_db = data;

  flux_object_unref (FLUX_OBJECT (file_io));
  flux_object_unref (FLUX_OBJECT (net_db));
}

/* Non-blocking I/O, done regularly during operation */
static void
save_address_list (TntNetDB *net_db)
{
  IOContext  *context;
  gchar      *path;
  gint        fd;

  path = full_address_list_path (net_db);
  fd = tnt_open (path, TRUE, TRUE, FALSE);
  g_free (path);
  if (fd < 0)
    return;

  context = g_new0 (IOContext, 1);
  context->net_db  = net_db;
  context->fd      = fd;
  context->file_io = flux_file_io_new (fd);
  flux_file_io_set_eow_func (context->file_io, save_address_list_done, net_db);
  flux_object_ref (FLUX_OBJECT (context->net_db));

  g_hash_table_foreach (net_db->blocked_addresses, save_address_entry, context);
  g_hash_table_foreach (net_db->unblocked_addresses, save_address_entry, context);

  g_free (context);
}

/* Blocking I/O, done once at init */
static void
load_address_list (TntNetDB *net_db)
{
  gchar         *path;
  gint           fd;
  AddressRecord  address_record;

  path = full_address_list_path (net_db);
  fd = tnt_open (path, FALSE, FALSE, FALSE);
  g_free (path);
  if (fd < 0)
    return;

  /* FIXME: Not interrupt-safe */
  while (read (fd, &address_record, sizeof (AddressRecord)) == sizeof (AddressRecord))
  {
    AddressRecord *record_added;

    if (find_address_record (net_db, &address_record.address, NULL))
      continue;

    record_added = g_memdup (&address_record, sizeof (AddressRecord));

    g_hash_table_insert (net_db->unblocked_addresses, record_added, record_added);
    address_subsets_insert_address (net_db, record_added);
  }

  tnt_close (fd);
}

static gboolean
database_changed_save_cb (gpointer data)
{
  TntNetDB *net_db = data;

  save_node_list (net_db);
  save_address_list (net_db);
  net_db->save_timeout_id = 0;
  return FALSE;
}

static void
database_changed (TntNetDB *net_db)
{
  if (net_db->save_timeout_id != 0)
    return;

  net_db->save_timeout_id = g_timeout_add (5000 + 100 * n_nodes_in_table (net_db) +
                                           100 * n_addresses_in_table (net_db),
                                           (GSourceFunc) database_changed_save_cb, net_db);
}

void
tnt_net_db_finalize (FluxObject *object)
{
  TntNetDB *net_db = TNT_NET_DB (object);
  GList    *l;

  assert_types (net_db);

  g_hash_table_destroy (net_db->node_meta_table);
  g_hash_table_destroy (net_db->blocked_addresses);
  g_hash_table_destroy (net_db->unblocked_addresses);

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    AddressSubset *subset = l->data;

    g_tree_destroy (subset->address_by_score);
    g_free (subset);
  }

  g_list_free (net_db->address_subsets);

  if (net_db->save_timeout_id != 0)
    g_source_remove (net_db->save_timeout_id);

  tnt_storage_close (net_db->storage);

  /* TODO: Free more stuff? */
}

gboolean
tnt_net_db_init (TntNetDB *net_db, TntStorage *storage)
{
  g_return_val_if_fail (storage != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (net_db));
  flux_object_set_type          (FLUX_OBJECT (net_db), TNT_NET_DB_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (net_db), tnt_net_db_finalize);

  net_db->storage             = storage;
  net_db->node_meta_table     = g_hash_table_new_full (hash_node_id, compare_node_ids, NULL, NULL);
  net_db->blocked_addresses   = g_hash_table_new_full ((GHashFunc) address_hash_address,
                                                       (GEqualFunc) address_equal_address,
                                                       (GDestroyNotify) address_destroy, NULL);
  net_db->unblocked_addresses = g_hash_table_new_full ((GHashFunc) address_hash_address,
                                                       (GEqualFunc) address_equal_address,
                                                       (GDestroyNotify) address_destroy, NULL);
  net_db->save_timeout_id     = 0;

  if (!node_record_mem_chunk)
    node_record_mem_chunk = flux_mem_chunk_new (256, sizeof (NodeRecord));

  load_address_list (net_db);
  load_node_list (net_db);

  tnt_net_db_add_address_named (net_db, DEFAULT_PEER_ADDRESS, DEFAULT_PEER_PORT);
  return TRUE;
}

TntNetDB *
tnt_net_db_new (TntStorage *storage)
{
  TntNetDB *net_db;

  net_db = g_new0 (TntNetDB, 1);
  if (!tnt_net_db_init (net_db, storage))
  {
    g_free (net_db);
    return NULL;
  }

  return net_db;
}

TntNodeMeta *
tnt_net_db_get_node_meta (TntNetDB *net_db, TntNodeID node_id)
{
  TntNodeMeta *node_meta;

  assert_types (net_db);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, NULL);

  node_meta = find_node_meta (net_db, node_id);
  if (!node_meta)
    return NULL;

  return tnt_node_meta_dup (node_meta);
}

void
tnt_net_db_set_node_meta (TntNetDB *net_db, TntNodeID node_id, TntNodeMeta *node_meta)
{
  assert_types (net_db);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);
  g_return_if_fail (node_meta != NULL);

  remove_node (net_db, node_id);
  add_node (net_db, node_id, tnt_node_meta_dup (node_meta));
  database_changed (net_db);
}

static void
addresses_remove_worst (TntNetDB *net_db)
{
  GTree         *biggest_tree = NULL;
  gint           biggest_n    = 0;
  AddressRecord *worst_record = NULL;
  GList         *l;

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    AddressSubset *subset = l->data;
    GTree         *address_by_score;
    gint           n_nodes;

    address_by_score = subset->address_by_score;
    n_nodes = g_tree_nnodes (address_by_score);

    if (n_nodes > biggest_n)
    {
      biggest_tree = address_by_score;
      biggest_n    = n_nodes;
    }
  }

  if (!biggest_tree)
    return;

  g_tree_search (biggest_tree, (GCompareFunc) address_search_worst, &worst_record);
  if (!worst_record)
    return;

  /* Remove from score trees */

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    AddressSubset *subset = l->data;
    g_tree_steal (subset->address_by_score, worst_record);
  }

  /* Remove from master hash - the record should be in only one of the hashes */

  g_hash_table_remove (net_db->blocked_addresses, worst_record);
  g_hash_table_remove (net_db->unblocked_addresses, worst_record);
}

void
tnt_net_db_add_address (TntNetDB *net_db, const TntNodeAddress4 *address)
{
  assert_types (net_db);
  g_return_if_fail (address != NULL);

  if (!find_address_record (net_db, address, NULL))
  {
    AddressRecord *record;

    record = g_new0 (AddressRecord, 1);
    memcpy (&record->address, address, sizeof (TntNodeAddress4));

    g_hash_table_insert (net_db->unblocked_addresses, record, record);
    address_subsets_insert_address (net_db, record);

    while (g_hash_table_size (net_db->blocked_addresses) +
           g_hash_table_size (net_db->unblocked_addresses) > ADDRESS_MAX)
      addresses_remove_worst (net_db);

    database_changed (net_db);
  }
}

void
tnt_net_db_block_address (TntNetDB *net_db, const TntNodeAddress4 *address)
{
  AddressRecord  lookup_record;
  AddressRecord *record;

  assert_types (net_db);
  g_return_if_fail (address != NULL);

  memset (&lookup_record, 0, sizeof (AddressRecord));
  memcpy (&lookup_record.address, address, sizeof (TntNodeAddress4));

  record = g_hash_table_lookup (net_db->unblocked_addresses, &lookup_record);
  if (!record)
    return;

  g_hash_table_steal (net_db->unblocked_addresses, record);
  g_hash_table_insert (net_db->blocked_addresses, record, record);
  address_subsets_remove_address (net_db, record);
}

void
tnt_net_db_unblock_address (TntNetDB *net_db, const TntNodeAddress4 *address)
{
  AddressRecord  lookup_record;
  AddressRecord *record;

  assert_types (net_db);
  g_return_if_fail (address != NULL);

  memset (&lookup_record, 0, sizeof (AddressRecord));
  memcpy (&lookup_record.address, address, sizeof (TntNodeAddress4));

  record = g_hash_table_lookup (net_db->blocked_addresses, &lookup_record);
  if (!record)
    return;

  g_hash_table_steal (net_db->blocked_addresses, record);
  g_hash_table_insert (net_db->unblocked_addresses, record, record);
  address_subsets_insert_address (net_db, record);
}

void
tnt_net_db_rate_address (TntNetDB *net_db, const TntNodeAddress4 *address, gboolean good)
{
  AddressRecord *record;

  assert_types (net_db);
  g_return_if_fail (address != NULL);

  if (!find_address_record (net_db, address, &record))
    return;

  address_subsets_remove_address (net_db, record);

  record->attempt_time = tnt_timestamp ();

  if ((record->attempt_success && good) ||
      (!record->attempt_success && !good))
  {
    if (record->successive_attempts < G_MAXUINT16)
      record->successive_attempts++;
  }
  else
  {
    record->attempt_success = good;
    record->successive_attempts = 1;
  }

  /* Insert into score trees unless blocked */
  if (!g_hash_table_lookup (net_db->blocked_addresses, record))
    address_subsets_insert_address (net_db, record);

  database_changed (net_db);
}

static void
populate_score_tree_cb (AddressRecord *record, gpointer value, AddressSubset *subset)
{
  if (ip_match (record->address.ip, subset->match_ip, subset->match_mask))
    g_tree_insert (subset->address_by_score, record, record);
}

static GTree *
addresses_get_score_tree (TntNetDB *net_db, TntNodeIP4 match_ip, TntNodeIP4 match_mask)
{
  AddressSubset *subset;
  GList         *l;

  for (l = net_db->address_subsets; l; l = g_list_next (l))
  {
    subset = l->data;

    if (subset->match_ip == match_ip &&
        subset->match_mask == match_mask)
      return subset->address_by_score;
  }

  /* No corresponding subset; create one */

  subset = g_new0 (AddressSubset, 1);
  subset->match_ip         = match_ip;
  subset->match_mask       = match_mask;
  subset->address_by_score = g_tree_new ((GCompareFunc) address_compare_score);

  net_db->address_subsets = g_list_prepend (net_db->address_subsets, subset);

  /* Insert matching, unblocked addresses into this score tree */

  g_hash_table_foreach (net_db->unblocked_addresses, (GHFunc) populate_score_tree_cb, subset);
  return subset->address_by_score;
}

gboolean
tnt_net_db_pick_address (TntNetDB *net_db, TntNodeIP4 match_ip, TntNodeIP4 match_mask,
                         TntNodeAddress4 *address)
{
  GTree         *score_tree;
  AddressRecord *best_record = NULL;

  assert_types (net_db);
  g_return_val_if_fail (address != NULL, FALSE);

  score_tree = addresses_get_score_tree (net_db, match_ip, match_mask);

  g_tree_search (score_tree, (GCompareFunc) address_search_best, &best_record);
  if (!best_record)
    return FALSE;

  if (address)
    memcpy (address, &best_record->address, sizeof (TntNodeAddress4));

  return TRUE;
}

typedef struct
{
  gint   n;
  GList *list;
}
RecordListContext;

static gboolean
build_record_list_cb (gpointer key, gpointer value, gpointer data)
{
  RecordListContext *context = data;
  AddressRecord     *record  = value;

  context->list = g_list_prepend (context->list, record);
  context->n--;

  if (context->n == 0)
    return TRUE;

  return FALSE;
}

GList *
tnt_net_db_pick_addresses (TntNetDB *net_db, TntNodeIP4 match_ip,
                           TntNodeIP4 match_mask, gint n_max)
{
  RecordListContext  context;
  GTree             *score_tree;
  GList             *address_list;
  GList             *l;

  assert_types (net_db);

  context.list = NULL;
  context.n    = n_max;

  score_tree = addresses_get_score_tree (net_db, match_ip, match_mask);
  g_tree_foreach (score_tree, (GTraverseFunc) build_record_list_cb, &context);

  address_list = context.list;

  for (l = address_list; l; l = g_list_next (l))
  {
    AddressRecord *record = l->data;
    l->data = g_memdup (&record->address, sizeof (TntNodeAddress4));
  }

  return address_list;
}

typedef struct
{
  TntNetDB *net_db;
  gchar    *name;
  guint     port;
}
NameLookupContext;

static void
name_lookup_cb (GList *ip_addr_list, GList *name_list, gpointer data)
{
  NameLookupContext *context = data;

  g_assert (context != NULL);
  assert_types (context->net_db);

  if (ip_addr_list != NULL)
  {
    FluxIPAddr      *ip_addr = ip_addr_list->data;
    TntNodeAddress4  node_address;

    node_address.ip   = tnt_ip_addr_to_ip (ip_addr);
    node_address.port = context->port;
    tnt_net_db_add_address (context->net_db, &node_address);
  }

  flux_object_unref (FLUX_OBJECT (context->net_db));
  g_free (context->name);
  g_free (context);
}

void
tnt_net_db_add_address_named (TntNetDB *net_db, const gchar *name, guint port)
{
  NameLookupContext *context;

  assert_types (net_db);
  g_return_if_fail (name != NULL);
  g_return_if_fail (port != 0);

  context = g_new0 (NameLookupContext, 1);
  context->net_db = net_db;
  context->name   = g_strdup (name);
  context->port   = port;

  flux_object_ref (FLUX_OBJECT (net_db));
  flux_ip_resolver_lookup (flux_ip_resolver_get_default (), name, name_lookup_cb, context);
}

gint
tnt_net_db_get_n_nodes (TntNetDB *net_db)
{
  return n_nodes_in_table (net_db);
}

gint
tnt_net_db_get_n_addresses (TntNetDB *net_db)
{
  return n_addresses_in_table (net_db);
}
