/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-network-core.h -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef TNT_NETWORK_CORE_H
#define TNT_NETWORK_CORE_H

/* These defaults are good for a 256/128 ADSL connection. They will leave
 * about 4000 CPS inbound and 2000 CPS outbound for other activities. Note
 * that these numbers are not simply BPS/8, as you never get full
 * utilization - the typical maximum is 85% of the bandwidth in
 * payload data. Outgoing control CPS should be equal to or higher than
 * incoming control CPS to avoid dropping too many routed packets. */

#define TNT_DEFAULT_CPS_DATA_IN_MAX     20000
#define TNT_DEFAULT_CPS_CONTROL_IN_MAX   4000

#define TNT_DEFAULT_CPS_DATA_OUT_MAX     6000
#define TNT_DEFAULT_CPS_CONTROL_OUT_MAX  4000

#define TNT_DEFAULT_CONNECTIONS_MAX        20

#define TNT_NETWORK_CORE(object) ((TntNetworkCore *) (object))

struct _TntNetworkCore
{
  TntCore                core;

  /* Privacy options */
  guint                  hide_my_ip           : 1;
  guint                  introduce_delays     : 1;
  guint                  forward_own_messages : 1;

  /* Other flags */
  guint                  bootstrap_only       : 1;

  guint64                my_fingerprint;
  TntNodeID              my_node_id;
  TntNodeAddress4        my_node_address;
  GList                 *my_alt_address_list;

  /* For end-to-end encryption */
  FluxRSAPublicKey      *my_public_key;
  FluxRSASecretKey      *my_secret_key;

  /* For channel encryption */
  FluxCipher            *cipher;

  TntNetDB              *net_db;
  TntBandwidthGroupList *bandwidth_group_list;
  TntMessageFilter      *filter;
  TntTimeWarp           *time_warp;

  TntTcpListener        *tcp_listener;
  GPtrArray             *pending_edge_array;
  GPtrArray             *edge_array;
  GHashTable            *outgoing_caps;

  TntCallbackList       *new_edge_cb_list;
  TntCallbackList       *starvation_cb_list;
  GHashTable            *message_cb_list_table;
  GHashTable            *pre_data_cb_list_table;

  /* Table mapping message types to gbooleans indicating whether the receipt of such
   * a message may trigger replies. */
  GHashTable            *message_replies_table;

  guint                  broadcast_node_info_timeout_id;
  guint                  broadcast_contact_info_timeout_id;
  guint                  bootstrap_edges_timeout_id;
};

/* Config validation errors */

enum
{
  TNT_NETWORK_ERROR_DATA_DIR_UNSET,
  TNT_NETWORK_ERROR_DATA_DIR_RELATIVE,
  TNT_NETWORK_ERROR_DATA_DIR_NONEXISTENT,
  TNT_NETWORK_ERROR_DATA_DIR_PERMS,

  TNT_NETWORK_ERROR_LISTENING_PORT_UNSET,
  TNT_NETWORK_ERROR_LISTENING_PORT_UNAVAIL,

  TNT_NETWORK_ERROR_BANDWIDTH_GROUPS_UNSET,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_DUP_NAME,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_NO_IP,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_NO_NETMASK,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_BAD_IP,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_BAD_NETMASK,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_DUP_RANGE,

  TNT_NETWORK_ERROR_DATA_DIR_FIRST = TNT_NETWORK_ERROR_DATA_DIR_UNSET,
  TNT_NETWORK_ERROR_DATA_DIR_LAST  = TNT_NETWORK_ERROR_DATA_DIR_PERMS,

  TNT_NETWORK_ERROR_LISTENING_PORT_FIRST = TNT_NETWORK_ERROR_LISTENING_PORT_UNSET,
  TNT_NETWORK_ERROR_LISTENING_PORT_LAST  = TNT_NETWORK_ERROR_LISTENING_PORT_UNAVAIL,

  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_FIRST = TNT_NETWORK_ERROR_BANDWIDTH_GROUPS_UNSET,
  TNT_NETWORK_ERROR_BANDWIDTH_GROUP_LAST  = TNT_NETWORK_ERROR_BANDWIDTH_GROUP_DUP_RANGE
};

gboolean         tnt_network_core_init                      (TntNetworkCore *network_core, TntConfig *config);
TntNetworkCore  *tnt_network_core_new                       (TntConfig *config);
void             tnt_network_core_finalize                  (FluxObject *network_core);

TntNetDB        *tnt_network_core_peek_net_db               (TntNetworkCore *network_core);
TntBandwidthGroupList *tnt_network_core_peek_bandwidth_groups (TntNetworkCore *network_core);
const GPtrArray *tnt_network_core_peek_edge_array           (TntNetworkCore *network_core);
gint             tnt_network_core_get_n_edges               (TntNetworkCore *network_core);
gint             tnt_network_core_get_n_pending_edges       (TntNetworkCore *network_core);

TntNodeID        tnt_network_core_get_node_id               (TntNetworkCore *network_core);

void             tnt_network_core_get_keypair               (TntNetworkCore *network_core,
                                                             FluxRSAPublicKey **public_key,
                                                             FluxRSASecretKey **secret_key);
void             tnt_network_core_set_keypair               (TntNetworkCore *network_core, 
                                                             FluxRSAPublicKey *public_key,
                                                             FluxRSASecretKey *secret_key);

TntEdge         *tnt_network_core_find_route                (TntNetworkCore *network_core, TntNodeID node_id);
void             tnt_network_core_connect_to_node           (TntNetworkCore *network_core, TntNodeID node_id);
TntEdge         *tnt_network_core_route_message             (TntNetworkCore *network_core, TntMessage *message);
gboolean         tnt_network_core_send_message              (TntNetworkCore *network_core, TntMessage *message, TntEdge *edge);

gboolean         tnt_network_core_have_edge                 (TntNetworkCore *network_core, TntEdge *edge);

void             tnt_network_core_set_outgoing_cap          (TntNetworkCore *network_core, TntMessageType type,
                                                             guint cap);

void             tnt_network_core_attach_new_edge_handler   (TntNetworkCore *network_core,
                                                             TntNewEdgeHandler *func, gpointer data);
void             tnt_network_core_detach_new_edge_handler   (TntNetworkCore *network_core,
                                                             TntNewEdgeHandler *func, gpointer data);

void             tnt_network_core_attach_starvation_handler (TntNetworkCore *network_core,
                                                             TntStarvationHandler *func, gpointer data);
void             tnt_network_core_detach_starvation_handler (TntNetworkCore *network_core,
                                                             TntStarvationHandler *func, gpointer data);

void             tnt_network_core_attach_message_handler    (TntNetworkCore *network_core,
                                                             TntMessageType type, gboolean may_generate_replies,
                                                             TntMessageHandler *func, gpointer data);
void             tnt_network_core_detach_message_handler    (TntNetworkCore *network_core,
                                                             TntMessageType type,
                                                             TntMessageHandler *func, gpointer data);

void             tnt_network_core_attach_pre_data_handler   (TntNetworkCore *network_core,
                                                             TntMessageType type,
                                                             TntPreDataHandler *func, gpointer data);
void             tnt_network_core_detach_pre_data_handler   (TntNetworkCore *network_core,
                                                             TntMessageType type,
                                                             TntPreDataHandler *func, gpointer data);

#endif
