/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-node-meta.c - Stores information on a network node.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <time.h>
#include "tnt.h"

static FluxMemChunk *node_meta_mem_chunk = NULL;

TntNodeMeta *
tnt_node_meta_new (void)
{
  if (!node_meta_mem_chunk)
  {
    node_meta_mem_chunk = flux_mem_chunk_new (64, sizeof (TntNodeMeta));
  }

  return flux_mem_chunk_alloc0 (node_meta_mem_chunk);
}

TntNodeMeta *
tnt_node_meta_dup (TntNodeMeta *node_meta)
{
  TntNodeMeta *new_meta;

  g_return_val_if_fail (node_meta != NULL, NULL);

  if (!node_meta_mem_chunk)
    node_meta_mem_chunk = flux_mem_chunk_new (64, sizeof (TntNodeMeta));

  new_meta = flux_mem_chunk_alloc (node_meta_mem_chunk);
  memcpy (new_meta, node_meta, sizeof (TntNodeMeta));
  return new_meta;
}

void
tnt_node_meta_free (TntNodeMeta *node_meta)
{
  g_return_if_fail (node_meta != NULL);

  flux_mem_chunk_free (node_meta_mem_chunk, node_meta);
}

void
tnt_node_meta_reclaim_memory (void)
{
  if (node_meta_mem_chunk)
    flux_mem_chunk_clean (node_meta_mem_chunk);
}

void
tnt_node_meta_seen (TntNodeMeta *node_meta)
{
  g_return_if_fail (node_meta != NULL);

  node_meta->last_seen = time (NULL);
}

TntTimeStamp
tnt_node_meta_get_last_seen (TntNodeMeta *node_meta)
{
  g_return_val_if_fail (node_meta != NULL, 0);

  return node_meta->last_seen;
}

void
tnt_node_meta_get_primary_address (TntNodeMeta *node_meta, TntNodeAddress4 *dest)
{
  g_return_if_fail (node_meta != NULL);
  g_return_if_fail (dest != NULL);

  memcpy (dest, &node_meta->primary_address, sizeof (TntNodeAddress4));
}

void
tnt_node_meta_set_primary_address (TntNodeMeta *node_meta, TntNodeAddress4 *src)
{
  g_return_if_fail (node_meta != NULL);
  g_return_if_fail (src != NULL);

  if (!memcmp (&node_meta->primary_address, src, sizeof (TntNodeAddress4)))
    return;

  node_meta->primary_rating = 0;
  memcpy (&node_meta->primary_address, src, sizeof (TntNodeAddress4));
}

void
tnt_node_meta_get_secondary_address (TntNodeMeta *node_meta, TntNodeAddress4 *dest)
{
  g_return_if_fail (node_meta != NULL);
  g_return_if_fail (dest != NULL);

  memcpy (dest, &node_meta->secondary_address, sizeof (TntNodeAddress4));
}

void
tnt_node_meta_set_secondary_address (TntNodeMeta *node_meta, TntNodeAddress4 *src)
{
  g_return_if_fail (node_meta != NULL);
  g_return_if_fail (src != NULL);

  if (!memcmp (&node_meta->secondary_address, src, sizeof (TntNodeAddress4)))
    return;

  node_meta->secondary_rating = 0;
  memcpy (&node_meta->secondary_address, src, sizeof (TntNodeAddress4));
}

gint
tnt_node_meta_get_primary_rating (TntNodeMeta *node_meta)
{
  g_return_val_if_fail (node_meta != NULL, 0);

  return node_meta->primary_rating;
}

gint
tnt_node_meta_get_secondary_rating (TntNodeMeta *node_meta)
{
  g_return_val_if_fail (node_meta != NULL, 0);

  return node_meta->secondary_rating;
}

void
tnt_node_meta_rate_primary_address (TntNodeMeta *node_meta, gint change)
{
  g_return_if_fail (node_meta != NULL);

  node_meta->primary_rating += change;
}

void
tnt_node_meta_rate_secondary_address (TntNodeMeta *node_meta, gint change)
{
  g_return_if_fail (node_meta != NULL);

  node_meta->secondary_rating += change;
}

void
tnt_node_meta_get_best_address (TntNodeMeta *node_meta, TntNodeAddress4 *dest)
{
  g_return_if_fail (node_meta != NULL);
  g_return_if_fail (dest != NULL);

  if (node_meta->primary_address.ip != TNT_NODE_IP_INVALID)
    tnt_node_meta_get_primary_address (node_meta, dest);
  else
    tnt_node_meta_get_secondary_address (node_meta, dest);
}
