/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-ogg.c - Ogg Vorbis metadata extraction.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "tnt.h"

#ifdef ENABLE_VORBIS

#include <string.h>
#include <vorbis/vorbisfile.h>
#include <glib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

static gchar *
get_comment (vorbis_comment *vc, const gchar *field)
{
  gint   i;
  gint   field_name_len;
  gchar *comment = NULL;

  field_name_len = strlen (field);

  /* Find field and extract comment */
  for (i = 0; i < vc->comments && !comment; i++)
  {
    gint len = vc->comment_lengths [i];

    if (len < field_name_len + 2)  /* FIELD=c (at least one byte) */
      continue;

    if (*(vc->user_comments [i] + field_name_len) != '=' ||
        strncasecmp (field, vc->user_comments [i], field_name_len))
      continue;

    comment = g_strndup (vc->user_comments [i] + field_name_len + 1,
                         len - (field_name_len + 1));
  }

  /* Convert to UTF-8 if possible */
  if (comment)
  {
    gchar *comment_utf8;

    comment_utf8 = g_locale_to_utf8 (comment, -1, NULL, NULL, NULL);

    g_free (comment);
    comment = comment_utf8;
  }

  if (!comment)
    comment = g_strdup ("");

  return comment;
}

static gchar *
join_comments (const gchar *c1, const gchar *c2)
{
  if (!c1 [0])
    return g_strdup (c2);
  if (!c2 [0])
    return g_strdup (c1);

  return g_strjoin (" ", c1, c2, NULL);
}

static gint
guess_year (const gchar *comment)
{
  gint  i;
  guint num_in_a_row = 0;
  gint  four_digit_start = 0;

  for (i = 0; comment [i]; i++)
  {
    if (comment [i] >= '0' && comment [i] <= '9')
    {
      if (!num_in_a_row)
        four_digit_start = i;
      num_in_a_row++;
    }
    else
    {
      num_in_a_row = 0;
    }

    if (num_in_a_row == 4)
    {
      gint year;

      year = comment [four_digit_start]     - '0' * 1000 +
             comment [four_digit_start + 1] - '0' *  100 +
             comment [four_digit_start + 2] - '0' *   10 +
             comment [four_digit_start + 3] - '0';

      /* Assume people will tire of Tentacles sometime this millennium */
      if (year >= 1900 && year <= 3000)
        return year;

      /* Not a plausible year */
      num_in_a_row = 0;
      i = four_digit_start;  /* Will advance by 1 */
    }
  }

  return -1;
}

gboolean
tnt_ogg_extract_file_meta (gint fd, TntFileMeta *file_meta)
{
  gboolean        result = FALSE;
  gint            ov_fd;
  FILE           *ov_fp;
  OggVorbis_File  vf;
  vorbis_comment *vc;

  g_return_val_if_fail (fd >= 0, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  ov_fd = dup (fd);
  if (ov_fd < 0)
	  return FALSE;

  lseek (ov_fd, 0, SEEK_SET);
  ov_fp = fdopen (ov_fd, "rb");
  if (!ov_fp)
  {
    close (ov_fd);
    return FALSE;
  }

  /* The Ogg Vorbis reference claims this is a faster way to check
   * for ogg-ness. However, it doesn't seem to work on my oggs.
   *
   *  if (ov_test      (ov_fp, &vf, NULL, 0) < 0 ||
   *      ov_test_open (&vf) < 0)
   */
  if (ov_open (ov_fp, &vf, NULL, 0))
  {
    fclose (ov_fp);
    return FALSE;
  }

  tnt_file_meta_set_format (file_meta, TNT_FORMAT_OGG);
  result = TRUE;

  tnt_file_meta_set_quality (file_meta, ov_bitrate (&vf, -1) / 1000);
  tnt_file_meta_set_duration (file_meta, (guint32) ov_time_total (&vf, -1)); 

  vc = ov_comment (&vf, -1);
  if (vc)
  {
    gchar *s0, *s1, *s2;
    gint   n;

    s0 = get_comment (vc, "title");
    s1 = get_comment (vc, "version");
    s2 = join_comments (s0, s1);
    tnt_file_meta_set_title (file_meta, s2);
    g_free (s0);
    g_free (s1);
    g_free (s2);

    s0 = get_comment (vc, "album");
    tnt_file_meta_set_collection (file_meta, s0);
    g_free (s0);

    s0 = get_comment (vc, "artist");
    if (!s0 [0])
    {
      g_free (s0);
      s0 = get_comment (vc, "organization");
    }
#if 0
    /* DISABLED: We probably don't want to fall back - it's better
     * to leave an empty field if unspecified. */
    if (!s0 [0])
    {
      /* Fall back to performer for both author/performer fields.
       * This makes the track artist the same as the album artist. */
      g_free (s0);
      s0 = get_comment (vc, "performer");
    }
#endif
    tnt_file_meta_set_author (file_meta, s0);
    g_free (s0);

    s0 = get_comment (vc, "performer");
    tnt_file_meta_set_performer (file_meta, s0);
    g_free (s0);

    s0 = get_comment (vc, "tracknumber");
    tnt_file_meta_set_seq (file_meta, s0);
    g_free (s0);

    s0 = get_comment (vc, "description");
    s1 = get_comment (vc, "genre");
    s2 = join_comments (s0, s1);
    tnt_file_meta_set_keywords (file_meta, s2);
    g_free (s0);
    g_free (s1);
    g_free (s2);

    s0 = get_comment (vc, "date");
    n = guess_year (s0);
    g_free (s0);
    if (n < 0)
    {
      s0 = get_comment (vc, "copyright");
      n = guess_year (s0);
      g_free (s0);
    }

    if (n > 0)
      tnt_file_meta_set_year (file_meta, n);
  }

  ov_clear (&vf);
  return result;
}

#else  /* !ENABLE_VORBIS */

gboolean
tnt_ogg_extract_file_meta (gint fd, TntFileMeta *file_meta)
{
  return FALSE;
}

#endif  /* ENABLE_VORBIS */
