/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-remote-query.c - Describes a metadata query made by a remote user.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_REMOTE_QUERY_TYPE)

typedef struct
{
  TntFileID    file_id;
  TntFileMeta *file_meta;
}
Match;

static void
free_match (Match *match)
{
  tnt_file_meta_free (match->file_meta);
  g_free (match);
}

static gboolean
query_matches_meta (TntRemoteQuery *query, TntFileMeta *file_meta)
{
  if (tnt_file_meta_matches_string (file_meta, query->query.match_field, query->query.string) &&
      (query->query.media    == TNT_MEDIA_UNKNOWN ||
       query->query.media    == tnt_file_meta_get_media (file_meta)) &&
      (query->query.category == 0 ||
       query->query.category == tnt_file_meta_get_category (file_meta, 0)))
    return TRUE;

  return FALSE;
}

static gboolean
try_match (TntQuery *query, const TntFileID *file_id, TntFileMeta *file_meta,
           TntNodeID node_id, gint remote_blocks)
{
  TntRemoteQuery *remote_query = TNT_REMOTE_QUERY (query);

  assert_types (remote_query);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  if (!tnt_query_get_match_local (TNT_QUERY (query)) && tnt_file_meta_is_local_file (file_meta))
    return FALSE;

  if (!tnt_query_get_match_partial (TNT_QUERY (query)) && !tnt_file_meta_is_local_file (file_meta) &&
      remote_blocks < TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len))
    return FALSE;

  if (query_matches_meta (remote_query, file_meta))
  {
    Match *match = g_new (Match, 1);

    memcpy (&match->file_id, file_id, sizeof (TntFileID));
    match->file_meta = tnt_file_meta_dup (file_meta);

    g_queue_push_tail (remote_query->match_queue, match);
    return TRUE;
  }

  return FALSE;
}

static gint
get_n_matches (TntQuery *query)
{
  TntRemoteQuery *remote_query = TNT_REMOTE_QUERY (query);

  assert_types (remote_query);
  return g_queue_get_length (remote_query->match_queue);
}

static void
filter_matches (TntQuery *query)
{
  /* FIXME: Not in use; not implemented */
}

void
tnt_remote_query_finalize (FluxObject *object)
{
  TntRemoteQuery *remote_query = TNT_REMOTE_QUERY (object);
  Match          *match;

  assert_types (object);

  while ((match = g_queue_pop_head (remote_query->match_queue)))
    free_match (match);

  g_queue_free (remote_query->match_queue);

  tnt_query_finalize (object);
}

gboolean
tnt_remote_query_init (TntRemoteQuery *remote_query, TntNodeID node_id)
{
  TntQuery *query;

  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, FALSE);

  tnt_query_init                (TNT_QUERY (remote_query));
  flux_object_set_type          (FLUX_OBJECT (remote_query), TNT_REMOTE_QUERY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (remote_query), tnt_remote_query_finalize);

  query = TNT_QUERY (remote_query);

  query->match_partial = FALSE;

  query->filter_func        = filter_matches;
  query->get_n_matches_func = get_n_matches;
  query->try_match_func     = try_match;

  remote_query->node_id = node_id;
  remote_query->match_queue = g_queue_new ();
  return TRUE;
}

TntRemoteQuery *
tnt_remote_query_new (TntNodeID node_id)
{
  TntRemoteQuery *remote_query;

  remote_query = g_new0 (TntRemoteQuery, 1);

  if (!tnt_remote_query_init (remote_query, node_id))
  {
    g_free (remote_query);
    return NULL;
  }

  return remote_query;
}

TntNodeID
tnt_remote_query_get_node_id (TntRemoteQuery *remote_query)
{
  assert_types (remote_query);

  return remote_query->node_id;
}

gboolean
tnt_remote_query_get_next_match (TntRemoteQuery *remote_query, TntFileID *file_id,
                                 TntFileMeta **file_meta)
{
  Match *match;

  assert_types (remote_query);

  match = g_queue_pop_head (remote_query->match_queue);
  if (!match)
    return FALSE;

  memcpy (file_id, &match->file_id, sizeof (TntFileID));
  *file_meta = match->file_meta;

  g_free (match);
  return TRUE;
}
