/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-tcp-io.c - TCP I/O channel.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <string.h>
#include <time.h>

#define DEBUG(x)
#undef  LOG_STREAMS

/* The Tentacles network packet
 * ============================
 *
 * Headers
 * -------
 *
 * mandatory  1 byte   flags                Determine what optional parts are present.
 * mandatory  2 bytes  type                 Message type.
 * optional   1 byte   ttl                  Time to live.
 * optional   8 bytes  source node id       Routing info.
 * optional   8 bytes  destination node id  Routing info.
 * optional   1 byte   signature length     ._____
 * optional   2 bytes  payload length       .___  |
 * optional   2 bytes  data block length    ._  | |
 *                                            | | |
 * Body                                       | | |  depends
 * ----                                       | | |
 *                                            | | |
 * optional   1-255 bytes     signature     ._|_|_|
 * optional   1-4095 bytes    payload       ._|_|
 * optional   1-262144 bytes  data block    ._|
 *
 * The smallest valid packet is 3 bytes (flags, type). I expect an average
 * packet to be about 25 bytes (flags, type, ttl, source, payload size,
 * payload) - with a payload of 11 bytes. Thus, 100 packets/s ~= 2.5KB/s.
 *
 * The payload is supplementary information which must be marshalled/demarshalled
 * according to packet type. The data block is an optional raw data block which
 * is not marshalled - its use depends on the message type and payload.
 *
 * If the payload is encrypted and signed, the signature pertains to the
 * unencrypted payload.
 */

#define SECONDS_PER_SLICE               1
#define DEFAULT_QUEUE_LEN               100
#define DEFAULT_BANDWIDTH_CAP_IN        10000  /* 1MB/s */
#define DEFAULT_BANDWIDTH_CAP_OUT       20000
#define CPS_TO_BYTES_PER_SLICE(cps)     ((cps) * SECONDS_PER_SLICE)
#define BYTES_PER_SLICE_TO_CPS(bpslice) ((bpslice) / SECONDS_PER_SLICE)
#define SHORT_TERM_WEIGHT               0.2
#define LONG_TERM_WEIGHT                0.05

#define DATA_BLOCK_SIZE_MAX             (1024 * 1024)  /* 1 MB */

#define assert_types(o)            flux_object_check (FLUX_OBJECT (o), TNT_TCP_IO_TYPE)

static void read_cb                (FluxIO *io, guint min_len, gpointer data);
static void write_cb               (FluxIO *io, guint min_len, gpointer data);
static void drop_all               (TntTcpIO *tcp_io);
static void end_read_state         (TntTcpIO *tcp_io);

DEBUG (

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void
dump_hex (guchar *buf, guint len)
{
  gint i;

  for (i = 0; i < len; i++)
    g_print ("%c%c ", int_to_hex (buf [i] >> 4), int_to_hex (buf [i] & 0x0f));

  g_print ("\n");
}

)  /* DEBUG */

typedef struct
{
  gint max_len;
  gint len;
}
QueueClassInfo;

/* FIXME: If we add a new QueueClassInfo struct, we need to recount the current messages
 * into the updated set of classes - otherwise a class count might never be decremented
 * due to a message being reclassed while in the queue. This is only a problem if queue
 * classes are changed after we start queuing messages, though. */

static void
set_queue_class_max_length (TntTcpIO *tcp_io, TntMessageType message_type, gint max_len)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
  {
    queue_info = g_new0 (QueueClassInfo, 1);
    g_hash_table_insert (tcp_io->queue_class_table,
                         GUINT_TO_POINTER (message_type), queue_info);
  }

  queue_info->max_len = max_len;
}

static gint
get_queue_class_max_length (TntTcpIO *tcp_io, TntMessageType message_type)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
    queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (TNT_MESSAGE_UNKNOWN));

  g_assert (queue_info);

  return queue_info->max_len;
}

static void
remove_queue_class (TntTcpIO *tcp_io, TntMessageType message_type)
{
  if (message_type == TNT_MESSAGE_UNKNOWN)
  {
    g_warning ("Can't remove default queue class.");
    return;
  }

  g_hash_table_remove (tcp_io->queue_class_table, GINT_TO_POINTER (message_type));
}

static void
increment_queue_class_length (TntTcpIO *tcp_io, TntMessageType message_type)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
    queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (TNT_MESSAGE_UNKNOWN));

  g_assert (queue_info);

  queue_info->len++;
}

static void
decrement_queue_class_length (TntTcpIO *tcp_io, TntMessageType message_type)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
    queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (TNT_MESSAGE_UNKNOWN));

  g_assert (queue_info);

  queue_info->len--;
  g_assert (queue_info->len >= 0);
}

static gint
get_queue_class_length (TntTcpIO *tcp_io, TntMessageType message_type)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
    queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (TNT_MESSAGE_UNKNOWN));
  if (!queue_info)
    return 0;

  return queue_info->len;
}

static gboolean
queue_class_is_full (TntTcpIO *tcp_io, TntMessageType message_type)
{
  QueueClassInfo *queue_info;

  queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (message_type));
  if (!queue_info)
    queue_info = g_hash_table_lookup (tcp_io->queue_class_table, GUINT_TO_POINTER (TNT_MESSAGE_UNKNOWN));

  g_assert (queue_info);

  if (queue_info->len >= queue_info->max_len)
    return TRUE;

  return FALSE;
}

/* Clears all info associated with read state-keeping. Note that pointers
 * are zeroed, but memory isn't freed; it is assumed that the data blocks
 * may be handed off to callbacks who become their new owners. */
static void
clear_read_state_info (TntTcpIO *tcp_io)
{
  tcp_io->read_info.flags           = 0;
  tcp_io->read_info.type            = TNT_MESSAGE_UNKNOWN;
  tcp_io->read_info.ttl             = 0;
  tcp_io->read_info.signature_size  = 0;
  tcp_io->read_info.payload         = NULL;
  tcp_io->read_info.payload_size    = 0;
  tcp_io->read_info.data_block      = NULL;
  tcp_io->read_info.data_block_size = 0;
  tcp_io->read_info.block_pos       = 0;

  tcp_io->read_info.source_id       = TNT_NODE_ID_INVALID;
  tcp_io->read_info.dest_id         = TNT_NODE_ID_INVALID;
}

static void
free_read_state_info (TntTcpIO *tcp_io)
{
  if (tcp_io->read_info.payload)
    g_free (tcp_io->read_info.payload);
  if (tcp_io->read_info.data_block)
    tnt_free_block (tcp_io->read_info.data_block);

  clear_read_state_info (tcp_io);
}

/* Clears all info associated with write state-keeping. Note that zeroed
 * pointers point to data that will be freed when message is destroyed. */
static void
clear_write_state_info (TntTcpIO *tcp_io)
{
  tcp_io->write_info.flags      = 0;
  tcp_io->write_info.type       = TNT_MESSAGE_UNKNOWN;
  tcp_io->write_info.ttl        = 0;
  tcp_io->write_info.block      = NULL;
  tcp_io->write_info.block_size = 0;
  tcp_io->write_info.block_pos  = 0;

  tcp_io->write_info.source_id  = TNT_NODE_ID_INVALID;
  tcp_io->write_info.dest_id    = TNT_NODE_ID_INVALID;

  if (tcp_io->write_info.message)
  {
    flux_object_unref (FLUX_OBJECT (tcp_io->write_info.message));
    tcp_io->write_info.message = NULL;
  }
}

static void
force_close (TntTcpIO *tcp_io)
{
  DEBUG (g_message ("TntTcpIO: Forced close."));

  if (tcp_io->read_info.payload)
    g_free (tcp_io->read_info.payload);
  if (tcp_io->read_info.data_block)
    tnt_free_block (tcp_io->read_info.data_block);

  drop_all (tcp_io);
  clear_read_state_info (tcp_io);
  clear_write_state_info (tcp_io);
  flux_tcp_io_disconnect (FLUX_TCP_IO (tcp_io));
}

#ifdef LOG_STREAMS
static void
log_stream (gpointer instance, const gchar *postfix, const guchar *data, gint len)
{
  gchar *file;
  FILE  *fhd;

  file = g_strdup_printf ("%p.%s", instance, postfix);
  fhd = fopen (file, "ab");
  g_free (file);

  if (!fhd)
    return;

  fwrite (data, len, 1, fhd);
  fclose (fhd);
}
#endif

static void
write_out (TntTcpIO *tcp_io, gpointer src, gint len)
{
  flux_dual_fifo_write (FLUX_DUAL_FIFO (tcp_io), src, len);
#ifdef LOG_STREAMS
  log_stream (tcp_io, "out", src, len);
#endif
}

static void
read_in (TntTcpIO *tcp_io, gpointer dest, gint len)
{
  flux_dual_fifo_read (FLUX_DUAL_FIFO (tcp_io), dest, len);
#ifdef LOG_STREAMS
  log_stream (tcp_io, "in", dest, len);
#endif
}

static TntMessage *
construct_message (TntTcpIO *tcp_io)
{
  TntMessage *message;

  /* TODO: Integrate signature and encryption */

  if (tcp_io->read_info.flags & (TNT_MESSAGE_FLAG_HAS_SOURCE | TNT_MESSAGE_FLAG_HAS_DEST))
    message = TNT_MESSAGE (tnt_routed_message_new (tcp_io->read_info.type,
                                                   tcp_io->read_info.source_id,
                                                   tcp_io->read_info.dest_id));
  else
    message = tnt_message_new (tcp_io->read_info.type);

  if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_IS_DIRECT)
    tnt_message_set_direct (message, TRUE);
  if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_TTL)
    tnt_message_set_ttl (message, tcp_io->read_info.ttl);

  if (tcp_io->read_info.payload)
    tnt_message_set_payload (message, tcp_io->read_info.payload, tcp_io->read_info.payload_size);
  if (tcp_io->read_info.data_block)
    tnt_message_set_data_block (message, tcp_io->read_info.data_block, tcp_io->read_info.data_block_size);

  return message;
}

/* Creates a Tnt(Routed)Message from stored state and sends it off
 * to waiting callback. */
static void
dispatch_message (TntTcpIO *tcp_io)
{
  TntMessage *message;

  if (!tcp_io->message_in_func)
  {
    /* Drop incoming messages if there's no handler attached. I guess a less
     * error-prone way to handle this would be to maintain an incoming queue
     * from which we can dispatch when a handler is attached, or that we can
     * free if we're unreffed immediately.
     *
     * This occurs in practice - we may get a message from the underlying
     * FluxTcpIO with no handler if we're ciphered and get immediately
     * unreffed by the layer above us. */

    if (tcp_io->read_info.payload)
      g_free (tcp_io->read_info.payload);
    if (tcp_io->read_info.data_block)
      tnt_free_block (tcp_io->read_info.data_block);
    return;
  }

  message = construct_message (tcp_io);

  DEBUG
  (
    g_print ("[recv] ");
    tnt_message_dump (message);
  )

  tcp_io->message_in_func (tcp_io, message, tcp_io->message_in_data);
}

/* Called on transition to new read state */
static void
begin_read_state (TntTcpIO *tcp_io, TntTcpIOReadState read_state)
{
  guint expected_size;

  tcp_io->read_info.state = read_state;

  switch (read_state)
  {
    case TNT_TCP_IO_READ_FLAGS:
      expected_size = TNT_MESSAGE_FLAGS_SIZE;
      break;

    case TNT_TCP_IO_READ_MESSAGE_HEADER:
      expected_size = TNT_MESSAGE_TYPE_SIZE;
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_TTL)        expected_size += TNT_MESSAGE_TTL_SIZE;
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_SOURCE)     expected_size += sizeof (TntNodeID);
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_DEST)       expected_size += sizeof (TntNodeID);
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_SIGNATURE)  expected_size += TNT_MESSAGE_SIGNATURE_SIZE_SIZE;
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_PAYLOAD)    expected_size += TNT_MESSAGE_PAYLOAD_SIZE_SIZE;
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK) expected_size += TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE;
      break;

    case TNT_TCP_IO_READ_SIGNATURE:
      expected_size = tcp_io->read_info.signature_size;

      if (expected_size == 0)
      {
        end_read_state (tcp_io);
        return;
      }
      break;

    case TNT_TCP_IO_READ_MESSAGE_PAYLOAD:
      expected_size = tcp_io->read_info.payload_size;

      if (expected_size == 0)
      {
        /* begin_read_state (tcp_io, TNT_TCP_IO_READ_FLAGS); */
        end_read_state (tcp_io);
        return;
      }

      if (expected_size > TNT_PAYLOAD_SIZE_MAX)
      {
        force_close (tcp_io);
        expected_size = 0;
        break;
      }

      tcp_io->read_info.payload = g_malloc (tcp_io->read_info.payload_size);
      tcp_io->read_info.block_pos = 0;
      break;

    case TNT_TCP_IO_READ_DATA_BLOCK:
      if (tcp_io->read_info.data_block_size > DATA_BLOCK_SIZE_MAX)
      {
        force_close (tcp_io);
        expected_size = 0;
        break;
      }

      if (tcp_io->data_in_begin_func)
      {
        TntMessage *message;
        gpointer    payload;

        /* Make a copy of payload for the message to take ownership of */
        payload = tcp_io->read_info.payload;
        if (tcp_io->read_info.payload)
        {
          tcp_io->read_info.payload = g_malloc (tcp_io->read_info.payload_size);
          memcpy (tcp_io->read_info.payload, payload, tcp_io->read_info.payload_size);
        }

        message = construct_message (tcp_io);

        tcp_io->read_info.payload = payload;
        tcp_io->data_in_begin_func (tcp_io, message, tcp_io->data_in_begin_data);
      }

      expected_size = TNT_DATA_STEP_SIZE < tcp_io->read_info.data_block_size ?
        TNT_DATA_STEP_SIZE : tcp_io->read_info.data_block_size;
      tcp_io->read_info.data_block = tnt_alloc_block (tcp_io->read_info.data_block_size);
      tcp_io->read_info.block_pos = 0;
      break;

    default:
      expected_size = 0;
      g_assert_not_reached ();
  }

  flux_io_set_read_threshold (FLUX_IO (tcp_io), expected_size);
}

/* Called when a read state ends. Decides what to do next, and performs
 * the transition to the appropriate next state. */
static void
end_read_state (TntTcpIO *tcp_io)
{
  TntTcpIOReadState next_state;

  switch (tcp_io->read_info.state)
  {
    case TNT_TCP_IO_READ_FLAGS:
      next_state = TNT_TCP_IO_READ_MESSAGE_HEADER;
      break;

    case TNT_TCP_IO_READ_MESSAGE_HEADER:
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_SIGNATURE)
        next_state = TNT_TCP_IO_READ_SIGNATURE;
      else if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_PAYLOAD)
        next_state = TNT_TCP_IO_READ_MESSAGE_PAYLOAD;
      else if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK)
        next_state = TNT_TCP_IO_READ_DATA_BLOCK;
      else
        next_state = TNT_TCP_IO_READ_FLAGS;
      break;

    case TNT_TCP_IO_READ_SIGNATURE:
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_PAYLOAD)
        next_state = TNT_TCP_IO_READ_MESSAGE_PAYLOAD;
      else if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK)
        next_state = TNT_TCP_IO_READ_DATA_BLOCK;
      else
        next_state = TNT_TCP_IO_READ_FLAGS;
      break;

    case TNT_TCP_IO_READ_MESSAGE_PAYLOAD:
      if (tcp_io->read_info.flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK)
        next_state = TNT_TCP_IO_READ_DATA_BLOCK;
      else
        next_state = TNT_TCP_IO_READ_FLAGS;
      break;

    case TNT_TCP_IO_READ_DATA_BLOCK:
      next_state = TNT_TCP_IO_READ_FLAGS;
      break;

    default:
      next_state = TNT_TCP_IO_READ_FLAGS;
      g_assert_not_reached ();
  }

  if (next_state == TNT_TCP_IO_READ_FLAGS)
  {
    dispatch_message (tcp_io);
    clear_read_state_info (tcp_io);
  }

  begin_read_state (tcp_io, next_state);
}

static void
read_cb (FluxIO *io, guint min_len, gpointer data)
{
  TntTcpIOReadState  state;
  guchar             flags;
  guchar            *temp_mem;
  guint16            type;

  flux_object_ref (FLUX_OBJECT (io));
  state = TNT_TCP_IO (io)->read_info.state;
  flags = TNT_TCP_IO (io)->read_info.flags;

  switch (state)
  {
    case TNT_TCP_IO_READ_FLAGS:
      read_in (TNT_TCP_IO (io), &TNT_TCP_IO (io)->read_info.flags, TNT_MESSAGE_FLAGS_SIZE);
      break;

    case TNT_TCP_IO_READ_MESSAGE_HEADER:
      read_in (TNT_TCP_IO (io), &type, TNT_MESSAGE_TYPE_SIZE);
      TNT_TCP_IO (io)->read_info.type = GUINT16_FROM_LE (type);

      if (flags & TNT_MESSAGE_FLAG_HAS_TTL)
      {
        read_in (TNT_TCP_IO (io), &TNT_TCP_IO (io)->read_info.ttl, TNT_MESSAGE_TTL_SIZE);
      }
      if (flags & TNT_MESSAGE_FLAG_HAS_SOURCE)
      {
        TntNodeID id;

        read_in (TNT_TCP_IO (io), &id, sizeof (TntNodeID));
        TNT_TCP_IO (io)->read_info.source_id = GUINT64_FROM_LE (id);
      }
      if (flags & TNT_MESSAGE_FLAG_HAS_DEST)
      {
        TntNodeID id;

        read_in (TNT_TCP_IO (io), &id, sizeof (TntNodeID));
        TNT_TCP_IO (io)->read_info.dest_id = GUINT64_FROM_LE (id);
      }
      if (flags & TNT_MESSAGE_FLAG_HAS_SIGNATURE)
      {
        guint8 val;

        /* Assumes TNT_MESSAGE_SIGNATURE_SIZE_SIZE is 1 */
        read_in (TNT_TCP_IO (io), &val, TNT_MESSAGE_SIGNATURE_SIZE_SIZE);
        TNT_TCP_IO (io)->read_info.signature_size = val;
      }
      if (flags & TNT_MESSAGE_FLAG_HAS_PAYLOAD)
      {
        guchar val [TNT_MESSAGE_PAYLOAD_SIZE_SIZE];

        read_in (TNT_TCP_IO (io), &val, TNT_MESSAGE_PAYLOAD_SIZE_SIZE);
        TNT_TCP_IO (io)->read_info.payload_size = GUINT16_FROM_LE (*((guint16 *) val));
      }
      if (flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK)
      {
        guchar val [TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE];

        read_in (TNT_TCP_IO (io), &val, TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE);
        TNT_TCP_IO (io)->read_info.data_block_size = GUINT32_FROM_LE (*((guint32 *) val));
      }
      break;

    case TNT_TCP_IO_READ_SIGNATURE:
      /* FIXME: Use drop_bytes here */
      temp_mem = g_malloc (TNT_TCP_IO (io)->read_info.signature_size);
      read_in (TNT_TCP_IO (io), temp_mem, TNT_TCP_IO (io)->read_info.signature_size);
      g_free (temp_mem);
      break;

    case TNT_TCP_IO_READ_MESSAGE_PAYLOAD:
      g_assert (TNT_TCP_IO (io)->read_info.payload);

      read_in (TNT_TCP_IO (io), (guchar *) TNT_TCP_IO (io)->read_info.payload +
               TNT_TCP_IO (io)->read_info.block_pos, min_len);
      TNT_TCP_IO (io)->read_info.block_pos += min_len;

      if (TNT_TCP_IO (io)->read_info.block_pos + TNT_DATA_STEP_SIZE >
          TNT_TCP_IO (io)->read_info.payload_size)
      {
        /* Last read increment in a block can be smaller than normal. This will also be invoked
         * when the block has been fully transferred (the threshold is set to 0), but I don't
         * think it matters. */
        flux_io_set_read_threshold (io, TNT_TCP_IO (io)->read_info.payload_size -
                                        TNT_TCP_IO (io)->read_info.block_pos);
      }
      break;

    case TNT_TCP_IO_READ_DATA_BLOCK:
      g_assert (TNT_TCP_IO (io)->read_info.data_block);

      read_in (TNT_TCP_IO (io),
               (guchar *) TNT_TCP_IO (io)->read_info.data_block +
               TNT_TCP_IO (io)->read_info.block_pos, min_len);
      TNT_TCP_IO (io)->read_info.block_pos += min_len;

      if (TNT_TCP_IO (io)->read_info.block_pos + TNT_DATA_STEP_SIZE >
          TNT_TCP_IO (io)->read_info.data_block_size)
      {
        /* Last read increment in a block can be smaller than normal. This will also be invoked
         * when the block has been fully transferred (the threshold is set to 0), but I don't
         * think it matters. */
        flux_io_set_read_threshold (io, TNT_TCP_IO (io)->read_info.data_block_size -
                                        TNT_TCP_IO (io)->read_info.block_pos);
      }
      break;

    default:
      g_assert_not_reached ();
  }

  if (state == TNT_TCP_IO_READ_DATA_BLOCK)
  {
    if (TNT_TCP_IO (io)->read_info.block_pos == TNT_TCP_IO (io)->read_info.data_block_size)
      end_read_state (TNT_TCP_IO (io));
  }
  else if (state == TNT_TCP_IO_READ_MESSAGE_PAYLOAD)
  {
    if (TNT_TCP_IO (io)->read_info.block_pos == TNT_TCP_IO (io)->read_info.payload_size)
      end_read_state (TNT_TCP_IO (io));
  }
  else
  {
    end_read_state (TNT_TCP_IO (io));
  }

  flux_object_unref (FLUX_OBJECT (io));
}

static TntMessage *
dequeue_message_out (TntTcpIO *tcp_io)
{
  GList      *node;
  TntMessage *message;

  if (!tcp_io->queue_out_len)
    return NULL;

  g_assert (tcp_io->queue_out);

  node = tcp_io->queue_out;
  message = node->data;
  tcp_io->queue_out = g_list_remove_link (tcp_io->queue_out, node);
  g_list_free_1 (node);

  if (!tcp_io->queue_out)
    tcp_io->queue_out_last = NULL;

  decrement_queue_class_length (tcp_io, tnt_message_get_type (message));
  tcp_io->queue_out_len--;

  DEBUG
  (
    g_print ("[send] ");
    tnt_message_dump (message);
  )

  return message;
}

static guchar
pack_message_flags (TntMessage *message)
{
  guchar flags = 0;

  if (tnt_message_get_ttl (message))
    flags |= TNT_MESSAGE_FLAG_HAS_TTL;

  if (flux_object_get_type (FLUX_OBJECT (message)) == TNT_ROUTED_MESSAGE_TYPE)
  {
    if (tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
      flags |= TNT_MESSAGE_FLAG_HAS_SOURCE;
    if (tnt_routed_message_has_destination (TNT_ROUTED_MESSAGE (message)))
      flags |= TNT_MESSAGE_FLAG_HAS_DEST;
  }

  if (tnt_message_get_direct (message))
    flags |= TNT_MESSAGE_FLAG_IS_DIRECT;

  if (tnt_message_has_payload (message))
    flags |= TNT_MESSAGE_FLAG_HAS_PAYLOAD;

  if (tnt_message_has_data_block (message))
    flags |= TNT_MESSAGE_FLAG_HAS_DATA_BLOCK;

  return flags;
}

/* Called on transition to new write state */
static void
begin_write_state (TntTcpIO *tcp_io, TntTcpIOReadState write_state)
{
  guint expected_size;

  tcp_io->write_info.state = write_state;

  switch (write_state)
  {
    case TNT_TCP_IO_WRITE_IDLE:
      expected_size = 0;
      break;

    case TNT_TCP_IO_WRITE_HEADERS:
      g_assert (tcp_io->write_info.message == NULL);
      tcp_io->write_info.message = dequeue_message_out (tcp_io);
      if (!tcp_io->write_info.message)
      {
        flux_dual_fifo_flush_output (FLUX_DUAL_FIFO (tcp_io));
        tcp_io->write_info.state = TNT_TCP_IO_WRITE_IDLE;
        expected_size = 0;
        break;
      }

      expected_size = TNT_MESSAGE_FLAGS_SIZE + TNT_MESSAGE_TYPE_SIZE;

      tcp_io->write_info.flags = pack_message_flags (tcp_io->write_info.message);
      tcp_io->write_info.type  = tnt_message_get_type (tcp_io->write_info.message);
      tcp_io->write_info.ttl   = tnt_message_get_ttl (tcp_io->write_info.message);

      if (tcp_io->write_info.ttl)
        expected_size += TNT_MESSAGE_TTL_SIZE;

      if (flux_object_get_type (FLUX_OBJECT (tcp_io->write_info.message)) == TNT_ROUTED_MESSAGE_TYPE)
      {
        if (tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (tcp_io->write_info.message)))
        {
          expected_size += sizeof (TntNodeID);
          tcp_io->write_info.source_id =
            tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (tcp_io->write_info.message));
        }
        if (tnt_routed_message_has_destination (TNT_ROUTED_MESSAGE (tcp_io->write_info.message)))
        {
          expected_size += sizeof (TntNodeID);
          tcp_io->write_info.dest_id =
            tnt_routed_message_get_destination (TNT_ROUTED_MESSAGE (tcp_io->write_info.message));
        }
      }

      if (tnt_message_has_payload (tcp_io->write_info.message))
        expected_size += TNT_MESSAGE_PAYLOAD_SIZE_SIZE;

      if (tnt_message_has_data_block (tcp_io->write_info.message))
        expected_size += TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE;

      break;

    case TNT_TCP_IO_WRITE_PAYLOAD:
      tnt_message_get_payload (tcp_io->write_info.message, &tcp_io->write_info.block,
                               &tcp_io->write_info.block_size);
      expected_size = TNT_DATA_STEP_SIZE < tcp_io->write_info.block_size ?
        TNT_DATA_STEP_SIZE : tcp_io->write_info.block_size;
      tcp_io->write_info.block_pos = 0;
      break;

    case TNT_TCP_IO_WRITE_DATA_BLOCK:
      tnt_message_get_data_block (tcp_io->write_info.message, &tcp_io->write_info.block,
                                  &tcp_io->write_info.block_size);
      expected_size = TNT_DATA_STEP_SIZE < tcp_io->write_info.block_size ?
        TNT_DATA_STEP_SIZE : tcp_io->write_info.block_size;
      tcp_io->write_info.block_pos = 0;
      break;

    default:
      expected_size = 0;
      g_assert_not_reached ();
  }

  flux_io_set_write_threshold (FLUX_IO (tcp_io), expected_size);
}

/* Called when a write state ends. Decides what to do next, and performs
 * the transition to the appropriate next state. */
static void
end_write_state (TntTcpIO *tcp_io)
{
  TntTcpIOWriteState next_state;
  gpointer           block;
  guint              block_len;

  switch (tcp_io->write_info.state)
  {
    case TNT_TCP_IO_WRITE_IDLE:
      next_state = TNT_TCP_IO_WRITE_HEADERS;
      break;

    case TNT_TCP_IO_WRITE_HEADERS:
      tnt_message_get_payload (tcp_io->write_info.message, &block, &block_len);
      if (block)
      {
        next_state = TNT_TCP_IO_WRITE_PAYLOAD;
        break;
      }

      /* Fall through to data block check */

    case TNT_TCP_IO_WRITE_PAYLOAD:
      tnt_message_get_data_block (tcp_io->write_info.message, &block, &block_len);
      if (block)
        next_state = TNT_TCP_IO_WRITE_DATA_BLOCK;
      else
        next_state = TNT_TCP_IO_WRITE_HEADERS;
      break;

    case TNT_TCP_IO_WRITE_DATA_BLOCK:
      next_state = TNT_TCP_IO_WRITE_HEADERS;
      break;

    default:
      next_state = TNT_TCP_IO_WRITE_IDLE;
      g_assert_not_reached ();
  }

  if (next_state == TNT_TCP_IO_WRITE_HEADERS)
  {
    clear_write_state_info (tcp_io);
  }

  begin_write_state (tcp_io, next_state);
}

static void
write_cb (FluxIO *io, guint min_len, gpointer data)
{
  TntTcpIOWriteState state;
  guint16            type;

  flux_object_ref (FLUX_OBJECT (io));
  state = TNT_TCP_IO (io)->write_info.state;

  switch (state)
  {
    case TNT_TCP_IO_WRITE_IDLE:
      /* This would actually be a bug. We can't get here if we're idling. */
      return;

    case TNT_TCP_IO_WRITE_HEADERS:
      write_out (TNT_TCP_IO (io), &TNT_TCP_IO (io)->write_info.flags, TNT_MESSAGE_FLAGS_SIZE);
      type = GUINT16_TO_LE (TNT_TCP_IO (io)->write_info.type);
      write_out (TNT_TCP_IO (io), &type, TNT_MESSAGE_TYPE_SIZE);

      if (TNT_TCP_IO (io)->write_info.flags & TNT_MESSAGE_FLAG_HAS_TTL)
        write_out (TNT_TCP_IO (io), &TNT_TCP_IO (io)->write_info.ttl, TNT_MESSAGE_TTL_SIZE);

      if (TNT_TCP_IO (io)->write_info.flags & TNT_MESSAGE_FLAG_HAS_SOURCE)
      {
        guint64 id = GUINT64_TO_LE (TNT_TCP_IO (io)->write_info.source_id);
        write_out (TNT_TCP_IO (io), &id, sizeof (guint64));
      }

      if (TNT_TCP_IO (io)->write_info.flags & TNT_MESSAGE_FLAG_HAS_DEST)
      {
        guint64 id = GUINT64_TO_LE (TNT_TCP_IO (io)->write_info.dest_id);
        write_out (TNT_TCP_IO (io), &id, sizeof (guint64));
      }

      if (TNT_TCP_IO (io)->write_info.flags & TNT_MESSAGE_FLAG_HAS_PAYLOAD)
      {
        gpointer payload;
        guint    size;
        guint16  size_16;

        tnt_message_get_payload (TNT_TCP_IO (io)->write_info.message, &payload, &size);
        size_16 = size;
        size_16 = GUINT16_TO_LE (size_16);
        write_out (TNT_TCP_IO (io), &size_16, TNT_MESSAGE_PAYLOAD_SIZE_SIZE);
      }

      if (TNT_TCP_IO (io)->write_info.flags & TNT_MESSAGE_FLAG_HAS_DATA_BLOCK)
      {
        gpointer block;
        guint32  size;

        tnt_message_get_data_block (TNT_TCP_IO (io)->write_info.message, &block, &size);
        size = GUINT32_TO_LE (size);
        write_out (TNT_TCP_IO (io), &size, TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE);
      }
      break;

    case TNT_TCP_IO_WRITE_PAYLOAD:
    case TNT_TCP_IO_WRITE_DATA_BLOCK:
      /* The block is fetched by begin_write_state () and placed in the block/size vars,
       * thus we don't have to differentiate between payloads and data blocks here. */
      write_out (TNT_TCP_IO (io),
                 (guchar *) TNT_TCP_IO (io)->write_info.block +
                 TNT_TCP_IO (io)->write_info.block_pos, min_len);
      TNT_TCP_IO (io)->write_info.block_pos += min_len;

      if (TNT_TCP_IO (io)->write_info.block_pos + TNT_DATA_STEP_SIZE >
          TNT_TCP_IO (io)->write_info.block_size)
      {
        flux_io_set_write_threshold (io, TNT_TCP_IO (io)->write_info.block_size -
                                     TNT_TCP_IO (io)->write_info.block_pos);
      }
      break;

    default:
      g_assert_not_reached ();
  }

  if (state == TNT_TCP_IO_WRITE_DATA_BLOCK || state == TNT_TCP_IO_WRITE_PAYLOAD)
  {
    if (TNT_TCP_IO (io)->write_info.block_pos == TNT_TCP_IO (io)->write_info.block_size)
      end_write_state (TNT_TCP_IO (io));
  }
  else
  {
    end_write_state (TNT_TCP_IO (io));
  }

  flux_object_unref (FLUX_OBJECT (io));
}

void
tnt_tcp_io_init_self (TntTcpIO *tcp_io)
{
  flux_object_set_type          (FLUX_OBJECT (tcp_io), TNT_TCP_IO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (tcp_io), tnt_tcp_io_finalize);

  tcp_io->queue_out_len = 0;
  tcp_io->queue_out = NULL;
  tcp_io->queue_out_last = NULL;
  tcp_io->message_in_func = NULL;
  tcp_io->data_in_begin_func = NULL;
  tcp_io->queue_class_table = g_hash_table_new_full ((GHashFunc)      g_direct_hash,
                                                     (GCompareFunc)   g_direct_equal,
                                                     (GDestroyNotify) NULL,
                                                     (GDestroyNotify) g_free);

  set_queue_class_max_length (tcp_io, TNT_MESSAGE_UNKNOWN, DEFAULT_QUEUE_LEN);

  flux_io_set_read_cps_goal  (FLUX_IO (tcp_io), DEFAULT_BANDWIDTH_CAP_IN);
  flux_io_set_write_cps_goal (FLUX_IO (tcp_io), DEFAULT_BANDWIDTH_CAP_OUT);

  flux_io_set_read_func (FLUX_IO (tcp_io), read_cb, NULL);
  flux_io_set_write_func (FLUX_IO (tcp_io), write_cb, NULL);
  begin_read_state (tcp_io, TNT_TCP_IO_READ_FLAGS);
  begin_write_state (tcp_io, TNT_TCP_IO_WRITE_IDLE);
}

void
tnt_tcp_io_init (TntTcpIO *tcp_io, FluxCipher *cipher)
{
  flux_tcp_io_init     (FLUX_TCP_IO (tcp_io), -1, cipher, 4, 4);
  tnt_tcp_io_init_self (TNT_TCP_IO (tcp_io));
}

TntTcpIO *
tnt_tcp_io_new (void)
{
  TntTcpIO *tcp_io;

  tcp_io = g_new0 (TntTcpIO, 1);
  tnt_tcp_io_init (tcp_io, NULL);

  return tcp_io;
}

TntTcpIO *
tnt_tcp_io_new_with_cipher (FluxCipher *cipher)
{
  TntTcpIO *tcp_io;

  g_return_val_if_fail (cipher != NULL, NULL);

  tcp_io = g_new0 (TntTcpIO, 1);
  tnt_tcp_io_init (tcp_io, cipher);

  return tcp_io;
}

DEBUG_CALLBACKS (

static void
_dead_cb (TntTcpIO *tcp_io, gpointer dummy, gchar *func_name)
{
  g_error ("Got callback on presumed dead socket (%s)!", func_name);
}

void
_tnt_tcp_io_be_dead_real (TntTcpIO *tcp_io, const gchar *func_name)
{
  gchar *func_name_dup = g_strdup (func_name);

  flux_io_set_read_func  (FLUX_IO (tcp_io), _dead_cb, func_name_dup);
  flux_io_set_write_func (FLUX_IO (tcp_io), _dead_cb, func_name_dup);
  flux_tcp_io_set_link_func (FLUX_TCP_IO (tcp_io), _dead_cb, func_name_dup);
}

)

void
tnt_tcp_io_finalize (FluxObject *tcp_io)
{
  DEBUG_CALLBACKS (_tnt_tcp_io_be_dead (TNT_TCP_IO (tcp_io)));

  drop_all (TNT_TCP_IO (tcp_io));
  g_hash_table_destroy (TNT_TCP_IO (tcp_io)->queue_class_table);
  free_read_state_info (TNT_TCP_IO (tcp_io));
  clear_write_state_info (TNT_TCP_IO (tcp_io));

  DEBUG (g_print ("TntTcpIO finalized.\n"));

  flux_tcp_io_finalize (tcp_io);
}

TntNodeIP4
tnt_tcp_io_get_remote_ip4 (TntTcpIO *tcp_io)
{
  TntNodeIP4  node_ip4;
  FluxIPAddr *ip_addr;

  ip_addr = flux_tcp_io_get_remote_ip_addr (FLUX_TCP_IO (tcp_io));
  if (!ip_addr)
    return TNT_NODE_IP_INVALID;

  node_ip4 = tnt_ip_addr_to_ip (ip_addr);
  flux_object_unref (FLUX_OBJECT (ip_addr));

  return node_ip4;
}

void
tnt_tcp_io_get_remote_address4 (TntTcpIO *tcp_io, TntNodeAddress4 *address)
{
  TntNodeIP4  node_ip4;
  FluxIPAddr *ip_addr;

  ip_addr = flux_tcp_io_get_remote_ip_addr (FLUX_TCP_IO (tcp_io));
  if (!ip_addr)
    address->ip = 0;

  address->ip = tnt_ip_addr_to_ip (ip_addr);
  flux_object_unref (FLUX_OBJECT (ip_addr));

  address->port = flux_tcp_io_get_remote_port (FLUX_TCP_IO (tcp_io));
}

void
tnt_tcp_io_set_message_in_func (TntTcpIO *tcp_io, TntMessageInFunc *func, gpointer data)
{
  assert_types (tcp_io);

  tcp_io->message_in_func = func;
  tcp_io->message_in_data = data;
}

void
tnt_tcp_io_set_data_in_begin_func (TntTcpIO *tcp_io, TntMessageInFunc *func, gpointer data)
{
  assert_types (tcp_io);

  tcp_io->data_in_begin_func = func;
  tcp_io->data_in_begin_data = data;
}

static void
queue_message (TntTcpIO *tcp_io, TntMessage *message)
{
  assert_types (tcp_io);
  g_return_if_fail (message != NULL);

  if (!tcp_io->queue_out)
  {
    tcp_io->queue_out = tcp_io->queue_out_last = g_list_append (tcp_io->queue_out_last, message);
  }
  else
  {
    g_list_append (tcp_io->queue_out_last, message);
    tcp_io->queue_out_last = g_list_next (tcp_io->queue_out_last);
  }

  increment_queue_class_length (tcp_io, tnt_message_get_type (message));
  tcp_io->queue_out_len++;
}

static void
drop_all (TntTcpIO *tcp_io)
{
  GList *l;

  assert_types (tcp_io);

  for (l = tcp_io->queue_out; l; l = l->next)
  {
    flux_object_unref (FLUX_OBJECT (l->data));
  }

  g_list_free (tcp_io->queue_out);
  tcp_io->queue_out      = NULL;
  tcp_io->queue_out_last = NULL;
  tcp_io->queue_out_len  = 0;
}

/* Remember to unref the message after queuing it, if you have no more
 * use for it. */
void
tnt_tcp_io_message_queue_out (TntTcpIO *tcp_io, TntMessage *message)
{
  assert_types (tcp_io);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  if (queue_class_is_full (tcp_io, tnt_message_get_type (message)))
    return;

  flux_object_ref (FLUX_OBJECT (message));
  queue_message (tcp_io, message);

  if (tcp_io->write_info.state == TNT_TCP_IO_WRITE_IDLE)
  {
    end_write_state (tcp_io);
  }
}

/* Use only on unconnected TcpIO */
void
tnt_tcp_io_reset (TntTcpIO *tcp_io)
{
  assert_types (tcp_io);

  drop_all (tcp_io);

  free_read_state_info (tcp_io);
  clear_write_state_info (tcp_io);
  begin_read_state (tcp_io, TNT_TCP_IO_READ_FLAGS);
  begin_write_state (tcp_io, TNT_TCP_IO_WRITE_IDLE);
}

gint
tnt_tcp_io_get_queue_out_max_len (TntTcpIO *tcp_io, TntMessageType message_type)
{
  assert_types (tcp_io);

  return get_queue_class_max_length (tcp_io, message_type);
}

void
tnt_tcp_io_set_queue_out_max_len (TntTcpIO *tcp_io, TntMessageType message_type, gint max_len)
{
  assert_types (tcp_io);

  if (max_len == -1)
    remove_queue_class (tcp_io, message_type);
  else
    set_queue_class_max_length (tcp_io, message_type, max_len);

  /* FIXME: Trim the queue? */
}

gint
tnt_tcp_io_get_queue_out_len (TntTcpIO *tcp_io, TntMessageType type)
{
  assert_types (tcp_io);

  if (type == TNT_MESSAGE_ANY)
    return tcp_io->queue_out_len;
  else
    return get_queue_class_length (tcp_io, type);
}

guint
tnt_tcp_io_get_cps (TntTcpIO *tcp_io, TntTransferDirection dir)
{
  assert_types (tcp_io);

  switch (dir)
  {
    case TNT_TRANSFER_DIRECTION_OUT:
      return flux_io_get_write_cps (FLUX_IO (tcp_io));

    case TNT_TRANSFER_DIRECTION_IN:
      return flux_io_get_read_cps (FLUX_IO (tcp_io));

    default:
      g_assert_not_reached ();
      return 0;
  }
}

guint
tnt_tcp_io_get_cps_goal (TntTcpIO *tcp_io, TntTransferDirection dir)
{
  assert_types (tcp_io);

  switch (dir)
  {
    case TNT_TRANSFER_DIRECTION_OUT:
      return flux_io_get_write_cps_goal (FLUX_IO (tcp_io));

    case TNT_TRANSFER_DIRECTION_IN:
      return flux_io_get_read_cps_goal (FLUX_IO (tcp_io));

    default:
      g_assert_not_reached ();
      return 0;
  }
}

void
tnt_tcp_io_set_cps_goal (TntTcpIO *tcp_io, TntTransferDirection dir, guint cps)
{
  assert_types (tcp_io);

  switch (dir)
  {
    case TNT_TRANSFER_DIRECTION_OUT:
      flux_io_set_write_cps_goal (FLUX_IO (tcp_io), cps);
      break;

    case TNT_TRANSFER_DIRECTION_IN:
      flux_io_set_read_cps_goal (FLUX_IO (tcp_io), cps);
      break;

    default:
      break;
  }
}

gboolean
tnt_tcp_io_get_read_info (TntTcpIO *tcp_io, TntMessageType *type, gpointer *payload, guint16 *payload_size)
{
  assert_types (tcp_io);

  if (tcp_io->read_info.state == TNT_TCP_IO_READ_FLAGS ||
      tcp_io->read_info.state == TNT_TCP_IO_READ_MESSAGE_HEADER)
    return FALSE;

  if (type)
    *type = tcp_io->read_info.type;

  if (tcp_io->read_info.state == TNT_TCP_IO_READ_DATA_BLOCK)
  {
    if (payload)
      *payload      = tcp_io->read_info.payload;
    if (payload_size)
      *payload_size = tcp_io->read_info.payload_size;
  }
  else
  {
    if (payload)
      *payload      = NULL;
    if (payload_size)
      *payload_size = 0;
  }

  return TRUE;
}

TntMessage *
tnt_tcp_io_get_write_message (TntTcpIO *tcp_io)
{
  assert_types (tcp_io);

  if (tcp_io->write_info.state == TNT_TCP_IO_WRITE_IDLE)
    return NULL;

  return tcp_io->write_info.message;
}
