/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-message.c - Network message.
 *
 * Copyright (C) 2002-2004 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <time.h>

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_TIME_WARP_TYPE)

#define DEFAULT_DELAY_MIN 2.0
#define DEFAULT_DELAY_MAX 6.0

typedef struct
{
  TntMessage *message;
  TntEdge    *source_edge;  /* No formal reference is kept */
  guint64     release_time_ms;
}
DelayedMessage;

static gboolean release_timeout_cb (TntTimeWarp *time_warp);

static gint
compare_release_time (const guint64 *a, const guint64 *b)
{
  if (*a < *b)
    return -1;
  if (*a > *b)
    return 1;

  return 0;
}

static gint
search_min_release_time (const guint64 *a, guint64 *min_release_time_ms)
{
  *min_release_time_ms = *a;
  return -1;
}

static void
free_struct (DelayedMessage *dm)
{
  flux_object_unref (FLUX_OBJECT (dm->message));
  g_free (dm);
}

static DelayedMessage *
find_first_struct (TntTimeWarp *time_warp)
{
  guint64 found_release_time_ms = 0;

  g_tree_search (time_warp->message_tree, (GCompareFunc) search_min_release_time,
                 &found_release_time_ms);

  if (found_release_time_ms == 0)
    return NULL;

  return g_tree_lookup (time_warp->message_tree, &found_release_time_ms);
}

static void
update_timeout (TntTimeWarp *time_warp)
{
  DelayedMessage *dm;

  dm = find_first_struct (time_warp);

  if (dm)
  {
    if (dm->release_time_ms != time_warp->timeout_ms)
    {
      GTimeVal time_val;
      gint     delta_ms;

      if (time_warp->timeout_id)
        g_source_remove (time_warp->timeout_id);

      g_get_current_time (&time_val);
      delta_ms = (gint64) dm->release_time_ms -
        ((gint64) time_val.tv_sec * (gint64) 1000) -
        (gint64) ((gint64) time_val.tv_usec / (gint64) 1000);

      if (delta_ms <= 0)
        time_warp->timeout_id = g_idle_add ((GSourceFunc) release_timeout_cb, time_warp);
      else
        time_warp->timeout_id = g_timeout_add (delta_ms, (GSourceFunc) release_timeout_cb, time_warp);

      time_warp->timeout_ms = dm->release_time_ms;
    }
  }
  else
  {
    if (time_warp->timeout_id)
    {
      g_source_remove (time_warp->timeout_id);
      time_warp->timeout_id = 0;
    }
  }
}

static gboolean
release_timeout_cb (TntTimeWarp *time_warp)
{
  DelayedMessage *dm;

  dm = find_first_struct (time_warp);
  if (!dm)
  {
    g_warning ("Message disappeared in time warp!");
    return FALSE;
  }

  tnt_message_set_time_warped (dm->message, TRUE);

  if (time_warp->message_handler)
  {
    time_warp->message_handler (time_warp->network_core, dm->source_edge, dm->message,
                                time_warp->message_handler_data);
  }

  g_tree_remove (time_warp->message_tree, &dm->release_time_ms);
  update_timeout (time_warp);
  return FALSE;
}

void
tnt_time_warp_finalize (FluxObject *object)
{
  TntTimeWarp *time_warp = TNT_TIME_WARP (object);

  assert_types (time_warp);

  g_tree_destroy (time_warp->message_tree);
  if (time_warp->timeout_id)
    g_source_remove (time_warp->timeout_id);
}

void
tnt_time_warp_init (TntTimeWarp *time_warp, TntNetworkCore *network_core)
{
  flux_object_init              (FLUX_OBJECT (time_warp));
  flux_object_set_type          (FLUX_OBJECT (time_warp), TNT_TIME_WARP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (time_warp), tnt_time_warp_finalize);

  time_warp->network_core = network_core;  /* No formal reference is kept */
  time_warp->message_tree = g_tree_new_full ((GCompareDataFunc) compare_release_time, NULL,
                                             NULL, (GDestroyNotify) free_struct);

  time_warp->max_delay = DEFAULT_DELAY_MAX;
  time_warp->min_delay = DEFAULT_DELAY_MIN;
}

TntTimeWarp *
tnt_time_warp_new (TntNetworkCore *network_core)
{
  TntTimeWarp *time_warp;

  time_warp = g_new0 (TntTimeWarp, 1);
  tnt_time_warp_init (time_warp, network_core);

  return time_warp;
}

void
tnt_time_warp_push_message (TntTimeWarp *time_warp, TntMessage *message, TntEdge *source_edge)
{
  DelayedMessage *dm;
  GTimeVal        time_val;

  assert_types (time_warp);

  dm = g_new0 (DelayedMessage, 1);
  dm->message = message;
  flux_object_ref (FLUX_OBJECT (message));
  dm->source_edge = source_edge;  /* No formal reference is kept */

  g_get_current_time (&time_val);

  do
  {
    dm->release_time_ms =
      ((guint64) time_val.tv_sec * (guint64) 1000) +
      ((guint64) time_val.tv_usec / (guint64) 1000) +
      (guint64) g_random_int_range (time_warp->min_delay * 1000, time_warp->max_delay * 1000);
  }
  while (g_tree_lookup (time_warp->message_tree, &dm->release_time_ms));

  g_tree_insert (time_warp->message_tree, &dm->release_time_ms, dm);
  update_timeout (time_warp);
}

void
tnt_time_warp_get_delay (TntTimeWarp *time_warp,
                         gdouble *min_delay, gdouble *max_delay)
{
  assert_types (time_warp);

  if (max_delay)
    *max_delay = time_warp->max_delay;
  if (min_delay)
    *min_delay = time_warp->min_delay;
}

void
tnt_time_warp_set_delay (TntTimeWarp *time_warp,
                         gdouble min_delay, gdouble max_delay)
{
  assert_types (time_warp);

  time_warp->max_delay = max_delay;
  time_warp->min_delay = min_delay;
}

void
tnt_time_warp_set_message_handler (TntTimeWarp *time_warp, TntMessageHandler *message_handler,
                                   gpointer data)
{
  assert_types (time_warp);

  time_warp->message_handler      = message_handler;
  time_warp->message_handler_data = data;

  update_timeout (time_warp);
}
