/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-util.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* Large file support */
#undef  _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#undef  __REDIRECT
#define __REDIRECT 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <dirent.h>
#include <ctype.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
# include <io.h>
# include <direct.h>
#endif

#include "tnt.h"

/* Max length of keyword string extracted from path */
#define EXTRACTED_KEYWORDS_LEN_MAX 128

static void
trim_path_in_place (gchar *path)
{
  gint len;
  gchar *s0;

  g_return_if_fail (path != NULL);

  g_strstrip (path);

  len = strlen (path);
  if (!len)
    return;

  /* Remove a trailing dir separator */
  for (s0 = path + len - 1; s0 > path; s0--)
  {
    if (*s0 == G_DIR_SEPARATOR)
      *s0 = '\0';
    else
      break;
  }
}

gchar *
tnt_nice_path (const gchar *path)
{
  gchar *new_path;
  gint   path_len;

  if (path == NULL)
    return NULL;

  /* Make new_path a copy of path, with two trailing null bytes */
  path_len = strlen (path);
  new_path = g_malloc (path_len + 2);
  strcpy (new_path, path);
  new_path [path_len + 1] = '\0';

  /* Remove leading/trailing spaces and trailing separator */
  trim_path_in_place (new_path);

#ifdef G_PLATFORM_WIN32

  /* Check for 'X:', 'X:\'. Windows paths are funny, in that they can't have a
   * trailing slash, except for the root drive letter, which needs one. */

  if (new_path [0] && new_path [1] == ':')
  {
    /* So far, so good - it has a drive letter */

    if (new_path [2] == '\0')
    {
      /* Path is 'X:'. Make it be 'X:\'. */
      new_path [2] = '\\';
    }
    else if (new_path [2] != '\\')
    {
      /* Failure: Path is 'X:something' */
      g_free (new_path);
      return NULL;
    }
  }
  else
  {
    /* Failure: Does not start in drive letter */
    g_free (new_path);
    return NULL;
  }

#else

  if (!g_path_is_absolute (new_path))
  {
    /* Failure: Unix path is not absolute */
    g_free (new_path);
    return NULL;
  }

#endif

  return new_path;
}

gboolean
tnt_path_exists (const gchar *path)
{
  struct stat sbuf;
  gint ret;

  g_return_val_if_fail (path != NULL, FALSE);

  ret = stat (path, &sbuf);
  if (ret < 0)
    return FALSE;

  return TRUE;
}

gboolean
tnt_path_is_dir (const gchar *path)
{
  struct stat sbuf;
  gint ret;

  g_return_val_if_fail (path != NULL, FALSE);

  ret = stat (path, &sbuf);
  if (ret < 0)
    return FALSE;

#ifndef G_PLATFORM_WIN32
  return S_ISDIR (sbuf.st_mode);
#else
  return sbuf.st_mode & _S_IFDIR ? TRUE : FALSE;
#endif
}

gboolean
tnt_path_is_file (const gchar *file)
{
  struct stat sbuf;
  gint ret;

  g_return_val_if_fail (file != NULL, FALSE);

  ret = stat (file, &sbuf);
  if (ret < 0)
    return FALSE;

#ifndef G_PLATFORM_WIN32
  return S_ISREG (sbuf.st_mode);
#else
  return sbuf.st_mode & _S_IFREG ? TRUE : FALSE;
#endif
}

#ifndef G_PLATFORM_WIN32

static gboolean
have_access_type_stat (struct stat *sbuf, mode_t mode)
{
  uid_t uid;
  gid_t gid;

  g_return_val_if_fail (sbuf != NULL, FALSE);

  uid = getuid ();
  gid = getgid ();

  if (uid == 0)
    return TRUE;

  if (sbuf->st_uid == uid)
    return TRUE;

  if (sbuf->st_gid == gid)
  {
    if ((sbuf->st_mode & S_IRWXG) >= (mode & S_IRWXG))
      return TRUE;
  }

  if ((sbuf->st_mode & S_IRWXO) >= (mode & S_IRWXO))
    return TRUE;

  return FALSE;
}

#else

static gboolean
have_access_type_stat (struct stat *sbuf, short mode)
{
  g_return_val_if_fail (sbuf != NULL, FALSE);

  if ((sbuf->st_mode & S_IRWXU) >= (mode & S_IRWXU))
    return TRUE;

  return FALSE;
}

#endif

static gboolean
have_access_type (const gchar *path, mode_t mode)
{
  gint access_flags = 0;

  g_return_val_if_fail (path != NULL, FALSE);

#ifndef G_PLATFORM_WIN32

  if (mode & (S_IRUSR | S_IRGRP | S_IROTH))
    access_flags |= R_OK;
  if (mode & (S_IWUSR | S_IWGRP | S_IWOTH))
    access_flags |= W_OK;
  if (mode & (S_IXUSR | S_IXGRP | S_IXOTH))
    access_flags |= X_OK;
#else

  if (mode & S_IRUSR)
    access_flags |= R_OK;
  if (mode & S_IWUSR)
    access_flags |= W_OK;
  if (mode & S_IXUSR)
    access_flags |= X_OK;

#endif

  return access (path, access_flags) == 0 ? TRUE : FALSE;
}

gboolean
tnt_dir_have_full_access (const gchar *dir)
{
#ifndef G_PLATFORM_WIN32
  /* Full access == read/write/execute for our user */
  return have_access_type (dir, S_IRWXU | S_IRWXG | S_IRWXO);
#else
  return TRUE;  /* Don't know how to handle dirs; mkdir() on win32 doesn't take perms */
#endif
}

gboolean
tnt_dir_have_full_access_stat (struct stat *sbuf)
{
#ifndef G_PLATFORM_WIN32
  /* Full access == read/write/execute for our user */
  return have_access_type_stat (sbuf, S_IRWXU | S_IRWXG | S_IRWXO);
#else
  return TRUE;  /* Don't know how to handle dirs; mkdir() on win32 doesn't take perms */
#endif
}

gboolean
tnt_dir_have_read_access (const gchar *dir)
{
#ifndef G_PLATFORM_WIN32
  /* Read access == read/execute for our user */
  return have_access_type (dir, S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#else
  return TRUE;  /* Don't know how to handle dirs; mkdir() on win32 doesn't take perms */
#endif
}

gboolean
tnt_dir_have_read_access_stat (struct stat *sbuf)
{
#ifndef G_PLATFORM_WIN32
  /* Read access == read/execute for our user */
  return have_access_type_stat (sbuf, S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#else
  return TRUE;  /* Don't know how to handle dirs; mkdir() on win32 doesn't take perms */
#endif
}

gboolean
tnt_file_have_full_access (const gchar *file)
{
#ifndef G_PLATFORM_WIN32
  /* Full access == read/write for our user */
  return have_access_type (file, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
#else
  return have_access_type (file, S_IRUSR | S_IWUSR);
#endif
}

gboolean
tnt_file_have_full_access_stat (struct stat *sbuf)
{
#ifndef G_PLATFORM_WIN32
  /* Full access == read/write for our user */
  return have_access_type_stat (sbuf, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
#else
  return have_access_type_stat (sbuf, S_IRUSR | S_IWUSR);
#endif
}

gboolean
tnt_file_have_read_access (const gchar *file)
{
#ifndef G_PLATFORM_WIN32
  /* Read access == read for our user */
  return have_access_type (file, S_IRUSR | S_IRGRP | S_IROTH);
#else
  return have_access_type (file, S_IRUSR);
#endif
}

gboolean
tnt_file_have_read_access_stat (struct stat *sbuf)
{
#ifndef G_PLATFORM_WIN32
  /* Read access == read for our user */
  return have_access_type_stat (sbuf, S_IRUSR | S_IRGRP | S_IROTH);
#else
  return have_access_type_stat (sbuf, S_IRUSR);
#endif
}

gboolean
tnt_dir_delete (const gchar *dir)
{
  gboolean       result = TRUE;
  DIR           *dir_hd;
  struct dirent *dir_entry;

  g_return_val_if_fail (dir != NULL, FALSE);

  dir_hd = opendir (dir);
  if (!dir_hd)
    return FALSE;

  while ((dir_entry = readdir (dir_hd)))
  {
    gchar *path;

    if (!strcmp (dir_entry->d_name, ".") ||
        !strcmp (dir_entry->d_name, ".."))
      continue;

    path = g_strjoin (G_DIR_SEPARATOR_S, dir, dir_entry->d_name, NULL);

    if (remove (path) != 0)
      result = FALSE;

    g_free (path);
  }

  closedir (dir_hd);
  rmdir (dir);
  return result;
}

gboolean
tnt_dir_create (const gchar *dir)
{
  char *copy, *p;

  copy = tnt_nice_path (dir);
  if (!copy)
    return FALSE;

  p = (gchar *) g_path_skip_root (copy);

  if (!p)
  {
    g_free (copy);
    return FALSE;
  }

  do
  {
    p = strchr (p, G_DIR_SEPARATOR);

    if (p)
      *p = '\0';

    if (access (copy, F_OK) == -1)
    {
      if (tnt_mkdir (copy) == -1)
      {
        g_free (copy);
        return FALSE;
      }
    }

    if (p)
    {
      *p = G_DIR_SEPARATOR;
      p++;
    }
  }
  while (p);

  g_free (copy);
  return TRUE;
}

gchar *
tnt_amount_print_dot_padded (guint64 amount)
{
  gchar *str, *s0_start, *s0, *s1;
  gint   dot_count, dot_dist;

  s0_start = s0 = g_strdup_printf ("%llu", amount);
  dot_count = (strlen (s0) - 1) / 3;
  dot_dist  = strlen (s0) % 3;

  if (!dot_dist)
    dot_dist = 3;

  str = s1 = g_malloc (strlen (s0) + dot_count + 1);

  for ( ; *s0; )
  {
    *s1++ = *s0++;
    dot_dist--;
    if (!dot_dist && dot_count)
    {
      *s1++ = ',';
      dot_count--;
      dot_dist = 3;
    }
  }

  *s1 = '\0';
  g_free (s0_start);
  return str;
}

#define PLURAL_S(n)       ((n) > 1 ? "s" : "")

gchar *
tnt_time_describe (guint seconds_abs)
{
  time_t     seconds_abs_tt;
  struct tm *tm;

  seconds_abs_tt = seconds_abs;
  tm = localtime (&seconds_abs_tt);

  return g_strdup_printf ("%02d:%02d", tm->tm_hour, tm->tm_min);
}

typedef struct
{
  gint   upper_limit;
  gint   divisor;
  gchar *unit;
}
TimeGranularityRule;

/* Note that seconds can be negative, which results in a reference
 * to the future. */
gchar *
tnt_time_describe_delta (gint seconds_ago)
{
  gchar *format;
  gchar *unit  = NULL;
  gint   value = 0;
  gint   seconds_abs;
  gint   i;

  TimeGranularityRule rules [] =
  {
    { SECONDS_IN_MINUTE,     1,                 "second" },
    { SECONDS_IN_HOUR,       SECONDS_IN_MINUTE, "minute" },
    { SECONDS_IN_DAY,        SECONDS_IN_HOUR,   "hour"   },
    { SECONDS_IN_WEEK  * 2,  SECONDS_IN_DAY,    "day"    },
    { SECONDS_IN_MONTH * 2,  SECONDS_IN_WEEK,   "week"   },
    { SECONDS_IN_YEAR  * 2,  SECONDS_IN_MONTH,  "month"  },
    { G_MAXINT,              SECONDS_IN_YEAR,   "year"   },
    { 0,                     0,                 NULL     }
  };

  seconds_abs = abs (seconds_ago);

  /* Very special cases */

  if (seconds_abs < 61)
  {
    if (seconds_ago > 0)
      return g_strdup ("recently");
    else
      return g_strdup ("imminent");
  }
  else if (seconds_abs > SECONDS_IN_HOUR * 16 &&
           seconds_abs < SECONDS_IN_HOUR * 40)
  {
    if (seconds_ago > 0)
      return g_strdup ("yesterday");
    else
      return g_strdup ("tomorrow");
  }

  /* General case */

  for (i = 0; rules [i].unit; i++)
  {
    if (seconds_abs < rules [i].upper_limit)
    {
      value = seconds_abs / rules [i].divisor;
      unit  = rules [i].unit;
      break;
    }
  }

  if (!unit)
  {
    return g_strdup ("unknown");
  }

  if (seconds_ago > 0)
    format = "%d %s%s ago";
  else
    format = "in %d %s%s";

  return g_strdup_printf (format, value, unit, PLURAL_S (value));
}

typedef struct
{
  guint64  upper_limit;
  guint64  divisor;
  gchar   *unit;
}
SizeGranularityRule;

gchar *
tnt_size_describe (guint64 size)
{
  gchar *unit  = NULL;
  gint   value = 0;
  gchar *value_str;
  gchar *ret;
  gint   i;

  SizeGranularityRule rules [] =
  {
    { BYTES_IN_K,      1,          "B"  },
    { BYTES_IN_M,      BYTES_IN_K, "kB" },
    { BYTES_IN_G * 10, BYTES_IN_M, "MB" },
    { BYTES_IN_T,      BYTES_IN_G, "GB" },
    { G_MAXUINT64,     BYTES_IN_T, "TB" },
    { 0,               0,          NULL }
  };

  /* General case */

  for (i = 0; rules [i].unit; i++)
  {
    if (size < rules [i].upper_limit)
    {
      value = (size + rules [i].divisor / 2) / rules [i].divisor;
      unit  = rules [i].unit;
      break;
    }
  }

  if (unit)
  {
    value_str = tnt_amount_print_dot_padded (value);
    ret = g_strdup_printf ("%s%s", value_str, unit);
    g_free (value_str);
  }
  else
    ret = g_strdup ("???");

  return ret;
}

gchar *
tnt_bandwidth_describe (guint cps)
{
  gchar *string, *description;

  string = tnt_size_describe (cps);
  description = g_strconcat (string, "/s", NULL);
  g_free (string);

  return description;
}

gint
tnt_strlen_limited (const gchar *string, gint limit)
{
  gint i;

  g_return_val_if_fail (string != NULL, 0);

  for (i = 0; i < limit; i++)
  {
    if (!string [i])
      break;
  }

  return i;
}

void
tnt_string_capitalize_first (gchar *string)
{
  g_return_if_fail (string != NULL);

  if (*string)
  {
    *string = toupper (*string);
  }
}

void
tnt_string_to_lower (gchar *string)
{
  g_return_if_fail (string != NULL);

  for ( ; *string; string++)
  {
    *string = tolower (*string);
  }
}

gchar *
tnt_utf8_strip_accents (const gchar *accented)
{
  gchar *stripped, *p;
  gint   len;

  g_return_val_if_fail (accented != NULL, NULL);

  /* Make sure we have enough space. Allocate for worst-case expansion, even
   * if the string shouldn't expand as much as contract from this operation. */
  len = strlen (accented);
  p = stripped = g_malloc (len * 6 + 1);

  for ( ; accented && *accented; accented = g_utf8_next_char (accented))
  {
    gunichar  c = g_utf8_get_char (accented);
    gunichar *decomposed;
    gsize     decomposed_len;
    gint      t;

    decomposed = g_unicode_canonical_decomposition (c, &decomposed_len);
    if (!decomposed)
      continue;

    /* Assumption: The first character in decomposed is the base character */
    t = g_unichar_to_utf8 (*decomposed, p);
    g_free (decomposed);

    if (t < 0)
      break;

    p += t;
  }

  *p = '\0';
  stripped = g_realloc (stripped, p - stripped + 1);
  return stripped;
}

void
tnt_utf8_strlcpy (gchar *dest, const gchar *src, gsize dest_size)
{
  const gchar *p, *last_p;
  gint         len;

  g_return_if_fail (src != NULL);
  g_return_if_fail (dest != NULL);
  g_return_if_fail (dest_size > 0);

  for (p = last_p = src; p; p = g_utf8_next_char (p))
  {
    if (p - src >= dest_size)
      break;

    last_p = p;

    if (!*p)
      break;
  }

  len = last_p - src;
  g_assert (len < dest_size);

  if (len > 0)
    memcpy (dest, src, len);

  *(dest + len) = '\0';
}

/* Doesn't do checking; string + len must be inside allocated block */
static GList *
add_string (GList *list, const gchar *string, gint len, GHashTable *string_table)
{
  gchar *word;

  word = g_malloc (len + 1);
  memcpy (word, string, len);
  word [len] = '\0';

  if (!g_hash_table_lookup (string_table, word))
  {
    list = g_list_prepend (list, word);
    g_hash_table_insert (string_table, word, list);
  }
  else
  {
    g_free (word);
  }

  return list;
}

/* get_min_string_end should be TRUE when indexing, FALSE when searching. I will explain
 * in detail some other time. */
GList *
tnt_string_extract_substrings (const gchar *string, gint min_token_len, gint max_token_len,
                               gboolean get_min_string_end)
{
  GList      *list = NULL;
  gint        len, i;
  guchar     *break_chars = " \n\t()[].-_:;+/&!@#=?<>\"\\";
  GHashTable *string_table;

  g_return_val_if_fail (string != NULL, NULL);
  g_return_val_if_fail (min_token_len <= max_token_len, NULL);

  string_table = g_hash_table_new (g_str_hash, g_str_equal);

  len = strlen (string);
  for (i = 0; i < len ; )
  {
    guint seg_len = strcspn (string + i, break_chars);

    if (seg_len < min_token_len)
    {
      i += seg_len;
    }
    else if (seg_len <= max_token_len)
    {
      list = add_string (list, string + i, seg_len, string_table);
      if (get_min_string_end)
        i += 1;
      else
        i += seg_len;
    }
    else /* seg_len > max_token_len */
    {
      list = add_string (list, string + i, max_token_len, string_table);
      i += 1;  /* Get all token-len substrings */
    }

    i += strspn (string + i, break_chars);
  }

  g_hash_table_destroy (string_table);
  return list;
}

/* Generate list of strings that are in list_2 but not in list_1. This is not an
 * exclusive or, since it only goes one way. */
GList *
tnt_string_list_get_additions (GList *list_1, GList *list_2)
{
  GHashTable *string_table;
  GList      *result_list = NULL;
  GList      *l;

  string_table = g_hash_table_new (g_str_hash, g_str_equal);

  for (l = list_1; l; l = g_list_next (l))
  {
    gchar *word = l->data;

    g_hash_table_insert (string_table, word, l);
  }

  for (l = list_2; l; l = g_list_next (l))
  {
    gchar *word = l->data;

    if (g_hash_table_lookup (string_table, word) == NULL)
      result_list = g_list_prepend (result_list, g_strdup (word));
  }

  g_hash_table_destroy (string_table);
  return result_list;
}

typedef struct
{
  gchar c;
  gint  score;
}
CharScore;

/* Letter scores based on frequencies in the US English Brown Corpus.
 * We're assuming that English will be the most common language. Higher
 * scores indicate lower frequency. Brown corpus source, most frequent
 * first: [etaoinsrhldcumfpgwybvkxjqz].
 *
 * If we collected stats for a two-letter combination table, that would
 * be a lot more accurate. It would be 28^2 = 784 entries long. */
static CharScore char_complexity_scores [26] =
{
  { 'a',  7 }, { 'b', 24 }, { 'c', 16 }, { 'd', 15 }, { 'e',  5 }, { 'f', 19 },
  { 'g', 21 }, { 'h', 13 }, { 'i',  9 }, { 'j', 28 }, { 'k', 26 }, { 'l', 14 },
  { 'm', 18 }, { 'n', 10 }, { 'o',  8 }, { 'p', 20 }, { 'q', 29 }, { 'r', 12 },
  { 's', 11 }, { 't',  6 }, { 'u', 17 }, { 'v', 25 }, { 'w', 22 }, { 'x', 27 },
  { 'y', 23 }, { 'z', 30 }
};

static gint
calc_string_complexity (guchar *string)
{
  gint score = 0;
  gint i;

  for (i = 0; string [i]; i++)
  {
    score += 100;

    if (string [i] - 'a' >= 0 && string [i] - 'a' < 26)
      score += char_complexity_scores [string [i] - 'a'].score;
    else if (string [i] - '0' >= 0 && string [i] - '0' < 10)
      score += 31;
    else
      score += 10;
  }

  return score;
}

gchar *
tnt_string_list_get_least_common (GList *list)
{
  gint   high_score  = -1;
  gchar *high_string = NULL;
  GList *l;

  for (l = list; l; l = g_list_next (l))
  {
    gint score;

    score = calc_string_complexity (l->data);

    if (score > high_score)
    {
      high_string = l->data;
      high_score  = score;
    }
  }

  return high_string;
}

#define VALID_IP4_CHARS "0123456789."

gboolean
tnt_string_is_valid_ipv4_address (const gchar *address)
{
  const gchar *s;
  gint         len;
  gint         n;
  gint         i;

  g_return_val_if_fail (address != NULL, FALSE);

  len = strlen (address);
  if (!len)
    return FALSE;

  if (strspn (address, VALID_IP4_CHARS) != len)
    return FALSE;

  s = address;

  for (i = 0; i < 4; i++)
  {
    if (*s == '\0')
      return FALSE;

    n = atoi (s);
    if (n < 0 || n > 255)
      return FALSE;

    s = strchr (s, '.');
    if (i < 3 && !s)
      return FALSE;

    s++;
  }

  return TRUE;
}

#define VALID_ADDRESS_CHARS        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-"
#define REQUIRED_ANY_ADDRESS_CHARS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

gboolean
tnt_string_is_valid_address (const gchar *address)
{
  gint len;

  g_return_val_if_fail (address != NULL, FALSE);

  if (tnt_string_is_valid_ipv4_address (address))
    return TRUE;

  len = strlen (address);
  if (!len)
    return FALSE;

  if (strspn (address, VALID_ADDRESS_CHARS) != len)
    return FALSE;

  if (strcspn (address, REQUIRED_ANY_ADDRESS_CHARS) == len)
    return FALSE;

  if (strstr (address, ".."))
    return FALSE;

  if (address [0] == '.' || address [len - 1] == '.')
    return FALSE;

  return TRUE;
}

TntNodeIP4
tnt_string_to_ip (const gchar *address)
{
  const gchar *s;
  gint         n;
  gint         i;
  TntNodeIP4    ip = 0;

  g_return_val_if_fail (address != NULL, TNT_NODE_IP_INVALID);

  if (!strlen (address))
    return 0;

  s = address;

  for (i = 0; i < 4; i++)
  {
    ip <<= 8;

    if (*s == '\0')
      return 0;

    n = atoi (s);
    if (n < 0 || n > 255)
      return 0;

    s = strchr (s, '.');
    if (i < 3 && !s)
      return 0;

    ip |= n;
    s++;
  }

  return ip;
}

gchar *
tnt_ip_to_string (TntNodeIP4 ip)
{
  return g_strdup_printf ("%d.%d.%d.%d", (ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff, ip & 0xff);
}

TntNodeIP4
tnt_ip_addr_to_ip (FluxIPAddr *ip_addr)
{
  TntNodeIP4  ip4;
  gchar      *ip_str;

  g_return_val_if_fail (ip_addr != NULL, TNT_NODE_IP_INVALID);

  ip_str = flux_ip_addr_get_string (ip_addr);
  if (!ip_str)
    return TNT_NODE_IP_INVALID;

  ip4 = tnt_string_to_ip (ip_str);
  g_free (ip_str);

  return ip4;
}

FluxIPAddr *
tnt_ip_to_ip_addr (TntNodeIP4 ip4)
{
  FluxIPAddr *ip_addr;
  gchar      *ip_str;

  ip_str = tnt_ip_to_string (ip4);
  if (!ip_str)
    return NULL;

  ip_addr = flux_ip_addr_new ();
  flux_ip_addr_set_string (ip_addr, ip_str);

  g_free (ip_str);
  return ip_addr;
}

gchar *
tnt_path_make_keyword_string (const gchar *path)
{
  gchar    *keywords;
  gboolean  last_was_space = TRUE;
  gint      len;
  gchar    *p0;
  gint      i;

  g_return_val_if_fail (path != NULL, NULL);
  g_return_val_if_fail (*path != '\0', NULL);

  len = strlen (path);
  if (!len)
    return NULL;

  /* Copy path, substituting spaces for path separators and limiting
   * to one space in a row. */

  keywords = g_malloc (len + 1);

  for (p0 = keywords, i = 0; i < len; i++)
  {
    gchar c = path [i];

    c = path [i];

    switch (c)
    {
      case G_DIR_SEPARATOR:
        c = ' ';
        /* Fall through */

      case ' ':
      case '\t':
      case '\n':
      case '\r':
        if (!last_was_space)
        {
          *(p0++) = ' ';
          last_was_space = TRUE;
        }
        break;

      default:
        *(p0++) = c;
        last_was_space = FALSE;
        break;
    }
  }

  *p0 = '\0';

  /* Limit length of string, if necessary */

  len = strlen (keywords);
  p0  = keywords;

  if (len > EXTRACTED_KEYWORDS_LEN_MAX)
  {
    while (len > EXTRACTED_KEYWORDS_LEN_MAX)
    {
      gchar *p1;

      p1 = strchr (p0, ' ');

      if (p1)
      {
        /* Got next space */
        p1++;  /* Skip the space */
        len -= (p1 - p0);
        p0 = p1;
      }
      else
      {
        /* Got end of string */
        p0 += (len - EXTRACTED_KEYWORDS_LEN_MAX);
        len = EXTRACTED_KEYWORDS_LEN_MAX;
      }
    }

    memmove (keywords, p0, len + 1);
    keywords = g_realloc (keywords, len + 1);
  }

  return keywords;
}

GSList *
tnt_path_split (const gchar *path)
{
  GSList *elem = NULL;
  gint    start, i;

  g_return_val_if_fail (path != NULL, NULL);

  for (start = i = 0; path [i]; i++)
  {
    if (path [i] == G_DIR_SEPARATOR)
    {
      if (start < i)
      {
        gchar *str;

        str = g_strndup (path + start, i - start);
        elem = g_slist_append (elem, str);
      }

      start = i + 1;
    }
  }

  if (start < i)
  {
    gchar *str;

    str = g_strndup (path + start, i - start);
    elem = g_slist_append (elem, str);
  }

  return elem;
}

static gchar *
join_pelems (GSList *first, GSList *last)
{
  gchar  *path;
  GSList *l;

  if (!first)
    return NULL;

  path = g_strdup (first->data);
  if (first == last)
    return path;

  for (l = g_slist_next (first); l; l = g_slist_next (l))
  {
    gchar *path_new;

    path_new = g_strjoin (G_DIR_SEPARATOR_S, path, l->data, NULL);
    g_free (path);
    path = path_new;

    if (l == last)
      break;
  }

  return path;
}

gchar *
tnt_path_join (GSList *pelem)
{
  return join_pelems (pelem, NULL);
}

gboolean
tnt_path_create (const gchar *path)
{
  GSList *elem_list, *l;

  g_return_val_if_fail (path != NULL, FALSE);

  elem_list = tnt_path_split (path);
  for (l = elem_list; l; l = g_slist_next (l))
  {
    gchar *part_path;

    part_path = join_pelems (elem_list, l);

    /* Add preceding '/' on Unix and trailing '\' for Win32 root 'C:\' */
#ifndef G_PLATFORM_WIN32
    {
      gchar *temp;

      temp = g_strconcat (G_DIR_SEPARATOR_S, part_path, NULL);
      g_free (part_path);
      part_path = temp;
    }
#else
    if (elem_list == l && strlen (part_path) == 2)
    {
      gchar *temp;

      temp = g_strconcat (part_path, G_DIR_SEPARATOR_S, NULL);
      g_free (part_path);
      part_path = temp;
    }
#endif

    if (!tnt_stat (part_path, NULL) && tnt_mkdir (part_path) < 0)
    {
      g_free (part_path);
      tnt_slist_free_with_data (elem_list);
      return FALSE;
    }

    g_free (part_path);
  }

  tnt_slist_free_with_data (elem_list);
  return TRUE;
}

gboolean
tnt_stat (const gchar *path, TntStat *tstat)
{
  struct stat sbuf;
  gint        result;

  g_return_val_if_fail (path != NULL, FALSE);

#ifndef G_PLATFORM_WIN32
  result = lstat (path, &sbuf);
#else
  result = stat (path, &sbuf);
#endif

  if (!result)
  {
    if (tstat)
    {
      tstat->len     = sbuf.st_size;
      tstat->mtime   = sbuf.st_mtime;
      tstat->ctime   = sbuf.st_ctime;
      tstat->is_file = S_ISREG (sbuf.st_mode) ? TRUE : FALSE;
      tstat->is_dir  = S_ISDIR (sbuf.st_mode) ? TRUE : FALSE;
    }

    return TRUE;
  }

  return FALSE;
}

gboolean
tnt_lstat (const gchar *path, TntStat *tstat)
{
  struct stat sbuf;
  gint        result;

  g_return_val_if_fail (path != NULL, FALSE);

#ifndef G_PLATFORM_WIN32
  result = lstat (path, &sbuf);
#else
  result = stat (path, &sbuf);
#endif

  if (!result)
  {
    if (tstat)
    {
      tstat->len     = sbuf.st_size;
      tstat->mtime   = sbuf.st_mtime;
      tstat->ctime   = sbuf.st_ctime;
      tstat->is_file = S_ISREG (sbuf.st_mode) ? TRUE : FALSE;
      tstat->is_dir  = S_ISDIR (sbuf.st_mode) ? TRUE : FALSE;
    }

    return TRUE;
  }

  return FALSE;
}

gboolean
tnt_fstat (gint fd, TntStat *tstat)
{
  struct stat sbuf;
  gint        result;

  g_return_val_if_fail (fd >= 0, FALSE);
  g_return_val_if_fail (tstat != NULL, FALSE);

  result = fstat (fd, &sbuf);

  if (!result)
  {
    if (tstat)
    {
      tstat->len     = sbuf.st_size;
      tstat->mtime   = sbuf.st_mtime;
      tstat->ctime   = sbuf.st_ctime;
      tstat->is_file = S_ISREG (sbuf.st_mode) ? TRUE : FALSE;
      tstat->is_dir  = S_ISDIR (sbuf.st_mode) ? TRUE : FALSE;
    }

    return TRUE;
  }

  return FALSE;
}

gint64
tnt_lseek (gint fd, gint64 offset, gint whence)
{
  gint64 result;

  result = lseek (fd, offset, whence);
  return result;
}

gint
tnt_open (const gchar *path, gboolean write, gboolean truncate, gboolean public)
{
  gint   flags;
  mode_t mode;

#ifndef G_PLATFORM_WIN32
  flags = (write ? O_RDWR | O_CREAT : O_RDONLY) | (truncate ? O_TRUNC : 0);
  mode = S_IRUSR | S_IWUSR | (public ? S_IRGRP | S_IWGRP | S_IROTH : 0);
#else
  flags = (write ? O_RDWR | O_CREAT : O_RDONLY) | (truncate ? O_TRUNC : 0) | _O_BINARY;
  mode = S_IRUSR | S_IWUSR;
#endif

  if (write || truncate)
    return open (path, flags, mode);

  return open (path, flags);
}

void
tnt_close (gint fd)
{
  close (fd);
}

gint
tnt_read (gint fd, gpointer buf, gint len)
{
  gint result;

  result = read (fd, buf, len);
  return result;
}

gint
tnt_mkdir (const gchar *path)
{
#ifndef G_PLATFORM_WIN32
  return mkdir (path, S_IRWXU | S_IRWXG);
#else
  return mkdir (path);
#endif
}

gint
tnt_rename (const gchar *old_path, const gchar *new_path)
{
#ifdef G_PLATFORM_WIN32
  remove (new_path);
#endif
  return rename (old_path, new_path);
}

gchar *
tnt_file_id_to_string (const TntFileID *file_id)
{
  gchar *str;
  gint   i;

  g_return_val_if_fail (file_id != NULL, NULL);

  str = g_malloc ((TNT_FILE_ID_LEN * 2) + 1 + 16 + 1);

  for (i = 0; i < TNT_FILE_ID_LEN; i++)
  {
    str [i * 2]       = NIBBLE_TO_HEX_ASCII (file_id->hash [i] >> 4);
    str [(i * 2) + 1] = NIBBLE_TO_HEX_ASCII (file_id->hash [i] & 0x0f);
  }

  str [TNT_FILE_ID_LEN * 2] = '-';
  g_snprintf (str + (TNT_FILE_ID_LEN * 2) + 1, 17, "%016llx", file_id->len);
  str [(TNT_FILE_ID_LEN * 2) + 1 + 16] = '\0';

  return str;
}

/* Does not do checking. Use caution. */
gboolean
tnt_string_to_file_id (const gchar *string, TntFileID *file_id)
{
  guint64 len = 0;
  gint    i;

  g_return_val_if_fail (string != NULL, FALSE);
  g_return_val_if_fail (file_id != NULL, FALSE);

  if (strlen (string) != TNT_FILE_ID_LEN * 2 + 1 + 16)
    return FALSE;

  for (i = 0; i < TNT_FILE_ID_LEN; i++)
  {
    gint b [2];

    b [0] = HEX_ASCII_TO_NIBBLE (string [i * 2]);
    b [1] = HEX_ASCII_TO_NIBBLE (string [(i * 2) + 1]);

    if (b [0] == -1 || b [1] == -1)
      return FALSE;

    file_id->hash [i] = (b [0] << 4) | b [1];
  }

  for (i = 0; i < 16; i++)
  {
    gchar c;
    gint  b;

    c = *(string + (TNT_FILE_ID_LEN * 2) + 1 + i);
    b = HEX_ASCII_TO_NIBBLE (c);
    if (b == -1)
      return FALSE;

    len <<= 4;
    len += b;
  }

  file_id->len = len;
  return TRUE;
}

gchar *
tnt_block_id_to_string (gint block_id)
{
  gchar *str;

  g_return_val_if_fail (block_id >= 0, NULL);

  str = g_strdup_printf ("%04x", block_id);
  return str;
}

gchar *
tnt_file_block_id_to_string (const TntFileID *file_id, gint block_id)
{
  gchar *str;
  gchar *file_str, *block_str;

  g_return_val_if_fail (file_id != NULL, NULL);
  g_return_val_if_fail (block_id >= 0, NULL);

  file_str  = tnt_file_id_to_string (file_id);
  block_str = tnt_block_id_to_string (block_id);

  str = g_strjoin (G_DIR_SEPARATOR_S, file_str, block_str, NULL);

  g_free (block_str);
  g_free (file_str);

  return str;
}

gboolean
tnt_is_valid_file_id_string (const gchar *string)
{
  g_return_val_if_fail (string != NULL, FALSE);

  if (strlen (string) != (TNT_FILE_ID_LEN * 2) + 1 + 16)
    return FALSE;

  if (strspn (string, "0123456789abcdef") != TNT_FILE_ID_LEN * 2)
    return FALSE;

  if (string [TNT_FILE_ID_LEN * 2] != '-')
    return FALSE;

  if (strspn (string + (TNT_FILE_ID_LEN * 2) + 1, "0123456789abcdef") != 16)
    return FALSE;

  return TRUE;
}

gboolean
tnt_is_valid_block_id_string (const gchar *string)
{
  g_return_val_if_fail (string != NULL, FALSE);

  if (strlen (string) != 4)
    return FALSE;

  if (strspn (string, "0123456789abcdef") != 4)
    return FALSE;

  return TRUE;
}

TntNodeID
tnt_public_key_to_node_id (FluxRSAPublicKey *public_key)
{
  TntNodeID  node_id;
  guchar    *raw_key = NULL;
  guint      length  = 0;
  guchar     hash [FLUX_RMD160_HASH_LEN];

  g_return_val_if_fail (public_key != NULL, TNT_NODE_ID_INVALID);

  flux_rsa_export_public_key (public_key, &raw_key, &length);
  flux_rmd160_hash_single (hash, raw_key, length);
  tnt_fold_mem (hash, FLUX_RMD160_HASH_LEN, &node_id, sizeof (TntNodeID));
  g_free (raw_key);

  return node_id;
}

void
tnt_fold_hash (const guchar *hash, guchar *dest, guint len)
{
  g_return_if_fail (hash != NULL);
  g_return_if_fail (dest != NULL);
  g_return_if_fail (len > 0);

  tnt_fold_mem (hash, 20, dest, len);
}

void
tnt_fold_mem (gconstpointer src, guint srclen, gpointer dest, guint destlen)
{
  gint i, j = 0;

  g_return_if_fail (src != NULL);
  g_return_if_fail (dest != NULL);
  g_return_if_fail (srclen > 0);
  g_return_if_fail (destlen > 0);

  memset (dest, 0, destlen);

  for (i = 0; i < srclen; i++)
  {
    ((guchar *) dest) [j++] ^= ((guchar *) src) [i];
    if (j == destlen)
      j = 0;
  }
}

TntTimeStamp
tnt_timestamp (void)
{
  return time (NULL);
}

void
tnt_slist_free_with_data (GSList *slist)
{
  GSList *l;

  for (l = slist; l; l = g_slist_next (l))
  {
    g_free (l->data);
  }

  g_slist_free (slist);
}

void
tnt_list_free_with_data (GList *list)
{
  GList *l;

  for (l = list; l; l = g_list_next (l))
  {
    g_free (l->data);
  }

  g_list_free (list);
}

/* Like uniq(1) */
GList *
tnt_list_unique (GList *list)
{
  GHashTable *hash;
  GList      *l;

  hash = g_hash_table_new (g_direct_hash, g_direct_equal);

  for (l = list; l; l = g_list_next (l))
  {
    GList *last_seen;

    last_seen = g_hash_table_lookup (hash, l->data);

    if (last_seen)
    {
      if (!last_seen->prev)
      {
        list = last_seen->next;
        g_list_delete_link (last_seen, last_seen);
      }
      else
      {
        g_list_delete_link (last_seen->prev, last_seen);
      }
    }
    else
    {
      g_hash_table_insert (hash, l->data, l);
    }
  }

  g_hash_table_destroy (hash);
  return list;
}

static gchar *filename_accept = " .abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789()_-";
static gchar *filename_reject = "/\\[]<>";

gchar *
tnt_filename_to_utf8_lossy (const gchar *filename)
{
  gchar *filename_utf8;

  g_return_val_if_fail (filename != NULL, NULL);

  /* Try the best solution first */
  filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
  if (filename_utf8)
    return filename_utf8;

  /* Emulate G_BROKEN_FILENAMES fallback */
  filename_utf8 = g_locale_to_utf8 (filename, -1, NULL, NULL, NULL);
  if (filename_utf8)
    return filename_utf8;

  /* Try some common codesets */
  filename_utf8 = g_convert (filename, -1, "utf-8", "iso-8859-1", NULL, NULL, NULL);

  if (!filename_utf8)
  {
    gchar *mangled;
    gint   mangled_len;
    gint   pos;

    /* Get rid of characters that are not [ .a-zA-Z0-9()_-] */
    mangled     = g_strdup (filename);
    mangled_len = strlen (mangled);

    while ((pos = strspn (mangled, filename_accept)) < mangled_len)
    {
      mangled [pos] = ' ';
    }

    filename_utf8 = g_filename_to_utf8 (mangled, -1, NULL, NULL, NULL);
    g_free (mangled);
  }

  g_assert (filename_utf8);
  return filename_utf8;
}

gchar *
tnt_filename_to_utf8_replace (gchar *filename)
{
  gchar *filename_utf8;

  g_return_val_if_fail (filename != NULL, NULL);

  filename_utf8 = tnt_filename_to_utf8_lossy (filename);
  g_free (filename);

  return filename_utf8;
}

static gchar *
strip_filename (const gchar *filename)
{
  gchar *mangled;
  gint   mangled_len;
  gint   pos;

  /* Get rid of characters we don't want in filenames */
  mangled     = g_strdup (filename);
  mangled_len = strlen (mangled);

  while ((pos = strcspn (mangled, filename_reject)) < mangled_len)
  {
    mangled [pos] = ' ';
  }

  return mangled;
}

gchar *
tnt_utf8_to_filename_lossy (gchar *string)
{
  gchar *filename_sys, *filename;

  g_return_val_if_fail (string != NULL, NULL);

  filename_sys = g_filename_from_utf8 (string, -1, NULL, NULL, NULL);
  if (!filename_sys)
    return NULL;

  filename = strip_filename (filename_sys);
  g_free (filename_sys);

  return filename;
}

gint
tnt_compare_paths (const gchar *path_1, const gchar *path_2)
{
#ifndef G_PLATFORM_WIN32
  return strcmp (path_1, path_2);
#else
  return strcasecmp (path_1, path_2);
#endif
}

gint
tnt_levenshtein_distance (const gchar *a, const gchar *b)
{
  gint k, i, j, n, m, cost, *d, distance;

  g_return_val_if_fail (a != NULL, -1);
  g_return_val_if_fail (b != NULL, -1);

  /* Step 1 */
  n = strlen (a); 
  m = strlen (b);

  if (n == 0 || m == 0)
    return -1;

  d = g_malloc ((sizeof (gint)) * (m + 1) * (n + 1));
  m++;
  n++;

  /* Step 2	*/
  for(k = 0; k < n; k++)
    d [k] = k;

  for (k = 0; k < m; k++)
    d [k * n] = k;

  /* Step 3 and 4	*/
  for (i = 1; i < n; i++)
  {
    for (j = 1; j < m; j++)
    {
      /* Step 5 */
      if (a [i - 1] == b [j - 1])
        cost = 0;
      else
        cost = 1;

      /* Step 6	*/
      d [j * n + i] = MIN (d [(j - 1) * n + i] + 1, MIN (
                           d [j * n + i - 1] + 1,
                           d [(j - 1) * n + i - 1] + cost));
    }
  }

  distance = d [n * m - 1];
  g_free (d);
  return distance;
}

static gint num_blocks = 0;

gpointer
tnt_alloc_block (gint len)
{
  gpointer block;

  g_return_val_if_fail (len > 0, NULL);

  block = g_malloc (len);
  num_blocks++;

  return block;
}

void
tnt_free_block (gpointer block)
{
  g_free (block);
  num_blocks--;
}

gint
tnt_get_blocks_allocated (void)
{
  return num_blocks;
}

#define PERSISTENT_RETRIES 30

typedef struct
{
  gint                     retries;

  gchar                   *old_path;
  gchar                   *new_path;

  TntPersistentRenameFunc *rename_done_func;
  gpointer                 user_data;
}
PersistentRename;

static gboolean
persistent_rename_cb (PersistentRename *pr)
{
  if (!tnt_rename (pr->old_path, pr->new_path))
  {
    /* Success */

    if (pr->rename_done_func)
      pr->rename_done_func (pr->old_path, pr->new_path, pr->user_data);

    g_free (pr->old_path);
    g_free (pr->new_path);
    g_free (pr);
  }
  else if (pr->retries < PERSISTENT_RETRIES)
  {
    /* Retry */

    pr->retries++;
    g_timeout_add (1000, (GSourceFunc) persistent_rename_cb, pr);
  }
  else
  {
    /* Failure */

    if (pr->rename_done_func)
      pr->rename_done_func (pr->old_path, NULL, pr->user_data);

    g_free (pr->old_path);
    g_free (pr->new_path);
    g_free (pr);
  }

  return FALSE;
}

void
tnt_persistent_rename (const gchar *old_path, const gchar *new_path,
                       TntPersistentRenameFunc *rename_done_func, gpointer data)
{
  PersistentRename *pr;

#if 0
  if (!tnt_rename (old_path, new_path))
    return;
#endif

  pr = g_new0 (PersistentRename, 1);

  pr->old_path = g_strdup (old_path);
  pr->new_path = g_strdup (new_path);

  pr->rename_done_func = rename_done_func;
  pr->user_data = data;

  g_idle_add ((GSourceFunc) persistent_rename_cb, pr);
}

void
tnt_debug_dump_address (const TntNodeAddress4 *address)
{
  g_print ("%d.%d.%d.%d:%d\n", (address->ip >> 24) & 0xff, (address->ip >> 16) & 0xff,
           (address->ip >> 8) & 0xff, (address->ip) & 0xff, address->port);
}
